@@ -1,7 +1,7 @@
 /* column-utils.c
  * Routines for column utilities.
  *
- * $Id: column-utils.c,v 1.32 2003/01/28 18:35:40 guy Exp $
+ * $Id: column-utils.c,v 1.47 2004/03/18 19:04:32 obiot Exp $
  *
  * Ethereal - Network traffic analyzer
  * By Gerald Combs <gerald@ethereal.com>
@@ -48,17 +48,42 @@
 /* Allocate all the data structures for constructing column data, given
    the number of columns. */
 void
-col_init(column_info *col_info, gint num_cols)
+col_setup(column_info *col_info, gint num_cols)
 {
+  int i;
+
   col_info->num_cols	= num_cols;
   col_info->col_fmt	= (gint *) g_malloc(sizeof(gint) * num_cols);
   col_info->fmt_matx	= (gboolean **) g_malloc(sizeof(gboolean *) * num_cols);
-  col_info->col_width	= (gint *) g_malloc(sizeof(gint) * num_cols);
+  col_info->col_first	= (int *) g_malloc(sizeof(int) * (NUM_COL_FMTS));
+  col_info->col_last 	= (int *) g_malloc(sizeof(int) * (NUM_COL_FMTS));
   col_info->col_title	= (gchar **) g_malloc(sizeof(gchar *) * num_cols);
   col_info->col_data	= (gchar **) g_malloc(sizeof(gchar *) * num_cols);
   col_info->col_buf	= (gchar **) g_malloc(sizeof(gchar *) * num_cols);
+  col_info->col_fence	= (int *) g_malloc(sizeof(int) * num_cols);
   col_info->col_expr	= (gchar **) g_malloc(sizeof(gchar *) * num_cols);
   col_info->col_expr_val = (gchar **) g_malloc(sizeof(gchar *) * num_cols);
+
+  for (i = 0; i < NUM_COL_FMTS; i++) {
+    col_info->col_first[i] = -1;
+    col_info->col_last[i] = -1;
+  }
+}
+
+/* Initialize the data structures for constructing column data. */
+void
+col_init(column_info *cinfo)
+{
+  int i;
+
+  for (i = 0; i < cinfo->num_cols; i++) {
+    cinfo->col_buf[i][0] = '\0';
+    cinfo->col_data[i] = cinfo->col_buf[i];
+    cinfo->col_fence[i] = 0;
+    cinfo->col_expr[i][0] = '\0';
+    cinfo->col_expr_val[i][0] = '\0';
+  }
+  cinfo->writable = TRUE;
 }
 
 gboolean
@@ -78,18 +103,35 @@ col_set_writable(column_info *cinfo, gboolean writable)
    the packet list */
 gint
 check_col(column_info *cinfo, gint el) {
-  int i;
 
   if (cinfo && cinfo->writable) {
-    for (i = 0; i < cinfo->num_cols; i++) {
-      if (cinfo->fmt_matx[i][el])
-        return TRUE;
+    /* We are constructing columns, and they're writable */
+    if (cinfo->col_first[el] >= 0) {
+      /* There is at least one column in that format */
+      return TRUE;
     }
   }
   return FALSE;
 }
 
+/* Sets the fence for a column to be at the end of the column. */
+void
+col_set_fence(column_info *cinfo, gint el)
+{
+  int i;
 
+  if (cinfo && cinfo->writable) {
+    /* We are constructing columns, and they're writable */
+    if (cinfo->col_first[el] >= 0) {
+      /* There is at least one column in that format */
+      for (i = cinfo->col_first[el]; i <= cinfo->col_last[el]; i++) {
+        if (cinfo->fmt_matx[i][el]) {
+          cinfo->col_fence[i] = strlen(cinfo->col_data[i]);
+        }
+      }
+    }
+  }
+}
 
 /* Use this to clear out a column, especially if you're going to be
    appending to it later; at least on some platforms, it's more
@@ -98,28 +140,90 @@ check_col(column_info *cinfo, gint el) {
    later append to it, as the later append will cause a string
    copy to be done. */
 void
-col_clear(column_info *cinfo, gint el) {
+col_clear(column_info *cinfo, gint el)
+{
   int    i;
+  int    fence;
 
-  for (i = 0; i < cinfo->num_cols; i++) {
+  g_assert(cinfo->col_first[el] >= 0);
+  for (i = cinfo->col_first[el]; i <= cinfo->col_last[el]; i++) {
     if (cinfo->fmt_matx[i][el]) {
-      cinfo->col_buf[i][0] = 0;
-      cinfo->col_data[i] = cinfo->col_buf[i];
+      /*
+       * At this point, either
+       *
+       *   1) col_data[i] is equal to col_buf[i], in which case we
+       *      don't have to worry about copying col_data[i] to
+       *      col_buf[i];
+       *
+       *   2) col_data[i] isn't equal to col_buf[i], in which case
+       *      the only thing that's been done to the column is
+       *      "col_set_str()" calls and possibly "col_set_fence()"
+       *      calls, in which case the fence is either unset and
+       *      at the beginning of the string or set and at the end
+       *      of the string - if it's at the beginning, we're just
+       *      going to clear the column, and if it's at the end,
+       *      we don't do anything.
+       */
+      fence = cinfo->col_fence[i];
+      if (fence == 0 || cinfo->col_buf[i] == cinfo->col_data[i]) {
+        /*
+         * The fence isn't at the end of the column, or the column wasn't
+         * last set with "col_set_str()", so clear the column out.
+         */
+        cinfo->col_buf[i][fence] = '\0';
+        cinfo->col_data[i] = cinfo->col_buf[i];
+      }
       cinfo->col_expr[i][0] = '\0';
       cinfo->col_expr_val[i][0] = '\0';
     }
   }
 }
 
+#define COL_CHECK_APPEND(cinfo, i, max_len) \
+  if (cinfo->col_data[i] != cinfo->col_buf[i]) {		\
+    /* This was set with "col_set_str()"; copy the string they	\
+       set it to into the buffer, so we can append to it. */	\
+    strncpy(cinfo->col_buf[i], cinfo->col_data[i], max_len);	\
+    cinfo->col_buf[i][max_len - 1] = '\0';			\
+    cinfo->col_data[i] = cinfo->col_buf[i];			\
+  }
+
 /* Use this if "str" points to something that will stay around (and thus
    needn't be copied). */
 void
-col_set_str(column_info *cinfo, gint el, gchar* str) {
+col_set_str(column_info *cinfo, gint el, gchar* str)
+{
   int i;
+  int fence;
+  size_t len, max_len;
 
-  for (i = 0; i < cinfo->num_cols; i++) {
-    if (cinfo->fmt_matx[i][el])
-      cinfo->col_data[i] = str;
+  if (el == COL_INFO)
+	max_len = COL_MAX_INFO_LEN;
+  else
+	max_len = COL_MAX_LEN;
+
+  g_assert(cinfo->col_first[el] >= 0);
+  for (i = cinfo->col_first[el]; i <= cinfo->col_last[el]; i++) {
+    if (cinfo->fmt_matx[i][el]) {
+      fence = cinfo->col_fence[i];
+      if (fence != 0) {
+        /*
+         * We will append the string after the fence.
+         * First arrange that we can append, if necessary.
+         */
+        COL_CHECK_APPEND(cinfo, i, max_len);
+
+        len = strlen(cinfo->col_buf[i]);
+        strncat(cinfo->col_buf[i], str, max_len - len);
+        cinfo->col_buf[i][max_len - 1] = 0;
+      } else {
+        /*
+         * There's no fence, so we can just set the column to point
+         * to the string.
+         */
+        cinfo->col_data[i] = str;
+      }
+    }
   }
 }
 
@@ -128,18 +232,32 @@ void
 col_add_fstr(column_info *cinfo, gint el, const gchar *format, ...) {
   va_list ap;
   int     i;
+  int     fence;
   size_t  max_len;
 
+  g_assert(cinfo->col_first[el] >= 0);
   if (el == COL_INFO)
 	max_len = COL_MAX_INFO_LEN;
   else
 	max_len = COL_MAX_LEN;
 
   va_start(ap, format);
-  for (i = 0; i < cinfo->num_cols; i++) {
+  for (i = cinfo->col_first[el]; i <= cinfo->col_last[el]; i++) {
     if (cinfo->fmt_matx[i][el]) {
-      vsnprintf(cinfo->col_buf[i], max_len, format, ap);
-      cinfo->col_data[i] = cinfo->col_buf[i];
+      fence = cinfo->col_fence[i];
+      if (fence != 0) {
+        /*
+         * We will append the string after the fence.
+         * First arrange that we can append, if necessary.
+         */
+        COL_CHECK_APPEND(cinfo, i, max_len);
+      } else {
+        /*
+         * There's no fence, so we can just write to the string.
+         */
+        cinfo->col_data[i] = cinfo->col_buf[i];
+      }
+      vsnprintf(&cinfo->col_buf[i][fence], max_len - fence, format, ap);
     }
   }
   va_end(ap);
@@ -147,33 +265,67 @@ col_add_fstr(column_info *cinfo, gint el, const gchar *format, ...) {
 
 /* Appends a vararg list to a packet info string. */
 void
-col_append_fstr(column_info *cinfo, gint el, const gchar *format, ...) {
+col_append_fstr(column_info *cinfo, gint el, const gchar *format, ...)
+{
   va_list ap;
   int     i;
   size_t  len, max_len;
 
+  g_assert(cinfo->col_first[el] >= 0);
   if (el == COL_INFO)
 	max_len = COL_MAX_INFO_LEN;
   else
 	max_len = COL_MAX_LEN;
 
   va_start(ap, format);
-  for (i = 0; i < cinfo->num_cols; i++) {
+  for (i = cinfo->col_first[el]; i <= cinfo->col_last[el]; i++) {
     if (cinfo->fmt_matx[i][el]) {
-      if (cinfo->col_data[i] != cinfo->col_buf[i]) {
-      	/* This was set with "col_set_str()"; copy the string they
-	   set it to into the buffer, so we can append to it. */
-	strncpy(cinfo->col_buf[i], cinfo->col_data[i], max_len);
-	cinfo->col_buf[i][max_len - 1] = '\0';
-      }
+      /*
+       * First arrange that we can append, if necessary.
+       */
+      COL_CHECK_APPEND(cinfo, i, max_len);
       len = strlen(cinfo->col_buf[i]);
       vsnprintf(&cinfo->col_buf[i][len], max_len - len, format, ap);
-      cinfo->col_data[i] = cinfo->col_buf[i];
     }
   }
   va_end(ap);
 }
 
+/* Appends a vararg list to a packet info string.
+ * Prefixes it with the given separator if the column is not empty. */
+void
+col_append_sep_fstr(column_info *cinfo, gint el, const gchar *separator,
+		const gchar *format, ...)
+{
+  va_list ap;
+  int     i;
+  size_t  len, max_len;
+
+  g_assert(cinfo->col_first[el] >= 0);
+  if (el == COL_INFO)
+	max_len = COL_MAX_INFO_LEN;
+  else
+	max_len = COL_MAX_LEN;
+
+  if (cinfo->col_buf[0]) {
+    col_append_str(cinfo, el, separator ? separator : ", ");
+  }
+  va_start(ap, format);
+  for (i = cinfo->col_first[el]; i <= cinfo->col_last[el]; i++) {
+    if (cinfo->fmt_matx[i][el]) {
+      /*
+       * First arrange that we can append, if necessary.
+       */
+      COL_CHECK_APPEND(cinfo, i, max_len);
+      len = strlen(cinfo->col_buf[i]);
+      vsnprintf(&cinfo->col_buf[i][len], max_len - len, format, ap);
+    }
+  }
+  va_end(ap);
+}
+
+
+
 /* Prepends a vararg list to a packet info string. */
 #define COL_BUF_MAX_LEN (((COL_MAX_INFO_LEN) > (COL_MAX_LEN)) ? \
 	(COL_MAX_INFO_LEN) : (COL_MAX_LEN))
@@ -186,13 +338,14 @@ col_prepend_fstr(column_info *cinfo, gint el, const gchar *format, ...)
   char   *orig;
   size_t  max_len;
 
+  g_assert(cinfo->col_first[el] >= 0);
   if (el == COL_INFO)
 	max_len = COL_MAX_INFO_LEN;
   else
 	max_len = COL_MAX_LEN;
 
   va_start(ap, format);
-  for (i = 0; i < cinfo->num_cols; i++) {
+  for (i = cinfo->col_first[el]; i <= cinfo->col_last[el]; i++) {
     if (cinfo->fmt_matx[i][el]) {
       if (cinfo->col_data[i] != cinfo->col_buf[i]) {
       	/* This was set with "col_set_str()"; which is effectively const */
@@ -203,6 +356,14 @@ col_prepend_fstr(column_info *cinfo, gint el, const gchar *format, ...)
 	orig[max_len - 1] = '\0';
       }
       vsnprintf(cinfo->col_buf[i], max_len, format, ap);
+      cinfo->col_buf[i][max_len - 1] = '\0';
+
+      /*
+       * Move the fence, unless it's at the beginning of the string.
+       */
+      if (cinfo->col_fence[i] > 0)
+        cinfo->col_fence[i] += strlen(cinfo->col_buf[i]);
+
       strncat(cinfo->col_buf[i], orig, max_len);
       cinfo->col_buf[i][max_len - 1] = '\0';
       cinfo->col_data[i] = cinfo->col_buf[i];
@@ -214,50 +375,74 @@ col_prepend_fstr(column_info *cinfo, gint el, const gchar *format, ...)
 /* Use this if "str" points to something that won't stay around (and
    must thus be copied). */
 void
-col_add_str(column_info *cinfo, gint el, const gchar* str) {
+col_add_str(column_info *cinfo, gint el, const gchar* str)
+{
   int    i;
+  int    fence;
   size_t max_len;
 
+  g_assert(cinfo->col_first[el] >= 0);
   if (el == COL_INFO)
 	max_len = COL_MAX_INFO_LEN;
   else
 	max_len = COL_MAX_LEN;
 
-  for (i = 0; i < cinfo->num_cols; i++) {
+  for (i = cinfo->col_first[el]; i <= cinfo->col_last[el]; i++) {
     if (cinfo->fmt_matx[i][el]) {
-      strncpy(cinfo->col_buf[i], str, max_len);
+      fence = cinfo->col_fence[i];
+      if (fence != 0) {
+        /*
+         * We will append the string after the fence.
+         * First arrange that we can append, if necessary.
+         */
+        COL_CHECK_APPEND(cinfo, i, max_len);
+      } else {
+        /*
+         * There's no fence, so we can just write to the string.
+         */
+        cinfo->col_data[i] = cinfo->col_buf[i];
+      }
+      strncpy(&cinfo->col_buf[i][fence], str, max_len - fence);
       cinfo->col_buf[i][max_len - 1] = 0;
-      cinfo->col_data[i] = cinfo->col_buf[i];
     }
   }
 }
 
 void
-col_append_str(column_info *cinfo, gint el, const gchar* str) {
+col_append_str(column_info *cinfo, gint el, const gchar* str)
+{
   int    i;
   size_t len, max_len;
 
+  g_assert(cinfo->col_first[el] >= 0);
   if (el == COL_INFO)
 	max_len = COL_MAX_INFO_LEN;
   else
 	max_len = COL_MAX_LEN;
 
-  for (i = 0; i < cinfo->num_cols; i++) {
+  for (i = cinfo->col_first[el]; i <= cinfo->col_last[el]; i++) {
     if (cinfo->fmt_matx[i][el]) {
-      if (cinfo->col_data[i] != cinfo->col_buf[i]) {
-      	/* This was set with "col_set_str()"; copy the string they
-	   set it to into the buffer, so we can append to it. */
-	strncpy(cinfo->col_buf[i], cinfo->col_data[i], max_len);
-	cinfo->col_buf[i][max_len - 1] = '\0';
-      }
+      /*
+       * First arrange that we can append, if necessary.
+       */
+      COL_CHECK_APPEND(cinfo, i, max_len);
       len = strlen(cinfo->col_buf[i]);
       strncat(cinfo->col_buf[i], str, max_len - len);
       cinfo->col_buf[i][max_len - 1] = 0;
-      cinfo->col_data[i] = cinfo->col_buf[i];
     }
   }
 }
 
+void
+col_append_sep_str(column_info *cinfo, gint el, const gchar* separator,
+    const gchar* str)
+{
+  if (cinfo->col_buf[0]) {
+    col_append_str(cinfo, el, separator ? separator : ", ");
+  }  
+  col_append_str(cinfo, el, str);
+}
+
 static void
 col_set_abs_date_time(frame_data *fd, column_info *cinfo, int col)
 {
@@ -338,20 +523,20 @@ col_set_abs_time(frame_data *fd, column_info *cinfo, int col)
 void
 col_set_cls_time(frame_data *fd, column_info *cinfo, int col)
 {
-  switch (timestamp_type) {
-    case ABSOLUTE:
+  switch (get_timestamp_setting()) {
+    case TS_ABSOLUTE:
       col_set_abs_time(fd, cinfo, col);
       break;
 
-    case ABSOLUTE_WITH_DATE:
+    case TS_ABSOLUTE_WITH_DATE:
       col_set_abs_date_time(fd, cinfo, col);
       break;
 
-    case RELATIVE:
+    case TS_RELATIVE:
       col_set_rel_time(fd, cinfo, col);
       break;
 
-    case DELTA:
+    case TS_DELTA:
       col_set_delta_time(fd, cinfo, col);
       break;
   }
@@ -363,21 +548,43 @@ col_set_addr(packet_info *pinfo, int col, address *addr, gboolean is_res,
 {
   guint32 ipv4_addr;
   struct e_in6_addr ipv6_addr;
-  struct atalk_ddp_addr ddp_addr;
-  struct sna_fid_type_4_addr sna_fid_type_4_addr;
-  guint8 fcid[3];
 
   pinfo->cinfo->col_expr[col][0] = '\0';
   pinfo->cinfo->col_expr_val[col][0] = '\0';
+  if (addr->type == AT_NONE)
+    return;	/* no address, nothing to do */
+  if (is_res) {
+    switch (addr->type) {
+
+    case AT_ETHER:
+      strncpy(pinfo->cinfo->col_buf[col], get_ether_name(addr->data), COL_MAX_LEN);
+      pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
+      pinfo->cinfo->col_data[col] = pinfo->cinfo->col_buf[col];
+      break;
+
+    case AT_IPv4:
+      memcpy(&ipv4_addr, addr->data, sizeof ipv4_addr);
+      strncpy(pinfo->cinfo->col_buf[col], get_hostname(ipv4_addr), COL_MAX_LEN);
+      pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
+      break;
+
+    case AT_IPv6:
+      memcpy(&ipv6_addr.s6_addr, addr->data, sizeof ipv6_addr.s6_addr);
+      strncpy(pinfo->cinfo->col_buf[col], get_hostname6(&ipv6_addr), COL_MAX_LEN);
+      pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
+      break;
+
+    default:
+      address_to_str_buf(addr, pinfo->cinfo->col_buf[col]);
+      break;
+    }
+  } else
+    address_to_str_buf(addr, pinfo->cinfo->col_buf[col]);
+  pinfo->cinfo->col_data[col] = pinfo->cinfo->col_buf[col];
+
   switch (addr->type) {
 
   case AT_ETHER:
-    if (is_res)
-      strncpy(pinfo->cinfo->col_buf[col], get_ether_name(addr->data), COL_MAX_LEN);
-    else
-      strncpy(pinfo->cinfo->col_buf[col], ether_to_str(addr->data), COL_MAX_LEN);
-    pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
-    pinfo->cinfo->col_data[col] = pinfo->cinfo->col_buf[col];
     if (is_src)
       strcpy(pinfo->cinfo->col_expr[col], "eth.src");
     else
@@ -387,13 +594,6 @@ col_set_addr(packet_info *pinfo, int col, address *addr, gboolean is_res,
     break;
 
   case AT_IPv4:
-    memcpy(&ipv4_addr, addr->data, sizeof ipv4_addr);
-    if (is_res)
-      strncpy(pinfo->cinfo->col_buf[col], get_hostname(ipv4_addr), COL_MAX_LEN);
-    else
-      strncpy(pinfo->cinfo->col_buf[col], ip_to_str(addr->data), COL_MAX_LEN);
-    pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
-    pinfo->cinfo->col_data[col] = pinfo->cinfo->col_buf[col];
     if (is_src)
       strcpy(pinfo->cinfo->col_expr[col], "ip.src");
     else
@@ -403,13 +603,6 @@ col_set_addr(packet_info *pinfo, int col, address *addr, gboolean is_res,
     break;
 
   case AT_IPv6:
-    memcpy(&ipv6_addr.s6_addr, addr->data, sizeof ipv6_addr.s6_addr);
-    if (is_res)
-      strncpy(pinfo->cinfo->col_buf[col], get_hostname6(&ipv6_addr), COL_MAX_LEN);
-    else
-      strncpy(pinfo->cinfo->col_buf[col], ip6_to_str(&ipv6_addr), COL_MAX_LEN);
-    pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
-    pinfo->cinfo->col_data[col] = pinfo->cinfo->col_buf[col];
     if (is_src)
       strcpy(pinfo->cinfo->col_expr[col], "ipv6.src");
     else
@@ -418,81 +611,22 @@ col_set_addr(packet_info *pinfo, int col, address *addr, gboolean is_res,
     pinfo->cinfo->col_expr_val[col][COL_MAX_LEN - 1] = '\0';
     break;
 
-  case AT_IPX:
-    strncpy(pinfo->cinfo->col_buf[col],
-      ipx_addr_to_str(pntohl(&addr->data[0]), &addr->data[4]), COL_MAX_LEN);
-    pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
-    pinfo->cinfo->col_data[col] = pinfo->cinfo->col_buf[col];
-    break;
-
-  case AT_SNA:
-    switch (addr->len) {
-
-    case 1:
-      snprintf(pinfo->cinfo->col_buf[col], COL_MAX_LEN, "%04X", addr->data[0]);
-      break;
-
-    case 2:
-      snprintf(pinfo->cinfo->col_buf[col], COL_MAX_LEN, "%04X",
-        pntohs(&addr->data[0]));
-      break;
-
-    case SNA_FID_TYPE_4_ADDR_LEN:
-      memcpy(&sna_fid_type_4_addr, addr->data, SNA_FID_TYPE_4_ADDR_LEN);
-      strncpy(pinfo->cinfo->col_buf[col],
-        sna_fid_type_4_addr_to_str(&sna_fid_type_4_addr), COL_MAX_LEN);
-      break;
-    }
-    pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
-    pinfo->cinfo->col_data[col] = pinfo->cinfo->col_buf[col];
-    break;
-
   case AT_ATALK:
-    memcpy(&ddp_addr, addr->data, sizeof ddp_addr);
-    strncpy(pinfo->cinfo->col_buf[col], atalk_addr_to_str(&ddp_addr),
-      COL_MAX_LEN);
-    pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
     if (is_src)
       strcpy(pinfo->cinfo->col_expr[col], "ddp.src");
     else
       strcpy(pinfo->cinfo->col_expr[col], "ddp.dst");
-    pinfo->cinfo->col_data[col] = pinfo->cinfo->col_buf[col];
-    strcpy(pinfo->cinfo->col_expr_val[col],pinfo->cinfo->col_buf[col]);
-    break;
-
-  case AT_VINES:
-    strncpy(pinfo->cinfo->col_buf[col], vines_addr_to_str(&addr->data[0]),
-      COL_MAX_LEN);
-    pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
-    pinfo->cinfo->col_data[col] = pinfo->cinfo->col_buf[col];
-    break;
-
-  case AT_OSI:
-    strncpy(pinfo->cinfo->col_buf[col], print_nsap_net(addr->data, addr->len),
-      COL_MAX_LEN);
-    pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
-    pinfo->cinfo->col_data[col] = pinfo->cinfo->col_buf[col];
+    strcpy(pinfo->cinfo->col_expr_val[col], pinfo->cinfo->col_buf[col]);
     break;
 
   case AT_ARCNET:
-    snprintf(pinfo->cinfo->col_buf[col], COL_MAX_LEN, "0x%02X",
-      addr->data[0]);
-    pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
     if (is_src)
       strcpy(pinfo->cinfo->col_expr[col], "arcnet.src");
     else
       strcpy(pinfo->cinfo->col_expr[col], "arcnet.dst");
-    pinfo->cinfo->col_data[col] = pinfo->cinfo->col_buf[col];
-    strcpy(pinfo->cinfo->col_expr_val[col],pinfo->cinfo->col_buf[col]);
+    strcpy(pinfo->cinfo->col_expr_val[col], pinfo->cinfo->col_buf[col]);
     break;
 
-  case AT_FC:
-    memcpy(fcid, addr->data, sizeof fcid);
-    strncpy(pinfo->cinfo->col_buf[col], fc_to_str(fcid), COL_MAX_LEN);
-    pinfo->cinfo->col_buf[col][COL_MAX_LEN - 1] = '\0';
-    pinfo->cinfo->col_data[col] = pinfo->cinfo->col_buf[col];
-    break;
-      
   default:
     break;
   }
@@ -637,6 +771,13 @@ col_set_circuit_id(packet_info *pinfo, int col)
     snprintf(pinfo->cinfo->col_buf[col], COL_MAX_LEN, "%u", pinfo->circuit_id);
     break;
 
+  case CT_ISUP:
+    snprintf(pinfo->cinfo->col_buf[col], COL_MAX_LEN, "%u", pinfo->circuit_id);
+    strcpy(pinfo->cinfo->col_expr[col], "isup.cic");
+    snprintf(pinfo->cinfo->col_expr_val[col], COL_MAX_LEN, "%u", pinfo->circuit_id);
+    pinfo->cinfo->col_expr_val[col][COL_MAX_LEN - 1] = '\0';
+    break;
+
   default:
     break;
   }
@@ -660,23 +801,48 @@ fill_in_columns(packet_info *pinfo)
       break;
 
     case COL_CLS_TIME:
-      col_set_cls_time(pinfo->fd, pinfo->cinfo, i);
+      if(pinfo->fd->flags.ref_time){
+         snprintf(pinfo->cinfo->col_buf[i], COL_MAX_LEN, "*REF*");
+         pinfo->cinfo->col_data[i] = pinfo->cinfo->col_buf[i];
+      } else {
+         col_set_cls_time(pinfo->fd, pinfo->cinfo, i);
+      }
       break;
 
     case COL_ABS_TIME:
-      col_set_abs_time(pinfo->fd, pinfo->cinfo, i);
+      if(pinfo->fd->flags.ref_time){
+         snprintf(pinfo->cinfo->col_buf[i], COL_MAX_LEN, "*REF*");
+         pinfo->cinfo->col_data[i] = pinfo->cinfo->col_buf[i];
+      } else {
+         col_set_abs_time(pinfo->fd, pinfo->cinfo, i);
+      }
       break;
 
     case COL_ABS_DATE_TIME:
-      col_set_abs_date_time(pinfo->fd, pinfo->cinfo, i);
+      if(pinfo->fd->flags.ref_time){
+         snprintf(pinfo->cinfo->col_buf[i], COL_MAX_LEN, "*REF*");
+         pinfo->cinfo->col_data[i] = pinfo->cinfo->col_buf[i];
+      } else {
+         col_set_abs_date_time(pinfo->fd, pinfo->cinfo, i);
+      }
       break;
 
     case COL_REL_TIME:
-      col_set_rel_time(pinfo->fd, pinfo->cinfo, i);
+      if(pinfo->fd->flags.ref_time){
+         snprintf(pinfo->cinfo->col_buf[i], COL_MAX_LEN, "*REF*");
+         pinfo->cinfo->col_data[i] = pinfo->cinfo->col_buf[i];
+      } else {
+         col_set_rel_time(pinfo->fd, pinfo->cinfo, i);
+      }
       break;
 
     case COL_DELTA_TIME:
-      col_set_delta_time(pinfo->fd, pinfo->cinfo, i);
+      if(pinfo->fd->flags.ref_time){
+         snprintf(pinfo->cinfo->col_buf[i], COL_MAX_LEN, "*REF*");
+         pinfo->cinfo->col_data[i] = pinfo->cinfo->col_buf[i];
+      } else {
+         col_set_delta_time(pinfo->fd, pinfo->cinfo, i);
+      }
       break;
 
     case COL_DEF_SRC:
@@ -762,6 +928,11 @@ fill_in_columns(packet_info *pinfo)
       strcpy(pinfo->cinfo->col_expr_val[i], pinfo->cinfo->col_buf[i]);
       break;
 
+    case COL_CUMULATIVE_BYTES:
+      snprintf(pinfo->cinfo->col_buf[i], COL_MAX_LEN, "%u", pinfo->fd->cum_bytes);
+      pinfo->cinfo->col_data[i] = pinfo->cinfo->col_buf[i];
+      break;
+
     case COL_OXID:
       snprintf(pinfo->cinfo->col_buf[i], COL_MAX_LEN, "0x%x", pinfo->oxid);
       pinfo->cinfo->col_buf[i][COL_MAX_LEN - 1] = '\0';
@@ -782,24 +953,22 @@ fill_in_columns(packet_info *pinfo)
       break;
 
     case COL_SRCIDX:
-        snprintf (pinfo->cinfo->col_buf[i], COL_MAX_LEN, "0x%x",
-                  pinfo->src_idx);
-        pinfo->cinfo->col_buf[i][COL_MAX_LEN - 1] = '\0';
-        pinfo->cinfo->col_data[i] = pinfo->cinfo->col_buf[i];
-        break;
+      snprintf(pinfo->cinfo->col_buf[i], COL_MAX_LEN, "0x%x", pinfo->src_idx);
+      pinfo->cinfo->col_buf[i][COL_MAX_LEN - 1] = '\0';
+      pinfo->cinfo->col_data[i] = pinfo->cinfo->col_buf[i];
+      break;
 
     case COL_DSTIDX:
-        snprintf (pinfo->cinfo->col_buf[i], COL_MAX_LEN, "0x%x",
-                  pinfo->dst_idx);
-        pinfo->cinfo->col_buf[i][COL_MAX_LEN - 1] = '\0';
-        pinfo->cinfo->col_data[i] = pinfo->cinfo->col_buf[i];
-        break;
+      snprintf(pinfo->cinfo->col_buf[i], COL_MAX_LEN, "0x%x", pinfo->dst_idx);
+      pinfo->cinfo->col_buf[i][COL_MAX_LEN - 1] = '\0';
+      pinfo->cinfo->col_data[i] = pinfo->cinfo->col_buf[i];
+      break;
 
     case COL_VSAN:
-        snprintf (pinfo->cinfo->col_buf[i], COL_MAX_LEN, "%d", pinfo->vsan);
-        pinfo->cinfo->col_buf[i][COL_MAX_LEN - 1] = '\0';
-        pinfo->cinfo->col_data[i] = pinfo->cinfo->col_buf[i];
-        break;
+      snprintf(pinfo->cinfo->col_buf[i], COL_MAX_LEN, "%u", pinfo->vsan);
+      pinfo->cinfo->col_buf[i][COL_MAX_LEN - 1] = '\0';
+      pinfo->cinfo->col_data[i] = pinfo->cinfo->col_buf[i];
+      break;
         
     case NUM_COL_FMTS:	/* keep compiler happy - shouldn't get here */
       g_assert_not_reached();
