@@ -1,7 +1,7 @@
 /* packet.c
  * Routines for packet disassembly
  *
- * $Id: packet.c,v 1.88 2003/03/01 09:38:41 guy Exp $
+ * $Id: packet.c,v 1.103 2004/02/28 20:59:23 guy Exp $
  *
  * Ethereal - Network traffic analyzer
  * By Gerald Combs <gerald@ethereal.com>
@@ -105,15 +105,16 @@ static GSList *init_routines;
 void
 register_init_routine(void (*func)(void))
 {
-	init_routines = g_slist_append(init_routines, func);
+	init_routines = g_slist_append(init_routines, (gpointer)func);
 }
 
+typedef void (*void_func_t)(void);
+
 /* Initialize all data structures used for dissection. */
 static void
 call_init_routine(gpointer routine, gpointer dummy _U_)
 {
-	void (*func)(void) = routine;
-
+	void_func_t func = (void_func_t)routine;
 	(*func)();
 }
 
@@ -143,18 +144,17 @@ init_dissection(void)
 static GSList *postseq_cleanup_routines;
 
 void
-register_postseq_cleanup_routine(void (*func)(void))
+register_postseq_cleanup_routine(void_func_t func)
 {
 	postseq_cleanup_routines = g_slist_append(postseq_cleanup_routines,
-			func);
+			(gpointer)func);
 }
 
 /* Call all the registered "postseq_cleanup" routines. */
 static void
 call_postseq_cleanup_routine(gpointer routine, gpointer dummy _U_)
 {
-	void (*func)(void) = routine;
-
+	void_func_t func = (void_func_t)routine;
 	(*func)();
 }
 
@@ -224,14 +224,14 @@ void
 register_final_registration_routine(void (*func)(void))
 {
 	final_registration_routines = g_slist_append(final_registration_routines,
-			func);
+			(gpointer)func);
 }
 
 /* Call all the registered "final_registration" routines. */
 static void
 call_final_registration_routine(gpointer routine, gpointer dummy _U_)
 {
-	void (*func)(void) = routine;
+	void_func_t func = (void_func_t)routine;
 
 	(*func)();
 }
@@ -249,18 +249,8 @@ void
 dissect_packet(epan_dissect_t *edt, union wtap_pseudo_header *pseudo_header,
 	       const guchar *pd, frame_data *fd, column_info *cinfo)
 {
-	int i;
-
-	if (cinfo != NULL) {
-		for (i = 0; i < cinfo->num_cols; i++) {
-			cinfo->col_buf[i][0] = '\0';
-			cinfo->col_data[i] = cinfo->col_buf[i];
-			cinfo->col_expr[i][0] = '\0';
-			cinfo->col_expr_val[i][0] = '\0';
-		}
-
-		col_set_writable(cinfo, TRUE);
-	}
+	if (cinfo != NULL)
+		col_init(cinfo);
 	edt->pi.current_proto = "<Missing Protocol Name>";
 	edt->pi.cinfo = cinfo;
 	edt->pi.fd = fd;
@@ -296,7 +286,9 @@ dissect_packet(epan_dissect_t *edt, union wtap_pseudo_header *pseudo_header,
 	edt->pi.srcport  = 0;
 	edt->pi.destport = 0;
 	edt->pi.match_port = 0;
+	edt->pi.match_string = NULL;
 	edt->pi.can_desegment = 0;
+	edt->pi.want_pdu_tracking = 0;
 	edt->pi.p2p_dir = P2P_DIR_UNKNOWN;
 	edt->pi.private_data = NULL;
         edt->pi.oxid = 0;
@@ -348,15 +340,23 @@ struct dissector_handle {
 		dissector_t	old;
 		new_dissector_t	new;
 	} dissector;
-	int		proto_index;
+	protocol_t	*protocol;
 };
 
 static int
 call_dissector_through_handle(dissector_handle_t handle, tvbuff_t *tvb,
     packet_info *pinfo, proto_tree *tree)
 {
+	const char *saved_proto;
 	int ret;
 
+	saved_proto = pinfo->current_proto;
+
+	if (handle->protocol != NULL) {
+		pinfo->current_proto =
+		    proto_get_protocol_short_name(handle->protocol);
+	}
+
 	if (handle->is_new)
 		ret = (*handle->dissector.new)(tvb, pinfo, tree);
 	else {
@@ -371,6 +371,9 @@ call_dissector_through_handle(dissector_handle_t handle, tvbuff_t *tvb,
 			ret = 1;
 		}
 	}
+
+	pinfo->current_proto = saved_proto;
+
 	return ret;
 }
 
@@ -397,8 +400,8 @@ call_dissector_work(dissector_handle_t handle, tvbuff_t *tvb,
 	volatile address save_src;
 	volatile address save_dst;
 
-	if (handle->proto_index != -1 &&
-	    !proto_is_protocol_enabled(handle->proto_index)) {
+	if (handle->protocol != NULL &&
+	    !proto_is_protocol_enabled(handle->protocol)) {
 		/*
 		 * The protocol isn't enabled.
 		 */
@@ -415,11 +418,16 @@ call_dissector_work(dissector_handle_t handle, tvbuff_t *tvb,
 	 * by one.
 	 * Thus only the subdissector immediately on top of whoever
 	 * offers this service can use it.
+	 * We save the current value of "can_desegment" for the
+	 * benefit of TCP proxying dissectors such as SOCKS, so they
+	 * can restore it and allow the dissectors they call to use
+	 * the desegmentation service.
 	 */
+	pinfo->saved_can_desegment = saved_can_desegment;
 	pinfo->can_desegment = saved_can_desegment-(saved_can_desegment>0);
-	if (handle->proto_index != -1) {
+	if (handle->protocol != NULL) {
 		pinfo->current_proto =
-		    proto_get_protocol_short_name(handle->proto_index);
+		    proto_get_protocol_short_name(handle->protocol);
 	}
 
 	if (pinfo->in_error_pkt) {
@@ -497,6 +505,7 @@ call_dissector_work(dissector_handle_t handle, tvbuff_t *tvb,
 		pinfo->net_dst = save_net_dst;
 		pinfo->src = save_src;
 		pinfo->dst = save_dst;
+		pinfo->want_pdu_tracking = 0;
 	} else {
 		/*
 		 * Just call the subdissector.
@@ -553,14 +562,63 @@ find_dissector_table(const char *name)
 	return g_hash_table_lookup( dissector_tables, name );
 }
 
+/* Find an entry in a uint dissector table. */ 
+static dtbl_entry_t *
+find_uint_dtbl_entry(dissector_table_t sub_dissectors, guint32 pattern)
+{
+	switch (sub_dissectors->type) {
+
+	case FT_UINT8:
+	case FT_UINT16:
+	case FT_UINT24:
+	case FT_UINT32:
+		/*
+		 * You can do a port lookup in these tables.
+		 */
+		break;
+
+	default:
+		/*
+		 * But you can't do a port lookup in any other types
+		 * of tables.
+		 */
+		g_assert_not_reached();
+	}
+
+	/*
+	 * Find the entry.
+	 */
+	return g_hash_table_lookup(sub_dissectors->hash_table,
+	    GUINT_TO_POINTER(pattern));
+}
+
+/* Add an entry to a uint dissector table. */
 void
 dissector_add(const char *name, guint32 pattern, dissector_handle_t handle)
 {
 	dissector_table_t sub_dissectors = find_dissector_table( name);
 	dtbl_entry_t *dtbl_entry;
 
-/* sanity check */
+/* sanity checks */
 	g_assert( sub_dissectors);
+	switch (sub_dissectors->type) {
+
+	case FT_UINT8:
+	case FT_UINT16:
+	case FT_UINT24:
+	case FT_UINT32:
+		/*
+		 * You can do a port lookup in these tables.
+		 */
+		break;
+
+	default:
+		/*
+		 * But you can't do a port lookup in any other types
+		 * of tables.
+		 */
+		g_assert_not_reached();
+	}
 
 	dtbl_entry = g_malloc(sizeof (dtbl_entry_t));
 	dtbl_entry->current = handle;
@@ -577,7 +635,8 @@ dissector_add(const char *name, guint32 pattern, dissector_handle_t handle)
 	dissector_add_handle(name, handle);
 }
 
-/* delete the entry for this dissector at this pattern */
+/* Delete the entry for a dissector in a uint dissector table
+   with a particular pattern. */
 
 /* NOTE: this doesn't use the dissector call variable. It is included to */
 /*	be consistant with the dissector_add and more importantly to be used */
@@ -597,8 +656,7 @@ dissector_delete(const char *name, guint32 pattern,
 	/*
 	 * Find the entry.
 	 */
-	dtbl_entry = g_hash_table_lookup(sub_dissectors->hash_table,
-	    GUINT_TO_POINTER(pattern));
+	dtbl_entry = find_uint_dtbl_entry(sub_dissectors, pattern);
 
 	if (dtbl_entry != NULL) {
 		/*
@@ -614,6 +672,8 @@ dissector_delete(const char *name, guint32 pattern,
 	}
 }
 
+/* Change the entry for a dissector in a uint dissector table
+   with a particular pattern to use a new dissector handle. */
 void
 dissector_change(const char *name, guint32 pattern, dissector_handle_t handle)
 {
@@ -626,8 +686,7 @@ dissector_change(const char *name, guint32 pattern, dissector_handle_t handle)
 	/*
 	 * See if the entry already exists. If so, reuse it.
 	 */
-	dtbl_entry = g_hash_table_lookup(sub_dissectors->hash_table,
-	    GUINT_TO_POINTER(pattern));
+	dtbl_entry = find_uint_dtbl_entry(sub_dissectors, pattern);
 	if (dtbl_entry != NULL) {
 	  dtbl_entry->current = handle;
 	  return;
@@ -650,7 +709,7 @@ dissector_change(const char *name, guint32 pattern, dissector_handle_t handle)
     	    GUINT_TO_POINTER( pattern), (gpointer)dtbl_entry);
 }
 
-/* Reset a dissector in a sub-dissector table to its initial value. */
+/* Reset an entry in a uint dissector table to its initial value. */
 void
 dissector_reset(const char *name, guint32 pattern)
 {
@@ -663,8 +722,7 @@ dissector_reset(const char *name, guint32 pattern)
 	/*
 	 * Find the entry.
 	 */
-	dtbl_entry = g_hash_table_lookup(sub_dissectors->hash_table,
-	    GUINT_TO_POINTER(pattern));
+	dtbl_entry = find_uint_dtbl_entry(sub_dissectors, pattern);
 
 	if (dtbl_entry == NULL)
 		return;
@@ -681,9 +739,9 @@ dissector_reset(const char *name, guint32 pattern)
 	}
 }
 
-/* Look for a given port in a given dissector table and, if found, call
-   the dissector with the arguments supplied, and return TRUE, otherwise
-   return FALSE. */
+/* Look for a given value in a given uint dissector table and, if found,
+   call the dissector with the arguments supplied, and return TRUE,
+   otherwise return FALSE. */
 gboolean
 dissector_try_port(dissector_table_t sub_dissectors, guint32 port,
     tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
@@ -693,8 +751,7 @@ dissector_try_port(dissector_table_t sub_dissectors, guint32 port,
 	guint32 saved_match_port;
 	int ret;
 
-	dtbl_entry = g_hash_table_lookup(sub_dissectors->hash_table,
-	    GUINT_TO_POINTER(port));
+	dtbl_entry = find_uint_dtbl_entry(sub_dissectors, port);
 	if (dtbl_entry != NULL) {
 		/*
 		 * Is there currently a dissector handle for this entry?
@@ -737,15 +794,258 @@ dissector_try_port(dissector_table_t sub_dissectors, guint32 port,
 	return FALSE;
 }
 
-/* Look for a given port in a given dissector table and, if found, return
-   the dissector handle for that port. */
+/* Look for a given value in a given uint dissector table and, if found,
+   return the dissector handle for that value. */
 dissector_handle_t
 dissector_get_port_handle(dissector_table_t sub_dissectors, guint32 port)
 {
 	dtbl_entry_t *dtbl_entry;
 
-	dtbl_entry = g_hash_table_lookup(sub_dissectors->hash_table,
-	    GUINT_TO_POINTER(port));
+	dtbl_entry = find_uint_dtbl_entry(sub_dissectors, port);
+	if (dtbl_entry != NULL)
+		return dtbl_entry->current;
+	else
+		return NULL;
+}
+
+/* Find an entry in a string dissector table. */ 
+static dtbl_entry_t *
+find_string_dtbl_entry(dissector_table_t sub_dissectors, const gchar *pattern)
+{
+	switch (sub_dissectors->type) {
+
+	case FT_STRING:
+	case FT_STRINGZ:
+		/*
+		 * You can do a string lookup in these tables.
+		 */
+		break;
+
+	default:
+		/*
+		 * But you can't do a string lookup in any other types
+		 * of tables.
+		 */
+		g_assert_not_reached();
+	}
+
+	/*
+	 * Find the entry.
+	 */
+	return g_hash_table_lookup(sub_dissectors->hash_table, pattern);
+}
+
+/* Add an entry to a string dissector table. */
+void
+dissector_add_string(const char *name, gchar *pattern,
+    dissector_handle_t handle)
+{
+	dissector_table_t sub_dissectors = find_dissector_table( name);
+	dtbl_entry_t *dtbl_entry;
+
+/* sanity check */
+	g_assert( sub_dissectors);
+
+	switch (sub_dissectors->type) {
+
+	case FT_STRING:
+	case FT_STRINGZ:
+		/*
+		 * You can do a string lookup in these tables.
+		 */
+		break;
+
+	default:
+		/*
+		 * But you can't do a string lookup in any other types
+		 * of tables.
+		 */
+		g_assert_not_reached();
+	}
+
+	dtbl_entry = g_malloc(sizeof (dtbl_entry_t));
+	dtbl_entry->current = handle;
+	dtbl_entry->initial = dtbl_entry->current;
+
+/* do the table insertion */
+    	g_hash_table_insert( sub_dissectors->hash_table, pattern,
+    	    (gpointer)dtbl_entry);
+
+	/*
+	 * Now add it to the list of handles that could be used with this
+	 * table, because it *is* being used with this table.
+	 */
+	dissector_add_handle(name, handle);
+}
+
+/* Delete the entry for a dissector in a string dissector table
+   with a particular pattern. */
+
+/* NOTE: this doesn't use the dissector call variable. It is included to */
+/*	be consistant with the dissector_add_string and more importantly to */
+/*      be used if the technique of adding a temporary dissector is */
+/*      implemented.  */
+/*	If temporary dissectors are deleted, then the original dissector must */
+/*	be available. */
+void
+dissector_delete_string(const char *name, const gchar *pattern,
+	dissector_handle_t handle _U_)
+{
+	dissector_table_t sub_dissectors = find_dissector_table( name);
+	dtbl_entry_t *dtbl_entry;
+
+/* sanity check */
+	g_assert( sub_dissectors);
+
+	/*
+	 * Find the entry.
+	 */
+	dtbl_entry = find_string_dtbl_entry(sub_dissectors, pattern);
+
+	if (dtbl_entry != NULL) {
+		/*
+		 * Found - remove it.
+		 */
+		g_hash_table_remove(sub_dissectors->hash_table, pattern);
+
+		/*
+		 * Now free up the entry.
+		 */
+		g_free(dtbl_entry);
+	}
+}
+
+/* Change the entry for a dissector in a string dissector table
+   with a particular pattern to use a new dissector handle. */
+void
+dissector_change_string(const char *name, gchar *pattern,
+    dissector_handle_t handle)
+{
+	dissector_table_t sub_dissectors = find_dissector_table( name);
+	dtbl_entry_t *dtbl_entry;
+
+/* sanity check */
+	g_assert( sub_dissectors);
+
+	/*
+	 * See if the entry already exists. If so, reuse it.
+	 */
+	dtbl_entry = find_string_dtbl_entry(sub_dissectors, pattern);
+	if (dtbl_entry != NULL) {
+	  dtbl_entry->current = handle;
+	  return;
+	}
+
+	/*
+	 * Don't create an entry if there is no dissector handle - I.E. the
+	 * user said not to decode something that wasn't being decoded
+	 * in the first place.
+	 */
+	if (handle == NULL)
+	  return;
+
+	dtbl_entry = g_malloc(sizeof (dtbl_entry_t));
+	dtbl_entry->initial = NULL;
+	dtbl_entry->current = handle;
+
+/* do the table insertion */
+    	g_hash_table_insert( sub_dissectors->hash_table, pattern,
+    	    (gpointer)dtbl_entry);
+}
+
+/* Reset an entry in a string sub-dissector table to its initial value. */
+void
+dissector_reset_string(const char *name, const gchar *pattern)
+{
+	dissector_table_t sub_dissectors = find_dissector_table( name);
+	dtbl_entry_t *dtbl_entry;
+
+/* sanity check */
+	g_assert( sub_dissectors);
+
+	/*
+	 * Find the entry.
+	 */
+	dtbl_entry = find_string_dtbl_entry(sub_dissectors, pattern);
+
+	if (dtbl_entry == NULL)
+		return;
+
+	/*
+	 * Found - is there an initial value?
+	 */
+	if (dtbl_entry->initial != NULL) {
+		dtbl_entry->current = dtbl_entry->initial;
+	} else {
+		g_hash_table_remove(sub_dissectors->hash_table, pattern);
+		g_free(dtbl_entry);
+	}
+}
+
+/* Look for a given string in a given dissector table and, if found, call
+   the dissector with the arguments supplied, and return TRUE, otherwise
+   return FALSE. */
+gboolean
+dissector_try_string(dissector_table_t sub_dissectors, const gchar *string,
+    tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
+{
+	dtbl_entry_t *dtbl_entry;
+	struct dissector_handle *handle;
+	int ret;
+	const gchar *saved_match_string;
+
+	dtbl_entry = find_string_dtbl_entry(sub_dissectors, string);
+	if (dtbl_entry != NULL) {
+		/*
+		 * Is there currently a dissector handle for this entry?
+		 */
+		handle = dtbl_entry->current;
+		if (handle == NULL) {
+			/*
+			 * No - pretend this dissector didn't exist,
+			 * so that other dissectors might have a chance
+			 * to dissect this packet.
+			 */
+			return FALSE;
+		}
+
+		/*
+		 * Save the current value of "pinfo->match_string",
+		 * set it to the string that matched, call the
+		 * dissector, and restore "pinfo->match_string".
+		 */
+		saved_match_string = pinfo->match_string;
+		pinfo->match_string = string;
+		ret = call_dissector_work(handle, tvb, pinfo, tree);
+		pinfo->match_string = saved_match_string;
+
+		/*
+		 * If a new-style dissector returned 0, it means that
+		 * it didn't think this tvbuff represented a packet for
+		 * its protocol, and didn't dissect anything.
+		 *
+		 * Old-style dissectors can't reject the packet.
+		 *
+		 * 0 is also returned if the protocol wasn't enabled.
+		 *
+		 * If the packet was rejected, we return FALSE, so that
+		 * other dissectors might have a chance to dissect this
+		 * packet, otherwise we return TRUE.
+		 */
+		return ret != 0;
+	}
+	return FALSE;
+}
+
+/* Look for a given value in a given string dissector table and, if found,
+   return the dissector handle for that value. */
+dissector_handle_t
+dissector_get_string_handle(dissector_table_t sub_dissectors,
+    const gchar *string)
+{
+	dtbl_entry_t *dtbl_entry;
+
+	dtbl_entry = find_string_dtbl_entry(sub_dissectors, string);
 	if (dtbl_entry != NULL)
 		return dtbl_entry->current;
 	else
@@ -800,6 +1100,7 @@ typedef struct dissector_foreach_info {
   DATFunc      caller_func;
   GHFunc       next_func;
   gchar       *table_name;
+  ftenum_t     selector_type;
 } dissector_foreach_info_t;
 
 /*
@@ -816,7 +1117,7 @@ dissector_table_foreach_func (gpointer key, gpointer value, gpointer user_data)
 
 	dtbl_entry = value;
 	if (dtbl_entry->current == NULL ||
-	    dtbl_entry->current->proto_index == -1) {
+	    dtbl_entry->current->protocol == NULL) {
 		/*
 		 * Either there is no dissector for this entry, or
 		 * the dissector doesn't have a protocol associated
@@ -828,7 +1129,8 @@ dissector_table_foreach_func (gpointer key, gpointer value, gpointer user_data)
 	}
 
 	info = user_data;
-	info->caller_func(info->table_name, key, value, info->caller_data);
+	info->caller_func(info->table_name, info->selector_type, key, value,
+	    info->caller_data);
 }
 
 /*
@@ -846,6 +1148,7 @@ dissector_all_tables_foreach_func (gpointer key, gpointer value, gpointer user_d
 	sub_dissectors = value;
 	info = user_data;
 	info->table_name = (gchar*) key;
+	info->selector_type = get_dissector_table_selector_type(info->table_name);
 	g_hash_table_foreach(sub_dissectors->hash_table, info->next_func, info);
 }
 
@@ -878,6 +1181,7 @@ dissector_table_foreach (char *name,
 	dissector_table_t sub_dissectors = find_dissector_table( name);
 
 	info.table_name = name;
+	info.selector_type = sub_dissectors->type;
 	info.caller_func = func;
 	info.caller_data = user_data;
 	g_hash_table_foreach(sub_dissectors->hash_table, dissector_table_foreach_func, &info);
@@ -888,7 +1192,9 @@ dissector_table_foreach (char *name,
  * function on each entry.
  */
 void
-dissector_table_foreach_handle(char *name, DATFunc_handle func, gpointer user_data)
+dissector_table_foreach_handle(char *name,
+			       DATFunc_handle func,
+			       gpointer user_data)
 {
 	dissector_table_t sub_dissectors = find_dissector_table( name);
 	GSList *tmp;
@@ -919,7 +1225,8 @@ dissector_table_foreach_changed_func (gpointer key, gpointer value, gpointer use
 	}
 
 	info = user_data;
-	info->caller_func(info->table_name, key, value, info->caller_data);
+	info->caller_func(info->table_name, info->selector_type, key, value,
+	    info->caller_data);
 }
 
 /*
@@ -951,12 +1258,47 @@ dissector_table_foreach_changed (char *name,
 	dissector_table_t sub_dissectors = find_dissector_table( name);
 
 	info.table_name = name;
+	info.selector_type = sub_dissectors->type;
 	info.caller_func = func;
 	info.caller_data = user_data;
 	g_hash_table_foreach(sub_dissectors->hash_table,
 	    dissector_table_foreach_changed_func, &info);
 }
 
+typedef struct dissector_foreach_table_info {
+  gpointer      caller_data;
+  DATFunc_table caller_func;
+} dissector_foreach_table_info_t;
+
+/*
+ * Called for each entry in the table of all dissector tables.
+ */
+static void
+dissector_all_tables_foreach_table_func (gpointer key, gpointer value, gpointer user_data)
+{
+	dissector_table_t table;
+	dissector_foreach_table_info_t *info;
+
+	table = value;
+	info = user_data;
+	(*info->caller_func)((gchar*)key, table->ui_name, info->caller_data);
+}
+
+/*
+ * Walk all dissector tables calling a user supplied function on each
+ * table.
+ */
+void
+dissector_all_tables_foreach_table (DATFunc_table func,
+				    gpointer user_data)
+{
+	dissector_foreach_table_info_t info;
+
+	info.caller_data = user_data;
+	info.caller_func = func;
+	g_hash_table_foreach(dissector_tables, dissector_all_tables_foreach_table_func, &info);
+}
+
 dissector_table_t
 register_dissector_table(const char *name, char *ui_name, ftenum_t type,
     int base)
@@ -975,8 +1317,29 @@ register_dissector_table(const char *name, char *ui_name, ftenum_t type,
 	/* Create and register the dissector table for this name; returns */
 	/* a pointer to the dissector table. */
 	sub_dissectors = g_malloc(sizeof (struct dissector_table));
-	sub_dissectors->hash_table = g_hash_table_new( g_direct_hash,
-	    g_direct_equal );
+	switch (type) {
+
+	case FT_UINT8:
+	case FT_UINT16:
+	case FT_UINT24:
+	case FT_UINT32:
+		/*
+		 * XXX - there's no "g_uint_hash()" or "g_uint_equal()",
+		 * so we use "g_direct_hash()" and "g_direct_equal()".
+		 */
+		sub_dissectors->hash_table = g_hash_table_new( g_direct_hash,
+		    g_direct_equal );
+		break;
+
+	case FT_STRING:
+	case FT_STRINGZ:
+		sub_dissectors->hash_table = g_hash_table_new( g_str_hash,
+		    g_str_equal );
+		break;
+
+	default:
+		g_assert_not_reached();
+	}
 	sub_dissectors->dissector_handles = NULL;
 	sub_dissectors->ui_name = ui_name;
 	sub_dissectors->type = type;
@@ -994,7 +1357,7 @@ get_dissector_table_ui_name(const char *name)
 }
 
 ftenum_t
-get_dissector_table_type(const char *name)
+get_dissector_table_selector_type(const char *name)
 {
 	dissector_table_t sub_dissectors = find_dissector_table( name);
 
@@ -1013,7 +1376,7 @@ static GHashTable *heur_dissector_lists = NULL;
 
 typedef struct {
 	heur_dissector_t dissector;
-	int	proto_index;
+	protocol_t *protocol;
 } heur_dtbl_entry_t;
 
 /* Finds a heuristic dissector table by field name. */
@@ -1035,7 +1398,7 @@ heur_dissector_add(const char *name, heur_dissector_t dissector, int proto)
 
 	dtbl_entry = g_malloc(sizeof (heur_dtbl_entry_t));
 	dtbl_entry->dissector = dissector;
-	dtbl_entry->proto_index = proto;
+	dtbl_entry->protocol = find_protocol_by_id(proto);
 
 	/* do the table insertion */
 	*sub_dissectors = g_slist_append(*sub_dissectors, (gpointer)dtbl_entry);
@@ -1055,26 +1418,32 @@ dissector_try_heuristic(heur_dissector_list_t sub_dissectors,
 	   then everytime a subdissector is called it is decremented by one.
 	   thus only the subdissector immediately ontop of whoever offers this
 	   service can use it.
+	   We save the current value of "can_desegment" for the
+	   benefit of TCP proxying dissectors such as SOCKS, so they
+	   can restore it and allow the dissectors they call to use
+	   the desegmentation service.
 	*/
 	saved_can_desegment=pinfo->can_desegment;
+	pinfo->saved_can_desegment = saved_can_desegment;
 	pinfo->can_desegment = saved_can_desegment-(saved_can_desegment>0);
 
 	status = FALSE;
 	saved_proto = pinfo->current_proto;
 	for (entry = sub_dissectors; entry != NULL; entry = g_slist_next(entry)) {
+		/* XXX - why set this now and above? */
 		pinfo->can_desegment = saved_can_desegment-(saved_can_desegment>0);
 		dtbl_entry = (heur_dtbl_entry_t *)entry->data;
-		if (dtbl_entry->proto_index != -1 &&
-		    !proto_is_protocol_enabled(dtbl_entry->proto_index)) {
+		if (dtbl_entry->protocol != NULL &&
+		    !proto_is_protocol_enabled(dtbl_entry->protocol)) {
 			/*
 			 * No - don't try this dissector.
 			 */
 			continue;
 		}
 
-		if (dtbl_entry->proto_index != -1) {
+		if (dtbl_entry->protocol != NULL) {
 			pinfo->current_proto =
-			    proto_get_protocol_short_name(dtbl_entry->proto_index);
+			    proto_get_protocol_short_name(dtbl_entry->protocol);
 		}
 		if ((*dtbl_entry->dissector)(tvb, pinfo, tree)) {
 			status = TRUE;
@@ -1115,18 +1484,38 @@ register_heur_dissector_list(const char *name, heur_dissector_list_t *sub_dissec
  */
 static GHashTable *registered_dissectors = NULL;
 
-/* Get the short name of the protocol for a dissector handle. */
+/* Get the short name of the protocol for a dissector handle, if it has
+   a protocol. */
 char *
 dissector_handle_get_short_name(dissector_handle_t handle)
 {
-	return proto_get_protocol_short_name(handle->proto_index);
+	if (handle->protocol == NULL) {
+		/*
+		 * No protocol (see, for example, the handle for
+		 * dissecting the set of protocols where the first
+		 * octet of the payload is an OSI network layer protocol
+		 * ID).
+		 */
+		return NULL;
+	}
+	return proto_get_protocol_short_name(handle->protocol);
 }
 
-/* Get the index of the protocol for a dissector handle. */
+/* Get the index of the protocol for a dissector handle, if it has
+   a protocol. */
 int
 dissector_handle_get_protocol_index(dissector_handle_t handle)
 {
-  return handle->proto_index;
+	if (handle->protocol == NULL) {
+		/*
+		 * No protocol (see, for example, the handle for
+		 * dissecting the set of protocols where the first
+		 * octet of the payload is an OSI network layer protocol
+		 * ID).
+		 */
+		return -1;
+	}
+	return proto_get_id(handle->protocol);
 }
 
 /* Find a registered dissector by name. */
@@ -1147,7 +1536,7 @@ create_dissector_handle(dissector_t dissector, int proto)
 	handle->name = NULL;
 	handle->is_new = FALSE;
 	handle->dissector.old = dissector;
-	handle->proto_index = proto;
+	handle->protocol = find_protocol_by_id(proto);
 
 	return handle;
 }
@@ -1161,7 +1550,7 @@ new_create_dissector_handle(new_dissector_t dissector, int proto)
 	handle->name = NULL;
 	handle->is_new = TRUE;
 	handle->dissector.new = dissector;
-	handle->proto_index = proto;
+	handle->protocol = find_protocol_by_id(proto);
 
 	return handle;
 }
@@ -1185,7 +1574,7 @@ register_dissector(const char *name, dissector_t dissector, int proto)
 	handle->name = name;
 	handle->is_new = FALSE;
 	handle->dissector.old = dissector;
-	handle->proto_index = proto;
+	handle->protocol = find_protocol_by_id(proto);
 
 	g_hash_table_insert(registered_dissectors, (gpointer)name,
 	    (gpointer) handle);
@@ -1209,7 +1598,7 @@ new_register_dissector(const char *name, new_dissector_t dissector, int proto)
 	handle->name = name;
 	handle->is_new = TRUE;
 	handle->dissector.new = dissector;
-	handle->proto_index = proto;
+	handle->protocol = find_protocol_by_id(proto);
 
 	g_hash_table_insert(registered_dissectors, (gpointer)name,
 	    (gpointer) handle);
@@ -1229,7 +1618,7 @@ call_dissector(dissector_handle_t handle, tvbuff_t *tvb,
 		 * it.  Just dissect this packet as data.
 		 */
 	        g_assert(data_handle != NULL);
-		g_assert(data_handle->proto_index != -1);
+		g_assert(data_handle->protocol != NULL);
 		call_dissector(data_handle, tvb, pinfo, tree);
 		return tvb_length(tvb);
 	}
