@@ -1,7 +1,7 @@
 /* proto.h
  * Definitions for protocol display
  *
- * $Id: proto.h,v 1.38 2003/01/31 03:17:52 guy Exp $
+ * $Id: proto.h,v 1.57 2004/03/25 23:55:21 guy Exp $
  *
  * Ethereal - Network traffic analyzer
  * By Gerald Combs <gerald@ethereal.com>
@@ -38,6 +38,9 @@
 #include "tvbuff.h"
 #include "ftypes/ftypes.h"
 
+/* The header-field index for the special text pseudo-field */
+extern int hf_text_only;
+
 struct _value_string;
 
 #define ITEM_LABEL_LENGTH	240
@@ -49,12 +52,16 @@ struct _value_string;
 /* ... and similarly, */
 #define TFS(x)	(const struct true_false_string*)(x)
 
+struct _protocol;
+
+typedef struct _protocol protocol_t;
+ 
 /* check protocol activation */
 #define CHECK_DISPLAY_AS_X(x_handle,index, tvb, pinfo, tree) {	\
-	if (!proto_is_protocol_enabled(index)) {		\
-		call_dissector(x_handle,tvb, pinfo, tree);	\
-		return;						\
-	}							\
+	if (!proto_is_protocol_enabled(find_protocol_by_id(index))) {	\
+		call_dissector(x_handle,tvb, pinfo, tree);		\
+		return;							\
+	}								\
   }
 
 enum {
@@ -99,16 +106,20 @@ typedef struct hf_register_info {
 } hf_register_info;
 
 
+typedef struct _item_label_t {
+	char representation[ITEM_LABEL_LENGTH];
+} item_label_t;
+
 /* Contains the field information for the proto_item. */
 typedef struct field_info {
 	header_field_info		*hfinfo;
 	gint				start;
 	gint				length;
 	gint				tree_type; /* ETT_* */
-	char				*representation; /* for GUI tree */
+	item_label_t			*rep; /* string for GUI tree */
 	int				visible;
-	fvalue_t			*value;
 	tvbuff_t			*ds_tvb;  /* data source tvbuff */
+	fvalue_t			value;
 } field_info;
 
 /* One of these exists for the entire protocol tree. Each proto_node
@@ -118,24 +129,29 @@ typedef struct {
     gboolean    visible;
 } tree_data_t;
 
-/* Each GNode (proto_tree, proto_item) points to one of
- * these. */
-typedef struct {
-    field_info  *finfo;
-    tree_data_t *tree_data;
+/* Each proto_tree, proto_item is one of these. */
+typedef struct _proto_node {
+	struct _proto_node *first_child;
+	struct _proto_node *last_child;
+	struct _proto_node *next;
+	struct _proto_node *parent;
+	field_info  *finfo;
+	tree_data_t *tree_data;
 } proto_node;
 
-typedef GNode proto_tree;
-typedef GNode proto_item;
+typedef proto_node proto_tree;
+typedef proto_node proto_item;
 
-/* Retrieve the proto_node from a GNode. */
-#define GNODE_PNODE(t)  ((proto_node*)((GNode*)(t))->data)
+typedef void (*proto_tree_foreach_func)(proto_node *, gpointer);
+
+extern void proto_tree_children_foreach(proto_tree *tree,
+    proto_tree_foreach_func func, gpointer data);
 
 /* Retrieve the field_info from a proto_item */
-#define PITEM_FINFO(t)  (GNODE_PNODE(t)->finfo)
+#define PITEM_FINFO(t)  ((t)->finfo)
 
 /* Retrieve the tree_data_t from a proto_tree */
-#define PTREE_DATA(t)   (GNODE_PNODE(t)->tree_data)
+#define PTREE_DATA(t)   ((t)->tree_data)
 
 /* Sets up memory used by proto routines. Called at program startup */
 extern void proto_init(const char *plugin_dir,
@@ -163,6 +179,15 @@ extern void proto_item_append_text(proto_item *ti, const char *format, ...);
 /* Set length of proto_item after having already been created. */
 extern void proto_item_set_len(proto_item *ti, gint length);
 
+/*
+ * Sets the length of the item based on its start and on the specified
+ * offset, which is the offset past the end of the item; as the start
+ * in the item is relative to the beginning of the data source tvbuff,
+ * we need to pass in a tvbuff - the end offset is relative to the beginning
+ * of that tvbuff.
+ */
+extern void proto_item_set_end(proto_item *pi, tvbuff_t *tvb, gint end);
+
 /* Get length of proto_item. Useful after using proto_tree_add_item()
  * to add a variable-length field (e.g., FT_NSTRING_UINT8) */
 extern int proto_item_get_len(proto_item *ti);
@@ -180,6 +205,18 @@ extern void proto_tree_free(proto_tree *tree);
 /* Create a subtree under an existing item; returns tree pointer */
 extern proto_tree* proto_item_add_subtree(proto_item *ti, gint idx);
 
+/* Get a subtree under an item; returns tree pointer */
+extern proto_tree* proto_item_get_subtree(proto_item *ti);
+
+/* Get a parent item; returns item pointer */
+extern proto_item* proto_item_get_parent(proto_item *ti);
+
+/* Get Nth generation parent item; returns item pointer */
+extern proto_item* proto_item_get_parent_nth(proto_item *ti, int gen);
+
+/* Get a parent item of subtree; returns item pointer */
+extern proto_item* proto_tree_get_parent(proto_tree *tree);
+
 extern int
 proto_register_field(char *name, char *abbrev, enum ftenum type, int parent,
 	struct _value_string* vals);
@@ -367,6 +404,9 @@ proto_tree_add_string_format(proto_tree *tree, int hfindex, tvbuff_t *tvb, gint
 	gint length, const char* value, const char *format, ...);
 #endif
 
+extern void
+proto_item_append_string(proto_item *pi, const char *str);
+
 /* Add a FT_BOOLEAN to a proto_tree */
 extern proto_item *
 proto_tree_add_boolean(proto_tree *tree, int hfindex, tvbuff_t *tvb, gint start,
@@ -511,7 +551,7 @@ extern char* proto_registrar_get_name(int n);
 extern char* proto_registrar_get_abbrev(int n);
 
 /* get the header field information based upon a field or protocol id */
-extern header_field_info* proto_registrar_get_nth(int hfindex);
+extern header_field_info* proto_registrar_get_nth(guint hfindex);
 
 /* get the header field information based upon a field name */
 extern header_field_info* proto_registrar_get_byname(char *field_name);
@@ -526,25 +566,34 @@ extern int proto_registrar_get_parent(int n);
 /* Is item #n a protocol? */
 extern gboolean proto_registrar_is_protocol(int n);
 
-/* Is item #n decoding enabled ? */
-extern gboolean proto_is_protocol_enabled(int proto_id);
+/* Is protocol's decoding enabled ? */
+extern gboolean proto_is_protocol_enabled(protocol_t *protocol);
 
 /* Can item #n decoding be disabled? */
-extern gboolean proto_can_disable_protocol(int proto_id);
+extern gboolean proto_can_toggle_protocol(int proto_id);
 
-/* Routines to use to iterate over the protocols; they return the item
- * number of the protocol in question, and keep state in "*cookie". */
+/* Routines to use to iterate over the protocols and their fields;
+ * they return the item number of the protocol in question or the
+ * appropriate hfinfo pointer, and keep state in "*cookie". */
 extern int proto_get_first_protocol(void **cookie);
 extern int proto_get_next_protocol(void **cookie);
+extern header_field_info *proto_get_first_protocol_field(int proto_id, void **cookle);
+extern header_field_info *proto_get_next_protocol_field(void **cookle);
 
-/* Given a protocol's filter_name, return it's proto_id */
+/* Given a protocol's "protocol_t", return its proto_id */
+extern int proto_get_id(protocol_t *protocol);
+
+/* Given a protocol's filter_name, return its proto_id */
 extern int proto_get_id_by_filter_name(gchar* filter_name);
 
+/* Given a protocol's item number, find the "protocol_t" structure for it */
+extern protocol_t *find_protocol_by_id(int proto_id);
+
 /* Given a protocol's item number, return its name. */
 extern char *proto_get_protocol_name(int n);
 
-/* Given a protocol's item number, return its short name. */
-extern char *proto_get_protocol_short_name(int proto_id);
+/* Given a protocol's "protocol_t", return its short name. */
+extern char *proto_get_protocol_short_name(protocol_t *protocol);
 
 /* Given a protocol's item number, return its filter name. */
 extern char *proto_get_protocol_filter_name(int proto_id);
@@ -552,8 +601,8 @@ extern char *proto_get_protocol_filter_name(int proto_id);
 /* Enable / Disable protocol */
 extern void proto_set_decoding(int proto_id, gboolean enabled);
 
-/* Disable disabling of protocol */
-extern void proto_set_cant_disable(int proto_id);
+/* Disable disabling/enabling of protocol */
+extern void proto_set_cant_toggle(int proto_id);
 
 /* Get length of registered field according to field type.
  * 0 means undeterminable at registration time.
@@ -566,9 +615,14 @@ extern gint proto_registrar_get_length(int n);
 extern gboolean proto_check_for_protocol_or_field(proto_tree* tree, int id);
 
 /* Return GPtrArray* of field_info pointers for all hfindex that appear in
- * tree. */
+ * tree. Only works with primed trees, and is fast. */
 extern GPtrArray* proto_get_finfo_ptr_array(proto_tree *tree, int hfindex);
 
+/* Return GPtrArray* of field_info pointers for all hfindex that appear in
+ * tree. Works with any tree, primed or unprimed, and is slower than
+ * proto_get_finfo_ptr_array because it has to search through the tree. */
+extern GPtrArray* proto_find_finfo(proto_tree *tree, int hfindex);
+
 /* Dumps a glossary of the protocol registrations to STDOUT */
 extern void proto_registrar_dump_protocols(void);
 
@@ -591,15 +645,17 @@ extern int           num_tree_types;
 extern int
 hfinfo_bitwidth(header_field_info *hfinfo);
 
+#include "epan.h"
+
 /*
  * Returns TRUE if we can do a "match selected" on the field, FALSE
  * otherwise.
  */
 extern gboolean
-proto_can_match_selected(field_info *finfo);
+proto_can_match_selected(field_info *finfo, epan_dissect_t *edt);
 
 extern char*
-proto_alloc_dfilter_string(field_info *finfo, guint8 *pd);
+proto_construct_dfilter_string(field_info *finfo, epan_dissect_t *edt);
 
 extern field_info*
 proto_find_field_from_offset(proto_tree *tree, guint offset, tvbuff_t *tvb);
