@@ -9,7 +9,7 @@
  * 		the data of a backing tvbuff, or can be a composite of
  * 		other tvbuffs.
  *
- * $Id: tvbuff.c,v 1.41 2003/02/24 01:22:26 guy Exp $
+ * $Id: tvbuff.c,v 1.61 2004/03/23 18:06:29 guy Exp $
  *
  * Copyright (c) 2000 by Gilbert Ramirez <gram@alumni.rice.edu>
  *
@@ -45,68 +45,11 @@
 #include "tvbuff.h"
 #include "strutil.h"
 
-typedef struct {
-	/* The backing tvbuff_t */
-	tvbuff_t	*tvb;
+static const guint8*
+ensure_contiguous_no_exception(tvbuff_t *tvb, gint offset, gint length,
+		int *exception);
 
-	/* The offset/length of 'tvb' to which I'm privy */
-	guint		offset;
-	guint		length;
-
-} tvb_backing_t;
-
-typedef struct {
-	GSList		*tvbs;
-
-	/* Used for quick testing to see if this
-	 * is the tvbuff that a COMPOSITE is
-	 * interested in. */
-	guint		*start_offsets;
-	guint		*end_offsets;
-
-} tvb_comp_t;
-
-struct tvbuff {
-	/* Record-keeping */
-	tvbuff_type		type;
-	gboolean		initialized;
-	guint			usage_count;
-	tvbuff_t*		ds_tvb;	  /* data source top-level tvbuff */
-
-	/* The tvbuffs in which this tvbuff is a member
-	 * (that is, a backing tvbuff for a TVBUFF_SUBSET
-	 * or a member for a TVB_COMPOSITE) */
-	GSList			*used_in;
-
-	/* TVBUFF_SUBSET and TVBUFF_COMPOSITE keep track
-	 * of the other tvbuff's they use */
-	union {
-		tvb_backing_t	subset;
-		tvb_comp_t	composite;
-	} tvbuffs;
-
-	/* We're either a TVBUFF_REAL_DATA or a
-	 * TVBUFF_SUBSET that has a backing buffer that
-	 * has real_data != NULL, or a TVBUFF_COMPOSITE
-	 * which has flattened its data due to a call
-	 * to tvb_get_ptr().
-	 */
-	guint8			*real_data;
-
-	/* Length of virtual buffer (and/or real_data). */
-	guint			length;
-
-	/* Reported length. */
-	guint			reported_length;
-
-	/* Offset from beginning of first TVBUFF_REAL. */
-	gint			raw_offset;
-
-	/* Func to call when actually freed */
-	tvbuff_free_cb_t	free_cb;
-};
-
-static guint8*
+static const guint8*
 ensure_contiguous(tvbuff_t *tvb, gint offset, gint length);
 
 /* We dole out tvbuff's from this memchunk. */
@@ -183,15 +126,6 @@ tvb_new(tvbuff_type type)
 	return tvb;
 }
 
-/* We accept a void* instead of a field_info* to satisfy CLEANUP_POP */
-static void
-tvb_free_void(void *tvb)
-{
-	tvb_free((tvbuff_t*)tvb);
-}
-
-
-
 void
 tvb_free(tvbuff_t* tvb)
 {
@@ -205,7 +139,10 @@ tvb_free(tvbuff_t* tvb)
 		switch (tvb->type) {
 		case TVBUFF_REAL_DATA:
 			if (tvb->free_cb) {
-				tvb->free_cb(tvb->real_data);
+				/*
+				 * XXX - do this with a union?
+				 */
+				tvb->free_cb((gpointer)tvb->real_data);
 			}
 			break;
 
@@ -230,8 +167,12 @@ tvb_free(tvbuff_t* tvb)
 				g_free(composite->start_offsets);
 			if (composite->end_offsets)
 				g_free(composite->end_offsets);
-			if (tvb->real_data)
-				g_free(tvb->real_data);
+			if (tvb->real_data) {
+				/*
+				 * XXX - do this with a union?
+				 */
+				g_free((gpointer)tvb->real_data);
+			}
 
 			break;
 		}
@@ -316,7 +257,7 @@ tvb_set_real_data(tvbuff_t* tvb, const guint8* data, guint length, gint reported
 		THROW(ReportedBoundsError);
 	}
 
-	tvb->real_data		= (gpointer) data;
+	tvb->real_data		= data;
 	tvb->length		= length;
 	tvb->reported_length	= reported_length;
 	tvb->initialized	= TRUE;
@@ -325,11 +266,18 @@ tvb_set_real_data(tvbuff_t* tvb, const guint8* data, guint length, gint reported
 tvbuff_t*
 tvb_new_real_data(const guint8* data, guint length, gint reported_length)
 {
+	static tvbuff_t	*last_tvb=NULL;
 	tvbuff_t	*tvb;
 
 	tvb = tvb_new(TVBUFF_REAL_DATA);
 
-	CLEANUP_PUSH(tvb_free_void, tvb);
+	if(last_tvb){
+		tvb_free(last_tvb);
+	}
+	/* remember this tvb in case we throw an exception and
+	 * lose the pointer to it.
+	 */
+	last_tvb=tvb;
 
 	tvb_set_real_data(tvb, data, length, reported_length);
 
@@ -339,7 +287,8 @@ tvb_new_real_data(const guint8* data, guint length, gint reported_length)
 	 */
 	tvb->ds_tvb = tvb;
 
-	CLEANUP_POP;
+	/* ok no exception so we dont need to remember it any longer */
+	last_tvb=NULL;
 
 	return tvb;
 }
@@ -404,6 +353,10 @@ compute_offset_length(tvbuff_t *tvb, gint offset, gint length,
 
 	/* Compute the length */
 	if (length < -1) {
+		if (exception) {
+			/* XXX - ReportedBoundsError? */
+			*exception = BoundsError;
+		}
 		return FALSE;
 	}
 	else if (length == -1) {
@@ -477,10 +430,6 @@ check_offset_length(tvbuff_t *tvb, gint offset, gint length,
 {
 	int exception = 0;
 
-	if (length < -1) {
-		THROW(BoundsError);
-	}
-
 	if (!check_offset_length_no_exception(tvb, offset, length, offset_ptr, length_ptr, &exception)) {
 		g_assert(exception > 0);
 		THROW(exception);
@@ -527,11 +476,18 @@ tvb_set_subset(tvbuff_t *tvb, tvbuff_t *backing,
 tvbuff_t*
 tvb_new_subset(tvbuff_t *backing, gint backing_offset, gint backing_length, gint reported_length)
 {
+	static tvbuff_t	*last_tvb=NULL;
 	tvbuff_t	*tvb;
 
 	tvb = tvb_new(TVBUFF_SUBSET);
 
-	CLEANUP_PUSH(tvb_free_void, tvb);
+	if(last_tvb){
+		tvb_free(last_tvb);
+	}
+	/* remember this tvb in case we throw an exception and
+	 * lose the pointer to it.
+	 */
+	last_tvb=tvb;
 
 	tvb_set_subset(tvb, backing, backing_offset, backing_length, reported_length);
 
@@ -541,7 +497,8 @@ tvb_new_subset(tvbuff_t *backing, gint backing_offset, gint backing_length, gint
 	 */
 	tvb->ds_tvb = backing->ds_tvb;
 
-	CLEANUP_POP;
+	/* ok no exception so we dont need to remember it any longer */
+	last_tvb=NULL;
 
 	return tvb;
 }
@@ -640,6 +597,17 @@ tvb_ensure_length_remaining(tvbuff_t *tvb, gint offset)
 	if (!compute_offset_length(tvb, offset, -1, &abs_offset, &abs_length, &exception)) {
 		THROW(exception);
 	}
+	if (abs_length == 0) {
+		/*
+		 * This routine ensures there's at least one byte available.
+		 * There aren't any bytes available, so throw the appropriate
+		 * exception.
+		 */
+		if (abs_offset >= tvb->reported_length)
+			THROW(ReportedBoundsError);
+		else
+			THROW(BoundsError);
+	}
 	return abs_length;
 }
 
@@ -754,7 +722,7 @@ tvb_set_reported_length(tvbuff_t* tvb, guint reported_length)
 }
 
 
-static guint8*
+static const guint8*
 first_real_data_ptr(tvbuff_t *tvb)
 {
 	tvbuff_t	*member;
@@ -774,7 +742,7 @@ first_real_data_ptr(tvbuff_t *tvb)
 	return NULL;
 }
 
-static int
+int
 offset_from_real_beginning(tvbuff_t *tvb, int counter)
 {
 	tvbuff_t	*member;
@@ -794,17 +762,9 @@ offset_from_real_beginning(tvbuff_t *tvb, int counter)
 	return 0;
 }
 
-gint
-tvb_raw_offset(tvbuff_t *tvb)
-{
-	if (tvb->raw_offset == -1) {
-		tvb->raw_offset = offset_from_real_beginning(tvb, 0);
-	}
-	return tvb->raw_offset;
-}
-
-static guint8*
-composite_ensure_contiguous(tvbuff_t *tvb, guint abs_offset, guint abs_length)
+static const guint8*
+composite_ensure_contiguous_no_exception(tvbuff_t *tvb, guint abs_offset,
+		guint abs_length)
 {
 	guint		i, num_members;
 	tvb_comp_t	*composite;
@@ -831,8 +791,11 @@ composite_ensure_contiguous(tvbuff_t *tvb, guint abs_offset, guint abs_length)
 	if (check_offset_length_no_exception(member_tvb, abs_offset - composite->start_offsets[i],
 				abs_length, &member_offset, &member_length, NULL)) {
 
+		/*
+		 * The range is, in fact, contiguous within member_tvb.
+		 */
 		g_assert(!tvb->real_data);
-		return ensure_contiguous(member_tvb, member_offset, member_length);
+		return ensure_contiguous_no_exception(member_tvb, member_offset, member_length, NULL);
 	}
 	else {
 		tvb->real_data = tvb_memdup(tvb, 0, -1);
@@ -843,13 +806,21 @@ composite_ensure_contiguous(tvbuff_t *tvb, guint abs_offset, guint abs_length)
 	return NULL;
 }
 
-static guint8*
-ensure_contiguous(tvbuff_t *tvb, gint offset, gint length)
+static const guint8*
+ensure_contiguous_no_exception(tvbuff_t *tvb, gint offset, gint length,
+		int *exception)
 {
 	guint	abs_offset, abs_length;
 
-	check_offset_length(tvb, offset, length, &abs_offset, &abs_length);
+	if (!check_offset_length_no_exception(tvb, offset, length,
+	    &abs_offset, &abs_length, exception)) {
+		return NULL;
+	}
 
+	/*
+	 * We know that all the data is present in the tvbuff, so
+	 * no exceptions should be thrown.
+	 */
 	if (tvb->real_data) {
 		return tvb->real_data + abs_offset;
 	}
@@ -858,11 +829,11 @@ ensure_contiguous(tvbuff_t *tvb, gint offset, gint length)
 			case TVBUFF_REAL_DATA:
 				g_assert_not_reached();
 			case TVBUFF_SUBSET:
-				return ensure_contiguous(tvb->tvbuffs.subset.tvb,
+				return ensure_contiguous_no_exception(tvb->tvbuffs.subset.tvb,
 						abs_offset - tvb->tvbuffs.subset.offset,
-						abs_length);
+						abs_length, NULL);
 			case TVBUFF_COMPOSITE:
-				return composite_ensure_contiguous(tvb, abs_offset, abs_length);
+				return composite_ensure_contiguous_no_exception(tvb, abs_offset, abs_length);
 		}
 	}
 
@@ -870,6 +841,20 @@ ensure_contiguous(tvbuff_t *tvb, gint offset, gint length)
 	return NULL;
 }
 
+static const guint8*
+ensure_contiguous(tvbuff_t *tvb, gint offset, gint length)
+{
+	int exception;
+	const guint8 *p;
+
+	p = ensure_contiguous_no_exception(tvb, offset, length, &exception);
+	if (p == NULL) {
+		g_assert(exception > 0);
+		THROW(exception);
+	}
+	return p;
+}
+
 static const guint8*
 guint8_find(const guint8* haystack, size_t haystacklen, guint8 needle)
 {
@@ -1003,10 +988,9 @@ tvb_memcpy(tvbuff_t *tvb, guint8* target, gint offset, gint length)
  * "tvb_ensure_bytes_exist()" and then allocates a buffer and copies
  * data to it.
  *
- * Does anything depend on this routine treating -1 as meaning "to the
- * end of the buffer"?  If so, perhaps we need two routines, one that
- * treats -1 as such, and one that checks for -1 and then calls this
- * routine.
+ * "composite_ensure_contiguous_no_exception()" depends on -1 not being
+ * an error; does anything else depend on this routine treating -1 as
+ * meaning "to the end of the buffer"?
  */
 guint8*
 tvb_memdup(tvbuff_t *tvb, gint offset, gint length)
@@ -1031,7 +1015,7 @@ tvb_get_ptr(tvbuff_t *tvb, gint offset, gint length)
 guint8
 tvb_get_guint8(tvbuff_t *tvb, gint offset)
 {
-	guint8* ptr;
+	const guint8* ptr;
 
 	ptr = ensure_contiguous(tvb, offset, sizeof(guint8));
 	return *ptr;
@@ -1040,7 +1024,7 @@ tvb_get_guint8(tvbuff_t *tvb, gint offset)
 guint16
 tvb_get_ntohs(tvbuff_t *tvb, gint offset)
 {
-	guint8* ptr;
+	const guint8* ptr;
 
 	ptr = ensure_contiguous(tvb, offset, sizeof(guint16));
 	return pntohs(ptr);
@@ -1049,7 +1033,7 @@ tvb_get_ntohs(tvbuff_t *tvb, gint offset)
 guint32
 tvb_get_ntoh24(tvbuff_t *tvb, gint offset)
 {
-	guint8* ptr;
+	const guint8* ptr;
 
 	ptr = ensure_contiguous(tvb, offset, 3);
 	return pntoh24(ptr);
@@ -1058,7 +1042,7 @@ tvb_get_ntoh24(tvbuff_t *tvb, gint offset)
 guint32
 tvb_get_ntohl(tvbuff_t *tvb, gint offset)
 {
-	guint8* ptr;
+	const guint8* ptr;
 
 	ptr = ensure_contiguous(tvb, offset, sizeof(guint32));
 	return pntohl(ptr);
@@ -1270,7 +1254,7 @@ tvb_get_ntohieee_double(tvbuff_t *tvb, int offset)
 guint16
 tvb_get_letohs(tvbuff_t *tvb, gint offset)
 {
-	guint8* ptr;
+	const guint8* ptr;
 
 	ptr = ensure_contiguous(tvb, offset, sizeof(guint16));
 	return pletohs(ptr);
@@ -1279,7 +1263,7 @@ tvb_get_letohs(tvbuff_t *tvb, gint offset)
 guint32
 tvb_get_letoh24(tvbuff_t *tvb, gint offset)
 {
-	guint8* ptr;
+	const guint8* ptr;
 
 	ptr = ensure_contiguous(tvb, offset, 3);
 	return pletoh24(ptr);
@@ -1288,7 +1272,7 @@ tvb_get_letoh24(tvbuff_t *tvb, gint offset)
 guint32
 tvb_get_letohl(tvbuff_t *tvb, gint offset)
 {
-	guint8* ptr;
+	const guint8* ptr;
 
 	ptr = ensure_contiguous(tvb, offset, sizeof(guint32));
 	return pletohl(ptr);
@@ -1541,14 +1525,14 @@ tvb_strnlen(tvbuff_t *tvb, gint offset, guint maxlength)
  * it returns 0 (meaning "equal") and -1 otherwise, otherwise return -1.
  */
 gint
-tvb_strneql(tvbuff_t *tvb, gint offset, const guint8 *str, gint size)
+tvb_strneql(tvbuff_t *tvb, gint offset, const gchar *str, gint size)
 {
-	guint8 *ptr;
+	const guint8 *ptr;
 
-	ptr = ensure_contiguous(tvb, offset, size);
+	ptr = ensure_contiguous_no_exception(tvb, offset, size, NULL);
 
 	if (ptr) {
-		int cmp = strncmp(ptr, str, size);
+		int cmp = strncmp((const char *)ptr, str, size);
 
 		/*
 		 * Return 0 if equal, -1 otherwise.
@@ -1568,14 +1552,14 @@ tvb_strneql(tvbuff_t *tvb, gint offset, const guint8 *str, gint size)
  * it returns 0 (meaning "equal") and -1 otherwise, otherwise return -1.
  */
 gint
-tvb_strncaseeql(tvbuff_t *tvb, gint offset, const guint8 *str, gint size)
+tvb_strncaseeql(tvbuff_t *tvb, gint offset, const gchar *str, gint size)
 {
-	guint8 *ptr;
+	const guint8 *ptr;
 
-	ptr = ensure_contiguous(tvb, offset, size);
+	ptr = ensure_contiguous_no_exception(tvb, offset, size, NULL);
 
 	if (ptr) {
-		int cmp = strncasecmp(ptr, str, size);
+		int cmp = strncasecmp((const char *)ptr, str, size);
 
 		/*
 		 * Return 0 if equal, -1 otherwise.
@@ -1597,9 +1581,9 @@ tvb_strncaseeql(tvbuff_t *tvb, gint offset, const guint8 *str, gint size)
 gint
 tvb_memeql(tvbuff_t *tvb, gint offset, const guint8 *str, gint size)
 {
-	guint8 *ptr;
+	const guint8 *ptr;
 
-	ptr = ensure_contiguous(tvb, offset, size);
+	ptr = ensure_contiguous_no_exception(tvb, offset, size, NULL);
 
 	if (ptr) {
 		int cmp = memcmp(ptr, str, size);
@@ -1629,12 +1613,8 @@ tvb_fake_unicode(tvbuff_t *tvb, int offset, int len, gboolean little_endian)
 	guint16 character;
 
 	/* Make sure we have enough data before allocating the buffer,
-	   so we don't blow up if the length is huge.
-	   We do so by attempting to fetch the last character; it'll
-	   throw an exception if it's past the end.
-
-           The byte order we use to fetch it is irrelevant here. */
-	tvb_get_letohs(tvb, offset + 2*(len - 1));
+	   so we don't blow up if the length is huge. */
+	tvb_ensure_bytes_exist(tvb, offset, 2*len);
 
 	/* We know we won't throw an exception, so we don't have to worry
 	   about leaking this buffer. */
@@ -1656,10 +1636,10 @@ tvb_fake_unicode(tvbuff_t *tvb, int offset, int len, gboolean little_endian)
  * Format the data in the tvb from offset for length ...
  */
 
-guint8 *
+gchar *
 tvb_format_text(tvbuff_t *tvb, gint offset, gint size)
 {
-  guint8 *ptr;
+  const guint8 *ptr;
   gint len = size;
 
   if ((ptr = ensure_contiguous(tvb, offset, size)) == NULL) {
@@ -1673,24 +1653,93 @@ tvb_format_text(tvbuff_t *tvb, gint offset, gint size)
 
 }
 
+/*
+ * Like "tvb_format_text()", but for null-padded strings; don't show
+ * the null padding characters as "\000".
+ */
+gchar *
+tvb_format_stringzpad(tvbuff_t *tvb, gint offset, gint size)
+{
+  const guint8 *ptr, *p;
+  gint len = size;
+  gint stringlen;
+
+  if ((ptr = ensure_contiguous(tvb, offset, size)) == NULL) {
+
+    len = tvb_length_remaining(tvb, offset);
+    ptr = ensure_contiguous(tvb, offset, len);
+
+  }
+
+  for (p = ptr, stringlen = 0; stringlen < len && *p != '\0'; p++, stringlen++)
+    ;
+  return format_text(ptr, stringlen);
+
+}
+
+/*
+ * Given a tvbuff, an offset, and a length, allocate a buffer big enough
+ * to hold a non-null-terminated string of that length at that offset,
+ * plus a trailing '\0', copy the string into it, and return a pointer
+ * to the string.
+ *
+ * Throws an exception if the tvbuff ends before the string does.
+ */
+guint8 *
+tvb_get_string(tvbuff_t *tvb, gint offset, gint length)
+{
+	const guint8 *ptr;
+	guint8 *strbuf;
+
+	ptr = ensure_contiguous(tvb, offset, length);
+	strbuf = g_malloc(length + 1);
+	if (length != 0)
+		memcpy(strbuf, ptr, length);
+	strbuf[length] = '\0';
+	return strbuf;
+}
+
+/*
+ * Given a tvbuff and an offset, with the offset assumed to refer to
+ * a null-terminated string, find the length of that string (and throw
+ * an exception if the tvbuff ends before we find the null), allocate
+ * a buffer big enough to hold the string, copy the string into it,
+ * and return a pointer to the string.  Also return the length of the
+ * string (including the terminating null) through a pointer.
+ */
+guint8 *
+tvb_get_stringz(tvbuff_t *tvb, gint offset, gint *lengthp)
+{
+	guint size;
+	guint8 *strptr;
+
+	size = tvb_strsize(tvb, offset);
+	strptr = g_malloc(size);
+	tvb_memcpy(tvb, strptr, offset, size);
+	*lengthp = size;
+	return strptr;
+}
+
 /* Looks for a stringz (NUL-terminated string) in tvbuff and copies
- * no more than maxlength number of bytes, including terminating NUL, to buffer.
+ * no more than bufsize number of bytes, including terminating NUL, to buffer.
  * Returns length of string (not including terminating NUL), or -1 if the string was
  * truncated in the buffer due to not having reached the terminating NUL.
  * In this way, it acts like snprintf().
  *
+ * bufsize MUST be greater than 0.
+ *
  * When processing a packet where the remaining number of bytes is less
- * than maxlength, an exception is not thrown if the end of the packet
+ * than bufsize, an exception is not thrown if the end of the packet
  * is reached before the NUL is found. If no NUL is found before reaching
  * the end of the short packet, -1 is still returned, and the string
- * is truncated with a NUL, albeit not at buffer[maxlength], but
+ * is truncated with a NUL, albeit not at buffer[bufsize - 1], but
  * at the correct spot, terminating the string.
  *
  * *bytes_copied will contain the number of bytes actually copied,
  * including the terminating-NUL.
  */
 static gint
-_tvb_get_nstringz(tvbuff_t *tvb, gint offset, guint maxlength, guint8* buffer,
+_tvb_get_nstringz(tvbuff_t *tvb, gint offset, guint bufsize, guint8* buffer,
 		gint *bytes_copied)
 {
 	gint	stringlen;
@@ -1700,8 +1749,13 @@ _tvb_get_nstringz(tvbuff_t *tvb, gint offset, guint maxlength, guint8* buffer,
 
 	check_offset_length(tvb, offset, 0, &abs_offset, &junk_length);
 
-	if (maxlength == 0) {
+	/* There must at least be room for the terminating NUL. */
+	g_assert(bufsize != 0);
+
+	/* If there's no room for anything else, just return the NUL. */
+	if (bufsize == 1) {
 		buffer[0] = 0;
+		*bytes_copied = 1;
 		return 0;
 	}
 
@@ -1715,19 +1769,25 @@ _tvb_get_nstringz(tvbuff_t *tvb, gint offset, guint maxlength, guint8* buffer,
 	}
 
 	/* This should not happen because check_offset_length() would
-	* have already thrown an exception if 'offset' were out-of-bounds.
-	*/
+	 * have already thrown an exception if 'offset' were out-of-bounds.
+	 */
 	g_assert(len != -1);
 
-	if ((guint)len < maxlength) {
+	/*
+	 * If we've been passed a negative number, bufsize will
+	 * be huge.
+	 */
+	g_assert(bufsize <= G_MAXINT);
+
+	if ((guint)len < bufsize) {
 		limit = len;
 		decreased_max = TRUE;
 	}
 	else {
-		limit = maxlength;
+		limit = bufsize;
 	}
 
-	stringlen = tvb_strnlen(tvb, abs_offset, limit);
+	stringlen = tvb_strnlen(tvb, abs_offset, limit - 1);
 	/* If NUL wasn't found, copy the data and return -1 */
 	if (stringlen == -1) {
 		tvb_memcpy(tvb, buffer, abs_offset, limit);
@@ -1750,24 +1810,24 @@ _tvb_get_nstringz(tvbuff_t *tvb, gint offset, guint maxlength, guint8* buffer,
 }
 
 /* Looks for a stringz (NUL-terminated string) in tvbuff and copies
- * no more than maxlength number of bytes, including terminating NUL, to buffer.
+ * no more than bufsize number of bytes, including terminating NUL, to buffer.
  * Returns length of string (not including terminating NUL), or -1 if the string was
  * truncated in the buffer due to not having reached the terminating NUL.
  * In this way, it acts like snprintf().
  *
  * When processing a packet where the remaining number of bytes is less
- * than maxlength, an exception is not thrown if the end of the packet
+ * than bufsize, an exception is not thrown if the end of the packet
  * is reached before the NUL is found. If no NUL is found before reaching
  * the end of the short packet, -1 is still returned, and the string
- * is truncated with a NUL, albeit not at buffer[maxlength], but
+ * is truncated with a NUL, albeit not at buffer[bufsize - 1], but
  * at the correct spot, terminating the string.
  */
 gint
-tvb_get_nstringz(tvbuff_t *tvb, gint offset, guint maxlength, guint8* buffer)
+tvb_get_nstringz(tvbuff_t *tvb, gint offset, guint bufsize, guint8* buffer)
 {
 	gint bytes_copied;
 
-	return _tvb_get_nstringz(tvb, offset, maxlength, buffer, &bytes_copied);
+	return _tvb_get_nstringz(tvb, offset, bufsize, buffer, &bytes_copied);
 }
 
 /* Like tvb_get_nstringz(), but never returns -1. The string is guaranteed to
@@ -1775,14 +1835,14 @@ tvb_get_nstringz(tvbuff_t *tvb, gint offset, guint maxlength, guint8* buffer)
  * a NUL is placed at the end of buffer to terminate it.
  */
 gint
-tvb_get_nstringz0(tvbuff_t *tvb, gint offset, guint maxlength, guint8* buffer)
+tvb_get_nstringz0(tvbuff_t *tvb, gint offset, guint bufsize, guint8* buffer)
 {
 	gint	len, bytes_copied;
 
-	len = _tvb_get_nstringz(tvb, offset, maxlength, buffer, &bytes_copied);
+	len = _tvb_get_nstringz(tvb, offset, bufsize, buffer, &bytes_copied);
 
 	if (len == -1) {
-		buffer[maxlength] = 0;
+		buffer[bufsize - 1] = 0;
 		return bytes_copied - 1;
 	}
 	else {
@@ -1828,7 +1888,7 @@ tvb_find_line_end(tvbuff_t *tvb, gint offset, int len, gint *next_offset,
 	/*
 	 * Look either for a CR or an LF.
 	 */
-	eol_offset = tvb_pbrk_guint8(tvb, offset, len, "\r\n");
+	eol_offset = tvb_pbrk_guint8(tvb, offset, len, (guint8 *)"\r\n");
 	if (eol_offset == -1) {
 		/*
 		 * No CR or LF - line is presumably continued in next packet.
@@ -1957,7 +2017,7 @@ tvb_find_line_end_unquoted(tvbuff_t *tvb, gint offset, int len,
 			 * Look either for a CR, an LF, or a '"'.
 			 */
 			char_offset = tvb_pbrk_guint8(tvb, cur_offset, len,
-			    "\r\n\"");
+			    (guint8 *)"\r\n\"");
 		}
 		if (char_offset == -1) {
 			/*
@@ -2056,8 +2116,36 @@ tvb_bytes_to_str(tvbuff_t *tvb, gint offset, gint len)
 	return bytes_to_str(tvb_get_ptr(tvb, offset, len), len);
 }
 
-tvbuff_t *
-tvb_get_ds_tvb(tvbuff_t *tvb)
+/* Find a needle tvbuff within a haystack tvbuff. */
+gint
+tvb_find_tvb(tvbuff_t *haystack_tvb, tvbuff_t *needle_tvb, gint haystack_offset)
 {
-	return tvb->ds_tvb;
+	guint		haystack_abs_offset, haystack_abs_length;
+	const guint8	*haystack_data;
+	const guint8	*needle_data;
+	const guint 	needle_len = needle_tvb->length;
+	const guint8	*location;
+
+	if (haystack_tvb->length < 1 || needle_tvb->length < 1) {
+		return -1;
+	}
+
+	/* Get pointers to the tvbuffs' data. */
+	haystack_data = tvb_get_ptr(haystack_tvb, 0, -1);
+	needle_data = tvb_get_ptr(needle_tvb, 0, -1);
+
+	check_offset_length(haystack_tvb, haystack_offset, -1,
+			&haystack_abs_offset, &haystack_abs_length);
+
+	location = epan_memmem(haystack_data + haystack_abs_offset, haystack_abs_length,
+			needle_data, needle_len);
+
+	if (location) {
+		return location - haystack_data;
+	}
+	else {
+		return -1;
+	}
+
+	return -1;
 }
