@@ -9,7 +9,7 @@
  * 		the data of a backing tvbuff, or can be a composite of
  * 		other tvbuffs.
  *
- * $Id: tvbuff.h,v 1.30 2003/02/24 01:22:30 guy Exp $
+ * $Id: tvbuff.h,v 1.41 2004/03/23 18:06:29 guy Exp $
  *
  * Copyright (c) 2000 by Gilbert Ramirez <gram@alumni.rice.edu>
  *
@@ -38,10 +38,6 @@
 #include <glib.h>
 #include "exceptions.h"
 
-typedef struct tvbuff tvbuff_t;
-
-typedef void (*tvbuff_free_cb_t)(void*);
-
 /* The different types of tvbuff's */
 typedef enum {
 	TVBUFF_REAL_DATA,
@@ -49,6 +45,71 @@ typedef enum {
 	TVBUFF_COMPOSITE
 } tvbuff_type;
 
+typedef struct {
+	/* The backing tvbuff_t */
+	struct tvbuff	*tvb;
+
+	/* The offset/length of 'tvb' to which I'm privy */
+	guint		offset;
+	guint		length;
+
+} tvb_backing_t;
+
+typedef struct {
+	GSList		*tvbs;
+
+	/* Used for quick testing to see if this
+	 * is the tvbuff that a COMPOSITE is
+	 * interested in. */
+	guint		*start_offsets;
+	guint		*end_offsets;
+
+} tvb_comp_t;
+
+typedef void (*tvbuff_free_cb_t)(void*);
+
+typedef struct tvbuff {
+	/* Record-keeping */
+	tvbuff_type		type;
+	gboolean		initialized;
+	guint			usage_count;
+	struct tvbuff		*ds_tvb;  /* data source top-level tvbuff */
+
+	/* The tvbuffs in which this tvbuff is a member
+	 * (that is, a backing tvbuff for a TVBUFF_SUBSET
+	 * or a member for a TVB_COMPOSITE) */
+	GSList			*used_in;
+
+	/* TVBUFF_SUBSET and TVBUFF_COMPOSITE keep track
+	 * of the other tvbuff's they use */
+	union {
+		tvb_backing_t	subset;
+		tvb_comp_t	composite;
+	} tvbuffs;
+
+	/* We're either a TVBUFF_REAL_DATA or a
+	 * TVBUFF_SUBSET that has a backing buffer that
+	 * has real_data != NULL, or a TVBUFF_COMPOSITE
+	 * which has flattened its data due to a call
+	 * to tvb_get_ptr().
+	 */
+	const guint8		*real_data;
+
+	/* Length of virtual buffer (and/or real_data). */
+	guint			length;
+
+	/* Reported length. */
+	guint			reported_length;
+
+	/* Offset from beginning of first TVBUFF_REAL. */
+	gint			raw_offset;
+
+	/* Func to call when actually freed */
+	tvbuff_free_cb_t	free_cb;
+} tvbuff_t;
+
+
+
 /* TVBUFF_REAL_DATA contains a guint8* that points to real data.
  * The data is allocated and contiguous.
  *
@@ -223,8 +284,11 @@ extern gint tvb_reported_length_remaining(tvbuff_t *tvb, gint offset);
    Also adjusts the data length. */
 extern void tvb_set_reported_length(tvbuff_t*, guint);
 
+extern int offset_from_real_beginning(tvbuff_t *tvb, int counter);
+
 /* Returns the offset from the first byte of real data. */
-extern gint tvb_raw_offset(tvbuff_t*);
+#define TVB_RAW_OFFSET(tvb)			\
+	((tvb->raw_offset==-1)?(tvb->raw_offset = offset_from_real_beginning(tvb, 0)):tvb->raw_offset)
 
 /************** START OF ACCESSORS ****************/
 /* All accessors will throw an exception if appropriate */
@@ -316,29 +380,57 @@ extern char *tvb_fake_unicode(tvbuff_t *tvb, int offset, int len,
 /*
  * Format the data in the tvb from offset for size ...
  */
-extern guint8 * tvb_format_text(tvbuff_t *tvb, gint offset, gint size);
+extern gchar * tvb_format_text(tvbuff_t *tvb, gint offset, gint size);
+
+/*
+ * Like "tvb_format_text()", but for null-padded strings; don't show
+ * the null padding characters as "\000".
+ */
+extern gchar *tvb_format_stringzpad(tvbuff_t *tvb, gint offset, gint size);
+
+/*
+ * Given a tvbuff, an offset, and a length, allocate a buffer big enough
+ * to hold a non-null-terminated string of that length at that offset,
+ * plus a trailing '\0', copy the string into it, and return a pointer
+ * to the string.
+ *
+ * Throws an exception if the tvbuff ends before the string does.
+ */
+extern guint8 *tvb_get_string(tvbuff_t *tvb, gint offset, gint length);
+
+/*
+ * Given a tvbuff and an offset, with the offset assumed to refer to
+ * a null-terminated string, find the length of that string (and throw
+ * an exception if the tvbuff ends before we find the null), allocate
+ * a buffer big enough to hold the string, copy the string into it,
+ * and return a pointer to the string.  Also return the length of the
+ * string (including the terminating null) through a pointer.
+ */
+extern guint8 *tvb_get_stringz(tvbuff_t *tvb, gint offset, gint *lengthp);
 
 /* Looks for a stringz (NUL-terminated string) in tvbuff and copies
- * no more than maxlength number of bytes, including terminating NUL, to buffer.
+ * no more than bufsize number of bytes, including terminating NUL, to buffer.
  * Returns length of string (not including terminating NUL), or -1 if the string was
  * truncated in the buffer due to not having reached the terminating NUL.
  * In this way, it acts like snprintf().
  *
  * When processing a packet where the remaining number of bytes is less
- * than maxlength, an exception is not thrown if the end of the packet
+ * than bufsize, an exception is not thrown if the end of the packet
  * is reached before the NUL is found. If no NUL is found before reaching
  * the end of the short packet, -1 is still returned, and the string
- * is truncated with a NUL, albeit not at buffer[maxlength], but
+ * is truncated with a NUL, albeit not at buffer[bufsize - 1], but
  * at the correct spot, terminating the string.
  */
-extern gint tvb_get_nstringz(tvbuff_t *tvb, gint offset, guint maxlength,
+extern gint tvb_get_nstringz(tvbuff_t *tvb, gint offset, guint bufsize,
     guint8* buffer);
 
 /* Like tvb_get_nstringz(), but never returns -1. The string is guaranteed to
  * have a terminating NUL. If the string was truncated when copied into buffer,
  * a NUL is placed at the end of buffer to terminate it.
+ *
+ * bufsize MUST be greater than 0.
  */
-extern gint tvb_get_nstringz0(tvbuff_t *tvb, gint offset, guint maxlength,
+extern gint tvb_get_nstringz0(tvbuff_t *tvb, gint offset, guint bufsize,
     guint8* buffer);
 
 /*
@@ -388,14 +480,14 @@ extern gint tvb_find_line_end_unquoted(tvbuff_t *tvb, gint offset, int len,
  * Call strncmp after checking if enough chars left, returning 0 if
  * it returns 0 (meaning "equal") and -1 otherwise, otherwise return -1.
  */
-extern gint tvb_strneql(tvbuff_t *tvb, gint offset, const guint8 *str,
+extern gint tvb_strneql(tvbuff_t *tvb, gint offset, const gchar *str,
     gint size);
 
 /*
  * Call strncasecmp after checking if enough chars left, returning 0 if
  * it returns 0 (meaning "equal") and -1 otherwise, otherwise return -1.
  */
-extern gint tvb_strncaseeql(tvbuff_t *tvb, gint offset, const guint8 *str,
+extern gint tvb_strncaseeql(tvbuff_t *tvb, gint offset, const gchar *str,
     gint size);
 
 /*
@@ -411,7 +503,15 @@ extern gint tvb_memeql(tvbuff_t *tvb, gint offset, const guint8 *str,
  */
 extern gchar *tvb_bytes_to_str(tvbuff_t *tvb, gint offset, gint len);
 
-extern tvbuff_t *tvb_get_ds_tvb(tvbuff_t *tvb);
+#define TVB_GET_DS_TVB(tvb)		\
+	(tvb->ds_tvb)
+
+/* Locate a sub-tvbuff within another tvbuff, starting at position
+ * 'haystack_offset'. Returns the index of the beginning of 'needle' within
+ * 'haystack', or -1 if 'needle' is not found. The index is relative
+ * to the start of 'haystack', not 'haystack_offset'. */
+extern gint tvb_find_tvb(tvbuff_t *haystack_tvb, tvbuff_t *needle_tvb,
+	gint haystack_offset);
 
 /************** END OF ACCESSORS ****************/
 
