@@ -1,6 +1,6 @@
 /* follow_dlg.c
  *
- * $Id: follow_dlg.c,v 1.31 2003/03/06 04:23:51 guy Exp $
+ * $Id: follow_dlg.c,v 1.53 2004/03/13 15:15:24 ulfl Exp $
  *
  * Ethereal - Network traffic analyzer
  * By Gerald Combs <gerald@ethereal.com>
@@ -38,10 +38,6 @@
 #include <unistd.h>
 #endif
 
-#ifdef NEED_SNPRINTF_H
-# include "snprintf.h"
-#endif
-
 #include <ctype.h>
 
 #include "color.h"
@@ -54,6 +50,7 @@
 #include "globals.h"
 #include "gtkglobals.h"
 #include "main.h"
+#include "alert_box.h"
 #include "simple_dialog.h"
 #include "packet-ipv6.h"
 #include "prefs.h"
@@ -63,6 +60,8 @@
 #include <epan/epan_dissect.h>
 #include <epan/filesystem.h>
 #include "compat_macros.h"
+#include "ipproto.h"
+#include "tap_menu.h"
 
 /* Show Stream */
 typedef enum {
@@ -90,11 +89,15 @@ typedef struct {
 	GtkWidget	*carray_bt;
 	GtkWidget	*follow_save_as_w;
 	gboolean        is_ipv6;
+	char		*filter_out_filter;
+	GtkWidget	*filter_te;
+	GtkWidget	*streamwindow;
 } follow_info_t;
 
 static void follow_destroy_cb(GtkWidget * win, gpointer data);
 static void follow_charset_toggle_cb(GtkWidget * w, gpointer parent_w);
 static void follow_load_text(follow_info_t *follow_info);
+static void follow_filter_out_stream(GtkWidget * w, gpointer parent_w);
 static void follow_print_stream(GtkWidget * w, gpointer parent_w);
 static void follow_save_as_cmd_cb(GtkWidget * w, gpointer data);
 static void follow_save_as_ok_cb(GtkWidget * w, GtkFileSelection * fs);
@@ -147,10 +150,14 @@ void
 follow_stream_cb(GtkWidget * w, gpointer data _U_)
 {
 	GtkWidget	*streamwindow, *vbox, *txt_scrollw, *text, *filter_te;
-	GtkWidget	*hbox, *button, *radio_bt;
+	GtkWidget	*hbox, *button_hbox, *button, *radio_bt;
+    GtkWidget   *stream_fr, *stream_vb;
 	GtkWidget	*stream_om, *stream_menu, *stream_mi;
-	int		tmp_fd;
+	GtkTooltips *tooltips;
+	int		    tmp_fd;
 	gchar		*follow_filter;
+	const gchar	*previous_filter;
+    int		    filter_out_filter_len;
 	const char	*hostname0, *hostname1;
 	char		*port0, *port1;
 	char		string[128];
@@ -159,7 +166,7 @@ follow_stream_cb(GtkWidget * w, gpointer data _U_)
 
 	/* we got tcp so we can follow */
 	if (cfile.edt->pi.ipproto != 6) {
-		simple_dialog(ESD_TYPE_CRIT, NULL,
+		simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK,
 			      "Error following stream.  Please make\n"
 			      "sure you have a TCP packet selected.");
 		return;
@@ -179,7 +186,7 @@ follow_stream_cb(GtkWidget * w, gpointer data _U_)
 			sizeof follow_info->data_out_filename, "follow");
 
 	if (tmp_fd == -1) {
-	    simple_dialog(ESD_TYPE_WARN, NULL,
+	    simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK,
 			  "Could not create temporary file %s: %s",
 			  follow_info->data_out_filename, strerror(errno));
 	    g_free(follow_info);
@@ -188,7 +195,7 @@ follow_stream_cb(GtkWidget * w, gpointer data _U_)
 
 	data_out_file = fdopen(tmp_fd, "wb");
 	if (data_out_file == NULL) {
-	    simple_dialog(ESD_TYPE_WARN, NULL,
+	    simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK,
 			  "Could not create temporary file %s: %s",
 			  follow_info->data_out_filename, strerror(errno));
 	    close(tmp_fd);
@@ -204,10 +211,34 @@ follow_stream_cb(GtkWidget * w, gpointer data _U_)
 
 	/* Set the display filter entry accordingly */
 	filter_te = OBJECT_GET_DATA(w, E_DFILTER_TE_KEY);
+
+	/* needed in follow_filter_out_stream(), is there a better way? */
+	follow_info->filter_te = filter_te;
+
+	/* save previous filter, const since we're not supposed to alter */
+	previous_filter =
+	    (const gchar *)gtk_entry_get_text(GTK_ENTRY(filter_te));
+
+	/* allocate our new filter. API claims g_malloc terminates program on failure */
+	/* my calc for max alloc needed is really +10 but when did a few extra bytes hurt ? */
+	filter_out_filter_len = strlen(follow_filter) + strlen(previous_filter) + 16;
+	follow_info->filter_out_filter = (gchar *)g_malloc(filter_out_filter_len);
+
+	/* append the negation */
+	if(strlen(previous_filter)) {
+	    g_snprintf(follow_info->filter_out_filter, filter_out_filter_len,
+            "%s and !(%s)", previous_filter, follow_filter);
+	} else {
+	    g_snprintf(follow_info->filter_out_filter, filter_out_filter_len,
+            "!(%s)", follow_filter);
+	}
+
+
 	gtk_entry_set_text(GTK_ENTRY(filter_te), follow_filter);
 
-	/* Run the display filter so it goes in effect. */
-	filter_packets(&cfile, follow_filter);
+	/* Run the display filter so it goes in effect - even if it's the
+	   same as the previous display filter. */
+	main_filter_packets(&cfile, follow_filter, TRUE);
 
 	/* Free the filter string, as we're done with it. */
 	g_free(follow_filter);
@@ -216,31 +247,43 @@ follow_stream_cb(GtkWidget * w, gpointer data _U_)
 	fclose(data_out_file);
 
 	/* The data_out_filename file now has all the text that was in the session */
-	streamwindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
+	streamwindow = window_new(GTK_WINDOW_TOPLEVEL, "Follow TCP stream");
+
+	/* needed in follow_filter_out_stream(), is there a better way? */
+	follow_info->streamwindow = streamwindow;
+
 	gtk_widget_set_name(streamwindow, "TCP stream window");
 
 	SIGNAL_CONNECT(streamwindow, "destroy", follow_destroy_cb, NULL);
-	SIGNAL_CONNECT(streamwindow, "realize", window_icon_realize_cb, NULL);
-	if (incomplete_tcp_stream) {
-	    gtk_window_set_title(GTK_WINDOW(streamwindow),
-				 "Contents of TCP stream (incomplete)");
-	} else {
-	    gtk_window_set_title(GTK_WINDOW(streamwindow),
-				 "Contents of TCP stream");
-	}
 	WIDGET_SET_SIZE(streamwindow, DEF_WIDTH, DEF_HEIGHT);
 	gtk_container_border_width(GTK_CONTAINER(streamwindow), 2);
 
 	/* setup the container */
+    tooltips = gtk_tooltips_new ();
+
 	vbox = gtk_vbox_new(FALSE, 0);
 	gtk_container_add(GTK_CONTAINER(streamwindow), vbox);
 
+    /* content frame */
+	if (incomplete_tcp_stream) {
+        stream_fr = gtk_frame_new("Stream Content (incomplete)");
+	} else {
+        stream_fr = gtk_frame_new("Stream Content");
+	}
+  	gtk_container_add(GTK_CONTAINER(vbox), stream_fr);
+  	gtk_widget_show(stream_fr);
+
+    stream_vb = gtk_vbox_new(FALSE, 0);
+	gtk_container_set_border_width( GTK_CONTAINER(stream_vb) , 5);
+	gtk_container_add(GTK_CONTAINER(stream_fr), stream_vb);
+
 	/* create a scrolled window for the text */
 	txt_scrollw = scrolled_window_new(NULL, NULL);
-	gtk_box_pack_start(GTK_BOX(vbox), txt_scrollw, TRUE, TRUE, 0);
-	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(txt_scrollw),
-				       GTK_POLICY_AUTOMATIC,
-				       GTK_POLICY_AUTOMATIC);
+#if GTK_MAJOR_VERSION >= 2
+    gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(txt_scrollw), 
+                                   GTK_SHADOW_IN);
+#endif
+	gtk_box_pack_start(GTK_BOX(stream_vb), txt_scrollw, TRUE, TRUE, 0);
 
 	/* create a text box */
 #if GTK_MAJOR_VERSION < 2
@@ -253,10 +296,22 @@ follow_stream_cb(GtkWidget * w, gpointer data _U_)
 	gtk_container_add(GTK_CONTAINER(txt_scrollw), text);
 	follow_info->text = text;
 
-	/* Create hbox */
-	hbox = gtk_hbox_new(FALSE, 1);
-	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
 
+	/* stream hbox */
+	hbox = gtk_hbox_new(FALSE, 1);
+	gtk_box_pack_start(GTK_BOX(stream_vb), hbox, FALSE, FALSE, 0);
+
+	/* Create Save As Button */
+    button = BUTTON_NEW_FROM_STOCK(GTK_STOCK_SAVE_AS);
+	SIGNAL_CONNECT(button, "clicked", follow_save_as_cmd_cb, follow_info);
+    gtk_tooltips_set_tip (tooltips, button, "Save the content as currently displayed ", NULL);
+	gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
+
+	/* Create Print Button */
+    button = BUTTON_NEW_FROM_STOCK(GTK_STOCK_PRINT);
+	SIGNAL_CONNECT(button, "clicked", follow_print_stream, follow_info);
+    gtk_tooltips_set_tip (tooltips, button, "Print the content as currently displayed", NULL);
+	gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
 
 	/* Stream to show */
 	follow_tcp_stats(&stats);
@@ -284,7 +339,7 @@ follow_stream_cb(GtkWidget * w, gpointer data _U_)
 	stream_menu = gtk_menu_new();
 
 	/* Both Hosts */
-	snprintf(string, sizeof(string),
+	g_snprintf(string, sizeof(string),
 		 "Entire conversation (%u bytes)",
 		 stats.bytes_written[0] + stats.bytes_written[1]);
 	stream_mi = gtk_menu_item_new_with_label(string);
@@ -295,7 +350,7 @@ follow_stream_cb(GtkWidget * w, gpointer data _U_)
 	follow_info->show_stream = BOTH_HOSTS;
 
 	/* Host 0 --> Host 1 */
-	snprintf(string, sizeof(string), "%s:%s --> %s:%s (%u bytes)",
+	g_snprintf(string, sizeof(string), "%s:%s --> %s:%s (%u bytes)",
 		 hostname0, port0, hostname1, port1,
 		 stats.bytes_written[0]);
 	stream_mi = gtk_menu_item_new_with_label(string);
@@ -305,7 +360,7 @@ follow_stream_cb(GtkWidget * w, gpointer data _U_)
 	gtk_widget_show(stream_mi);
 
 	/* Host 1 --> Host 0 */
-	snprintf(string, sizeof(string), "%s:%s --> %s:%s (%u bytes)",
+	g_snprintf(string, sizeof(string), "%s:%s --> %s:%s (%u bytes)",
 		 hostname1, port1, hostname0, port0,
 		 stats.bytes_written[1]);
 	stream_mi = gtk_menu_item_new_with_label(string);
@@ -317,6 +372,8 @@ follow_stream_cb(GtkWidget * w, gpointer data _U_)
 	gtk_option_menu_set_menu(GTK_OPTION_MENU(stream_om), stream_menu);
 	/* Set history to 0th item, i.e., the first item. */
 	gtk_option_menu_set_history(GTK_OPTION_MENU(stream_om), 0);
+    gtk_tooltips_set_tip (tooltips, stream_om, 
+        "Select the stream direction to display", NULL);
 	gtk_box_pack_start(GTK_BOX(hbox), stream_om, FALSE, FALSE, 0);
 
 	/* ASCII radio button */
@@ -358,39 +415,32 @@ follow_stream_cb(GtkWidget * w, gpointer data _U_)
                        follow_info);
 	follow_info->carray_bt = radio_bt;
 
+	/* button hbox */
+	button_hbox = gtk_hbutton_box_new();
+	gtk_box_pack_start(GTK_BOX(vbox), button_hbox, FALSE, FALSE, 0);
+    gtk_button_box_set_layout (GTK_BUTTON_BOX(button_hbox), GTK_BUTTONBOX_END);
+    gtk_button_box_set_spacing(GTK_BUTTON_BOX(button_hbox), 5);
+
+	/* Create exclude stream button */
+	button = gtk_button_new_with_label("Filter out this stream");
+	SIGNAL_CONNECT(button, "clicked", follow_filter_out_stream, follow_info);
+    gtk_tooltips_set_tip (tooltips, button, 
+        "Build a display filter which cuts this stream from the capture", NULL);
+	gtk_box_pack_start(GTK_BOX(button_hbox), button, FALSE, FALSE, 0);
+
 	/* Create Close Button */
-#if GTK_MAJOR_VERSION < 2
-	button = gtk_button_new_with_label("Close");
-#else
-        button = gtk_button_new_from_stock(GTK_STOCK_CLOSE);
-#endif
+    button = BUTTON_NEW_FROM_STOCK(GTK_STOCK_CLOSE);
 	SIGNAL_CONNECT_OBJECT(button, "clicked", gtk_widget_destroy,
                               streamwindow);
-	gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
+    gtk_tooltips_set_tip (tooltips, button, 
+        "Close the dialog and keep the current display filter", NULL);
+	gtk_box_pack_start(GTK_BOX(button_hbox), button, FALSE, FALSE, 0);
 
 	/* Catch the "key_press_event" signal in the window, so that we can catch
 	the ESC key being pressed and act as if the "Cancel" button had
 	been selected. */
 	dlg_set_cancel(streamwindow, button);
 
-	/* Create Save As Button */
-#if GTK_MAJOR_VERSION < 2
-	button = gtk_button_new_with_label("Save As");
-#else
-        button = gtk_button_new_from_stock(GTK_STOCK_SAVE_AS);
-#endif
-	SIGNAL_CONNECT(button, "clicked", follow_save_as_cmd_cb, follow_info);
-	gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
-
-	/* Create Print Button */
-#if GTK_MAJOR_VERSION < 2
-	button = gtk_button_new_with_label("Print");
-#else
-        button = gtk_button_new_from_stock(GTK_STOCK_PRINT);
-#endif
-	SIGNAL_CONNECT(button, "clicked", follow_print_stream, follow_info);
-	gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
-
 
 	/* Tuck away the follow_info object into the window */
 	OBJECT_SET_DATA(streamwindow, E_FOLLOW_INFO_KEY, follow_info);
@@ -408,7 +458,8 @@ follow_stream_cb(GtkWidget * w, gpointer data _U_)
 }
 
 /* The destroy call back has the responsibility of
- * unlinking the temporary file */
+ * unlinking the temporary file
+ * and freeing the filter_out_filter */
 static void
 follow_destroy_cb(GtkWidget *w, gpointer data _U_)
 {
@@ -416,6 +467,7 @@ follow_destroy_cb(GtkWidget *w, gpointer data _U_)
 
 	follow_info = OBJECT_GET_DATA(w, E_FOLLOW_INFO_KEY);
 	unlink(follow_info->data_out_filename);
+	g_free(follow_info->filter_out_filter);
 	gtk_widget_destroy(w);
 	forget_follow_info(follow_info);
 	g_free(follow_info);
@@ -469,9 +521,17 @@ follow_charset_toggle_cb(GtkWidget * w _U_, gpointer data)
 }
 
 #define FLT_BUF_SIZE 1024
-static void
+
+typedef enum {
+	FRS_OK,
+	FRS_OPEN_ERROR,
+	FRS_READ_ERROR,
+	FRS_PRINT_ERROR
+} frs_return_t;
+
+static frs_return_t
 follow_read_stream(follow_info_t *follow_info,
-		   void (*print_line) (char *, int, gboolean, void *),
+		   gboolean (*print_line) (char *, int, gboolean, void *),
 		   void *arg)
 {
     tcp_stream_chunk	sc;
@@ -485,170 +545,184 @@ follow_read_stream(follow_info_t *follow_info,
     gchar               initbuf[256];
     guint32             server_packet_count = 0;
     guint32             client_packet_count = 0;
+    char                buffer[FLT_BUF_SIZE];
+    int                 nchars;
 
     iplen = (follow_info->is_ipv6) ? 16 : 4;
 
     data_out_file = fopen(follow_info->data_out_filename, "rb");
-    if (data_out_file) {
-	char buffer[FLT_BUF_SIZE];
-	int nchars;
-	while (fread(&sc, 1, sizeof(sc), data_out_file)) {
-	    if (client_port == 0) {
-		memcpy(client_addr, sc.src_addr, iplen);
-		client_port = sc.src_port;
-	    }
-	    skip = FALSE;
-	    if (memcmp(client_addr, sc.src_addr, iplen) == 0 &&
-		client_port == sc.src_port) {
-		is_server = FALSE;
-		global_pos = &global_client_pos;
-		if (follow_info->show_stream == FROM_SERVER) {
-			skip = TRUE;
-		}
-	    }
-	    else {
-		is_server = TRUE;
-		global_pos = &global_server_pos;
-		if (follow_info->show_stream == FROM_CLIENT) {
-			skip = TRUE;
-		}
-	    }
-
-	    while (sc.dlen > 0) {
-		bcount = (sc.dlen < FLT_BUF_SIZE) ? sc.dlen : FLT_BUF_SIZE;
-		nchars = fread(buffer, 1, bcount, data_out_file);
-		if (nchars == 0)
-		    break;
-		sc.dlen -= bcount;
-		if (!skip) {
-		    switch (follow_info->show_type) {
-			case SHOW_EBCDIC:
-			    /* If our native arch is ASCII, call: */
-			    EBCDIC_to_ASCII(buffer, nchars);
-			    (*print_line) (buffer, nchars, is_server, arg);
-			    break;
-			case SHOW_ASCII:
-			    /* If our native arch is EBCDIC, call:
-			     * ASCII_TO_EBCDIC(buffer, nchars);
-			     */
-			    (*print_line) (buffer, nchars, is_server, arg);
-			    break;
-			case SHOW_HEXDUMP:
-			    current_pos = 0;
-			    while (current_pos < nchars) {
-				gchar hexbuf[256];
-				gchar hexchars[] = "0123456789abcdef";
-				int i, cur;
-				/* is_server indentation : put 63 spaces at the begenning
-				 * of the string */
-				sprintf(hexbuf, (is_server &&
-					follow_info->show_stream == BOTH_HOSTS) ?
-					"                                 "
-					"                                             %08X  " :
-					"%08X  ", *global_pos);
-				cur = strlen(hexbuf);
-				for (i = 0; i < 16 && current_pos + i < nchars;
-				     i++) {
-				    hexbuf[cur++] =
-					hexchars[(buffer[current_pos + i] & 0xf0)
-						 >> 4];
-				    hexbuf[cur++] =
-					hexchars[buffer[current_pos + i] & 0x0f];
-				    if (i == 7) {
-					hexbuf[cur++] = ' ';
-					hexbuf[cur++] = ' ';
-				    } else if (i != 15)
-					hexbuf[cur++] = ' ';
-				}
-				/* Fill it up if column isn't complete */
-				if (i < 16) {
-				    int j;
-
-				    for (j = i; j < 16; j++) {
-					if (j == 7)
-					    hexbuf[cur++] = ' ';
-					hexbuf[cur++] = ' ';
-					hexbuf[cur++] = ' ';
-					hexbuf[cur++] = ' ';
-				    }
-				} else
-				    hexbuf[cur++] = ' ';
-
-				/* Now dump bytes as text */
-				for (i = 0; i < 16 && current_pos + i < nchars;
-				     i++) {
-				    hexbuf[cur++] =
-				        (isprint((guchar)buffer[current_pos + i]) ?
-				        buffer[current_pos + i] : '.' );
-				    if (i == 7) {
-					hexbuf[cur++] = ' ';
-				    }
-				}
-				current_pos += i;
-				(*global_pos) += i;
-				hexbuf[cur++] = '\n';
-				hexbuf[cur] = 0;
-				(*print_line) (hexbuf, strlen(hexbuf), is_server, arg);
-			    }
-			    break;
-			case SHOW_CARRAY:
-			    current_pos = 0;
-			    sprintf(initbuf, "char peer%d_%d[] = {\n", is_server ? 1 : 0,
-				    is_server ? server_packet_count++ : client_packet_count++);
-			    (*print_line) (initbuf, strlen(initbuf), is_server, arg);
-			    while (current_pos < nchars) {
-				gchar hexbuf[256];
-				gchar hexchars[] = "0123456789abcdef";
-				int i, cur;
-
-				cur = 0;
-				for (i = 0; i < 8 && current_pos + i < nchars;
-				     i++) {
-				  /* Prepend entries with "0x" */
-				  hexbuf[cur++] = '0';
-				  hexbuf[cur++] = 'x';
-				    hexbuf[cur++] =
-					hexchars[(buffer[current_pos + i] & 0xf0)
-						 >> 4];
-				    hexbuf[cur++] =
-					hexchars[buffer[current_pos + i] & 0x0f];
-
-				    /* Delimit array entries with a comma */
-				    if (current_pos + i + 1 < nchars)
-				      hexbuf[cur++] = ',';
-
-				    hexbuf[cur++] = ' ';
-				}
-
-				/* Terminate the array if we are at the end */
-				if (current_pos + i == nchars) {
-				  hexbuf[cur++] = '}';
-				  hexbuf[cur++] = ';';
-				}
-
-				current_pos += i;
-				(*global_pos) += i;
-				hexbuf[cur++] = '\n';
-				hexbuf[cur] = 0;
-				(*print_line) (hexbuf, strlen(hexbuf), is_server, arg);
-			    }
-			    break;
-		    }
-		}
-	    }
-	}
-	if (ferror(data_out_file)) {
-	    simple_dialog(ESD_TYPE_WARN, NULL,
-			  "Error reading temporary file %s: %s", follow_info->data_out_filename,
-			  strerror(errno));
-	}
-	fclose(data_out_file);
-	data_out_file = NULL;
-    } else {
-	simple_dialog(ESD_TYPE_WARN, NULL,
+    if (data_out_file == NULL) {
+	simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK,
 		      "Could not open temporary file %s: %s", follow_info->data_out_filename,
 		      strerror(errno));
+	return FRS_OPEN_ERROR;
     }
+
+    while (fread(&sc, 1, sizeof(sc), data_out_file)) {
+	if (client_port == 0) {
+	    memcpy(client_addr, sc.src_addr, iplen);
+	    client_port = sc.src_port;
+	}
+	skip = FALSE;
+	if (memcmp(client_addr, sc.src_addr, iplen) == 0 &&
+	    client_port == sc.src_port) {
+	    is_server = FALSE;
+	    global_pos = &global_client_pos;
+	    if (follow_info->show_stream == FROM_SERVER) {
+		skip = TRUE;
+	    }
+	}
+	else {
+	    is_server = TRUE;
+	    global_pos = &global_server_pos;
+	    if (follow_info->show_stream == FROM_CLIENT) {
+		skip = TRUE;
+	    }
+	}
+
+	while (sc.dlen > 0) {
+	    bcount = (sc.dlen < FLT_BUF_SIZE) ? sc.dlen : FLT_BUF_SIZE;
+	    nchars = fread(buffer, 1, bcount, data_out_file);
+	    if (nchars == 0)
+		break;
+	    sc.dlen -= bcount;
+	    if (!skip) {
+		switch (follow_info->show_type) {
+
+		case SHOW_EBCDIC:
+		    /* If our native arch is ASCII, call: */
+		    EBCDIC_to_ASCII(buffer, nchars);
+		    if (!(*print_line) (buffer, nchars, is_server, arg))
+			goto print_error;
+		    break;
+
+		case SHOW_ASCII:
+		    /* If our native arch is EBCDIC, call:
+		     * ASCII_TO_EBCDIC(buffer, nchars);
+		     */
+		    if (!(*print_line) (buffer, nchars, is_server, arg))
+			goto print_error;
+		    break;
+
+		case SHOW_HEXDUMP:
+		    current_pos = 0;
+		    while (current_pos < nchars) {
+			gchar hexbuf[256];
+			gchar hexchars[] = "0123456789abcdef";
+			int i, cur;
+
+			/* is_server indentation : put 63 spaces at the
+			 * beginning of the string */
+			g_snprintf(hexbuf, sizeof(hexbuf),
+                (is_server && follow_info->show_stream == BOTH_HOSTS) ?
+				"                                 "
+				"                                             %08X  " :
+				"%08X  ", *global_pos);
+			cur = strlen(hexbuf);
+			for (i = 0; i < 16 && current_pos + i < nchars; i++) {
+			    hexbuf[cur++] =
+				hexchars[(buffer[current_pos + i] & 0xf0) >> 4];
+			    hexbuf[cur++] =
+				hexchars[buffer[current_pos + i] & 0x0f];
+			    if (i == 7) {
+				hexbuf[cur++] = ' ';
+				hexbuf[cur++] = ' ';
+			    } else if (i != 15)
+				hexbuf[cur++] = ' ';
+			}
+			/* Fill it up if column isn't complete */
+			if (i < 16) {
+			    int j;
+
+			    for (j = i; j < 16; j++) {
+				if (j == 7)
+				    hexbuf[cur++] = ' ';
+				hexbuf[cur++] = ' ';
+				hexbuf[cur++] = ' ';
+				hexbuf[cur++] = ' ';
+			    }
+			} else
+			    hexbuf[cur++] = ' ';
+
+			/* Now dump bytes as text */
+			for (i = 0; i < 16 && current_pos + i < nchars; i++) {
+			    hexbuf[cur++] =
+			        (isprint((guchar)buffer[current_pos + i]) ?
+			        buffer[current_pos + i] : '.' );
+			    if (i == 7) {
+				hexbuf[cur++] = ' ';
+			    }
+			}
+			current_pos += i;
+			(*global_pos) += i;
+			hexbuf[cur++] = '\n';
+			hexbuf[cur] = 0;
+			if (!(*print_line) (hexbuf, strlen(hexbuf), is_server, arg))
+			    goto print_error;
+		    }
+		    break;
+
+		case SHOW_CARRAY:
+		    current_pos = 0;
+		    g_snprintf(initbuf, 256, "char peer%d_%d[] = {\n", is_server ? 1 : 0,
+			    is_server ? server_packet_count++ : client_packet_count++);
+		    if (!(*print_line) (initbuf, strlen(initbuf), is_server, arg))
+			goto print_error;
+		    while (current_pos < nchars) {
+			gchar hexbuf[256];
+			gchar hexchars[] = "0123456789abcdef";
+			int i, cur;
+
+			cur = 0;
+			for (i = 0; i < 8 && current_pos + i < nchars; i++) {
+			  /* Prepend entries with "0x" */
+			  hexbuf[cur++] = '0';
+			  hexbuf[cur++] = 'x';
+			    hexbuf[cur++] =
+				hexchars[(buffer[current_pos + i] & 0xf0) >> 4];
+			    hexbuf[cur++] =
+				hexchars[buffer[current_pos + i] & 0x0f];
+
+			    /* Delimit array entries with a comma */
+			    if (current_pos + i + 1 < nchars)
+			      hexbuf[cur++] = ',';
+
+			    hexbuf[cur++] = ' ';
+			}
+
+			/* Terminate the array if we are at the end */
+			if (current_pos + i == nchars) {
+			    hexbuf[cur++] = '}';
+			    hexbuf[cur++] = ';';
+			}
+
+			current_pos += i;
+			(*global_pos) += i;
+			hexbuf[cur++] = '\n';
+			hexbuf[cur] = 0;
+			if (!(*print_line) (hexbuf, strlen(hexbuf), is_server, arg))
+			    goto print_error;
+		    }
+		    break;
+		}
+	    }
+	}
+    }
+    if (ferror(data_out_file)) {
+	simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK,
+		      "Error reading temporary file %s: %s", follow_info->data_out_filename,
+		      strerror(errno));
+	fclose(data_out_file);
+	data_out_file = NULL;
+	return FRS_READ_ERROR;
+    }
+
+    return FRS_OK;
+
+print_error:
+    fclose(data_out_file);
+    data_out_file = NULL;
+    return FRS_PRINT_ERROR;
 }
 
 /*
@@ -660,12 +734,34 @@ follow_read_stream(follow_info_t *follow_info,
  *
  * For now, we support only text printing.
  */
-static void
+static gboolean
 follow_print_text(char *buffer, int nchars, gboolean is_server _U_, void *arg)
 {
     FILE *fh = arg;
 
-    fwrite(buffer, nchars, 1, fh);
+    if (fwrite(buffer, nchars, 1, fh) != 1)
+      return FALSE;
+    return TRUE;
+}
+
+static void
+follow_filter_out_stream(GtkWidget * w _U_, gpointer data) 
+{
+    follow_info_t	*follow_info = data;
+
+    /* Lock out user from messing with us. (ie. don't free our data!) */
+    gtk_widget_set_sensitive(follow_info->streamwindow, FALSE);
+
+    /* Set the display filter. */
+    gtk_entry_set_text(GTK_ENTRY(follow_info->filter_te), follow_info->filter_out_filter);
+
+    /* Run the display filter so it goes in effect. */
+    main_filter_packets(&cfile, follow_info->filter_out_filter, FALSE);
+
+    /* we force a subsequent close */
+    gtk_widget_destroy(follow_info->streamwindow);
+  
+    return;
 }
 
 static void
@@ -687,7 +783,7 @@ follow_print_stream(GtkWidget * w _U_, gpointer data)
 	to_file = TRUE;
 	break;
     default:			/* "Can't happen" */
-	simple_dialog(ESD_TYPE_CRIT, NULL,
+	simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK,
 		      "Couldn't figure out where to send the print "
 		      "job. Check your preferences.");
 	return;
@@ -695,27 +791,57 @@ follow_print_stream(GtkWidget * w _U_, gpointer data)
 
     fh = open_print_dest(to_file, print_dest);
     if (fh == NULL) {
-	switch (to_file) {
-	case FALSE:
-	    simple_dialog(ESD_TYPE_WARN, NULL,
+	if (to_file)
+	    open_failure_alert_box(prefs.pr_file, errno, TRUE);
+	else {
+	    simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK,
 			  "Couldn't run print command %s.", prefs.pr_cmd);
-	    break;
-
-	case TRUE:
-	    simple_dialog(ESD_TYPE_WARN, NULL,
-			  file_write_error_message(errno), prefs.pr_file);
-	    break;
 	}
 	return;
     }
 
     print_preamble(fh, PR_FMT_TEXT);
-    follow_read_stream(follow_info, follow_print_text, fh);
+    if (ferror(fh))
+	goto print_error;
+    switch (follow_read_stream(follow_info, follow_print_text, fh)) {
+
+    case FRS_OK:
+	break;
+
+    case FRS_OPEN_ERROR:
+    case FRS_READ_ERROR:
+	/* XXX - cancel printing? */
+	close_print_dest(to_file, fh);
+	return;
+
+    case FRS_PRINT_ERROR:
+	goto print_error;
+    }
     print_finale(fh, PR_FMT_TEXT);
+    if (ferror(fh))
+	goto print_error;
+    if (!close_print_dest(to_file, fh)) {
+	if (to_file)
+	    write_failure_alert_box(prefs.pr_file, errno);
+	else {
+	    simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK,
+			  "Error closing print destination.");
+	}
+    }
+    return;
+
+print_error:
+    if (to_file)
+	write_failure_alert_box(prefs.pr_file, errno);
+    else {
+	simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK,
+		      "Error writing to print command: %s", strerror(errno));
+    }
+    /* XXX - cancel printing? */
     close_print_dest(to_file, fh);
 }
 
-static void
+static gboolean
 follow_add_to_gtk_text(char *buffer, int nchars, gboolean is_server,
 		       void *arg)
 {
@@ -729,7 +855,7 @@ follow_add_to_gtk_text(char *buffer, int nchars, gboolean is_server,
     gchar         *convbuf;
 #endif
 
-#ifdef _WIN32
+#if GTK_MAJOR_VERSION >= 2 || GTK_MINOR_VERSION >= 3
     /* While our isprint() hack is in place, we
      * have to use convert some chars to '.' in order
      * to be able to see the data we *should* see
@@ -738,18 +864,7 @@ follow_add_to_gtk_text(char *buffer, int nchars, gboolean is_server,
     int i;
 
     for (i = 0; i < nchars; i++) {
-	    if (buffer[i] == 0x0a || buffer[i] == 0x0d) {
-		    continue;
-	    }
-	    else if (! isprint(buffer[i])) {
-		    buffer[i] = '.';
-	    }
-    }
-#elif GTK_MAJOR_VERSION >= 2
-    int i;
-
-    for (i = 0; i < nchars; i++) {
-        if (buffer[i] == '\n')
+        if (buffer[i] == '\n' || buffer[i] == '\r')
             continue;
         if (! isprint(buffer[i])) {
             buffer[i] = '.';
@@ -776,6 +891,7 @@ follow_add_to_gtk_text(char *buffer, int nchars, gboolean is_server,
                                      NULL);
     g_free(convbuf);
 #endif
+    return TRUE;
 }
 
 static void
@@ -884,8 +1000,7 @@ follow_save_as_ok_cb(GtkWidget * w _U_, GtkFileSelection * fs)
 
 	fh = fopen(to_name, "wb");
 	if (fh == NULL) {
-		simple_dialog(ESD_TYPE_WARN, NULL,
-			file_write_error_message(errno), to_name);
+		open_failure_alert_box(to_name, errno, TRUE);
 		g_free(to_name);
 		return;
 	}
@@ -894,8 +1009,23 @@ follow_save_as_ok_cb(GtkWidget * w _U_, GtkFileSelection * fs)
 	follow_info = OBJECT_GET_DATA(fs, E_FOLLOW_INFO_KEY);
 	gtk_widget_destroy(GTK_WIDGET(fs));
 
-	follow_read_stream(follow_info, follow_print_text, fh);
-	fclose(fh);
+	switch (follow_read_stream(follow_info, follow_print_text, fh)) {
+
+	case FRS_OK:
+		if (fclose(fh) == EOF)
+			write_failure_alert_box(to_name, errno);
+		break;
+
+	case FRS_OPEN_ERROR:
+	case FRS_READ_ERROR:
+		fclose(fh);
+		break;
+
+	case FRS_PRINT_ERROR:
+		write_failure_alert_box(to_name, errno);
+		fclose(fh);
+		break;
+	}
 
 	/* Save the directory name for future file dialogs. */
 	dirname = get_dirname(to_name);  /* Overwrites to_name */
