@@ -1,7 +1,7 @@
 /* prefs_dlg.c
  * Routines for handling preferences
  *
- * $Id: prefs_dlg.c,v 1.58 2002/12/27 18:32:55 oabad Exp $
+ * $Id: prefs_dlg.c,v 1.80 2004/03/13 15:15:25 ulfl Exp $
  *
  * Ethereal - Network traffic analyzer
  * By Gerald Combs <gerald@ethereal.com>
@@ -70,15 +70,15 @@ static void	prefs_tree_select_cb(GtkCTree *, GtkCTreeNode *, gint,
 static void	prefs_tree_select_cb(GtkTreeSelection *, gpointer);
 #endif
 
-#define E_PRINT_PAGE_KEY   "printer_options_page"
-#define E_COLUMN_PAGE_KEY  "column_options_page"
-#define E_STREAM_PAGE_KEY  "tcp_stream_options_page"
-#define E_GUI_PAGE_KEY	   "gui_options_page"
-#define E_CAPTURE_PAGE_KEY "capture_options_page"
-#define E_NAMERES_PAGE_KEY "nameres_options_page"
-#define E_TOOLTIPS_KEY     "tooltips"
-
-static int first_proto_prefs_page = -1;
+#define E_GUI_PAGE_KEY	        "gui_options_page"
+#define E_GUI_COLUMN_PAGE_KEY   "gui_column_options_page"
+#define E_GUI_FONT_PAGE_KEY     "gui_font_options_page"
+#define E_GUI_STREAM_PAGE_KEY   "gui_tcp_stream_options_page"
+#define E_CAPTURE_PAGE_KEY      "capture_options_page"
+#define E_PRINT_PAGE_KEY        "printer_options_page"
+#define E_NAMERES_PAGE_KEY      "nameres_options_page"
+#define E_TOOLTIPS_KEY          "tooltips"
+#define E_PAGE_MODULE_KEY       "page_module"
 
 /*
  * Keep a static pointer to the notebook to be able to choose the
@@ -115,7 +115,13 @@ struct ct_struct {
   gboolean     is_protocol;
 };
 
-static void
+static guint
+pref_exists(pref_t *pref _U_, gpointer user_data _U_)
+{
+  return 1;
+}
+
+static guint
 pref_show(pref_t *pref, gpointer user_data)
 {
   GtkWidget *main_tb = user_data;
@@ -145,15 +151,15 @@ pref_show(pref_t *pref, gpointer user_data)
     switch (pref->info.base) {
 
     case 10:
-      sprintf(uint_str, "%u", pref->saved_val.uint);
+      g_snprintf(uint_str, 10+1, "%u", pref->saved_val.uint);
       break;
 
     case 8:
-      sprintf(uint_str, "%o", pref->saved_val.uint);
+      g_snprintf(uint_str, 10+1, "%o", pref->saved_val.uint);
       break;
 
     case 16:
-      sprintf(uint_str, "%x", pref->saved_val.uint);
+      g_snprintf(uint_str, 10+1, "%x", pref->saved_val.uint);
       break;
     }
     pref->control = create_preference_entry(main_tb, pref->ordinal,
@@ -199,6 +205,8 @@ pref_show(pref_t *pref, gpointer user_data)
     break;
   }
   g_free(label_string);
+
+  return 0;
 }
 
 #define MAX_TREE_NODE_NAME_LEN 64
@@ -217,6 +225,25 @@ module_prefs_show(module_t *module, gpointer user_data)
   GtkTreeIter      iter;
 #endif
 
+  /*
+   * Is this module a subtree, with modules underneath it?
+   */
+  if (!module->is_subtree) {
+    /*
+     * No.
+     * Does it have any preferences (other than possibly obsolete ones)?
+     */
+    if (prefs_pref_foreach(module, pref_exists, NULL) == 0) {
+      /*
+       * No.  Don't put the module into the preferences window.
+       * XXX - we should do the same for subtrees; if a subtree has
+       * nothing under it that will be displayed, don't put it into
+       * the window.
+       */
+      return;
+    }
+  }
+
   /*
    * Add this module to the tree.
    */
@@ -263,8 +290,7 @@ module_prefs_show(module_t *module, gpointer user_data)
     prefs_module_list_foreach(module->prefs, module_prefs_show, &child_cts);
   } else {
     /*
-     * No.
-     * Create a notebook page for it.
+     * No.  Create a notebook page for it.
      */
 
     /* Frame */
@@ -286,6 +312,9 @@ module_prefs_show(module_t *module, gpointer user_data)
     /* Add items for each of the preferences */
     prefs_pref_foreach(module, pref_show, main_tb);
 
+    /* Associate this module with the page's frame. */
+    OBJECT_SET_DATA(frame, E_PAGE_MODULE_KEY, module);
+
     /* Add the page to the notebook */
     gtk_notebook_append_page(GTK_NOTEBOOK(cts->notebook), frame, NULL);
 
@@ -297,9 +326,6 @@ module_prefs_show(module_t *module, gpointer user_data)
     gtk_tree_store_set(model, &iter, 0, label_str, 1, cts->page, -1);
 #endif
 
-    /* If this is the first protocol page, remember its page number */
-    if (first_proto_prefs_page == -1)
-      first_proto_prefs_page = cts->page;
     cts->page++;
 
     /* Show 'em what we got */
@@ -312,16 +338,17 @@ prefs_cb(GtkWidget *w _U_, gpointer dummy _U_)
 {
   GtkWidget         *top_hb, *bbox, *prefs_nb, *ct_sb, *frame,
                     *ok_bt, *apply_bt, *save_bt, *cancel_bt;
-  GtkWidget         *print_pg, *column_pg, *stream_pg, *gui_pg;
+  GtkWidget         *gui_pg, *gui_font_pg, *gui_column_pg, *gui_stream_pg;
 #ifdef HAVE_LIBPCAP
   GtkWidget         *capture_pg;
 #endif
-  GtkWidget         *nameres_pg;
+  GtkWidget         *print_pg, *nameres_pg;
   gchar             label_str[MAX_TREE_NODE_NAME_LEN];
   struct ct_struct  cts;
 #if GTK_MAJOR_VERSION < 2
   gchar             *label_ptr = label_str;
   GtkCTreeNode      *ct_node;
+  GtkCTreeNode      *ct_base_node;
 #else
   GtkTreeStore      *store;
   GtkTreeSelection  *selection;
@@ -329,6 +356,10 @@ prefs_cb(GtkWidget *w _U_, gpointer dummy _U_)
   GtkTreeViewColumn *column;
   gint              col_offset;
   GtkTreeIter       iter;
+  GtkTreeIter       base_iter;
+#endif
+#if GTK_MAJOR_VERSION < 2
+	static gchar *fixedwidths[] = { "c", "m", NULL };
 #endif
 
   if (prefs_w != NULL) {
@@ -366,6 +397,10 @@ prefs_cb(GtkWidget *w _U_, gpointer dummy _U_)
 
   /* Place a Ctree on the left for preference categories */
   ct_sb = scrolled_window_new(NULL, NULL);
+#if GTK_MAJOR_VERSION >= 2
+  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(ct_sb), 
+                                   GTK_SHADOW_IN);
+#endif
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(ct_sb),
   	GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
   gtk_container_add(GTK_CONTAINER(top_hb), ct_sb);
@@ -407,63 +442,6 @@ prefs_cb(GtkWidget *w _U_, gpointer dummy _U_)
   gtk_container_add(GTK_CONTAINER(top_hb), prefs_nb);
   gtk_widget_show(prefs_nb);
 
-  /* Printing prefs */
-  frame = gtk_frame_new("Printing");
-  gtk_widget_show(GTK_WIDGET(frame));
-  print_pg = printer_prefs_show();
-  gtk_container_add(GTK_CONTAINER(frame), print_pg);
-  OBJECT_SET_DATA(prefs_w, E_PRINT_PAGE_KEY, print_pg);
-  gtk_notebook_append_page (GTK_NOTEBOOK(prefs_nb), frame, NULL);
-  strcpy(label_str, "Printing");
-#if GTK_MAJOR_VERSION < 2
-  ct_node = gtk_ctree_insert_node(GTK_CTREE(cts.tree), NULL, NULL,
-  		&label_ptr, 5, NULL, NULL, NULL, NULL, TRUE, TRUE);
-  gtk_ctree_node_set_row_data(GTK_CTREE(cts.tree), ct_node,
-  		GINT_TO_POINTER(cts.page));
-#else
-  gtk_tree_store_append(store, &iter, NULL);
-  gtk_tree_store_set(store, &iter, 0, label_str, 1, cts.page, -1);
-#endif
-  cts.page++;
-
-  /* Column prefs */
-  frame = gtk_frame_new("Columns");
-  gtk_widget_show(GTK_WIDGET(frame));
-  column_pg = column_prefs_show();
-  gtk_container_add(GTK_CONTAINER(frame), column_pg);
-  OBJECT_SET_DATA(prefs_w, E_COLUMN_PAGE_KEY, column_pg);
-  gtk_notebook_append_page (GTK_NOTEBOOK(prefs_nb), frame, NULL);
-  strcpy(label_str, "Columns");
-#if GTK_MAJOR_VERSION < 2
-  ct_node = gtk_ctree_insert_node(GTK_CTREE(cts.tree), NULL, NULL,
-  		&label_ptr, 5, NULL, NULL, NULL, NULL, TRUE, TRUE);
-  gtk_ctree_node_set_row_data(GTK_CTREE(cts.tree), ct_node,
-  		GINT_TO_POINTER(cts.page));
-#else
-  gtk_tree_store_append(store, &iter, NULL);
-  gtk_tree_store_set(store, &iter, 0, label_str, 1, cts.page, -1);
-#endif
-  cts.page++;
-
-  /* TCP Streams prefs */
-  frame = gtk_frame_new("TCP Streams");
-  gtk_widget_show(GTK_WIDGET(frame));
-  stream_pg = stream_prefs_show();
-  gtk_container_add(GTK_CONTAINER(frame), stream_pg);
-  OBJECT_SET_DATA(prefs_w, E_STREAM_PAGE_KEY, stream_pg);
-  gtk_notebook_append_page (GTK_NOTEBOOK(prefs_nb), frame, NULL);
-  strcpy(label_str, "TCP Streams");
-#if GTK_MAJOR_VERSION < 2
-  ct_node = gtk_ctree_insert_node(GTK_CTREE(cts.tree), NULL, NULL,
-  		&label_ptr, 5, NULL, NULL, NULL, NULL, TRUE, TRUE);
-  gtk_ctree_node_set_row_data(GTK_CTREE(cts.tree), ct_node,
-  		GINT_TO_POINTER(cts.page));
-#else
-  gtk_tree_store_append(store, &iter, NULL);
-  gtk_tree_store_set(store, &iter, 0, label_str, 1, cts.page, -1);
-#endif
-  cts.page++;
-
   /* GUI prefs */
   frame = gtk_frame_new("User Interface");
   gtk_widget_show(GTK_WIDGET(frame));
@@ -474,11 +452,102 @@ prefs_cb(GtkWidget *w _U_, gpointer dummy _U_)
   strcpy(label_str, "User Interface");
 #if GTK_MAJOR_VERSION < 2
   ct_node = gtk_ctree_insert_node(GTK_CTREE(cts.tree), NULL, NULL,
+  		&label_ptr, 5, NULL, NULL, NULL, NULL, FALSE /*TRUE*/, TRUE);
+  gtk_ctree_node_set_row_data(GTK_CTREE(cts.tree), ct_node,
+  		GINT_TO_POINTER(cts.page));
+  ct_base_node = ct_node;
+  gtk_ctree_select(GTK_CTREE(cts.tree), ct_node);
+#else
+  gtk_tree_store_append(store, &iter, NULL);
+  gtk_tree_store_set(store, &iter, 0, label_str, 1, cts.page, -1);
+  base_iter = iter;
+  gtk_tree_selection_select_iter(selection, &iter);
+#endif
+  cts.page++;
+
+  /* GUI Column prefs */
+  frame = gtk_frame_new("Columns");
+  gtk_widget_show(GTK_WIDGET(frame));
+  gui_column_pg = column_prefs_show();
+  gtk_container_add(GTK_CONTAINER(frame), gui_column_pg);
+  OBJECT_SET_DATA(prefs_w, E_GUI_COLUMN_PAGE_KEY, gui_column_pg);
+  gtk_notebook_append_page (GTK_NOTEBOOK(prefs_nb), frame, NULL);
+  strcpy(label_str, "Columns");
+#if GTK_MAJOR_VERSION < 2
+  ct_node = gtk_ctree_insert_node(GTK_CTREE(cts.tree), ct_base_node, NULL,
   		&label_ptr, 5, NULL, NULL, NULL, NULL, TRUE, TRUE);
   gtk_ctree_node_set_row_data(GTK_CTREE(cts.tree), ct_node,
   		GINT_TO_POINTER(cts.page));
 #else
-  gtk_tree_store_append(store, &iter, NULL);
+  gtk_tree_store_append(store, &iter, &base_iter);
+  gtk_tree_store_set(store, &iter, 0, label_str, 1, cts.page, -1);
+  /* expand the parent */
+  gtk_tree_view_expand_all(GTK_TREE_VIEW(cts.tree));
+#endif
+  cts.page++;
+
+  /* GUI Font prefs */
+  frame = gtk_frame_new("Font");
+  gtk_widget_show(GTK_WIDGET(frame));
+  gui_font_pg = gui_font_prefs_show();
+  gtk_container_add(GTK_CONTAINER(frame), gui_font_pg);
+  OBJECT_SET_DATA(prefs_w, E_GUI_FONT_PAGE_KEY, gui_font_pg);
+  gtk_notebook_append_page (GTK_NOTEBOOK(prefs_nb), frame, NULL);
+
+  /* We set the current font and, for GTK+ 1.2[.x], the font filter
+     now, because they appear not to work when run before appending
+     the frame to the notebook. */
+
+  /* Set the font to the current font.
+     XXX - GTK+ 1.2.8, and probably earlier versions, have a bug
+     wherein that doesn't necessarily cause that font to be
+     selected in the dialog box.  I've sent to the GTK+ folk
+     a fix; hopefully, it'll show up in 1.2.9 if, as, and when
+     they put out a 1.2.9 release. */
+  gtk_font_selection_set_font_name(
+	    GTK_FONT_SELECTION(gui_font_pg), prefs.PREFS_GUI_FONT_NAME);
+
+#if GTK_MAJOR_VERSION < 2
+  /* Set its filter to show only fixed_width fonts. */
+  gtk_font_selection_set_filter(
+	    GTK_FONT_SELECTION(gui_font_pg),
+	    GTK_FONT_FILTER_BASE, /* user can't change the filter */
+	    GTK_FONT_ALL,	  /* bitmap or scalable are fine */
+	    NULL,		  /* all foundries are OK */
+	    NULL,		  /* all weights are OK (XXX - normal only?) */
+	    NULL,		  /* all slants are OK (XXX - Roman only?) */
+	    NULL,		  /* all setwidths are OK */
+	    fixedwidths,	  /* ONLY fixed-width fonts */
+	    NULL);	/* all charsets are OK (XXX - ISO 8859/1 only?) */
+#endif  
+  
+  strcpy(label_str, "Font");
+#if GTK_MAJOR_VERSION < 2
+  ct_node = gtk_ctree_insert_node(GTK_CTREE(cts.tree), ct_base_node, NULL,
+  		&label_ptr, 5, NULL, NULL, NULL, NULL, TRUE, TRUE);
+  gtk_ctree_node_set_row_data(GTK_CTREE(cts.tree), ct_node,
+  		GINT_TO_POINTER(cts.page));
+#else
+  gtk_tree_store_append(store, &iter, &base_iter);
+  gtk_tree_store_set(store, &iter, 0, label_str, 1, cts.page, -1);
+#endif
+  cts.page++;
+
+  /* GUI Colors prefs */
+  frame = gtk_frame_new("Colors");
+  gtk_widget_show(GTK_WIDGET(frame));
+  gui_stream_pg = stream_prefs_show();
+  gtk_container_add(GTK_CONTAINER(frame), gui_stream_pg);
+  OBJECT_SET_DATA(prefs_w, E_GUI_STREAM_PAGE_KEY, gui_stream_pg);
+  gtk_notebook_append_page (GTK_NOTEBOOK(prefs_nb), frame, NULL);
+  strcpy(label_str, "Colors");
+#if GTK_MAJOR_VERSION < 2
+  ct_node = gtk_ctree_insert_node(GTK_CTREE(cts.tree), ct_base_node, NULL,
+  		&label_ptr, 5, NULL, NULL, NULL, NULL, TRUE, TRUE);
+  gtk_ctree_node_set_row_data(GTK_CTREE(cts.tree), ct_node,
+  		GINT_TO_POINTER(cts.page));
+#else
+  gtk_tree_store_append(store, &iter, &base_iter);
   gtk_tree_store_set(store, &iter, 0, label_str, 1, cts.page, -1);
 #endif
   cts.page++;
@@ -511,14 +580,33 @@ prefs_cb(GtkWidget *w _U_, gpointer dummy _U_)
 #endif /* _WIN32 */
 #endif /* HAVE_LIBPCAP */
 
+  /* Printing prefs */
+  frame = gtk_frame_new("Printing");
+  gtk_widget_show(GTK_WIDGET(frame));
+  print_pg = printer_prefs_show();
+  gtk_container_add(GTK_CONTAINER(frame), print_pg);
+  OBJECT_SET_DATA(prefs_w, E_PRINT_PAGE_KEY, print_pg);
+  gtk_notebook_append_page (GTK_NOTEBOOK(prefs_nb), frame, NULL);
+  strcpy(label_str, "Printing");
+#if GTK_MAJOR_VERSION < 2
+  ct_node = gtk_ctree_insert_node(GTK_CTREE(cts.tree), NULL, NULL,
+  		&label_ptr, 5, NULL, NULL, NULL, NULL, TRUE, TRUE);
+  gtk_ctree_node_set_row_data(GTK_CTREE(cts.tree), ct_node,
+  		GINT_TO_POINTER(cts.page));
+#else
+  gtk_tree_store_append(store, &iter, NULL);
+  gtk_tree_store_set(store, &iter, 0, label_str, 1, cts.page, -1);
+#endif
+  cts.page++;
+
   /* Name resolution prefs */
-  frame = gtk_frame_new("Name resolution");
+  frame = gtk_frame_new("Name Resolution");
   gtk_widget_show(GTK_WIDGET(frame));
   nameres_pg = nameres_prefs_show();
   gtk_container_add(GTK_CONTAINER(frame), nameres_pg);
   OBJECT_SET_DATA(prefs_w, E_NAMERES_PAGE_KEY, nameres_pg);
   gtk_notebook_append_page (GTK_NOTEBOOK(prefs_nb), frame, NULL);
-  strcpy(label_str, "Name resolution");
+  strcpy(label_str, "Name Resolution");
 #if GTK_MAJOR_VERSION < 2
   ct_node = gtk_ctree_insert_node(GTK_CTREE(cts.tree), NULL, NULL,
   		&label_ptr, 5, NULL, NULL, NULL, NULL, TRUE, TRUE);
@@ -536,52 +624,22 @@ prefs_cb(GtkWidget *w _U_, gpointer dummy _U_)
   prefs_module_list_foreach(NULL, module_prefs_show, &cts);
 
   /* Button row: OK and cancel buttons */
-  bbox = gtk_hbutton_box_new();
-  gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_END);
-  gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 5);
-  gtk_container_add(GTK_CONTAINER(cts.main_vb), bbox);
+  bbox = dlg_button_row_new(GTK_STOCK_OK, GTK_STOCK_APPLY, GTK_STOCK_SAVE, GTK_STOCK_CANCEL, NULL);
+  gtk_box_pack_start(GTK_BOX(cts.main_vb), bbox, FALSE, FALSE, 0);
   gtk_widget_show(bbox);
 
-#if GTK_MAJOR_VERSION < 2
-  ok_bt = gtk_button_new_with_label ("OK");
-#else
-  ok_bt = gtk_button_new_from_stock(GTK_STOCK_OK);
-#endif
+  ok_bt = OBJECT_GET_DATA(bbox, GTK_STOCK_OK);
   SIGNAL_CONNECT(ok_bt, "clicked", prefs_main_ok_cb, prefs_w);
-  GTK_WIDGET_SET_FLAGS(ok_bt, GTK_CAN_DEFAULT);
-  gtk_box_pack_start (GTK_BOX (bbox), ok_bt, TRUE, TRUE, 0);
   gtk_widget_grab_default(ok_bt);
-  gtk_widget_show(ok_bt);
 
-#if GTK_MAJOR_VERSION < 2
-  apply_bt = gtk_button_new_with_label ("Apply");
-#else
-  apply_bt = gtk_button_new_from_stock(GTK_STOCK_APPLY);
-#endif
+  apply_bt = OBJECT_GET_DATA(bbox, GTK_STOCK_APPLY);
   SIGNAL_CONNECT(apply_bt, "clicked", prefs_main_apply_cb, prefs_w);
-  GTK_WIDGET_SET_FLAGS(apply_bt, GTK_CAN_DEFAULT);
-  gtk_box_pack_start(GTK_BOX (bbox), apply_bt, TRUE, TRUE, 0);
-  gtk_widget_show(apply_bt);
 
-#if GTK_MAJOR_VERSION < 2
-  save_bt = gtk_button_new_with_label ("Save");
-#else
-  save_bt = gtk_button_new_from_stock(GTK_STOCK_SAVE);
-#endif
+  save_bt = OBJECT_GET_DATA(bbox, GTK_STOCK_SAVE);
   SIGNAL_CONNECT(save_bt, "clicked", prefs_main_save_cb, prefs_w);
-  GTK_WIDGET_SET_FLAGS(save_bt, GTK_CAN_DEFAULT);
-  gtk_box_pack_start (GTK_BOX (bbox), save_bt, TRUE, TRUE, 0);
-  gtk_widget_show(save_bt);
 
-#if GTK_MAJOR_VERSION < 2
-  cancel_bt = gtk_button_new_with_label ("Cancel");
-#else
-  cancel_bt = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
-#endif
+  cancel_bt = OBJECT_GET_DATA(bbox, GTK_STOCK_CANCEL);
   SIGNAL_CONNECT(cancel_bt, "clicked", prefs_main_cancel_cb, prefs_w);
-  GTK_WIDGET_SET_FLAGS(cancel_bt, GTK_CAN_DEFAULT);
-  gtk_box_pack_start (GTK_BOX (bbox), cancel_bt, TRUE, TRUE, 0);
-  gtk_widget_show(cancel_bt);
 
   /* Catch the "key_press_event" signal in the window, so that we can catch
      the ESC key being pressed and act as if the "Cancel" button had
@@ -661,8 +719,7 @@ create_preference_radio_buttons(GtkWidget *main_tb, int table_position,
 		button = gtk_radio_button_new_with_label(rb_group,
 		    enum_valp->name);
 		gtk_widget_show(button);
-		if (rb_group == NULL)
-			rb_group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
+		rb_group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
 		gtk_box_pack_start(GTK_BOX(radio_button_hbox), button, FALSE,
 		    FALSE, 10);
 		if (enum_valp->value == current_val) {
@@ -819,10 +876,10 @@ create_preference_entry(GtkWidget *main_tb, int table_position,
 	return entry;
 }
 
-static void
+static guint
 pref_fetch(pref_t *pref, gpointer user_data)
 {
-  char *str_val;
+  const char *str_val;
   char *p;
   guint uval;
   gboolean bval;
@@ -871,10 +928,9 @@ pref_fetch(pref_t *pref, gpointer user_data)
 
   case PREF_STRING:
     str_val = gtk_entry_get_text(GTK_ENTRY(pref->control));
-    if (*pref->varp.string == NULL || strcmp(*pref->varp.string, str_val) != 0) {
+    if (strcmp(*pref->varp.string, str_val) != 0) {
       *pref_changed_p = TRUE;
-      if (*pref->varp.string != NULL)
-        g_free(*pref->varp.string);
+      g_free(*pref->varp.string);
       *pref->varp.string = g_strdup(str_val);
     }
     break;
@@ -883,6 +939,7 @@ pref_fetch(pref_t *pref, gpointer user_data)
     g_assert_not_reached();
     break;
   }
+  return 0;
 }
 
 static void
@@ -902,7 +959,7 @@ module_prefs_fetch(module_t *module, gpointer user_data)
     *must_redissect_p = TRUE;
 }
 
-static void
+static guint
 pref_clean(pref_t *pref, gpointer user_data _U_)
 {
   switch (pref->type) {
@@ -927,6 +984,7 @@ pref_clean(pref_t *pref, gpointer user_data _U_)
     g_assert_not_reached();
     break;
   }
+  return 0;
 }
 
 static void
@@ -946,8 +1004,8 @@ prefs_main_ok_cb(GtkWidget *ok_bt _U_, gpointer parent_w)
      the preferences panes have been copied to "prefs" and the registered
      preferences). */
   printer_prefs_fetch(OBJECT_GET_DATA(parent_w, E_PRINT_PAGE_KEY));
-  column_prefs_fetch(OBJECT_GET_DATA(parent_w, E_COLUMN_PAGE_KEY));
-  stream_prefs_fetch(OBJECT_GET_DATA(parent_w, E_STREAM_PAGE_KEY));
+  column_prefs_fetch(OBJECT_GET_DATA(parent_w, E_GUI_COLUMN_PAGE_KEY));
+  stream_prefs_fetch(OBJECT_GET_DATA(parent_w, E_GUI_STREAM_PAGE_KEY));
   gui_prefs_fetch(OBJECT_GET_DATA(parent_w, E_GUI_PAGE_KEY));
 #ifdef HAVE_LIBPCAP
 #ifdef _WIN32
@@ -964,8 +1022,8 @@ prefs_main_ok_cb(GtkWidget *ok_bt _U_, gpointer parent_w)
 
   /* Now apply those preferences. */
   printer_prefs_apply(OBJECT_GET_DATA(parent_w, E_PRINT_PAGE_KEY));
-  column_prefs_apply(OBJECT_GET_DATA(parent_w, E_COLUMN_PAGE_KEY));
-  stream_prefs_apply(OBJECT_GET_DATA(parent_w, E_STREAM_PAGE_KEY));
+  column_prefs_apply(OBJECT_GET_DATA(parent_w, E_GUI_COLUMN_PAGE_KEY));
+  stream_prefs_apply(OBJECT_GET_DATA(parent_w, E_GUI_STREAM_PAGE_KEY));
   gui_prefs_apply(OBJECT_GET_DATA(parent_w, E_GUI_PAGE_KEY));
 #ifdef HAVE_LIBPCAP
 #ifdef _WIN32
@@ -998,8 +1056,8 @@ prefs_main_apply_cb(GtkWidget *apply_bt _U_, gpointer parent_w)
      the preferences panes have been copied to "prefs" and the registered
      preferences). */
   printer_prefs_fetch(OBJECT_GET_DATA(parent_w, E_PRINT_PAGE_KEY));
-  column_prefs_fetch(OBJECT_GET_DATA(parent_w, E_COLUMN_PAGE_KEY));
-  stream_prefs_fetch(OBJECT_GET_DATA(parent_w, E_STREAM_PAGE_KEY));
+  column_prefs_fetch(OBJECT_GET_DATA(parent_w, E_GUI_COLUMN_PAGE_KEY));
+  stream_prefs_fetch(OBJECT_GET_DATA(parent_w, E_GUI_STREAM_PAGE_KEY));
   gui_prefs_fetch(OBJECT_GET_DATA(parent_w, E_GUI_PAGE_KEY));
 #ifdef HAVE_LIBPCAP
 #ifdef _WIN32
@@ -1016,8 +1074,8 @@ prefs_main_apply_cb(GtkWidget *apply_bt _U_, gpointer parent_w)
 
   /* Now apply those preferences. */
   printer_prefs_apply(OBJECT_GET_DATA(parent_w, E_PRINT_PAGE_KEY));
-  column_prefs_apply(OBJECT_GET_DATA(parent_w, E_COLUMN_PAGE_KEY));
-  stream_prefs_apply(OBJECT_GET_DATA(parent_w, E_STREAM_PAGE_KEY));
+  column_prefs_apply(OBJECT_GET_DATA(parent_w, E_GUI_COLUMN_PAGE_KEY));
+  stream_prefs_apply(OBJECT_GET_DATA(parent_w, E_GUI_STREAM_PAGE_KEY));
   gui_prefs_apply(OBJECT_GET_DATA(parent_w, E_GUI_PAGE_KEY));
 #ifdef HAVE_LIBPCAP
 #ifdef _WIN32
@@ -1050,8 +1108,8 @@ prefs_main_save_cb(GtkWidget *save_bt _U_, gpointer parent_w)
      the preferences panes have been copied to "prefs" and the registered
      preferences). */
   printer_prefs_fetch(OBJECT_GET_DATA(parent_w, E_PRINT_PAGE_KEY));
-  column_prefs_fetch(OBJECT_GET_DATA(parent_w, E_COLUMN_PAGE_KEY));
-  stream_prefs_fetch(OBJECT_GET_DATA(parent_w, E_STREAM_PAGE_KEY));
+  column_prefs_fetch(OBJECT_GET_DATA(parent_w, E_GUI_COLUMN_PAGE_KEY));
+  stream_prefs_fetch(OBJECT_GET_DATA(parent_w, E_GUI_STREAM_PAGE_KEY));
   gui_prefs_fetch(OBJECT_GET_DATA(parent_w, E_GUI_PAGE_KEY));
 #ifdef HAVE_LIBPCAP
 #ifdef _WIN32
@@ -1069,7 +1127,7 @@ prefs_main_save_cb(GtkWidget *save_bt _U_, gpointer parent_w)
   /* Create the directory that holds personal configuration files, if
      necessary.  */
   if (create_persconffile_dir(&pf_dir_path) == -1) {
-     simple_dialog(ESD_TYPE_WARN, NULL,
+     simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK,
       "Can't create directory\n\"%s\"\nfor preferences file: %s.", pf_dir_path,
       strerror(errno));
      g_free(pf_dir_path);
@@ -1077,7 +1135,7 @@ prefs_main_save_cb(GtkWidget *save_bt _U_, gpointer parent_w)
     /* Write the preferencs out. */
     err = write_prefs(&pf_path);
     if (err != 0) {
-       simple_dialog(ESD_TYPE_WARN, NULL,
+       simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK,
         "Can't open preferences file\n\"%s\": %s.", pf_path,
         strerror(err));
        g_free(pf_path);
@@ -1099,8 +1157,8 @@ prefs_main_save_cb(GtkWidget *save_bt _U_, gpointer parent_w)
 
 	4) we did apply the protocol preferences, at least, in the past. */
   printer_prefs_apply(OBJECT_GET_DATA(parent_w, E_PRINT_PAGE_KEY));
-  column_prefs_apply(OBJECT_GET_DATA(parent_w, E_COLUMN_PAGE_KEY));
-  stream_prefs_apply(OBJECT_GET_DATA(parent_w, E_STREAM_PAGE_KEY));
+  column_prefs_apply(OBJECT_GET_DATA(parent_w, E_GUI_COLUMN_PAGE_KEY));
+  stream_prefs_apply(OBJECT_GET_DATA(parent_w, E_GUI_STREAM_PAGE_KEY));
   gui_prefs_apply(OBJECT_GET_DATA(parent_w, E_GUI_PAGE_KEY));
 #ifdef HAVE_LIBPCAP
 #ifdef _WIN32
@@ -1121,7 +1179,7 @@ prefs_main_save_cb(GtkWidget *save_bt _U_, gpointer parent_w)
   }
 }
 
-static void
+static guint
 pref_revert(pref_t *pref, gpointer user_data)
 {
   gboolean *pref_changed_p = user_data;
@@ -1151,13 +1209,9 @@ pref_revert(pref_t *pref, gpointer user_data)
     break;
 
   case PREF_STRING:
-    if (*pref->varp.string != pref->saved_val.string &&
-	(*pref->varp.string == NULL ||
-	 pref->saved_val.string == NULL ||
-	 strcmp(*pref->varp.string, pref->saved_val.string) != 0)) {
+    if (strcmp(*pref->varp.string, pref->saved_val.string) != 0) {
       *pref_changed_p = TRUE;
-      if (*pref->varp.string != NULL)
-        g_free(*pref->varp.string);
+      g_free(*pref->varp.string);
       *pref->varp.string = g_strdup(pref->saved_val.string);
     }
     break;
@@ -1166,6 +1220,7 @@ pref_revert(pref_t *pref, gpointer user_data)
     g_assert_not_reached();
     break;
   }
+  return 0;
 }
 
 static void
@@ -1201,8 +1256,8 @@ prefs_main_cancel_cb(GtkWidget *cancel_bt _U_, gpointer parent_w)
 
   /* Now apply the reverted-to preferences. */
   printer_prefs_apply(OBJECT_GET_DATA(parent_w, E_PRINT_PAGE_KEY));
-  column_prefs_apply(OBJECT_GET_DATA(parent_w, E_COLUMN_PAGE_KEY));
-  stream_prefs_apply(OBJECT_GET_DATA(parent_w, E_STREAM_PAGE_KEY));
+  column_prefs_apply(OBJECT_GET_DATA(parent_w, E_GUI_COLUMN_PAGE_KEY));
+  stream_prefs_apply(OBJECT_GET_DATA(parent_w, E_GUI_STREAM_PAGE_KEY));
   gui_prefs_apply(OBJECT_GET_DATA(parent_w, E_GUI_PAGE_KEY));
   nameres_prefs_apply(OBJECT_GET_DATA(parent_w, E_NAMERES_PAGE_KEY));
   prefs_apply_all();
@@ -1231,8 +1286,8 @@ prefs_main_destroy_cb(GtkWidget *win _U_, gpointer user_data _U_)
 {
   /* Let the preference tabs clean up anything they've done. */
   printer_prefs_destroy(OBJECT_GET_DATA(prefs_w, E_PRINT_PAGE_KEY));
-  column_prefs_destroy(OBJECT_GET_DATA(prefs_w, E_COLUMN_PAGE_KEY));
-  stream_prefs_destroy(OBJECT_GET_DATA(prefs_w, E_STREAM_PAGE_KEY));
+  column_prefs_destroy(OBJECT_GET_DATA(prefs_w, E_GUI_COLUMN_PAGE_KEY));
+  stream_prefs_destroy(OBJECT_GET_DATA(prefs_w, E_GUI_STREAM_PAGE_KEY));
   gui_prefs_destroy(OBJECT_GET_DATA(prefs_w, E_GUI_PAGE_KEY));
 #ifdef HAVE_LIBPCAP
 #ifdef _WIN32
@@ -1253,63 +1308,80 @@ prefs_main_destroy_cb(GtkWidget *win _U_, gpointer user_data _U_)
 
   /* Note that we no longer have a "Preferences" dialog box. */
   prefs_w = NULL;
-  first_proto_prefs_page = -1;
 }
 
 struct properties_data {
-  GtkWidget *w;
-  int page_num;
   const char *title;
+  module_t *module;
 };
 
-/* XXX this way of searching the correct page number is really ugly ... */
 static void
 module_search_properties(module_t *module, gpointer user_data)
 {
   struct properties_data *p = (struct properties_data *)user_data;
 
-  if (p->title == NULL) return;
-  if (strcmp(module->title, p->title) == 0) {
-    /* found it */
-    gtk_notebook_set_page(GTK_NOTEBOOK(p->w), p->page_num);
-    p->title = NULL;
-  } else {
-    p->page_num++;
-  }
+  /* If this module has the specified title, remember it. */
+  if (strcmp(module->title, p->title) == 0)
+    p->module = module;
 }
 
 void
 properties_cb(GtkWidget *w, gpointer dummy)
 {
-  const gchar *title = NULL;
+  header_field_info *hfinfo;
+  const gchar *title;
   struct properties_data p;
+  int page_num;
+  GtkWidget *frame;
+  module_t *page_module;
 
-  if (finfo_selected) {
-    header_field_info *hfinfo = finfo_selected->hfinfo;
-    if (hfinfo->parent == -1) {
-      title = prefs_get_title_by_name(hfinfo->abbrev);
-    } else {
-      title =
-	prefs_get_title_by_name(proto_registrar_get_abbrev(hfinfo->parent));
-    }
-  } else {
+  if (cfile.finfo_selected == NULL) {
+    /* There is no field selected */
     return;
   }
 
-  if (!title) return;
+  /* Find the title for the protocol for the selected field. */
+  hfinfo = cfile.finfo_selected->hfinfo;
+  if (hfinfo->parent == -1)
+    title = prefs_get_title_by_name(hfinfo->abbrev);
+  else
+    title = prefs_get_title_by_name(proto_registrar_get_abbrev(hfinfo->parent));
+  if (!title)
+    return;	/* Couldn't find it. XXX - just crash? "Can't happen"? */
 
+  /* Find the module for that protocol by searching for one with that title.
+     XXX - should we just associate protocols with modules directly? */
+  p.title = title;
+  p.module = NULL;
+  prefs_module_list_foreach(protocols_module->prefs, module_search_properties,
+                            &p);
+  if (p.module == NULL) {
+    /* We didn't find it - that protocol probably has no preferences. */
+    return;
+  }
+
+  /* Create a preferences window, or pop up an existing one. */
   if (prefs_w != NULL) {
     reactivate_window(prefs_w);
   } else {
     prefs_cb(w, dummy);
   }
 
-  p.w = notebook;
-  p.page_num = first_proto_prefs_page;
-  p.title = title;
-
-  prefs_module_list_foreach(protocols_module->prefs, module_search_properties,
-      &p);
+  /* Search all the pages in that window for the one with the specified
+     module. */
+  for (page_num = 0;
+       (frame = gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page_num)) != NULL;
+       page_num++) {
+    /* Get the module for this page. */
+    page_module = OBJECT_GET_DATA(frame, E_PAGE_MODULE_KEY);
+    if (page_module == NULL)
+      continue;	/* It doesn't have one. */
+    if (page_module == p.module) {
+      /* We found it.  Select that page. */
+      gtk_notebook_set_page(GTK_NOTEBOOK(notebook), page_num);
+      break;
+    }
+  }
 }
 
 /* Prefs tree selection callback.  The node data has been loaded with
