@@ -1,7 +1,7 @@
 /* proto_draw.c
  * Routines for GTK+ packet display
  *
- * $Id: proto_draw.c,v 1.61 2002/11/11 15:39:06 oabad Exp $
+ * $Id: proto_draw.c,v 1.90 2004/03/20 06:34:09 guy Exp $
  *
  * Ethereal - Network traffic analyzer
  * By Gerald Combs <gerald@ethereal.com>
@@ -31,9 +31,23 @@
 
 #include <ctype.h>
 
+#ifdef HAVE_IO_H
+#include <io.h>			/* open/close on win32 */
+#endif
+
+#ifdef HAVE_FCNTL_H
+#include <fcntl.h>
+#endif
+
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
 #include <gtk/gtk.h>
 #include <gdk/gdkkeysyms.h>
 
+#include <string.h>
+
 #include <epan/epan_dissect.h>
 
 #include "main.h"
@@ -44,11 +58,21 @@
 
 #include "colors.h"
 #include "prefs.h"
+#include "file_dlg.h"
 #include "proto_draw.h"
 #include "packet_win.h"
+#include "dlg_utils.h"
 #include "ui_util.h"
 #include "gtkglobals.h"
 #include "compat_macros.h"
+#include "alert_box.h"
+#include "simple_dialog.h"
+
+
+/* Win32 needs the O_BINARY flag for open() */
+#ifndef O_BINARY
+#define O_BINARY	0
+#endif
 
 #define BYTE_VIEW_WIDTH    16
 #define BYTE_VIEW_SEP      8
@@ -61,12 +85,52 @@
 #define E_BYTE_VIEW_END_KEY       "byte_view_end"
 #define E_BYTE_VIEW_ENCODE_KEY    "byte_view_encode"
 
+/* gtk_tree_view_expand_to_path doesn't exist in gtk+ v2.0 so we must include it
+ * when building with this version (taken from gtk+ v2.2.4) */
+#if GTK_MAJOR_VERSION >= 2 && GTK_MINOR_VERSION == 0
+/**
+ * gtk_tree_view_expand_to_path:
+ * @tree_view: A #GtkTreeView.
+ * @path: path to a row.
+ *
+ * Expands the row at @path. This will also expand all parent rows of
+ * @path as necessary.
+ *
+ * Since: 2.2
+ **/
+void
+gtk_tree_view_expand_to_path (GtkTreeView *tree_view,
+                              GtkTreePath *path)
+{
+  gint i, depth;
+  gint *indices;
+  GtkTreePath *tmp;
+
+  g_return_if_fail (GTK_IS_TREE_VIEW (tree_view));
+  g_return_if_fail (path != NULL);
+
+  depth = gtk_tree_path_get_depth (path);
+  indices = gtk_tree_path_get_indices (path);
+
+  tmp = gtk_tree_path_new ();
+  g_return_if_fail (tmp != NULL);
+
+  for (i = 0; i < depth; i++)
+    {
+      gtk_tree_path_append_index (tmp, indices[i]);
+      gtk_tree_view_expand_row (tree_view, tmp, FALSE);
+    }
+
+  gtk_tree_path_free (tmp);
+}
+#endif
+
 static GtkWidget *
 add_byte_tab(GtkWidget *byte_nb, const char *name, tvbuff_t *tvb,
     proto_tree *tree, GtkWidget *tree_view);
 
 static void
-proto_tree_draw_node(GNode *node, gpointer data);
+proto_tree_draw_node(proto_node *node, gpointer data);
 
 /* Get the current text window for the notebook. */
 GtkWidget *
@@ -151,7 +215,7 @@ void
 redraw_hex_dump_all(void)
 {
   if (cfile.current_frame != NULL)
-    redraw_hex_dump( byte_nb_ptr, cfile.current_frame, finfo_selected);
+    redraw_hex_dump( byte_nb_ptr, cfile.current_frame, cfile.finfo_selected);
 
   redraw_hex_dump_packet_wins();
 }
@@ -273,12 +337,26 @@ lookup_finfo(GtkTreeModel *model, GtkTreePath *path _U_, GtkTreeIter *iter,
     }
     return FALSE;
 }
+
+GtkTreePath *tree_find_by_field_info(GtkTreeView *tree_view, field_info *finfo) {
+  GtkTreeModel *model;
+  struct field_lookup_info fli;
+
+  model = gtk_tree_view_get_model(tree_view);
+  fli.fi = finfo;
+  gtk_tree_model_foreach(model, lookup_finfo, &fli);
+
+  return gtk_tree_model_get_path(model, &fli.iter);
+}
+
 #endif
 
 /* If the user selected a certain byte in the byte view, try to find
- * the item in the GUI proto_tree that corresponds to that byte, and
- * select it. */
-static gint
+ * the item in the GUI proto_tree that corresponds to that byte, and:
+ *
+ *	if we succeed, select it, and return TRUE;
+ *	if we fail, return FALSE. */
+gboolean
 byte_view_select(GtkWidget *widget, GdkEventButton *event)
 {
     proto_tree   *tree;
@@ -403,13 +481,13 @@ byte_view_select(GtkWidget *widget, GdkEventButton *event)
      * which text column/row the user selected. This could be off
      * if the bold version of the font is bigger than the
      * regular version of the font. */
-    column = (bv->hadj->value + event->x) / m_font_width;
-    row = (bv->vadj->value + event->y) / m_font_height;
+    column = (int) ((bv->hadj->value + event->x) / m_font_width);
+    row = (int) ((bv->vadj->value + event->y) / m_font_height);
 #else
     /* get the row/column selected */
     gtk_text_view_window_to_buffer_coords(bv,
                          gtk_text_view_get_window_type(bv, event->window),
-                         event->x, event->y, &x, &y);
+                         (gint) event->x, (gint) event->y, &x, &y);
     gtk_text_view_get_iter_at_location(bv, &iter, x, y);
     row = gtk_text_iter_get_line(&iter);
     column = gtk_text_iter_get_line_offset(&iter);
@@ -525,6 +603,8 @@ byte_view_button_press_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
 	if(event->type == GDK_BUTTON_PRESS) {
 		event_button = (GdkEventButton *) event;
 
+        /* To qoute the "Gdk Event Structures" doc:
+         * "Normally button 1 is the left mouse button, 2 is the middle button, and 3 is the right button" */
 		switch(event_button->button) {
 
 		case 1:
@@ -540,20 +620,15 @@ byte_view_button_press_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
 }
 
 GtkWidget *
-create_byte_view(gint bv_size, GtkWidget *pane)
+byte_view_new(void)
 {
   GtkWidget *byte_nb;
 
   byte_nb = gtk_notebook_new();
   gtk_notebook_set_tab_pos(GTK_NOTEBOOK(byte_nb), GTK_POS_BOTTOM);
 
-  gtk_paned_pack2(GTK_PANED(pane), byte_nb, FALSE, FALSE);
-#if GTK_MAJOR_VERSION < 2
-  WIDGET_SET_SIZE(byte_nb, -1, bv_size);
-#else
-  gtk_widget_set_size_request(byte_nb, -1, bv_size);
-#endif
-  gtk_widget_show(byte_nb);
+  /* this will only have an effect, if no tabs are shown */
+  gtk_notebook_set_show_border(GTK_NOTEBOOK(byte_nb), FALSE);
 
   /* Add a placeholder byte view so that there's at least something
      displayed in the byte view notebook. */
@@ -594,11 +669,6 @@ add_byte_tab(GtkWidget *byte_nb, const char *name, tvbuff_t *tvb,
 
   /* Byte view.  Create a scrolled window for the text. */
   byte_scrollw = scrolled_window_new(NULL, NULL);
-
-  /* Add scrolled pane to tabbed window */
-  label = gtk_label_new(name);
-  gtk_notebook_append_page(GTK_NOTEBOOK(byte_nb), byte_scrollw, label);
-
 #if GTK_MAJOR_VERSION < 2
   /* The horizontal scrollbar of the scroll-window doesn't seem
    * to affect the GtkText widget at all, even when line wrapping
@@ -608,10 +678,13 @@ add_byte_tab(GtkWidget *byte_nb, const char *name, tvbuff_t *tvb,
 			/* Horizontal */GTK_POLICY_NEVER,
 			/* Vertical*/	GTK_POLICY_ALWAYS);
 #else
-  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(byte_scrollw),
-                                 GTK_POLICY_AUTOMATIC,
-                                 GTK_POLICY_AUTOMATIC);
+  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(byte_scrollw), 
+                                   GTK_SHADOW_IN);
 #endif
+  /* Add scrolled pane to tabbed window */
+  label = gtk_label_new(name);
+  gtk_notebook_append_page(GTK_NOTEBOOK(byte_nb), byte_scrollw, label);
+
   gtk_widget_show(byte_scrollw);
 
 #if GTK_MAJOR_VERSION < 2
@@ -689,6 +762,200 @@ add_byte_views(epan_dissect_t *edt, GtkWidget *tree_view,
 	gtk_notebook_set_page(GTK_NOTEBOOK(byte_nb_ptr), 0);
 }
 
+static GtkWidget *savehex_dlg=NULL;
+static GtkWidget *file_entry=NULL;
+
+static void
+savehex_dlg_destroy_cb(void)
+{
+        savehex_dlg = NULL;
+}
+
+static void
+savehex_dlg_cancel_cb(GtkWidget *cancel_bt _U_, gpointer parent_w)
+{
+	gtk_widget_destroy(GTK_WIDGET(parent_w));
+}
+
+/* Forward declaration */
+static void
+savehex_save_clicked_cb(GtkWidget * w, gpointer data);
+
+/* Launch the dialog box to put up the file selection box etc */
+void savehex_cb(GtkWidget * w _U_, gpointer data _U_)
+{
+	int start, end, len;
+	const guint8 *data_p = NULL;
+    gchar label[200];
+
+    GtkWidget   *bv;
+	GtkWidget   *dlg_box, *dlg_lb;
+	GtkWidget   *file_box, *file_bt = NULL;
+	GtkWidget   *bbox, *save_button, *cancel_button;
+	GtkTooltips *tooltips;
+
+
+    /* don't show up the dialog, if no data has to be saved */
+	bv = get_notebook_bv_ptr(byte_nb_ptr);
+	if (bv == NULL) {
+		/* shouldn't happen */
+		simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK, "Could not find the corresponding text window!");
+		return;
+	}
+	end = GPOINTER_TO_INT(OBJECT_GET_DATA(bv, E_BYTE_VIEW_START_KEY));
+	start = GPOINTER_TO_INT(OBJECT_GET_DATA(bv, E_BYTE_VIEW_END_KEY));
+	data_p = get_byte_view_data_and_length(GTK_WIDGET(bv), &len);
+
+	if (data_p == NULL || start == -1 || start > end) {
+		simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK, "No data selected to save!");
+		return;
+	}
+
+    /* if the window is already open, bring it to front */
+	if(savehex_dlg){
+		gdk_window_raise(savehex_dlg->window);
+		return;
+	}
+
+	tooltips = gtk_tooltips_new();
+
+	/*
+	 * Build the dialog box we need ... a text entry field and a
+	 * browse button, along with OK and Cancel
+	 */
+
+	savehex_dlg=dlg_window_new("Ethereal: Export Selected Packet Bytes");
+	SIGNAL_CONNECT(savehex_dlg, "destroy", savehex_dlg_destroy_cb, NULL);
+
+	dlg_box=gtk_vbox_new(FALSE, 10);
+	gtk_container_border_width(GTK_CONTAINER(dlg_box), 10);
+	gtk_container_add(GTK_CONTAINER(savehex_dlg), dlg_box);
+	gtk_widget_show(dlg_box);
+
+    /* label */
+    g_snprintf(label, sizeof(label), "Will save %u %s of raw binary data to file:", 
+        end - start, plurality(end - start, "byte", "bytes"));
+    dlg_lb = gtk_label_new(label);
+	gtk_box_pack_start(GTK_BOX(dlg_box), dlg_lb, FALSE, FALSE, 0);
+	gtk_widget_show(dlg_lb);
+
+	/* File entry box */
+	file_box=gtk_hbox_new(FALSE, 3);
+
+	/* File entry */
+	file_entry=gtk_entry_new();
+	WIDGET_SET_SIZE(file_entry, 300, -1);
+	gtk_tooltips_set_tip(tooltips, file_entry, ("Enter Save Data filename"), NULL);
+	gtk_box_pack_start(GTK_BOX(file_box), file_entry, TRUE, TRUE, 0);
+	gtk_widget_show(file_entry);
+
+	/* File Browse button */
+	file_bt=BUTTON_NEW_FROM_STOCK(ETHEREAL_STOCK_BROWSE);
+	SIGNAL_CONNECT(file_bt, "clicked", select_file_cb,
+		       "Ethereal: Export Selected Packet Bytes");
+
+	/* file entry for print dialog */
+	OBJECT_SET_DATA(file_bt, E_FILE_TE_PTR_KEY, file_entry);
+	/* file entry for print dialog */
+
+	gtk_tooltips_set_tip (tooltips, file_bt, ("Browse output filename in filesystem"), NULL);
+	gtk_box_pack_start(GTK_BOX(file_box), file_bt, FALSE, TRUE, 0);
+	gtk_widget_show(file_bt);
+
+	gtk_box_pack_start(GTK_BOX(dlg_box), file_box, TRUE, TRUE, 0);
+	gtk_widget_show(file_box);
+
+	/* Now, the button box */
+    bbox = dlg_button_row_new(GTK_STOCK_SAVE, GTK_STOCK_CANCEL, NULL);
+    gtk_box_pack_start(GTK_BOX(dlg_box), bbox, FALSE, FALSE, 0);
+    gtk_widget_show(bbox);
+
+	/* the save button */
+    save_button = OBJECT_GET_DATA(bbox, GTK_STOCK_SAVE);
+    SIGNAL_CONNECT_OBJECT(save_button, "clicked",
+                              savehex_save_clicked_cb, NULL);
+	gtk_widget_grab_default(save_button);
+	gtk_tooltips_set_tip (tooltips, save_button, ("Save the data to the specified file"), NULL);
+
+	cancel_button = OBJECT_GET_DATA(bbox, GTK_STOCK_CANCEL);
+	SIGNAL_CONNECT(cancel_button, "clicked", savehex_dlg_cancel_cb, savehex_dlg);
+	gtk_tooltips_set_tip (tooltips, cancel_button, ("Cancel save and quit dialog"), NULL);
+
+	/* Catch the "activate" signal on the filter text entry, so that
+	   if the user types Return there, we act as if the "Create Stat"
+	   button had been selected, as happens if Return is typed if some
+	   widget that *doesn't* handle the Return key has the input
+	   focus. */
+	dlg_set_activate(file_entry, save_button);
+
+	/* Catch the "key_press_event" signal in the window, so that we can
+	   catch the ESC key being pressed and act as if the "Cancel" button
+	   had been selected. */
+	dlg_set_cancel(savehex_dlg, cancel_button);
+
+	/* Give the initial focus to the "File" entry box. */
+	gtk_widget_grab_focus(file_entry);
+
+	gtk_widget_show_all(savehex_dlg);
+}
+
+/* save the current highlighted hex data as hex_raw.dat */
+static void
+savehex_save_clicked_cb(GtkWidget * w _U_, gpointer data _U_)
+{
+        GtkWidget *bv;
+	int fd, start, end, len;
+	const guint8 *data_p = NULL;
+	char *file = NULL;
+
+	file = (char *)gtk_entry_get_text(GTK_ENTRY(file_entry));
+
+	if (!file ||! *file) {
+		simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK, "Please enter a filename!");
+		return;
+	}
+
+	/* Must check if file name exists first */
+
+	bv = get_notebook_bv_ptr(byte_nb_ptr);
+	if (bv == NULL) {
+		/* shouldn't happen */
+		simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK, "Could not find the corresponding text window!");
+		return;
+	}
+	/*
+	 * Retrieve the info we need 
+	 */
+	end = GPOINTER_TO_INT(OBJECT_GET_DATA(bv, E_BYTE_VIEW_START_KEY));
+	start = GPOINTER_TO_INT(OBJECT_GET_DATA(bv, E_BYTE_VIEW_END_KEY));
+	data_p = get_byte_view_data_and_length(GTK_WIDGET(bv), &len);
+
+	if (data_p == NULL || start == -1 || start > end) {
+		simple_dialog(ESD_TYPE_ERROR, ESD_BTN_OK,
+		    "No data selected to save!");
+		return;
+	}
+
+	fd = open(file, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0666);
+	if (fd == -1) {
+		open_failure_alert_box(file, errno, TRUE);
+		return;
+	}
+	if (write(fd, data_p + start, end - start) < 0) {
+		write_failure_alert_box(file, errno);
+		close(fd);
+		return;
+	}
+	if (close(fd) < 0) {
+		write_failure_alert_box(file, errno);
+		return;
+	}
+
+	/* Get rid of the dialog box */
+	gtk_widget_destroy(GTK_WIDGET(savehex_dlg));
+}
+
+
 #if GTK_MAJOR_VERSION < 2
 static void
 packet_hex_print_common(GtkText *bv, const guint8 *pd, int len, int bstart,
@@ -1242,9 +1509,8 @@ set_ptree_font_all(PangoFontDescription *font)
 #endif
 }
 
-void
-create_tree_view(gint tv_size, e_prefs *prefs, GtkWidget *pane,
-		GtkWidget **tv_scrollw_p, GtkWidget **tree_view_p)
+GtkWidget *
+main_tree_view_new(e_prefs *prefs, GtkWidget **tree_view_p)
 {
   GtkWidget *tv_scrollw, *tree_view;
 #if GTK_MAJOR_VERSION >= 2
@@ -1256,16 +1522,10 @@ create_tree_view(gint tv_size, e_prefs *prefs, GtkWidget *pane,
 
   /* Tree view */
   tv_scrollw = scrolled_window_new(NULL, NULL);
-#if GTK_MAJOR_VERSION < 2
-  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(tv_scrollw),
-                                  GTK_POLICY_ALWAYS, GTK_POLICY_ALWAYS);
-#else
-  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(tv_scrollw),
-                                  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
+#if GTK_MAJOR_VERSION >= 2
+  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(tv_scrollw), 
+                                   GTK_SHADOW_IN);
 #endif
-  gtk_paned_pack1(GTK_PANED(pane), tv_scrollw, TRUE, TRUE);
-  WIDGET_SET_SIZE(tv_scrollw, -1, tv_size);
-  gtk_widget_show(tv_scrollw);
 
 #if GTK_MAJOR_VERSION < 2
   tree_view = ctree_new(1, 0);
@@ -1301,7 +1561,8 @@ create_tree_view(gint tv_size, e_prefs *prefs, GtkWidget *pane,
   remember_ptree_widget(tree_view);
 
   *tree_view_p = tree_view;
-  *tv_scrollw_p = tv_scrollw;
+
+  return tv_scrollw;
 }
 
 #if GTK_MAJOR_VERSION < 2
@@ -1356,6 +1617,51 @@ main_proto_tree_draw(proto_tree *protocol_tree)
     proto_tree_draw(protocol_tree, tree_view);
 }
 
+/* If the user selected a position in the tree view, try to find
+ * the item in the GUI proto_tree that corresponds to that byte, and
+ * select it. */
+gboolean
+tree_view_select(GtkWidget *widget, GdkEventButton *event)
+{
+#if GTK_MAJOR_VERSION < 2
+        GtkCTree     *ctree;
+        GtkCTreeNode *node;
+        gint         row;
+        gint         column;
+
+
+        if(gtk_clist_get_selection_info(GTK_CLIST(widget), 
+            (gint) (((GdkEventButton *)event)->x),
+            (gint) (((GdkEventButton *)event)->y),
+            &row, &column))
+        {
+            ctree = GTK_CTREE(widget);
+
+            node = gtk_ctree_node_nth(ctree, row);
+            g_assert(node);
+
+            gtk_ctree_select(ctree, node);
+        } else {
+            return FALSE;
+        }
+#else
+        GtkTreeSelection    *sel;
+        GtkTreePath         *path;
+
+        if (gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(widget),
+                                          (gint) (((GdkEventButton *)event)->x),
+                                          (gint) (((GdkEventButton *)event)->y),
+                                          &path, NULL, NULL, NULL))
+        {
+            sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(widget));
+            gtk_tree_selection_select_path(sel, path);
+        } else {
+            return FALSE;
+        }
+#endif
+    return TRUE;
+}
+
 void
 proto_tree_draw(proto_tree *protocol_tree, GtkWidget *tree_view)
 {
@@ -1387,15 +1693,15 @@ proto_tree_draw(proto_tree *protocol_tree, GtkWidget *tree_view)
     gtk_tree_store_clear(store);
 #endif
 
-    g_node_children_foreach((GNode*) protocol_tree, G_TRAVERSE_ALL,
-                            proto_tree_draw_node, &info);
+    proto_tree_children_foreach(protocol_tree, proto_tree_draw_node, &info);
+
 #if GTK_MAJOR_VERSION < 2
     gtk_clist_thaw(GTK_CLIST(tree_view));
 #endif
 }
 
 static void
-proto_tree_draw_node(GNode *node, gpointer data)
+proto_tree_draw_node(proto_node *node, gpointer data)
 {
     struct proto_tree_draw_info	info;
     struct proto_tree_draw_info	*parent_info = (struct proto_tree_draw_info*) data;
@@ -1409,21 +1715,22 @@ proto_tree_draw_node(GNode *node, gpointer data)
 #else
     GtkTreeStore *store;
     GtkTreeIter   iter;
+    GtkTreePath  *path;
 #endif
 
     if (!fi->visible)
         return;
 
     /* was a free format label produced? */
-    if (fi->representation) {
-        label_ptr = fi->representation;
+    if (fi->rep) {
+        label_ptr = fi->rep->representation;
     }
     else { /* no, make a generic label */
         label_ptr = label_str;
         proto_item_fill_label(fi, label_str);
     }
 
-    if (g_node_n_children(node) > 0) {
+    if (node->first_child != NULL) {
         is_leaf = FALSE;
         g_assert(fi->tree_type >= 0 && fi->tree_type < num_tree_types);
         if (tree_is_expanded[fi->tree_type]) {
@@ -1458,18 +1765,20 @@ proto_tree_draw_node(GNode *node, gpointer data)
 #else
         info.iter = &iter;
 #endif
-        g_node_children_foreach(node, G_TRAVERSE_ALL,
-                                proto_tree_draw_node, &info);
-    }
+        proto_tree_children_foreach(node, proto_tree_draw_node, &info);
 #if GTK_MAJOR_VERSION >= 2
-    if (is_expanded == TRUE)
-    {
-        GtkTreePath *path;
         path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter);
-        gtk_tree_view_expand_row(info.tree_view, path, FALSE);
+        if (is_expanded)
+/* #if GTK_MINOR_VERSION >= 2 */
+            gtk_tree_view_expand_to_path(info.tree_view, path);
+/*#else
+            gtk_tree_view_expand_row(info.tree_view, path, FALSE);
+#endif*/
+        else
+            gtk_tree_view_collapse_row(info.tree_view, path);
         gtk_tree_path_free(path);
-    }
 #endif
+    }
 }
 
 /*
@@ -1494,3 +1803,4 @@ clear_tree_and_hex_views(void)
   gtk_tree_store_clear(GTK_TREE_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(tree_view))));
 #endif
 }
+
