@@ -2,7 +2,7 @@
  * Routines for AppleTalk packet disassembly: LLAP, DDP, NBP, ATP, ASP,
  * RTMP.
  *
- * $Id: packet-atalk.c,v 1.87 2003/03/05 07:17:49 guy Exp $
+ * $Id: packet-atalk.c,v 1.96 2004/02/25 09:31:05 guy Exp $
  *
  * Simon Wilkinson <sxw@dcs.ed.ac.uk>
  *
@@ -73,6 +73,7 @@ static int hf_ddp_dst_socket = -1;
 static int hf_ddp_src_socket = -1;
 static int hf_ddp_type = -1;
 
+static dissector_handle_t ddp_handle;
 
 /* --------------------------------------
  * ATP protocol parameters
@@ -127,6 +128,7 @@ static int hf_atp_segment_overlap_conflict = -1;
 static int hf_atp_segment_multiple_tails = -1;
 static int hf_atp_segment_too_long_segment = -1;
 static int hf_atp_segment_error = -1;
+static int hf_atp_reassembled_in = -1;
 
 /* ------------------------- */
 static int proto_zip = -1;
@@ -251,6 +253,7 @@ static gint ett_asp_vers   = -1;
 static gint ett_asp_addr   = -1;
 static gint ett_asp_addr_line = -1;
 static gint ett_asp_directory = -1;
+static gint ett_asp_utf8_name = -1;
 static gint ett_asp_status_server_flag = -1;
 
 static guint asp_packet_init_count = 200;
@@ -346,6 +349,7 @@ static const fragment_items atp_frag_items = {
 	&hf_atp_segment_multiple_tails,
 	&hf_atp_segment_too_long_segment,
 	&hf_atp_segment_error,
+	&hf_atp_reassembled_in,
 	"segments"
 };
 
@@ -514,9 +518,7 @@ static int dissect_pascal_string(tvbuff_t *tvb, int offset, proto_tree *tree,
 		 * code, we could perhaps avoid allocating and freeing
 		 * this string buffer.
 		 */
-		tmp = g_malloc( len+1 );
-		tvb_memcpy(tvb, tmp, offset, len);
-		tmp[len] = 0;
+		tmp = tvb_get_string(tvb, offset, len);
 		item = proto_tree_add_string(tree, hf_index, tvb, offset-1, len+1, tmp);
 
 		subtree = proto_item_add_subtree(item, ett_pstring);
@@ -855,23 +857,9 @@ dissect_atp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree) {
 				     frag_number,
 				     len,
 				     more_fragment);
-     if (fd_head != NULL) {
-	if (fd_head->next != NULL) {
-            new_tvb = tvb_new_real_data(fd_head->data, fd_head->len, fd_head->len);
-            tvb_set_child_real_data_tvbuff(tvb, new_tvb);
-            add_new_data_source(pinfo, new_tvb, "Reassembled ATP");
-	    /* Show all fragments. */
-	    if (tree) {
-		    show_fragment_seq_tree(fd_head, &atp_frag_items,
-		        atp_tree, pinfo, new_tvb);
-	    }
-        }
-        else
-      	    new_tvb = tvb_new_subset(tvb, ATP_HDRSIZE -1, -1, -1);
-     }
-     else {
-	new_tvb = NULL;
-     }
+     new_tvb = process_reassembled_data(tvb, ATP_HDRSIZE -1, pinfo,
+				"Reassembled ATP", fd_head, &atp_frag_items,
+				NULL, atp_tree);
   }
   else {
       /* full packet */
@@ -897,6 +885,8 @@ dissect_atp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree) {
 
 /*
 	copy and paste from dsi
+	XXX - is the format of this reply dependent on the type of server,
+	with this format being the format for AFP servers?
 */
 static gint
 dissect_asp_reply_get_status(tvbuff_t *tvb, packet_info *pinfo _U_, proto_tree *tree, gint offset)
@@ -994,9 +984,9 @@ dissect_asp_reply_get_status(tvbuff_t *tvb, packet_info *pinfo _U_, proto_tree *
 		ofs++;
 		sub_tree = proto_item_add_subtree(ti, ett_asp_vers);
 		for (i = 0; i < nbe; i++) {
-			len = tvb_get_guint8(tvb, ofs) +1;
+			len = tvb_get_guint8(tvb, ofs);
 			proto_tree_add_item(sub_tree, hf_asp_server_vers, tvb, ofs, 1, FALSE);
-			ofs += len;
+			ofs += len + 1;
 		}
 	}
 
@@ -1007,9 +997,9 @@ dissect_asp_reply_get_status(tvbuff_t *tvb, packet_info *pinfo _U_, proto_tree *
 		ofs++;
 		sub_tree = proto_item_add_subtree(ti, ett_asp_uams);
 		for (i = 0; i < nbe; i++) {
-			len = tvb_get_guint8(tvb, ofs) +1;
+			len = tvb_get_guint8(tvb, ofs);
 			proto_tree_add_item(sub_tree, hf_asp_server_uams, tvb, ofs, 1, FALSE);
-			ofs += len;
+			ofs += len + 1;
 		}
 	}
 
@@ -1063,9 +1053,7 @@ dissect_asp_reply_get_status(tvbuff_t *tvb, packet_info *pinfo _U_, proto_tree *
 				break;
 			case 4: /* DNS */
 				if (len > 2) {
-					tmp = g_malloc( len -1);
-					tvb_memcpy(tvb, tmp, ofs +2, len -2);
-					tmp[len -2] = 0;
+					tmp = tvb_get_string(tvb, ofs +2, len -2);
 					ti = proto_tree_add_text(adr_tree, tvb, ofs, len, "dns %s", tmp);
 					g_free(tmp);
 					break;
@@ -1093,18 +1081,25 @@ dissect_asp_reply_get_status(tvbuff_t *tvb, packet_info *pinfo _U_, proto_tree *
 		ofs++;
 		sub_tree = proto_item_add_subtree(ti, ett_asp_directory);
 		for (i = 0; i < nbe; i++) {
-			len = tvb_get_guint8(tvb, ofs) +1;
+			len = tvb_get_guint8(tvb, ofs);
 			proto_tree_add_item(sub_tree, hf_asp_server_directory, tvb, ofs, 1, FALSE);
-			ofs += len;
+			ofs += len + 1;
 		}
 	}
 	if (utf_ofs) {
+		guint16 ulen;
+		char *tmp;
+
 		ofs = utf_ofs;
-		len  = tvb_get_ntohs(tvb, ofs);
-		proto_tree_add_item(tree, hf_asp_server_utf8_name_len, tvb, ofs, 2, FALSE);
+		ulen = tvb_get_ntohs(tvb, ofs);
+		tmp = tvb_get_string(tvb, ofs + 2, ulen);
+		ti = proto_tree_add_text(tree, tvb, ofs, ulen +2, "UTF8 server name: %s", tmp);
+		sub_tree = proto_item_add_subtree(ti, ett_asp_utf8_name);
+		proto_tree_add_uint(sub_tree, hf_asp_server_utf8_name_len, tvb, ofs, 2, ulen);
 		ofs += 2;		
-		proto_tree_add_item(tree, hf_asp_server_utf8_name, tvb, ofs, len, FALSE);
-		ofs += len;
+		proto_tree_add_string(sub_tree, hf_asp_server_utf8_name, tvb, ofs, ulen, tmp);
+		ofs += ulen;
+		g_free(tmp);
 	}
 	/* FIXME: offset is not updated */
 	return offset;
@@ -1176,7 +1171,7 @@ dissect_asp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
   if (!aspinfo)
      return;
 
-  fn = aspinfo->command;
+  fn = (guint8) aspinfo->command;
 
   if (check_col(pinfo->cinfo, COL_INFO)) {
 	if (aspinfo->reply)
@@ -1278,7 +1273,7 @@ dissect_asp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 	case ASPFUNC_CMD:
   	case ASPFUNC_WRITE:
     		proto_item_set_len(asp_tree, 4);
-		aspinfo->code = tvb_get_ntohl(tvb, offset);;
+		aspinfo->code = tvb_get_ntohl(tvb, offset);
 		proto_tree_add_item(asp_tree, hf_asp_error, tvb, offset, 4, FALSE);
 		offset += 4;
 		len = tvb_reported_length_remaining(tvb,offset);
@@ -1324,7 +1319,7 @@ dissect_atp_zip(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
   if (!aspinfo)
      return;
 
-  fn = aspinfo->command;
+  fn = (guint8) aspinfo->command;
 
   if (check_col(pinfo->cinfo, COL_INFO)) {
 	if (aspinfo->reply)
@@ -1703,18 +1698,18 @@ dissect_llap(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 
   new_tvb = tvb_new_subset(tvb, 3, -1, -1);
 
-  if (proto_is_protocol_enabled(proto_ddp)) {
-    pinfo->current_proto = "DDP";
-    switch (type) {
+  switch (type) {
 
-    case 0x01:
+  case 0x01:
+    if (proto_is_protocol_enabled(find_protocol_by_id(proto_ddp))) {
+      pinfo->current_proto = "DDP";
       dissect_ddp_short(new_tvb, pinfo, dnode, snode, tree);
       return;
-
-    case 0x02:
-      dissect_ddp(new_tvb, pinfo, tree);
-      return;
     }
+
+  case 0x02:
+    if (call_dissector(ddp_handle, new_tvb, pinfo, tree))
+      return;
   }
   call_dissector(data_handle,new_tvb, pinfo, tree);
 }
@@ -1965,6 +1960,10 @@ proto_register_atalk(void)
     { &hf_atp_segments,
       { "ATP Fragments",	"atp.fragments", FT_NONE, BASE_NONE,
 		NULL, 0x0, "ATP Fragments", HFILL }},
+
+    { &hf_atp_reassembled_in,
+      { "Reassembled ATP in frame", "atp.reassembled_in", FT_FRAMENUM, BASE_NONE, NULL, 0x0,
+       "This ATP packet is reassembled in this frame", HFILL }}
   };
 
   static hf_register_info hf_asp[] = {
@@ -2209,6 +2208,7 @@ proto_register_atalk(void)
 	&ett_asp_addr,
 	&ett_asp_addr_line,
 	&ett_asp_directory,
+	&ett_asp_utf8_name,
 
 	&ett_nbp,
 	&ett_nbp_info,
@@ -2262,7 +2262,7 @@ proto_register_atalk(void)
 void
 proto_reg_handoff_atalk(void)
 {
-  dissector_handle_t ddp_handle, nbp_handle, rtmp_request_handle;
+  dissector_handle_t nbp_handle, rtmp_request_handle;
   dissector_handle_t atp_handle;
   dissector_handle_t zip_ddp_handle;
   dissector_handle_t rtmp_data_handle, llap_handle;
