@@ -1,7 +1,7 @@
 /* packet-clnp.c
  * Routines for ISO/OSI network and transport protocol packet disassembly
  *
- * $Id: packet-clnp.c,v 1.68 2003/03/04 06:47:08 guy Exp $
+ * $Id: packet-clnp.c,v 1.83 2004/02/24 17:49:06 ulfl Exp $
  * Laurent Deniel <laurent.deniel@free.fr>
  * Ralf Schneider <Ralf.Schneider@t-online.de>
  *
@@ -27,7 +27,6 @@
 #ifdef HAVE_CONFIG_H
 # include "config.h"
 #endif
-#include <signal.h>
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
@@ -40,6 +39,7 @@
 #include "packet-isis.h"
 #include "packet-esis.h"
 #include "nlpid.h"
+#include "ipproto.h"
 
 /* protocols and fields */
 
@@ -49,10 +49,6 @@ static gint ett_clnp_type      = -1;
 static gint ett_clnp_segments  = -1;
 static gint ett_clnp_segment   = -1;
 static gint ett_clnp_disc_pdu  = -1;
-static int  proto_cotp         = -1;
-static gint ett_cotp           = -1;
-static int  proto_cltp         = -1;
-static gint ett_cltp           = -1;
 
 static int hf_clnp_id          = -1;
 static int hf_clnp_length      = -1;
@@ -72,6 +68,29 @@ static int hf_clnp_segment_overlap_conflict = -1;
 static int hf_clnp_segment_multiple_tails = -1;
 static int hf_clnp_segment_too_long_segment = -1;
 static int hf_clnp_segment_error = -1;
+static int hf_clnp_reassembled_in = -1;
+
+static int  proto_cotp         = -1;
+static gint ett_cotp           = -1;
+static gint ett_cotp_segments  = -1;
+static gint ett_cotp_segment   = -1;
+
+static int hf_cotp_srcref      = -1;
+static int hf_cotp_destref     = -1;
+static int hf_cotp_type        = -1;
+static int hf_cotp_segments    = -1;
+static int hf_cotp_segment     = -1;
+static int hf_cotp_segment_overlap = -1;
+static int hf_cotp_segment_overlap_conflict = -1;
+static int hf_cotp_segment_multiple_tails = -1;
+static int hf_cotp_segment_too_long_segment = -1;
+static int hf_cotp_segment_error = -1;
+static int hf_cotp_reassembled_in = -1;
+
+static int  proto_cltp         = -1;
+static gint ett_cltp           = -1;
+
+static int hf_cltp_type = -1;
 
 static const fragment_items clnp_frag_items = {
 	&ett_clnp_segment,
@@ -83,6 +102,21 @@ static const fragment_items clnp_frag_items = {
 	&hf_clnp_segment_multiple_tails,
 	&hf_clnp_segment_too_long_segment,
 	&hf_clnp_segment_error,
+	&hf_clnp_reassembled_in,
+	"segments"
+};
+
+static const fragment_items cotp_frag_items = {
+	&ett_cotp_segment,
+	&ett_cotp_segments,
+	&hf_cotp_segments,
+	&hf_cotp_segment,
+	&hf_cotp_segment_overlap,
+	&hf_cotp_segment_overlap_conflict,
+	&hf_cotp_segment_multiple_tails,
+	&hf_cotp_segment_too_long_segment,
+	&hf_cotp_segment_error,
+	&hf_cotp_reassembled_in,
 	"segments"
 };
 
@@ -183,6 +217,25 @@ struct clnp_segment {
 #define CR_TPDU        		0xE	/* COTP */
 #define DT_TPDU        		0xF	/* COTP */
 
+static const value_string cotp_tpdu_type_abbrev_vals[] = {
+  { ED_TPDU,	"ED" },
+  { EA_TPDU,	"EA" },
+  { RJ_TPDU,	"RJ" },
+  { AK_TPDU,	"AK" },
+  { ER_TPDU,	"ER" },
+  { DR_TPDU,	"DR" },
+  { DC_TPDU,	"DC" },
+  { CC_TPDU,	"CC" },
+  { CR_TPDU,	"CR" },
+  { DT_TPDU,	"DT" },
+  { 0,		NULL }
+};
+
+static const value_string cltp_tpdu_type_abbrev_vals[] = {
+  { UD_TPDU,	"UD" },
+  { 0,		NULL }
+};
+
 /* field position */
 
 #define P_LI           		0
@@ -285,11 +338,19 @@ static heur_dissector_list_t clnp_heur_subdissector_list;
  * Reassembly of CLNP.
  */
 static GHashTable *clnp_segment_table = NULL;
+static GHashTable *clnp_reassembled_table = NULL;
+
+/*
+ * Reassembly of COTP.
+ */
+static GHashTable *cotp_segment_table = NULL;
+static GHashTable *cotp_reassembled_table = NULL;
 
 /* options */
 static guint tp_nsap_selector = NSEL_TP;
 static gboolean always_decode_transport = FALSE;
 static gboolean clnp_reassemble = FALSE;
+static gboolean cotp_reassemble = FALSE;
 
 /* function definitions */
 
@@ -297,7 +358,7 @@ static gboolean clnp_reassemble = FALSE;
 static gchar *print_tsap(const guchar *tsap, int length)
 {
 
-  static gchar  str[3][MAX_TSAP_LEN * 2 + 1];
+  static gchar  str[3][MAX_TSAP_LEN * 2 + 3];	/* TSAP in hex + '0x' + NULL */
   static gchar *cur;
   gchar tmp[3];
   gboolean allprintable;
@@ -683,9 +744,21 @@ static int ositp_decode_DR(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
     return -1;
 
   dst_ref = tvb_get_ntohs(tvb, offset + P_DST_REF);
+
   src_ref = tvb_get_ntohs(tvb, offset + P_SRC_REF);
+
   reason  = tvb_get_guint8(tvb, offset + P_REASON_IN_DR);
 
+  /* the settings of the TCP srcport and destport are currently disables,
+   * for the following reasons:
+   * a) only used for ISO conversation handling (which currently doesn't work)
+   * b) will prevent "ISO on TCP" (RFC1006) packets from using "follow TCP stream" correctly
+   *
+   * A future conversation handling might be able to handle different kinds of conversations
+   * (TCP, ISO, TCP on TCP, ...), but in that case this has to be fixed in any case.
+   */
+  /*pinfo->srcport = src_ref;*/
+  /*pinfo->destport = dst_ref;*/
   switch(reason) {
     case (128+0): str = "Normal Disconnect"; break;
     case (128+1): str = "Remote transport entity congestion"; break;
@@ -715,12 +788,10 @@ static int ositp_decode_DR(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
     cotp_tree = proto_item_add_subtree(ti, ett_cotp);
     proto_tree_add_text(cotp_tree, tvb, offset,      1,
 			"Length indicator: %u", li);
-    proto_tree_add_text(cotp_tree, tvb, offset +  1, 1,
-			"TPDU code: 0x%x (DR)", tpdu);
-    proto_tree_add_text(cotp_tree, tvb, offset +  2, 2,
-			"Destination reference: 0x%04x", dst_ref);
-    proto_tree_add_text(cotp_tree, tvb, offset +  4, 2,
-			"Source reference: 0x%04x", src_ref);
+    proto_tree_add_uint_format(cotp_tree, hf_cotp_type, tvb, offset +  1, 1, tpdu,
+			       "TPDU code: 0x%x (DR)", tpdu);
+    proto_tree_add_uint(cotp_tree, hf_cotp_destref, tvb, offset +  2, 2, dst_ref);
+    proto_tree_add_uint(cotp_tree, hf_cotp_srcref, tvb, offset +  4, 2, src_ref);
     proto_tree_add_text(cotp_tree, tvb, offset +  6, 1,
 			"Cause: %s", str);
   }
@@ -748,7 +819,10 @@ static int ositp_decode_DT(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   guint16  dst_ref;
   guint    tpdu_nr;
   guint    fragment = 0;
+  guint32  fragment_length = 0;
   tvbuff_t *next_tvb;
+  tvbuff_t *reassembled_tvb = NULL;
+  fragment_data *fd_head;
 
   /* VP_CHECKSUM is the only parameter allowed in the variable part.
      (This means we may misdissect this if the packet is bad and
@@ -804,6 +878,9 @@ static int ositp_decode_DT(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
       break;
   }
 
+  /* pinfo->destport = dst_ref; */
+  /* pinfo->srcport = 0; */
+  pinfo->fragmented = fragment;
   if (check_col(pinfo->cinfo, COL_INFO)) {
     if (is_class_234) {
       col_append_fstr(pinfo->cinfo, COL_INFO, "DT TPDU (%u) dst-ref: 0x%04x %s",
@@ -826,18 +903,15 @@ static int ositp_decode_DT(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   offset += 1;
 
   if (tree) {
-    proto_tree_add_text(cotp_tree, tvb, offset, 1,
+    proto_tree_add_uint_format(cotp_tree, hf_cotp_type, tvb, offset, 1, tpdu, 
 			"TPDU code: 0x%x (DT)", tpdu);
-
   }
   offset += 1;
   li -= 1;
 
   if (is_class_234) {
-    if (tree) {
-      proto_tree_add_text(cotp_tree, tvb, offset, 2,
-			  "Destination reference: 0x%04x", dst_ref);
-    }
+    if (tree)
+      proto_tree_add_uint(cotp_tree, hf_cotp_destref, tvb, offset, 2, dst_ref);
     offset += 2;
     li -= 2;
   }
@@ -867,22 +941,71 @@ static int ositp_decode_DT(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   offset += li;
 
   next_tvb = tvb_new_subset(tvb, offset, -1, -1);
-  if (uses_inactive_subset){
-	if (dissector_try_heuristic(cotp_is_heur_subdissector_list, next_tvb,
-					pinfo, tree)) {
-		*subdissector_found = TRUE;
-	} else {
-	  /* Fill in other Dissectors using inactive subset here */
-	  call_dissector(data_handle,next_tvb, pinfo, tree);
-	}
+  if (cotp_reassemble) {
+    fragment_length = tvb_length(next_tvb);
+    /*
+     * XXX - these sequence numbers are connection sequence number,
+     * not segment sequence numbers - the first segment of a
+     * segmented packet doesn't have a specific sequence number (e.g., 0
+     * or 1), it has whatever the appropriate sequence number is for
+     * it in the connection.
+     */
+    fd_head = fragment_add_seq_check(next_tvb, 0, pinfo, dst_ref,
+				     cotp_segment_table, 
+				     cotp_reassembled_table,
+				     tpdu_nr, 
+				     fragment_length, fragment);
+    if (fd_head) {
+      if (fd_head->next) {
+	/* This is the last packet */
+	reassembled_tvb = tvb_new_real_data(fd_head->data,
+					    fd_head->len,
+					    fd_head->len);
+	tvb_set_child_real_data_tvbuff(next_tvb, reassembled_tvb);
+	add_new_data_source(pinfo, reassembled_tvb, "Reassembled COTP");
+	
+	show_fragment_seq_tree(fd_head,
+			       &cotp_frag_items,
+			       cotp_tree,
+			       pinfo, reassembled_tvb);
+	pinfo->fragmented = fragment;
+	next_tvb = reassembled_tvb;
+      }
+    }
+    if (fragment && reassembled_tvb == NULL) {
+      proto_tree_add_text(cotp_tree, tvb, offset, -1,
+			  "User data (%u byte%s)", fragment_length,
+			  plurality(fragment_length, "", "s"));
+    } 
+
+  } 
+
+  if (uses_inactive_subset) {
+    if (dissector_try_heuristic(cotp_is_heur_subdissector_list, next_tvb,
+				pinfo, tree)) {
+      *subdissector_found = TRUE;
+    } else {
+      /* Fill in other Dissectors using inactive subset here */
+      call_dissector(data_handle,next_tvb, pinfo, tree);
+    }
   } else {
-        if (dissector_try_heuristic(cotp_heur_subdissector_list, next_tvb,
-				    pinfo, tree)) {
-		*subdissector_found = TRUE;
-	} else {
-		call_dissector(data_handle,next_tvb, pinfo, tree);
-	}
-  }
+    /*
+     * We dissect payload if one of the following is TRUE: 
+     *
+     * - Reassembly option for COTP in preferences is unchecked 
+     * - Reassembly option is checked and this packet is the last fragment
+     */
+    if ( (!cotp_reassemble) ||
+	 ((cotp_reassemble) && (!fragment))) {
+      if (dissector_try_heuristic(cotp_heur_subdissector_list, next_tvb,
+				  pinfo, tree)) {
+        *subdissector_found = TRUE;
+      } else {
+        call_dissector(data_handle,next_tvb, pinfo, tree);
+      }
+    }
+  }   
+
   offset += tvb_length_remaining(tvb, offset);
      /* we dissected all of the containing PDU */
 
@@ -942,6 +1065,9 @@ static int ositp_decode_ED(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   } /* li */
 
   dst_ref = tvb_get_ntohs(tvb, offset + P_DST_REF);
+
+  /* pinfo->destport = dst_ref; */
+  /* pinfo->srcport = 0; */
   if (check_col(pinfo->cinfo, COL_INFO))
     col_append_fstr(pinfo->cinfo, COL_INFO, "ED TPDU (%u) dst-ref: 0x%04x",
 		 tpdu_nr, dst_ref);
@@ -955,16 +1081,14 @@ static int ositp_decode_ED(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   offset += 1;
 
   if (tree) {
-    proto_tree_add_text(cotp_tree, tvb, offset, 1,
+    proto_tree_add_uint_format(cotp_tree, hf_cotp_type, tvb, offset, 1, tpdu,
 			"TPDU code: 0x%x (ED)", tpdu);
   }
   offset += 1;
   li -= 1;
 
-  if (tree) {
-    proto_tree_add_text(cotp_tree, tvb, offset, 2,
-			"Destination reference: 0x%04x", dst_ref);
-  }
+  if (tree)
+    proto_tree_add_uint(cotp_tree, hf_cotp_destref, tvb, offset, 2, dst_ref);
   offset += 2;
   li -= 2;
 
@@ -1022,6 +1146,9 @@ static int ositp_decode_RJ(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   }
 
   dst_ref = tvb_get_ntohs(tvb, offset + P_DST_REF);
+
+  /* pinfo->destport = dst_ref; */
+  /* pinfo->srcport = 0; */
   if (check_col(pinfo->cinfo, COL_INFO))
     col_append_fstr(pinfo->cinfo, COL_INFO, "RJ TPDU (%u) dst-ref: 0x%04x",
 		 tpdu_nr, dst_ref);
@@ -1031,13 +1158,12 @@ static int ositp_decode_RJ(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
     cotp_tree = proto_item_add_subtree(ti, ett_cotp);
     proto_tree_add_text(cotp_tree, tvb, offset,      1,
 			"Length indicator: %u", li);
-    proto_tree_add_text(cotp_tree, tvb, offset +  1, 1,
+    proto_tree_add_uint_format(cotp_tree, hf_cotp_type, tvb, offset +  1, 1, tpdu,
 			"TPDU code: 0x%x (RJ)", tpdu);
     if (li == LI_NORMAL_RJ)
       proto_tree_add_text(cotp_tree, tvb, offset +  1, 1,
 			  "Credit: %u", cdt);
-    proto_tree_add_text(cotp_tree, tvb, offset +  2, 2,
-			"Destination reference: 0x%04x", dst_ref);
+    proto_tree_add_uint(cotp_tree, hf_cotp_destref, tvb, offset +  2, 2, dst_ref);
     if (li == LI_NORMAL_RJ)
       proto_tree_add_text(cotp_tree, tvb, offset +  4, 1,
 			  "Your TPDU number: 0x%02x", tpdu_nr);
@@ -1070,11 +1196,14 @@ static int ositp_decode_CC(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   tvbuff_t *next_tvb;
 
   src_ref = tvb_get_ntohs(tvb, offset + P_SRC_REF);
+  
   class_option = (tvb_get_guint8(tvb, offset + P_CLASS_OPTION) >> 4 ) & 0x0F;
   if (class_option > 4)
     return -1;
 
   dst_ref = tvb_get_ntohs(tvb, offset + P_DST_REF);
+  /* pinfo->srcport = src_ref; */
+  /* pinfo->destport = dst_ref; */
   if (check_col(pinfo->cinfo, COL_INFO))
     col_append_fstr(pinfo->cinfo, COL_INFO,
 		 "%s TPDU src-ref: 0x%04x dst-ref: 0x%04x",
@@ -1091,24 +1220,20 @@ static int ositp_decode_CC(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   offset += 1;
 
   if (tree) {
-    proto_tree_add_text(cotp_tree, tvb, offset, 1,
+    proto_tree_add_uint_format(cotp_tree, hf_cotp_type, tvb, offset, 1, tpdu,
 			"TPDU code: 0x%x (%s)", tpdu,
 			(tpdu == CR_TPDU) ? "CR" : "CC");
   }
   offset += 1;
   li -= 1;
 
-  if (tree) {
-    proto_tree_add_text(cotp_tree, tvb, offset, 2,
-			"Destination reference: 0x%04x", dst_ref);
-  }
+  if (tree)
+    proto_tree_add_uint(cotp_tree, hf_cotp_destref, tvb, offset, 2, dst_ref);
   offset += 2;
   li -= 2;
 
-  if (tree) {
-    proto_tree_add_text(cotp_tree, tvb, offset, 2,
-			"Source reference: 0x%04x", src_ref);
-  }
+  if (tree)
+    proto_tree_add_uint(cotp_tree, hf_cotp_srcref, tvb, offset, 2, src_ref);
   offset += 2;
   li -= 2;
 
@@ -1154,6 +1279,8 @@ static int ositp_decode_DC(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   dst_ref = tvb_get_ntohs(tvb, offset + P_DST_REF);
   src_ref = tvb_get_ntohs(tvb, offset + P_SRC_REF);
 
+  /* pinfo->srcport = src_ref; */
+  /* pinfo->destport = dst_ref; */
   if (check_col(pinfo->cinfo, COL_INFO))
     col_append_fstr(pinfo->cinfo, COL_INFO,
 		 "DC TPDU src-ref: 0x%04x dst-ref: 0x%04x",
@@ -1169,23 +1296,19 @@ static int ositp_decode_DC(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   offset += 1;
 
   if (tree) {
-    proto_tree_add_text(cotp_tree, tvb, offset, 1,
+    proto_tree_add_uint_format(cotp_tree, hf_cotp_type, tvb, offset, 1, tpdu, 
 			"TPDU code: 0x%x (DC)", tpdu);
   }
   offset += 1;
   li -= 1;
 
-  if (tree) {
-    proto_tree_add_text(cotp_tree, tvb, offset, 2,
-			"Destination reference: 0x%04x", dst_ref);
-  }
+  if (tree)
+    proto_tree_add_uint(cotp_tree, hf_cotp_destref, tvb, offset, 2, dst_ref);
   offset += 2;
   li -= 2;
 
-  if (tree) {
-    proto_tree_add_text(cotp_tree, tvb, offset, 2,
-			"Source reference: 0x%04x", src_ref);
-  }
+  if (tree)
+    proto_tree_add_uint(cotp_tree, hf_cotp_srcref, tvb, offset, 2, src_ref);
   offset += 2;
   li -= 2;
 
@@ -1214,6 +1337,8 @@ static int ositp_decode_AK(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
     dst_ref = tvb_get_ntohs(tvb, offset + P_DST_REF);
     tpdu_nr = tvb_get_guint8(tvb, offset + P_TPDU_NR_234);
 
+    /* pinfo->srcport = 0; */
+    /* pinfo->destport = dst_ref; */
     if (check_col(pinfo->cinfo, COL_INFO))
       col_append_fstr(pinfo->cinfo, COL_INFO, "AK TPDU (%u) dst-ref: 0x%04x",
 		   tpdu_nr, dst_ref);
@@ -1227,7 +1352,7 @@ static int ositp_decode_AK(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
     offset += 1;
 
     if (tree) {
-      proto_tree_add_text(cotp_tree, tvb, offset, 1,
+      proto_tree_add_uint_format(cotp_tree, hf_cotp_type, tvb, offset, 1, tpdu, 
 			  "TPDU code: 0x%x (AK)", tpdu);
       proto_tree_add_text(cotp_tree, tvb, offset, 1,
 			  "Credit: %u", cdt);
@@ -1235,10 +1360,8 @@ static int ositp_decode_AK(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
     offset += 1;
     li -= 1;
 
-    if (tree) {
-      proto_tree_add_text(cotp_tree, tvb, offset, 2,
-			  "Destination reference: 0x%04x", dst_ref);
-    }
+    if (tree)
+      proto_tree_add_uint(cotp_tree, hf_cotp_destref, tvb, offset, 2, dst_ref);
     offset += 2;
     li -= 2;
 
@@ -1272,16 +1395,14 @@ static int ositp_decode_AK(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
     offset += 1;
 
     if (tree) {
-      proto_tree_add_text(cotp_tree, tvb, offset, 1,
+      proto_tree_add_uint_format(cotp_tree, hf_cotp_type, tvb, offset, 1, tpdu, 
 			  "TPDU code: 0x%x (AK)", tpdu);
     }
     offset += 1;
     li -= 1;
 
-    if (tree) {
-      proto_tree_add_text(cotp_tree, tvb, offset, 2,
-			  "Destination reference: 0x%04x", dst_ref);
-    }
+    if (tree)
+      proto_tree_add_uint(cotp_tree, hf_cotp_destref, tvb, offset, 2, dst_ref);
     offset += 2;
     li -= 2;
 
@@ -1355,6 +1476,8 @@ static int ositp_decode_EA(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   } /* li */
 
   dst_ref = tvb_get_ntohs(tvb, offset + P_DST_REF);
+  /* pinfo->srcport = 0; */
+  /* pinfo->destport = dst_ref; */
   if (check_col(pinfo->cinfo, COL_INFO))
     col_append_fstr(pinfo->cinfo, COL_INFO,
 		 "EA TPDU (%u) dst-ref: 0x%04x", tpdu_nr, dst_ref);
@@ -1368,16 +1491,14 @@ static int ositp_decode_EA(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   offset += 1;
 
   if (tree) {
-    proto_tree_add_text(cotp_tree, tvb, offset, 1,
+    proto_tree_add_uint_format(cotp_tree, hf_cotp_type, tvb, offset, 1, tpdu,
 			"TPDU code: 0x%x (EA)", tpdu);
   }
   offset += 1;
   li -= 1;
 
-  if (tree) {
-    proto_tree_add_text(cotp_tree, tvb, offset, 2,
-			"Destination reference: 0x%04x", dst_ref);
-  }
+  if (tree)
+    proto_tree_add_uint(cotp_tree, hf_cotp_destref, tvb, offset, 2, dst_ref);
   offset += 2;
   li -= 2;
 
@@ -1436,6 +1557,8 @@ static int ositp_decode_ER(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   }
 
   dst_ref = tvb_get_ntohs(tvb, offset + P_DST_REF);
+  /* pinfo->srcport = 0; */
+  /* pinfo->destport = dst_ref; */
   if (check_col(pinfo->cinfo, COL_INFO))
     col_append_fstr(pinfo->cinfo, COL_INFO, "ER TPDU dst-ref: 0x%04x", dst_ref);
 
@@ -1444,10 +1567,9 @@ static int ositp_decode_ER(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
     cotp_tree = proto_item_add_subtree(ti, ett_cotp);
     proto_tree_add_text(cotp_tree, tvb, offset,      1,
 			"Length indicator: %u", li);
-    proto_tree_add_text(cotp_tree, tvb, offset +  1, 1,
+    proto_tree_add_uint_format(cotp_tree, hf_cotp_type, tvb, offset +  1, 1, tpdu,
 			"TPDU code: 0x%x (ER)", tpdu);
-    proto_tree_add_text(cotp_tree, tvb, offset +  2, 2,
-			"Destination reference: 0x%04x", dst_ref);
+    proto_tree_add_uint(cotp_tree, hf_cotp_destref, tvb, offset +  2, 2, dst_ref);
     proto_tree_add_text(cotp_tree, tvb, offset +  4, 1,
 			"Reject cause: %s", str);
   }
@@ -1477,7 +1599,7 @@ static int ositp_decode_UD(tvbuff_t *tvb, int offset, guint8 li, guint8 tpdu,
   offset += 1;
 
   if (tree) {
-    proto_tree_add_text(cltp_tree, tvb, offset, 1,
+    proto_tree_add_uint_format(cltp_tree, hf_cltp_type, tvb, offset, 1, tpdu, 
 			"TPDU code: 0x%x (UD)", tpdu);
   }
   offset += 1;
@@ -1515,7 +1637,7 @@ static gboolean dissect_ositp_internal(tvbuff_t *tvb, packet_info *pinfo,
   gboolean is_cltp = FALSE;
   gboolean subdissector_found = FALSE;
 
-  if (!proto_is_protocol_enabled(proto_cotp))
+  if (!proto_is_protocol_enabled(find_protocol_by_id(proto_cotp)))
     return FALSE;	/* COTP has been disabled */
   /* XXX - what about CLTP? */
 
@@ -1647,8 +1769,6 @@ static void dissect_clnp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
   gboolean    update_col_info = TRUE;
   gboolean    save_fragmented;
 
-  signal(SIGSEGV, SIG_DFL);
-
   if (check_col(pinfo->cinfo, COL_PROTOCOL))
     col_set_str(pinfo->cinfo, COL_PROTOCOL, "CLNP");
   if (check_col(pinfo->cinfo, COL_INFO))
@@ -1872,30 +1992,13 @@ static void dissect_clnp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 	((cnf_type & CNF_MORE_SEGS) || segment_offset != 0) &&
 	tvb_bytes_exist(tvb, offset, segment_length - cnf_hdr_len) &&
 	cksum_status != CKSUM_NOT_OK) {
-    fd_head = fragment_add(tvb, offset, pinfo, du_id, clnp_segment_table,
-			   segment_offset, segment_length - cnf_hdr_len,
+    fd_head = fragment_add_check(tvb, offset, pinfo, du_id, clnp_segment_table,
+			   clnp_reassembled_table, segment_offset,
+			   segment_length - cnf_hdr_len,
 			   cnf_type & CNF_MORE_SEGS);
 
-    if (fd_head != NULL) {
-      /* OK, we have the complete reassembled payload.
-         Allocate a new tvbuff, referring to the reassembled payload. */
-      next_tvb = tvb_new_real_data(fd_head->data, fd_head->datalen,
-	fd_head->datalen);
-
-      /* Add the tvbuff to the list of tvbuffs to which the tvbuff we
-         were handed refers, so it'll get cleaned up when that tvbuff
-         is cleaned up. */
-      tvb_set_child_real_data_tvbuff(tvb, next_tvb);
-
-      /* Add the defragmented data to the data source list. */
-      add_new_data_source(pinfo, next_tvb, "Reassembled CLNP");
-
-      update_col_info = !show_fragment_tree(fd_head, &clnp_frag_items,
-        clnp_tree, pinfo, next_tvb);
-    } else {
-      /* We don't have the complete reassembled payload. */
-      next_tvb = NULL;
-    }
+    next_tvb = process_reassembled_data(tvb, offset, pinfo, "Reassembled CLNP",
+        fd_head, &clnp_frag_items, &update_col_info, clnp_tree);
   } else {
     /* If this is the first segment, dissect its contents, otherwise
        just show it as a segment.
@@ -2005,6 +2108,14 @@ static void
 clnp_reassemble_init(void)
 {
   fragment_table_init(&clnp_segment_table);
+  reassembled_table_init(&clnp_reassembled_table);
+}
+
+static void
+cotp_reassemble_init(void)
+{
+  fragment_table_init(&cotp_segment_table);
+  reassembled_table_init(&cotp_reassembled_table);
 }
 
 void proto_register_clnp(void)
@@ -2071,6 +2182,10 @@ void proto_register_clnp(void)
     { &hf_clnp_segments,
       { "CLNP Segments", "clnp.segments", FT_NONE, BASE_DEC, NULL, 0x0,
 	"CLNP Segments", HFILL }},
+
+    { &hf_clnp_reassembled_in,
+      { "Reassembled CLNP in frame", "clnp.reassembled_in", FT_FRAMENUM, BASE_NONE, NULL, 0x0,
+	"This CLNP packet is reassembled in this frame", HFILL }}
   };
   static gint *ett[] = {
     &ett_clnp,
@@ -2087,11 +2202,13 @@ void proto_register_clnp(void)
   proto_register_subtree_array(ett, array_length(ett));
   register_dissector("clnp", dissect_clnp, proto_clnp);
   register_heur_dissector_list("clnp", &clnp_heur_subdissector_list);  
+  register_init_routine(clnp_reassemble_init);
+  register_init_routine(cotp_reassemble_init);
 
   clnp_module = prefs_register_protocol(proto_clnp, NULL);
   prefs_register_uint_preference(clnp_module, "tp_nsap_selector",
-	"NSAP selector for Transport Protocol (last byte in hexa)",
-	"NSAP selector for Transport Protocol (last byte in hexa)",
+	"NSAP selector for Transport Protocol (last byte in hex)",
+	"NSAP selector for Transport Protocol (last byte in hex)",
        	16, &tp_nsap_selector);
   prefs_register_bool_preference(clnp_module, "always_decode_transport",
 	"Always try to decode NSDU as transport PDUs",
@@ -2103,53 +2220,104 @@ void proto_register_clnp(void)
 	&clnp_reassemble);
 }
 
+void
+proto_reg_handoff_clnp(void)
+{
+  data_handle = find_dissector("data");
+
+  clnp_handle = create_dissector_handle(dissect_clnp, proto_clnp);
+  dissector_add("osinl", NLPID_ISO8473_CLNP, clnp_handle);
+  dissector_add("osinl", NLPID_NULL, clnp_handle); /* Inactive subset */
+  dissector_add("x.25.spi", NLPID_ISO8473_CLNP, clnp_handle);
+}
+
 void proto_register_cotp(void)
 {
-  /*        static hf_register_info hf[] = {
-                { &variable,
-                { "Name",           "cotp.abbreviation", TYPE, VALS_POINTER }},
-        };*/
-	static gint *ett[] = {
-		&ett_cotp,
-	};
+  static hf_register_info hf[] = {
+    { &hf_cotp_srcref,
+      { "Source reference", "cotp.srcref", FT_UINT16, BASE_HEX, NULL, 0x0,
+        "Source address reference", HFILL}},
+    { &hf_cotp_destref,
+      { "Destination reference", "cotp.destref", FT_UINT16, BASE_HEX, NULL, 0x0,
+        "Destination address reference", HFILL}}, 
+    { &hf_cotp_type,
+      { "COTP PDU Type", "cotp.type", FT_UINT8, BASE_HEX, VALS(cotp_tpdu_type_abbrev_vals), 0x0,
+        "COTP PDU Type", HFILL}},
+    { &hf_cotp_segment_overlap,
+      { "Segment overlap", "cotp.segment.overlap", FT_BOOLEAN, BASE_NONE, NULL, 0x0,
+	"Segment overlaps with other segments", HFILL }},
+    { &hf_cotp_segment_overlap_conflict,
+      { "Conflicting data in segment overlap", "cotp.segment.overlap.conflict", FT_BOOLEAN, BASE_NONE, NULL, 0x0,
+	"Overlapping segments contained conflicting data", HFILL }},
+    { &hf_cotp_segment_multiple_tails,
+      { "Multiple tail segments found", "cotp.segment.multipletails", FT_BOOLEAN, BASE_NONE, NULL, 0x0,
+	"Several tails were found when reassembling the packet", HFILL }},
+    { &hf_cotp_segment_too_long_segment,
+      { "Segment too long", "cotp.segment.toolongsegment", FT_BOOLEAN, BASE_NONE, NULL, 0x0,
+	"Segment contained data past end of packet", HFILL }},
+    { &hf_cotp_segment_error,
+      { "Reassembly error", "cotp.segment.error", FT_FRAMENUM, BASE_NONE, NULL, 0x0,
+	"Reassembly error due to illegal segments", HFILL }},
+    { &hf_cotp_segment,
+      { "COTP Segment", "cotp.segment", FT_FRAMENUM, BASE_NONE, NULL, 0x0,
+	"COTP Segment", HFILL }},
+    { &hf_cotp_segments,
+      { "COTP Segments", "cotp.segments", FT_NONE, BASE_DEC, NULL, 0x0,
+	"COTP Segments", HFILL }},
+    { &hf_cotp_reassembled_in,
+      { "Reassembled COTP in frame", "cotp.reassembled_in", FT_FRAMENUM, BASE_NONE, NULL, 0x0,
+	"This COTP packet is reassembled in this frame", HFILL }},
 
-        proto_cotp = proto_register_protocol(PROTO_STRING_COTP, "COTP", "cotp");
- /*       proto_register_field_array(proto_cotp, hf, array_length(hf));*/
-	proto_register_subtree_array(ett, array_length(ett));
+  };
+  static gint *ett[] = {
+	&ett_cotp,
+	&ett_cotp_segment,
+	&ett_cotp_segments,
+  };
 
-	/* subdissector code in inactive subset */
-	register_heur_dissector_list("cotp_is", &cotp_is_heur_subdissector_list);
+  module_t *cotp_module;
 
-	/* other COTP/ISO 8473 subdissectors */
-	register_heur_dissector_list("cotp", &cotp_heur_subdissector_list);
+  proto_cotp = proto_register_protocol(PROTO_STRING_COTP, "COTP", "cotp");
+  proto_register_field_array(proto_cotp, hf, array_length(hf));
+  proto_register_subtree_array(ett, array_length(ett));
+  cotp_module = prefs_register_protocol(proto_cotp, NULL);
 
-	/* XXX - what about CLTP and proto_cltp? */
-	register_dissector("ositp", dissect_ositp, proto_cotp);
+  prefs_register_bool_preference(cotp_module, "reassemble",
+	 "Reassemble segmented COTP datagrams",
+	 "Whether segmented COTP datagrams should be reassembled",
+	&cotp_reassemble);
+
+  /* subdissector code in inactive subset */
+  register_heur_dissector_list("cotp_is", &cotp_is_heur_subdissector_list);
+
+  /* other COTP/ISO 8473 subdissectors */
+  register_heur_dissector_list("cotp", &cotp_heur_subdissector_list);
+
+  /* XXX - what about CLTP and proto_cltp? */
+  register_dissector("ositp", dissect_ositp, proto_cotp);
+}
+
+void
+proto_reg_handoff_cotp(void)
+{
+  dissector_handle_t ositp_handle;
+
+  ositp_handle = find_dissector("ositp");
+  dissector_add("ip.proto", IP_PROTO_TP, ositp_handle);
 }
 
 void proto_register_cltp(void)
 {
-  /*        static hf_register_info hf[] = {
-                { &variable,
-                { "Name",           "cltp.abbreviation", TYPE, VALS_POINTER }},
-        };*/
-	static gint *ett[] = {
-		&ett_cltp,
-	};
+  static hf_register_info hf[] = {
+    { &hf_cltp_type,
+      { "CLTP PDU Type", "cltp.type", FT_UINT8, BASE_HEX, VALS(cltp_tpdu_type_abbrev_vals), 0x0,
+        "CLTP PDU Type", HFILL}},
+  };
+  static gint *ett[] = {
+	&ett_cltp,
+  };
 
-        proto_cltp = proto_register_protocol(PROTO_STRING_CLTP, "CLTP", "cltp");
- /*       proto_register_field_array(proto_cotp, hf, array_length(hf));*/
-	proto_register_subtree_array(ett, array_length(ett));
-	register_init_routine(clnp_reassemble_init);
-}
-
-void
-proto_reg_handoff_clnp(void)
-{
-        data_handle = find_dissector("data");
-
-	clnp_handle = create_dissector_handle(dissect_clnp, proto_clnp);
-	dissector_add("osinl", NLPID_ISO8473_CLNP, clnp_handle);
-	dissector_add("osinl", NLPID_NULL, clnp_handle); /* Inactive subset */
-	dissector_add("x.25.spi", NLPID_ISO8473_CLNP, clnp_handle);
+  proto_cltp = proto_register_protocol(PROTO_STRING_CLTP, "CLTP", "cltp");
+  proto_register_field_array(proto_cltp, hf, array_length(hf));
+  proto_register_subtree_array(ett, array_length(ett));
 }
