@@ -2,7 +2,7 @@
  * Routines for DCERPC Browser packet disassembly
  * Copyright 2001, Ronnie Sahlberg
  *
- * $Id: packet-dcerpc-browser.c,v 1.10 2003/01/30 08:19:37 guy Exp $
+ * $Id: packet-dcerpc-browser.c,v 1.14 2004/01/19 20:10:33 jmayer Exp $
  *
  * Ethereal - Network traffic analyzer
  * By Gerald Combs <gerald@ethereal.com>
@@ -39,6 +39,7 @@
 #include "smb.h"
 
 static int proto_dcerpc_browser = -1;
+static int hf_browser_opnum = -1;
 static int hf_browser_rc = -1;
 static int hf_browser_unknown_long = -1;
 static int hf_browser_unknown_hyper = -1;
@@ -52,7 +53,7 @@ static gint ett_dcerpc_browser = -1;
 static int
 dissect_browser_long_pointer(tvbuff_t *tvb, int offset,
                              packet_info *pinfo, proto_tree *tree,
-                             char *drep)
+                             guint8 *drep)
 {
 	dcerpc_info *di;
 
@@ -89,7 +90,7 @@ static guint16 ver_dcerpc_browser = 0;
 static int
 dissect_browser_TYPE_4_data(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	guint32 len;
 	dcerpc_info *di;
@@ -113,7 +114,7 @@ dissect_browser_TYPE_4_data(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_4(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_unknown_long, NULL);
@@ -135,7 +136,7 @@ dissect_browser_TYPE_4(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_3_data(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	guint32 len;
 	dcerpc_info *di;
@@ -160,7 +161,7 @@ dissect_browser_TYPE_3_data(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_3(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_unknown_long, NULL);
@@ -183,7 +184,7 @@ dissect_browser_TYPE_3(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_2(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	guint32 level;
 
@@ -219,7 +220,7 @@ dissect_browser_TYPE_2(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_1(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_unknown_long, NULL);
@@ -232,7 +233,7 @@ dissect_browser_TYPE_1(tvbuff_t *tvb, int offset,
 
 
 /*
- IDL  long Function_00(
+ IDL  long BrowserrServerEnum(
  IDL        [in] [unique] [string] wchar_t *element_9,
  IDL        [in] [unique] [string] wchar_t *element_10,
  IDL        [in] [unique] [string] wchar_t *element_11,
@@ -245,9 +246,9 @@ dissect_browser_TYPE_1(tvbuff_t *tvb, int offset,
  IDL  );
 */
 static int
-dissect_browser_UNKNOWN_00_rqst(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_server_enum_rqst(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_str_pointer_item(tvb, offset, pinfo, tree, drep,
 			NDR_POINTER_UNIQUE, "unknown string", 
@@ -282,9 +283,9 @@ dissect_browser_UNKNOWN_00_rqst(tvbuff_t *tvb, int offset,
 	return offset;
 }
 static int
-dissect_browser_UNKNOWN_00_reply(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_server_enum_reply(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_pointer(tvb, offset, pinfo, tree, drep,
 			dissect_browser_TYPE_1, NDR_POINTER_REF,
@@ -304,16 +305,16 @@ dissect_browser_UNKNOWN_00_reply(tvbuff_t *tvb, int offset,
 }
 
 /*
-  IDL long Function_01(
+  IDL long BrowserrDebugCall(
   IDL       [in] [unique] [string] wchar_t *element_18,
   IDL       [in] long element_19,
   IDL       [in] long element_20
   IDL );
 */
 static int
-dissect_browser_UNKNOWN_01_rqst(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_debug_call_rqst(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_str_pointer_item(tvb, offset, pinfo, tree, drep,
 			NDR_POINTER_UNIQUE, "unknown string", 
@@ -328,9 +329,9 @@ dissect_browser_UNKNOWN_01_rqst(tvbuff_t *tvb, int offset,
 	return offset;
 }
 static int
-dissect_browser_UNKNOWN_01_reply(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_debug_call_reply(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_rc, NULL);
@@ -340,16 +341,16 @@ dissect_browser_UNKNOWN_01_reply(tvbuff_t *tvb, int offset,
 
 
 /*
-  IDL long Function_02(
+  IDL long BrowserrQueryOtherDomains(
   IDL       [in] [unique] [string] wchar_t *element_21,
   IDL       [in,out] [ref] TYPE_1 *element_22,
   IDL       [out] long element_23
   IDL );
 */
 static int
-dissect_browser_UNKNOWN_02_rqst(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_query_other_domains_rqst(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_str_pointer_item(tvb, offset, pinfo, tree, drep,
 			NDR_POINTER_UNIQUE, "unknown string", 
@@ -362,9 +363,9 @@ dissect_browser_UNKNOWN_02_rqst(tvbuff_t *tvb, int offset,
 	return offset;
 }
 static int
-dissect_browser_UNKNOWN_02_reply(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_query_other_domains_reply(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_unknown_long, NULL);
@@ -377,14 +378,14 @@ dissect_browser_UNKNOWN_02_reply(tvbuff_t *tvb, int offset,
 
 
 /*
-  IDL long Function_03(
+  IDL long BrowserrResetNetlogonState(
   IDL       [in] [unique] [string] wchar_t *element_24
   IDL );
 */
 static int
-dissect_browser_UNKNOWN_03_rqst(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_reset_netlogon_state_rqst(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_str_pointer_item(tvb, offset, pinfo, tree, drep,
 			NDR_POINTER_UNIQUE, "unknown string", 
@@ -393,9 +394,9 @@ dissect_browser_UNKNOWN_03_rqst(tvbuff_t *tvb, int offset,
 	return offset;
 }
 static int
-dissect_browser_UNKNOWN_03_reply(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_reset_netlogon_state_reply(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_rc, NULL);
@@ -405,15 +406,15 @@ dissect_browser_UNKNOWN_03_reply(tvbuff_t *tvb, int offset,
 
 
 /*
-  IDL long Function_04(
+  IDL long BrowserrDebugTrace(
   IDL       [in] [unique] [string] wchar_t *element_25,
   IDL       [in] [string] char element_26
   IDL );
 */
 static int
-dissect_browser_UNKNOWN_04_rqst(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_debug_trace_rqst(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_str_pointer_item(tvb, offset, pinfo, tree, drep,
 			NDR_POINTER_UNIQUE, "unknown string", 
@@ -426,9 +427,9 @@ dissect_browser_UNKNOWN_04_rqst(tvbuff_t *tvb, int offset,
 	return offset;
 }
 static int
-dissect_browser_UNKNOWN_04_reply(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_debug_trace_reply(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_rc, NULL);
@@ -468,7 +469,7 @@ dissect_browser_UNKNOWN_04_reply(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_5(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint64(tvb, offset, pinfo, tree, drep,
 		hf_browser_unknown_hyper, NULL);
@@ -532,15 +533,15 @@ dissect_browser_TYPE_5(tvbuff_t *tvb, int offset,
 
 
 /*
-  IDL long Function_05(
+  IDL long BrowserrQueryStatistics(
   IDL       [in] [unique] [string] wchar_t *element_47,
   IDL       [out] [ref] TYPE_5 **element_48
   IDL );
 */
 static int
-dissect_browser_UNKNOWN_05_rqst(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_query_statistics_rqst(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_str_pointer_item(tvb, offset, pinfo, tree, drep,
 			NDR_POINTER_UNIQUE, "unknown string", 
@@ -549,9 +550,9 @@ dissect_browser_UNKNOWN_05_rqst(tvbuff_t *tvb, int offset,
 	return offset;
 }
 static int
-dissect_browser_UNKNOWN_05_reply(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_query_statistics_reply(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_pointer(tvb, offset, pinfo, tree, drep,
 			dissect_browser_TYPE_5, NDR_POINTER_UNIQUE,
@@ -565,14 +566,14 @@ dissect_browser_UNKNOWN_05_reply(tvbuff_t *tvb, int offset,
 
 
 /*
-  IDL long Function_06(
+  IDL long BrowserrResetStatistics(
   IDL       [in] [unique] [string] wchar_t *element_49
   IDL );
 */
 static int
-dissect_browser_UNKNOWN_06_rqst(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_reset_statistics_rqst(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_str_pointer_item(tvb, offset, pinfo, tree, drep,
 			NDR_POINTER_UNIQUE, "unknown string", 
@@ -581,9 +582,9 @@ dissect_browser_UNKNOWN_06_rqst(tvbuff_t *tvb, int offset,
 	return offset;
 }
 static int
-dissect_browser_UNKNOWN_06_reply(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_reset_statistics_reply(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_rc, NULL);
@@ -593,14 +594,14 @@ dissect_browser_UNKNOWN_06_reply(tvbuff_t *tvb, int offset,
 
 
 /*
-  IDL long Function_07(
+  IDL long NetrBrowserStatisticsClear(
   IDL       [in] [unique] [string] wchar_t *element_49
   IDL );
 */
 static int
-dissect_browser_UNKNOWN_07_rqst(tvbuff_t *tvb, int offset,
+dissect_browser_netr_browser_statistics_clear_rqst(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_str_pointer_item(tvb, offset, pinfo, tree, drep,
 			NDR_POINTER_UNIQUE, "unknown string", 
@@ -609,9 +610,9 @@ dissect_browser_UNKNOWN_07_rqst(tvbuff_t *tvb, int offset,
 	return offset;
 }
 static int
-dissect_browser_UNKNOWN_07_reply(tvbuff_t *tvb, int offset,
+dissect_browser_netr_browser_statistics_clear_reply(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_rc, NULL);
@@ -643,7 +644,7 @@ dissect_browser_UNKNOWN_07_reply(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_11(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint64(tvb, offset, pinfo, tree, drep,
 		hf_browser_unknown_hyper, NULL);
@@ -705,7 +706,7 @@ dissect_browser_TYPE_11(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_11_array(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_ucarray(tvb, offset, pinfo, tree, drep,
 		dissect_browser_TYPE_11);
@@ -716,7 +717,7 @@ dissect_browser_TYPE_11_array(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_10(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_unknown_long, NULL);
@@ -738,7 +739,7 @@ dissect_browser_TYPE_10(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_9_data(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	guint32 len;
 	dcerpc_info *di;
@@ -764,7 +765,7 @@ dissect_browser_TYPE_9_data(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_9(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_unknown_long, NULL);
@@ -786,7 +787,7 @@ dissect_browser_TYPE_9(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_8(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	guint32 level;
 
@@ -822,7 +823,7 @@ dissect_browser_TYPE_8(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_7(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_unknown_long, NULL);
@@ -834,16 +835,16 @@ dissect_browser_TYPE_7(tvbuff_t *tvb, int offset,
 
 
 /*
-  IDL long Function_08(
+  IDL long NetrBrowserStatisticsGet(
   IDL       [in] [unique] [string] wchar_t *element_75,
   IDL       [in] long element_76,
   IDL   [in,out] [ref] TYPE_7 *element_77
   IDL );
 */
 static int
-dissect_browser_UNKNOWN_08_rqst(tvbuff_t *tvb, int offset,
+dissect_browser_netr_browser_statistics_get_rqst(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_str_pointer_item(tvb, offset, pinfo, tree, drep,
 			NDR_POINTER_UNIQUE, "unknown string", 
@@ -859,9 +860,9 @@ dissect_browser_UNKNOWN_08_rqst(tvbuff_t *tvb, int offset,
 	return offset;
 }
 static int
-dissect_browser_UNKNOWN_08_reply(tvbuff_t *tvb, int offset,
+dissect_browser_netr_browser_statistics_get_reply(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_pointer(tvb, offset, pinfo, tree, drep,
 			dissect_browser_TYPE_7, NDR_POINTER_REF,
@@ -875,7 +876,7 @@ dissect_browser_UNKNOWN_08_reply(tvbuff_t *tvb, int offset,
 
 
 /*
-  IDL long Function_09(
+  IDL long BrowserrSetNetlogonState(
   IDL       [in] [unique] [string] wchar_t *element_78,
   IDL       [in] [ref] [string] wchar_t *element_79,
   IDL       [in] [unique] [string] wchar_t *element_80,
@@ -883,9 +884,9 @@ dissect_browser_UNKNOWN_08_reply(tvbuff_t *tvb, int offset,
   IDL );
 */
 static int
-dissect_browser_UNKNOWN_09_rqst(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_set_netlogon_state_rqst(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_str_pointer_item(tvb, offset, pinfo, tree, drep,
 			NDR_POINTER_UNIQUE, "unknown string", 
@@ -905,9 +906,9 @@ dissect_browser_UNKNOWN_09_rqst(tvbuff_t *tvb, int offset,
 	return offset;
 }
 static int
-dissect_browser_UNKNOWN_09_reply(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_set_netlogon_state_reply(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_rc, NULL);
@@ -927,7 +928,7 @@ dissect_browser_UNKNOWN_09_reply(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_12_data(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	guint32 len;
 	dcerpc_info *di;
@@ -953,7 +954,7 @@ dissect_browser_TYPE_12_data(tvbuff_t *tvb, int offset,
 static int
 dissect_browser_TYPE_12(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_uint32(tvb, offset, pinfo, tree, drep,
 		hf_browser_unknown_long, NULL);
@@ -967,15 +968,15 @@ dissect_browser_TYPE_12(tvbuff_t *tvb, int offset,
 
 
 /*
-  IDL long Function_0a(
+  IDL long BrowserrQueryEmulatedDomains(
   IDL       [in] [unique] [string] wchar_t *element_84,
   IDL       [in,out] [ref] TYPE_12 *element_85
   );
 */
 static int
-dissect_browser_UNKNOWN_0a_rqst(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_query_emulated_domains_rqst(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_str_pointer_item(tvb, offset, pinfo, tree, drep,
 			NDR_POINTER_UNIQUE, "unknown string", 
@@ -988,9 +989,9 @@ dissect_browser_UNKNOWN_0a_rqst(tvbuff_t *tvb, int offset,
 	return offset;
 }
 static int
-dissect_browser_UNKNOWN_0a_reply(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_query_emulated_domains_reply(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_pointer(tvb, offset, pinfo, tree, drep,
 			dissect_browser_TYPE_12, NDR_POINTER_REF,
@@ -1004,7 +1005,7 @@ dissect_browser_UNKNOWN_0a_reply(tvbuff_t *tvb, int offset,
 
 
 /*
-  IDL long Function_0b(
+  IDL long BrowserrServerEnumEx(
   IDL       [in] [unique] [string] wchar_t *element_86,
   IDL       [in] [unique] [string] wchar_t *element_87,
   IDL       [in] [unique] [string] wchar_t *element_88,
@@ -1017,9 +1018,9 @@ dissect_browser_UNKNOWN_0a_reply(tvbuff_t *tvb, int offset,
   IDL );
 */
 static int
-dissect_browser_UNKNOWN_0b_rqst(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_server_enum_ex_rqst(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_str_pointer_item(tvb, offset, pinfo, tree, drep,
 			NDR_POINTER_UNIQUE, "unknown string", 
@@ -1054,9 +1055,9 @@ dissect_browser_UNKNOWN_0b_rqst(tvbuff_t *tvb, int offset,
 	return offset;
 }
 static int
-dissect_browser_UNKNOWN_0b_reply(tvbuff_t *tvb, int offset,
+dissect_browser_browserr_server_enum_ex_reply(tvbuff_t *tvb, int offset,
 			packet_info *pinfo, proto_tree *tree,
-			char *drep)
+			guint8 *drep)
 {
 	offset = dissect_ndr_pointer(tvb, offset, pinfo, tree, drep,
 			dissect_browser_TYPE_1, NDR_POINTER_REF,
@@ -1077,42 +1078,52 @@ dissect_browser_UNKNOWN_0b_reply(tvbuff_t *tvb, int offset,
   IDL }
 */
 static dcerpc_sub_dissector dcerpc_browser_dissectors[] = {
-        { BROWSER_UNKNOWN_00, "BROWSER_UNKNOWN_00",
-		dissect_browser_UNKNOWN_00_rqst,
-		dissect_browser_UNKNOWN_00_reply },
-        { BROWSER_UNKNOWN_01, "BROWSER_UNKNOWN_01",
-		dissect_browser_UNKNOWN_01_rqst,
-		dissect_browser_UNKNOWN_01_reply },
-        { BROWSER_UNKNOWN_02, "BROWSER_UNKNOWN_02",
-		dissect_browser_UNKNOWN_02_rqst,
-		dissect_browser_UNKNOWN_02_reply },
-        { BROWSER_UNKNOWN_03, "BROWSER_UNKNOWN_03",
-		dissect_browser_UNKNOWN_03_rqst,
-		dissect_browser_UNKNOWN_03_reply },
-        { BROWSER_UNKNOWN_04, "BROWSER_UNKNOWN_04",
-		dissect_browser_UNKNOWN_04_rqst,
-		dissect_browser_UNKNOWN_04_reply },
-        { BROWSER_UNKNOWN_05, "BROWSER_UNKNOWN_05",
-		dissect_browser_UNKNOWN_05_rqst,
-		dissect_browser_UNKNOWN_05_reply },
-        { BROWSER_UNKNOWN_06, "BROWSER_UNKNOWN_06",
-		dissect_browser_UNKNOWN_06_rqst,
-		dissect_browser_UNKNOWN_06_reply },
-        { BROWSER_UNKNOWN_07, "BROWSER_UNKNOWN_07",
-		dissect_browser_UNKNOWN_07_rqst,
-		dissect_browser_UNKNOWN_07_reply },
-        { BROWSER_UNKNOWN_08, "BROWSER_UNKNOWN_08",
-		dissect_browser_UNKNOWN_08_rqst,
-		dissect_browser_UNKNOWN_08_reply },
-        { BROWSER_UNKNOWN_09, "BROWSER_UNKNOWN_09",
-		dissect_browser_UNKNOWN_09_rqst,
-		dissect_browser_UNKNOWN_09_reply },
-        { BROWSER_UNKNOWN_0a, "BROWSER_UNKNOWN_0a",
-		dissect_browser_UNKNOWN_0a_rqst,
-		dissect_browser_UNKNOWN_0a_reply },
-        { BROWSER_UNKNOWN_0b, "BROWSER_UNKNOWN_0b",
-		dissect_browser_UNKNOWN_0b_rqst,
-		dissect_browser_UNKNOWN_0b_reply },
+        { BROWSER_BROWSERR_SERVER_ENUM, "BrowserrServerEnum",
+		dissect_browser_browserr_server_enum_rqst,
+		dissect_browser_browserr_server_enum_reply },
+        { BROWSER_BROWSERR_DEBUG_CALL, "BrowserrDebugCall",
+		dissect_browser_browserr_debug_call_rqst,
+		dissect_browser_browserr_debug_call_reply },
+        { BROWSER_BROWSERR_QUERY_OTHER_DOMAINS,	 
+	  "BrowserrQueryOtherDomains",
+		dissect_browser_browserr_query_other_domains_rqst,
+		dissect_browser_browserr_query_other_domains_reply },
+        { BROWSER_BROWSERR_RESET_NETLOGON_STATE, 
+	  "BrowserrResetNetlogonState",
+		dissect_browser_browserr_reset_netlogon_state_rqst,
+		dissect_browser_browserr_reset_netlogon_state_reply },
+        { BROWSER_BROWSERR_DEBUG_TRACE, 
+	  "BrowserrDebugTrace",
+		dissect_browser_browserr_debug_trace_rqst,
+		dissect_browser_browserr_debug_trace_reply },
+        { BROWSER_BROWSERR_QUERY_STATISTICS, 
+	  "BrowserrQueryStatistics",
+		dissect_browser_browserr_query_statistics_rqst,
+		dissect_browser_browserr_query_statistics_reply },
+        { BROWSER_BROWSERR_RESET_STATISTICS, 
+	  "BrowserrResetStatistics",
+		dissect_browser_browserr_reset_statistics_rqst,
+		dissect_browser_browserr_reset_statistics_reply },
+        { BROWSER_NETR_BROWSER_STATISTICS_CLEAR, 
+	  "NetrBrowserStatisticsClear",
+		dissect_browser_netr_browser_statistics_clear_rqst,
+		dissect_browser_netr_browser_statistics_clear_reply },
+        { BROWSER_NETR_BROWSER_STATISTICS_GET, 
+	  "NetrBrowserStatisticsGet",
+		dissect_browser_netr_browser_statistics_get_rqst,
+		dissect_browser_netr_browser_statistics_get_reply },
+        { BROWSER_BROWSERR_SET_NETLOGON_STATE, 
+	  "BrowserrSetNetlogonState",
+		dissect_browser_browserr_set_netlogon_state_rqst,
+		dissect_browser_browserr_set_netlogon_state_reply },
+        { BROWSER_BROWSERR_QUERY_EMULATED_DOMAINS, 
+	  "BrowserrQueryEmulatedDomains",
+		dissect_browser_browserr_query_emulated_domains_rqst,
+		dissect_browser_browserr_query_emulated_domains_reply },
+        { BROWSER_BROWSERR_SERVER_ENUM_EX, 
+	  "BrowserrServerEnumEx",
+		dissect_browser_browserr_server_enum_ex_rqst,
+		dissect_browser_browserr_server_enum_ex_reply },
 
         {0, NULL, NULL,  NULL }
 };
@@ -1121,6 +1132,11 @@ void
 proto_register_dcerpc_browser(void)
 {
 static hf_register_info hf[] = {
+
+	{ &hf_browser_opnum, { 
+		"Operation", "rpc_browser.opnum", FT_UINT16, BASE_DEC,
+		NULL, 0x0, "Operation", HFILL }},
+
 	{ &hf_browser_rc, {
 		"Return code", "rpc_browser.rc", FT_UINT32, BASE_HEX,
 		VALS(NT_errors), 0x0, "Browser return code", HFILL }},
@@ -1161,5 +1177,5 @@ proto_reg_handoff_dcerpc_browser(void)
 
         dcerpc_init_uuid(proto_dcerpc_browser, ett_dcerpc_browser,
                          &uuid_dcerpc_browser, ver_dcerpc_browser,
-                         dcerpc_browser_dissectors, -1);
+                         dcerpc_browser_dissectors, hf_browser_opnum);
 }
