@@ -1,7 +1,7 @@
 /* packet-diameter.c
  * Routines for Diameter packet disassembly
  *
- * $Id: packet-diameter.c,v 1.51 2002/12/02 23:43:26 guy Exp $
+ * $Id: packet-diameter.c,v 1.65 2004/03/23 21:19:55 guy Exp $
  *
  * Copyright (c) 2001 by David Frascone <dave@frascone.com>
  *
@@ -22,6 +22,15 @@
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
+ * References:
+ * 2004-03-11
+ * http://www.ietf.org/rfc/rfc3588.txt
+ * http://www.iana.org/assignments/radius-types
+ * http://www.ietf.org/internet-drafts/draft-ietf-aaa-diameter-cc-03.txt
+ * http://www.ietf.org/internet-drafts/draft-ietf-aaa-diameter-nasreq-14.txt
+ * http://www.ietf.org/internet-drafts/draft-ietf-aaa-diameter-mobileip-16.txt
+ * http://www.ietf.org/internet-drafts/draft-ietf-aaa-diameter-sip-app-01.txt
+ * http://www.ietf.org/html.charters/aaa-charter.html
  */
 
 #ifdef HAVE_CONFIG_H
@@ -31,6 +40,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <errno.h>
 #include <ctype.h>
 #include <time.h>
 #include <glib.h>
@@ -38,7 +48,9 @@
 #include "xmlstub.h"
 #include <epan/packet.h>
 #include <epan/resolv.h>
+#include <epan/report_err.h>
 #include "prefs.h"
+#include "packet-tcp.h"
 
 #ifdef NEED_SNPRINTF_H
 # include "snprintf.h"
@@ -68,8 +80,10 @@ typedef enum {
   DIAMETER_IP_FILTER_RULE,     /* OctetString */
   DIAMETER_QOS_FILTER_RULE,    /* OctetString */
   DIAMETER_MIP_REG_REQ,        /* OctetString */
-  DIAMETER_VENDOR_ID,           /* Integer32  */
-  DIAMETER_APPLICATION_ID
+  DIAMETER_VENDOR_ID,          /* Integer32  */
+  DIAMETER_APPLICATION_ID,
+  DIAMETER_URI,                /* OctetString */
+  DIAMETER_SESSION_ID          /* OctetString */
 
 } diameterDataType;
 
@@ -94,6 +108,9 @@ static value_string TypeValues[]={
   {  DIAMETER_MIP_REG_REQ,     "MIPRegistrationRequest"},
   {  DIAMETER_VENDOR_ID,       "VendorId"},
   {  DIAMETER_APPLICATION_ID,  "AppId"},
+  {  DIAMETER_URI,             "DiameterURI"},
+  {  DIAMETER_SESSION_ID,      "Session-Id"},
+	
   {0, (char *)NULL}
 };
 
@@ -152,15 +169,11 @@ static ApplicationId   *ApplicationIdHead=NULL;
 #define TCP_PORT_DIAMETER	1812
 #define SCTP_PORT_DIAMETER	1812
 
-static const true_false_string flags_set_truth = {
-  "Set",
-  "Not set"
-};
-
 static const true_false_string reserved_set = {
   "*** Error! Reserved Bit is Set",
   "Ok"
 };
+
 static int proto_diameter = -1;
 static int hf_diameter_length = -1;
 static int hf_diameter_code = -1;
@@ -172,7 +185,7 @@ static int hf_diameter_flags = -1;
 static int hf_diameter_flags_request = -1;
 static int hf_diameter_flags_proxyable = -1;
 static int hf_diameter_flags_error = -1;
-static int hf_diameter_flags_reserved3 = -1;
+static int hf_diameter_flags_T		= -1;
 static int hf_diameter_flags_reserved4 = -1;
 static int hf_diameter_flags_reserved5 = -1;
 static int hf_diameter_flags_reserved6 = -1;
@@ -201,22 +214,24 @@ static int hf_diameter_avp_data_string = -1;
 static int hf_diameter_avp_data_v4addr = -1;
 static int hf_diameter_avp_data_v6addr = -1;
 static int hf_diameter_avp_data_time = -1;
-
+static int hf_diameter_avp_session_id		= -1;
 static gint ett_diameter = -1;
 static gint ett_diameter_flags = -1;
 static gint ett_diameter_avp = -1;
 static gint ett_diameter_avp_flags = -1;
 static gint ett_diameter_avpinfo = -1;
 
-static char gbl_diameterString[200];
-static int gbl_diameterTcpPort=TCP_PORT_DIAMETER;
-static int gbl_diameterSctpPort=SCTP_PORT_DIAMETER;
+static guint gbl_diameterTcpPort=TCP_PORT_DIAMETER;
+static guint gbl_diameterSctpPort=SCTP_PORT_DIAMETER;
 
 /* desegmentation of Diameter over TCP */
 static gboolean gbl_diameter_desegment = TRUE;
 
+/* Allow zero as a valid application ID */
+static gboolean allow_zero_as_app_id = FALSE;
+
 #define DICT_FN  "diameter/dictionary.xml"
-static gchar *gbl_diameterDictionary = NULL;
+static gchar *gbl_diameterDictionary;
 
 typedef struct _e_diameterhdr {
   guint32  versionLength;
@@ -237,12 +252,12 @@ typedef struct _e_avphdr {
 #define DIAM_FLAGS_R 0x80
 #define DIAM_FLAGS_P 0x40
 #define DIAM_FLAGS_E 0x20
-#define DIAM_FLAGS_RESERVED3 0x10
+#define DIAM_FLAGS_T 0x10
 #define DIAM_FLAGS_RESERVED4 0x08
 #define DIAM_FLAGS_RESERVED5 0x04
 #define DIAM_FLAGS_RESERVED6 0x02
 #define DIAM_FLAGS_RESERVED7 0x01
-#define DIAM_FLAGS_RESERVED  0x1f
+#define DIAM_FLAGS_RESERVED  0x0f
 
 #define DIAM_LENGTH_MASK  0x00ffffffl
 #define DIAM_COMMAND_MASK DIAM_LENGTH_MASK
@@ -287,7 +302,7 @@ xmlParseFilePush( char *filename, int checkValid) {
 
   f = fopen(filename, "r");
   if (f == NULL) {
-	g_warning("Diameter: Unable to open %s", filename);
+	report_open_failure(filename, errno, FALSE);
 	return NULL;
   }
 
@@ -346,7 +361,7 @@ addStaticAVP(int code, gchar *name, diameterDataType type, value_string *values)
   entry->type = type;
   entry->values = vEntry;
   if (vEntry)
-	entry->type = DIAMETER_INTEGER32;
+	entry->type = DIAMETER_ENUMERATED;
 
   /* And, add it to the list */
   entry->next = avpListHead;
@@ -526,8 +541,8 @@ dictionaryAddApplication(char *name, int id)
 {
   ApplicationId *entry;
 
-  if (!name || (id <= 0)) {
-	g_warning( "Diameter Error: Inavlid application (name=%p, id=%d)",
+  if (!name || (id < 0) || (id == 0 && !allow_zero_as_app_id)) {
+	g_warning( "Diameter Error: Invalid application (name=%p, id=%d)",
 			   name, id);
 	return (-1);
   } /* Sanity Checks */
@@ -658,7 +673,7 @@ xmlDictionaryParseSegment(xmlNodePtr cur, int base)
 static int
 xmlDictionaryParse(xmlNodePtr cur)
 {
-  /* We should expect a base protocol, followed by multiple applicaitons */
+  /* We should expect a base protocol, followed by multiple applications */
   while (cur != NULL) {
 	if (strcasecmp(cur->name, "base") == 0) {
 	  /* Base protocol.  Descend and parse */
@@ -668,6 +683,8 @@ xmlDictionaryParse(xmlNodePtr cur)
 	  xmlDictionaryParseSegment(cur, 0);
 	} else if (strcasecmp(cur->name, "text") == 0) {
 	  /* Ignore text */
+	} else if (strcasecmp(cur->name, "comment") == 0) {
+	  /* Ignore text */
 	} else {
 	  g_warning( "Diameter: XML Expecting a base or an application  (got \"%s\")",
 				 cur->name);
@@ -684,7 +701,7 @@ xmlDictionaryParse(xmlNodePtr cur)
  * This routine will call libxml to parse in the dictionary.
  */
 static int
-loadXMLDictionary()
+loadXMLDictionary(void)
 {
   xmlDocPtr doc;
   xmlNodePtr cur;
@@ -742,7 +759,7 @@ loadXMLDictionary()
  * with the old static data from packet-diameter-defs.h
  */
 static void
-initializeDictionaryDefaults()
+initializeDictionaryDefaults(void)
 {
   int i;
 
@@ -774,7 +791,7 @@ initializeDictionaryDefaults()
  * our static dictionary.
  */
 static void
-initializeDictionary()
+initializeDictionary(void)
 {
   /*
    * Using ugly ordering here.  If loadLibXML succeeds, then
@@ -992,11 +1009,25 @@ diameter_avp_get_value(guint32 avpCode, guint32 vendorId, guint32 avpValue)
 
 /* Code to actually dissect the packets */
 
+static gboolean
+check_diameter(tvbuff_t *tvb)
+{
+  if (!tvb_bytes_exist(tvb, 0, 1))
+	return FALSE;	/* not enough bytes to check the version */
+  if (tvb_get_guint8(tvb, 0) != 1)
+	return FALSE;	/* not version 1 */
+
+  /* XXX - fetch length and make sure it's at least MIN_DIAMETER_SIZE?
+     Fetch flags and check that none of the DIAM_FLAGS_RESERVED bits
+     are set? */
+  return TRUE;
+}
+
 /*
  * Main dissector
  */
-static guint32 dissect_diameter_common(tvbuff_t *tvb, size_t start, packet_info *pinfo,
-									   proto_tree *tree)
+static void
+dissect_diameter_common(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 {
 
   /* Set up structures needed to add the protocol subtree and manage it */
@@ -1006,7 +1037,7 @@ static guint32 dissect_diameter_common(tvbuff_t *tvb, size_t start, packet_info
   tvbuff_t        *avp_tvb;
   proto_tree      *diameter_tree;
   e_diameterhdr    dh;
-  size_t           offset=0;
+  int              offset=0;
   size_t           avplength;
   proto_tree      *avp_tree;
   proto_item      *avptf;
@@ -1020,9 +1051,6 @@ static guint32 dissect_diameter_common(tvbuff_t *tvb, size_t start, packet_info
   guint      bpos;
   static  int initialized=FALSE;
 
-  /* set our offset */
-  offset=start;
-
   /*
    * Only parse in dictionary if there are diameter packets to
    * dissect.
@@ -1150,7 +1178,7 @@ static guint32 dissect_diameter_common(tvbuff_t *tvb, size_t start, packet_info
 	proto_tree_add_boolean(flags_tree, hf_diameter_flags_request, tvb, offset, 1, flags);
 	proto_tree_add_boolean(flags_tree, hf_diameter_flags_proxyable, tvb, offset, 1, flags);
 	proto_tree_add_boolean(flags_tree, hf_diameter_flags_error, tvb, offset, 1, flags);
-	proto_tree_add_boolean(flags_tree, hf_diameter_flags_reserved3, tvb, offset, 1, flags);
+	proto_tree_add_boolean(flags_tree, hf_diameter_flags_T, tvb, offset, 1, flags);
 	proto_tree_add_boolean(flags_tree, hf_diameter_flags_reserved4, tvb, offset, 1, flags);
 	proto_tree_add_boolean(flags_tree, hf_diameter_flags_reserved5, tvb, offset, 1, flags);
 	proto_tree_add_boolean(flags_tree, hf_diameter_flags_reserved6, tvb, offset, 1, flags);
@@ -1180,7 +1208,7 @@ static guint32 dissect_diameter_common(tvbuff_t *tvb, size_t start, packet_info
 
 	/* If we have a bad packet, don't bother trying to parse the AVPs */
 	if (BadPacket) {
-	  return (offset + MAX(pktLength,MIN_DIAMETER_SIZE));
+	  return;
 	}
 
 	/* Start looking at the AVPS */
@@ -1199,59 +1227,35 @@ static guint32 dissect_diameter_common(tvbuff_t *tvb, size_t start, packet_info
 	if (avp_tree != NULL) {
 	  dissect_avps( avp_tvb, pinfo, avp_tree);
 	}
-	return MAX((offset + avplength), MIN_DIAMETER_SIZE);
+	return;
   }
-  return (offset + MAX(pktLength, MIN_DIAMETER_SIZE));
-
 } /* dissect_diameter_common */
 
-static void
-dissect_diameter_sctp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
+
+static guint
+get_diameter_pdu_len(tvbuff_t *tvb, int offset)
 {
-	dissect_diameter_common(tvb, 0, pinfo, tree);
+  /* Get the length of the Diameter packet. */
+  return tvb_get_ntoh24(tvb, offset + 1);
 }
 
-static void
+static int
+dissect_diameter(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
+{
+  if (!check_diameter(tvb))
+	return 0;
+  dissect_diameter_common(tvb, pinfo, tree);
+  return tvb_length(tvb);
+}
+
+static int
 dissect_diameter_tcp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 {
-	size_t offset = 0;
-	guint32 plen;
-	guint32 available_bytes;
-/* 	guint32 noffset; */
-
-	/* Loop through the packet, dissecting multiple diameter messages */
-	do {
-		available_bytes = tvb_length_remaining(tvb, offset);
-		if (available_bytes < 4) {
-			g_warning("Diameter:  Bailing because only %d bytes of packet are available",
-					  available_bytes);
-			return; /* Bail.  We can't even get our length */
-		}
-
-		/* get our packet length */
-		plen = tvb_get_ntohl(tvb, offset);
-		plen &= 0x00ffffff; /* get rid of the flags */
-
-		/*Desegmentation */
-		if (gbl_diameter_desegment) {
-			if (pinfo->can_desegment
-				&& plen > available_bytes) {
-				pinfo->desegment_offset = offset;
-				pinfo->desegment_len = plen - available_bytes;
-/* 				g_warning("Diameter: Bailing for deseg because plen(%d) > available(%d)", */
-/* 						  plen, available_bytes); */
-				return;
-			}
-		}
-
-		/* Otherwise, dissect our packet */
-		offset = dissect_diameter_common(tvb, offset, pinfo, tree);
-
-/* 		g_warning("dissected from %d to %d bytes out of %d (available was %d plen was %d)", */
-/* 				  offset, noffset, tvb_length(tvb), available_bytes, plen); */
-/* 		offset=noffset; */
-	} while (offset < tvb_reported_length(tvb));
-
+  if (!check_diameter(tvb))
+	return 0;
+  tcp_dissect_pdus(tvb, pinfo, tree, gbl_diameter_desegment, 4,
+	get_diameter_pdu_len, dissect_diameter_common);
+  return tvb_length(tvb);
 } /* dissect_diameter_tcp */
 
 /*
@@ -1430,8 +1434,9 @@ static void dissect_avps(tvbuff_t *tvb, packet_info *pinfo, proto_tree *avp_tree
 	  g_warning("Diameter: Invalid AVP: Reserved bit set.  flags = 0x%x,"
 				" resFl=0x%x",
 				flags, AVP_FLAGS_RESERVED);
-	  /* For now, don't set bad packet, since I'm accidentally setting a wrong bit */
-	  /* BadPacket = TRUE; */
+	  /* For now, don't set bad packet, since I'm accidentally setting a wrong bit 
+	   BadPacket = TRUE; 
+	   */
 	}
 
 	/*
@@ -1663,6 +1668,7 @@ static void dissect_avps(tvbuff_t *tvb, packet_info *pinfo, proto_tree *avp_tree
 					      "Error!  Bad Enumerated Length");
 		}
 		break;
+
 	  case DIAMETER_VENDOR_ID:
 		if (avpDataLength == 4) {
 		  guint32 data;
@@ -1680,6 +1686,7 @@ static void dissect_avps(tvbuff_t *tvb, packet_info *pinfo, proto_tree *avp_tree
 					      "Error!  Bad Vendor ID Length");
 		}
 		break;
+
 	  case DIAMETER_APPLICATION_ID:
 		if (avpDataLength == 4) {
 		  guint32 data;
@@ -1697,10 +1704,16 @@ static void dissect_avps(tvbuff_t *tvb, packet_info *pinfo, proto_tree *avp_tree
 					      "Error!  Bad Application ID Length");
 		}
 		break;
+
 	  case DIAMETER_MIP_REG_REQ:
 		safe_dissect_mip(tvb, pinfo, avpi_tree, offset, avpDataLength);
 		break;
 
+	  case DIAMETER_SESSION_ID:
+		proto_tree_add_item(avpi_tree, hf_diameter_avp_session_id,
+				    tvb, offset, avpDataLength, FALSE);
+		break;
+
 	  default:
 	  case DIAMETER_OCTET_STRING:
 		proto_tree_add_bytes_format(avpi_tree, hf_diameter_avp_data_bytes,
@@ -1725,29 +1738,27 @@ proto_reg_handoff_diameter(void)
   static int TcpPort=0;
   static int SctpPort=0;
   static dissector_handle_t diameter_tcp_handle;
-  static dissector_handle_t diameter_sctp_handle;
+  static dissector_handle_t diameter_handle;
 
   if (!Initialized) {
-	diameter_tcp_handle = create_dissector_handle(dissect_diameter_tcp,
+	diameter_tcp_handle = new_create_dissector_handle(dissect_diameter_tcp,
 	    proto_diameter);
-	diameter_sctp_handle = create_dissector_handle(dissect_diameter_sctp,
+	diameter_handle = new_create_dissector_handle(dissect_diameter,
 	    proto_diameter);
 	Initialized=TRUE;
   } else {
 	dissector_delete("tcp.port", TcpPort, diameter_tcp_handle);
-	dissector_delete("sctp.port", SctpPort, diameter_sctp_handle);
+	dissector_delete("sctp.port", SctpPort, diameter_handle);
   }
 
   /* set port for future deletes */
   TcpPort=gbl_diameterTcpPort;
   SctpPort=gbl_diameterSctpPort;
 
-  strcpy(gbl_diameterString, "Diameter Protocol");
-
   /* g_warning ("Diameter: Adding tcp dissector to port %d",
 	 gbl_diameterTcpPort); */
   dissector_add("tcp.port", gbl_diameterTcpPort, diameter_tcp_handle);
-  dissector_add("sctp.port", gbl_diameterSctpPort, diameter_sctp_handle);
+  dissector_add("sctp.port", gbl_diameterSctpPort, diameter_handle);
 }
 
 /* registration with the filtering engine */
@@ -1774,9 +1785,9 @@ proto_register_diameter(void)
 		{ &hf_diameter_flags_error,
 		  { "Error","diameter.flags.error", FT_BOOLEAN, 8, TFS(&flags_set_truth), DIAM_FLAGS_E,
 			"", HFILL }},
-		{ &hf_diameter_flags_reserved3,
-		  { "Reserved","diameter.flags.reserved3", FT_BOOLEAN, 8, TFS(&reserved_set),
-			DIAM_FLAGS_RESERVED3, "", HFILL }},
+		{ &hf_diameter_flags_T,
+		  { "T(Potentially re-transmitted message)","diameter.flags.T", FT_BOOLEAN, 8, TFS(&flags_set_truth),DIAM_FLAGS_T,
+			"", HFILL }},
 		{ &hf_diameter_flags_reserved4,
 		  { "Reserved","diameter.flags.reserved4", FT_BOOLEAN, 8, TFS(&reserved_set),
 			DIAM_FLAGS_RESERVED4, "", HFILL }},
@@ -1868,6 +1879,9 @@ proto_register_diameter(void)
 		{ &hf_diameter_avp_data_time,
 		  { "Time","diameter.avp.data.time", FT_ABSOLUTE_TIME, BASE_NONE,
 		    NULL, 0x0, "", HFILL }},
+		{ &hf_diameter_avp_session_id,
+		  { "Session ID","diameter.avp.session_id", FT_STRING, BASE_NONE,
+		    NULL, 0x0, "", HFILL }},
 
 	};
 	static gint *ett[] = {
@@ -1878,9 +1892,9 @@ proto_register_diameter(void)
 		&ett_diameter_avpinfo
 	};
 	module_t *diameter_module;
+	gchar *default_diameterDictionary;
 
-	proto_diameter = proto_register_protocol (gbl_diameterString,
-											  "Diameter", "diameter");
+	proto_diameter = proto_register_protocol ("Diameter Protocol", "Diameter", "diameter");
 	proto_register_field_array(proto_diameter, hf, array_length(hf));
 	proto_register_subtree_array(ett, array_length(ett));
 
@@ -1900,26 +1914,39 @@ proto_register_diameter(void)
 	/*
 	 * Build our default dictionary filename
 	 */
-	if (! gbl_diameterDictionary) {
-		gbl_diameterDictionary = (gchar *) g_malloc(strlen(get_datafile_dir()) +
-													1 + strlen(DICT_FN) + 1); /* slash + fn + null */
-		sprintf(gbl_diameterDictionary, "%s" G_DIR_SEPARATOR_S "%s",
-				get_datafile_dir(), DICT_FN );
-	}
-	/* Now register its preferences so it can be changed. */
+	default_diameterDictionary = get_datafile_path(DICT_FN);
+
+	/*
+	 * Now register the dictionary filename as a preference,
+	 * so it can be changed.
+	 */
+	gbl_diameterDictionary = default_diameterDictionary;
 	prefs_register_string_preference(diameter_module, "dictionary.name",
 									 "Diameter XML Dictionary",
 									 "Set the dictionary used for Diameter messages",
 									 &gbl_diameterDictionary);
 
+	/*
+	 * We don't need the default dictionary, so free it (a copy was made
+	 * of it in "gbl_diameterDictionary" by
+	 * "prefs_register_string_preference()").
+	 */
+	g_free(default_diameterDictionary);
+
 	/* Desegmentation */
 	prefs_register_bool_preference(diameter_module, "desegment",
-								   "Desegment all Diameter messages spanning multiple TCP segments",
+								   "Desegment all Diameter messages\nspanning multiple TCP segments",
 								   "Whether the Diameter dissector should desegment all messages spanning multiple TCP segments",
 								   &gbl_diameter_desegment);
-
+	/* Allow zero as valid application ID */
+	prefs_register_bool_preference(diameter_module, "allow_zero_as_app_id",
+			"Allow 0 as valid application ID",
+			"If set, the value 0 (zero) can be used as a valid "
+			"application ID. This is used in experimental cases.",
+			&allow_zero_as_app_id);
 	/* Register some preferences we no longer support, so we can report
 	   them as obsolete rather than just illegal. */
 	prefs_register_obsolete_preference(diameter_module, "udp.port");
 	prefs_register_obsolete_preference(diameter_module, "command_in_header");
 } /* proto_register_diameter */
+
