@@ -1,7 +1,7 @@
 /* packet-dns.c
  * Routines for DNS packet disassembly
  *
- * $Id: packet-dns.c,v 1.99 2003/01/31 08:29:09 guy Exp $
+ * $Id: packet-dns.c,v 1.120 2004/01/23 19:16:52 guy Exp $
  *
  * Ethereal - Network traffic analyzer
  * By Gerald Combs <gerald@ethereal.com>
@@ -35,6 +35,7 @@
 #endif
 
 #include <glib.h>
+#include <epan/ipv6-utils.h>
 #include <epan/packet.h>
 #include "ipproto.h"
 #include <epan/resolv.h>
@@ -51,6 +52,7 @@ static int hf_dns_flags_authoritative = -1;
 static int hf_dns_flags_truncated = -1;
 static int hf_dns_flags_recdesired = -1;
 static int hf_dns_flags_recavail = -1;
+static int hf_dns_flags_z = -1;
 static int hf_dns_flags_authenticated = -1;
 static int hf_dns_flags_checkdisable = -1;
 static int hf_dns_flags_rcode = -1;
@@ -135,6 +137,10 @@ static dissector_handle_t gssapi_handle;
 #define T_A6		38              /* IPv6 address with indirection (RFC 2874) */
 #define T_DNAME         39              /* Non-terminal DNS name redirection (RFC 2672) */
 #define T_OPT		41		/* OPT pseudo-RR (RFC 2671) */
+#define T_DS            43		/* Delegation Signature(RFC 3658) */
+#define T_RRSIG         46              /* future RFC 2535bis */
+#define T_NSEC          47              /* future RFC 2535bis */
+#define T_DNSKEY        48              /* future RFC 2535bis */
 #define T_TKEY		249		/* Transaction Key (RFC 2930) */
 #define T_TSIG		250		/* Transaction Signature (RFC 2845) */
 #define T_WINS		65281		/* Microsoft's WINS RR */
@@ -157,6 +163,7 @@ static dissector_handle_t gssapi_handle;
 #define F_TRUNCATED     (1<<9)          /* response is truncated */
 #define F_RECDESIRED    (1<<8)          /* recursion desired */
 #define F_RECAVAIL      (1<<7)          /* recursion available */
+#define F_Z		(1<<6)		/* Z */
 #define F_AUTHENTIC     (1<<5)          /* authentic data (RFC2535) */
 #define F_CHECKDISABLE  (1<<4)          /* checking disabled (RFC2535) */
 #define F_RCODE         (0xF<<0)        /* reply code */
@@ -186,6 +193,11 @@ static const true_false_string tfs_flags_recavail = {
 	"Server can't do recursive queries"
 };
 
+static const true_false_string tfs_flags_z = {
+	"reserved - incorrect!",
+	"reserved (0)"
+};
+
 static const true_false_string tfs_flags_authenticated = {
 	"Answer/authority portion was authenticated by the server",
 	"Answer/authority portion was not authenticated by the server"
@@ -261,6 +273,9 @@ static const value_string tsigerror_vals[] = {
 #define TKEYMODE_RESOLVERASSIGNED           (4)
 #define TKEYMODE_DELETE                     (5)
 
+#define TDSDIGEST_RESERVED (0)
+#define TDSDIGEST_SHA1     (1)
+
 /* See RFC 1035 for all RR types for which no RFC is listed, except for
    the ones with "???", and for the Microsoft WINS and WINS-R RRs, for
    which one should look at
@@ -317,7 +332,14 @@ dns_type_name (guint type)
     "A6",				/* RFC 2874 */
     "DNAME",				/* RFC 2672 */
     NULL,
-    "OPT"				/* RFC 2671 */
+    "OPT",				/* RFC 2671 */
+    NULL,
+    "DS",				/* RFC 3658 */
+    NULL,
+    NULL,
+    "RRSIG",                            /* future RFC 2535bis */
+    "NSEC",                             /* future RFC 2535bis */
+    "DNSKEY"                            /* future RFC 2535bis */
   };
 
   if (type < sizeof(type_names)/sizeof(type_names[0]))
@@ -409,7 +431,14 @@ dns_long_type_name (guint type)
     "IPv6 address with indirection",	/* RFC 2874 */
     "Non-terminal DNS name redirection", /* RFC 2672 */
     NULL,
-    "EDNS0 option"			/* RFC 2671 */
+    "EDNS0 option",			/* RFC 2671 */
+    NULL,
+    "Delegation Signer",                /* RFC 3658 */
+    NULL,
+    NULL,
+    "RR signature",                     /* future RFC 2535bis */
+    "Next secured",                     /* future RFC 2535bis */
+    "DNS public key"                    /* future RFC 2535bis */
   };
   static char unkbuf[7+1+2+1+4+1+1+10+1+1];	/* "Unknown RR type (%u)" */
 
@@ -554,9 +583,9 @@ get_dns_name(tvbuff_t *tvb, int offset, int dns_data_offset,
 
 	  if (maxname > 0) {
 	    print_len = snprintf(np, maxname + 1, "\\[x");
-	    if (print_len != -1) {
+	    if (print_len != -1 && print_len <= maxname) {
 	      /* Some versions of snprintf return -1 if they'd truncate
-	         the output. */
+	         the output.  Others return <buf_size> or greater. */
 	      np += print_len;
 	      maxname -= print_len;
 	    } else {
@@ -569,9 +598,9 @@ get_dns_name(tvbuff_t *tvb, int offset, int dns_data_offset,
 	    if (maxname > 0) {
 	      print_len = snprintf(np, maxname + 1, "%02x",
 	        tvb_get_guint8(tvb, offset));
-	      if (print_len != -1) {
+	      if (print_len != -1 && print_len <= maxname) {
 		/* Some versions of snprintf return -1 if they'd truncate
-		 the output. */
+		 the output.  Others return <buf_size> or greater. */
 		np += print_len;
 		maxname -= print_len;
 	      } else {
@@ -584,9 +613,9 @@ get_dns_name(tvbuff_t *tvb, int offset, int dns_data_offset,
 	  }
 	  if (maxname > 0) {
 	    print_len = snprintf(np, maxname + 1, "/%d]", bit_count);
-	    if (print_len != -1) {
+	    if (print_len != -1 && print_len <= maxname) {
 	      /* Some versions of snprintf return -1 if they'd truncate
-	         the output. */
+	         the output.  Others return <buf_size> or greater. */
 	      np += print_len;
 	      maxname -= print_len;
 	    } else {
@@ -803,7 +832,8 @@ add_opt_rr_to_tree(proto_item *trr, int rr_type, tvbuff_t *tvb, int offset,
   const char *name, int namelen, const char *type_name, int class,
   guint ttl, gushort data_len)
 {
-  proto_tree *rr_tree;
+  proto_tree *rr_tree, *Z_tree;
+  proto_item *Z_item = NULL;
 
   rr_tree = proto_item_add_subtree(trr, rr_type);
   proto_tree_add_text(rr_tree, tvb, offset, namelen, "Name: %s", name);
@@ -819,7 +849,12 @@ add_opt_rr_to_tree(proto_item *trr, int rr_type, tvbuff_t *tvb, int offset,
   proto_tree_add_text(rr_tree, tvb, offset, 1, "EDNS0 version: %u",
       (ttl >> 16) & 0xff);
   offset++;
-  proto_tree_add_text(rr_tree, tvb, offset, 2, "Must be zero: 0x%x", ttl & 0xffff);
+  Z_item = proto_tree_add_text(rr_tree, tvb, offset, 2, "Z: 0x%x", ttl & 0xffff);
+  if (ttl & 0x8000) {
+     Z_tree = proto_item_add_subtree(Z_item, rr_type);
+     proto_tree_add_text(Z_tree, tvb, offset, 2, "Bit 0 (DO bit): 1 (Accepts DNSSEC security RRs)");
+     proto_tree_add_text(Z_tree, tvb, offset, 2, "Bits 1-15: 0x%x (reserved)", (ttl >> 17) & 0xff);
+  }
   offset += 2;
   proto_tree_add_text(rr_tree, tvb, offset, 2, "Data length: %u", data_len);
   return rr_tree;
@@ -832,6 +867,8 @@ add_opt_rr_to_tree(proto_item *trr, int rr_type, tvbuff_t *tvb, int offset,
 #define	DNS_ALGO_DH		2	/* Diffie-Hellman */
 #define	DNS_ALGO_DSA		3	/* DSA */
 #define	DNS_ALGO_ECC		4	/* Elliptic curve crypto */
+#define DNS_ALGO_RSASHA1        5	/* RSA/SHA1 */
+#define DNS_ALGO_HMACMD5        157	/* HMAC/MD5 */
 #define	DNS_ALGO_INDIRECT	252	/* Indirect key */
 #define	DNS_ALGO_PRIVATEDNS	253	/* Private, domain name  */
 #define	DNS_ALGO_PRIVATEOID	254	/* Private, OID */
@@ -841,6 +878,8 @@ static const value_string algo_vals[] = {
 	  { DNS_ALGO_DH,         "Diffie-Hellman" },
 	  { DNS_ALGO_DSA,        "DSA" },
 	  { DNS_ALGO_ECC,        "Elliptic curve crypto" },
+	  { DNS_ALGO_RSASHA1,    "RSA/SHA1" },
+	  { DNS_ALGO_HMACMD5,    "HMAC/MD5" },
 	  { DNS_ALGO_INDIRECT,   "Indirect key" },
 	  { DNS_ALGO_PRIVATEDNS, "Private, domain name" },
 	  { DNS_ALGO_PRIVATEOID, "Private, OID" },
@@ -862,6 +901,36 @@ static const value_string cert_vals[] = {
 	  { 0,                   NULL }
 };
 
+/**
+ *   Compute the key id of a KEY RR depending of the algorithm used.
+ */
+static guint16
+compute_key_id(tvbuff_t *tvb, int offset, int size, guint8 algo) 
+{
+  guint32 ac;
+  guint8 c1, c2;
+
+  g_assert(size >= 4);
+  
+  switch( algo ) {
+     case DNS_ALGO_RSAMD5:
+       return (guint16)(tvb_get_guint8(tvb, offset + size - 3) << 8) + tvb_get_guint8( tvb, offset + size - 2 );
+     default:
+       for (ac = 0; size > 1; size -= 2, offset += 2) {
+	 c1 = tvb_get_guint8( tvb, offset );
+	 c2 = tvb_get_guint8( tvb, offset + 1 );
+	 ac +=  (c1 << 8) + c2 ;
+       }
+       if (size > 0) {
+	 c1 = tvb_get_guint8( tvb, offset );
+	 ac += c1 << 8;
+       }
+       ac += (ac >> 16) & 0xffff;
+       return (guint16)(ac & 0xffff);
+  }
+}
+
+
 static int
 dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
   column_info *cinfo, proto_tree *dns_tree, packet_info *pinfo)
@@ -1051,9 +1120,16 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
       int mask;
       int port_num;
       int i;
-      char bitnames[128+1];
-      char portnumstring[10+1];
+      static GString *bitnames = NULL;
 
+      if (bitnames == NULL)
+	bitnames = g_string_sized_new(128);
+
+      if (rr_len < 4) {
+	if (dns_tree != NULL)
+	  goto bad_rr;
+	break;
+      }
       wks_addr = tvb_get_ptr(tvb, cur_offset, 4);
       if (cinfo != NULL)
 	col_append_fstr(cinfo, COL_INFO, " %s", ip_to_str(wks_addr));
@@ -1064,6 +1140,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += 4;
 	rr_len -= 4;
 
+	if (rr_len < 1)
+	  goto bad_rr;
 	protocol = tvb_get_guint8(tvb, cur_offset);
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 1, "Protocol: %s",
 		     ipprotostr(protocol));
@@ -1075,24 +1153,23 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	  bits = tvb_get_guint8(tvb, cur_offset);
 	  if (bits != 0) {
 	    mask = 1<<7;
-	    bitnames[0] = '\0';
+	    g_string_truncate(bitnames, 0);
 	    for (i = 0; i < 8; i++) {
 	      if (bits & mask) {
-		if (bitnames[0] != '\0')
-		  strcat(bitnames, ", ");
+		if (bitnames->len != 0)
+		  g_string_append(bitnames, ", ");
 		switch (protocol) {
 
 		case IP_PROTO_TCP:
-		  strcat(bitnames, get_tcp_port(port_num));
+		  g_string_append(bitnames, get_tcp_port(port_num));
 		  break;
 
 		case IP_PROTO_UDP:
-		  strcat(bitnames, get_udp_port(port_num));
+		  g_string_append(bitnames, get_udp_port(port_num));
 		  break;
 
 		default:
-		  sprintf(portnumstring, "%u", port_num);
-		  strcat(bitnames, portnumstring);
+		  g_string_sprintfa(bitnames, "%u", port_num);
 		  break;
 	        }
 	      }
@@ -1100,7 +1177,7 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	      port_num++;
 	    }
 	    proto_tree_add_text(rr_tree, tvb, cur_offset, 1,
-		"Bits: 0x%02x (%s)", bits, bitnames);
+		"Bits: 0x%02x (%s)", bits, bitnames->str);
 	  } else
 	    port_num += 8;
 	  cur_offset += 1;
@@ -1179,6 +1256,7 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
     }
     break;
 
+  case T_RRSIG:
   case T_SIG:
     {
       int rr_len = data_len;
@@ -1188,6 +1266,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
       int signer_name_len;
 
       if (dns_tree != NULL) {
+	if (rr_len < 2)
+	  goto bad_rr;
 	type_covered = tvb_get_ntohs(tvb, cur_offset);
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 2, "Type covered: %s (%s)",
 		dns_type_name(type_covered),
@@ -1195,22 +1275,30 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += 2;
 	rr_len -= 2;
 
+	if (rr_len < 1)
+	  goto bad_rr;
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 1, "Algorithm: %s",
 		val_to_str(tvb_get_guint8(tvb, cur_offset), algo_vals,
 	            "Unknown (0x%02X)"));
 	cur_offset += 1;
 	rr_len -= 1;
 
+	if (rr_len < 1)
+	  goto bad_rr;
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 1, "Labels: %u",
 		tvb_get_guint8(tvb, cur_offset));
 	cur_offset += 1;
 	rr_len -= 1;
 
+	if (rr_len < 4)
+	  goto bad_rr;
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "Original TTL: %s",
 		time_secs_to_str(tvb_get_ntohl(tvb, cur_offset)));
 	cur_offset += 4;
 	rr_len -= 4;
 
+	if (rr_len < 4)
+	  goto bad_rr;
 	nstime.secs = tvb_get_ntohl(tvb, cur_offset);
 	nstime.nsecs = 0;
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "Signature expiration: %s",
@@ -1218,6 +1306,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += 4;
 	rr_len -= 4;
 
+	if (rr_len < 4)
+	  goto bad_rr;
 	nstime.secs = tvb_get_ntohl(tvb, cur_offset);
 	nstime.nsecs = 0;
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "Time signed: %s",
@@ -1225,7 +1315,9 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += 4;
 	rr_len -= 4;
 
-	proto_tree_add_text(rr_tree, tvb, cur_offset, 2, "Key footprint: 0x%04x",
+	if (rr_len < 2)
+	  goto bad_rr;
+	proto_tree_add_text(rr_tree, tvb, cur_offset, 2, "Id of signing key(footprint): %u",
 		tvb_get_ntohs(tvb, cur_offset));
 	cur_offset += 2;
 	rr_len -= 2;
@@ -1236,19 +1328,25 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += signer_name_len;
 	rr_len -= signer_name_len;
 
-	proto_tree_add_text(rr_tree, tvb, cur_offset, rr_len, "Signature");
+	if (rr_len != 0)
+	  proto_tree_add_text(rr_tree, tvb, cur_offset, rr_len, "Signature");
       }
     }
     break;
 
+  case T_DNSKEY:  
   case T_KEY:
     {
       int rr_len = data_len;
       guint16 flags;
       proto_item *tf;
       proto_tree *flags_tree;
+      guint8 algo;
+      guint16 key_id;
 
       if (dns_tree != NULL) {
+	if (rr_len < 2)
+	  goto bad_rr;
         flags = tvb_get_ntohs(tvb, cur_offset);
 	tf = proto_tree_add_text(rr_tree, tvb, cur_offset, 2, "Flags: 0x%04X", flags);
 	flags_tree = proto_item_add_subtree(tf, ett_t_key_flags);
@@ -1286,25 +1384,38 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 		decode_boolean_bitfield(flags, 0x0040,
 		  2*8, "Key is valid for use with MIME security multiparts",
 		       "Key is not valid for use with MIME security multiparts"));
-	  proto_tree_add_text(flags_tree, tvb, cur_offset, 2, "%s",
-		decode_numeric_bitfield(flags, 0x000F,
-		  2*8, "Signatory = %u"));
+	  if( type != T_DNSKEY )
+	        proto_tree_add_text(flags_tree, tvb, cur_offset, 2, "%s",
+		    decode_numeric_bitfield(flags, 0x000F,
+		       2*8, "Signatory = %u"));
+	    else proto_tree_add_text(flags_tree, tvb, cur_offset, 2, "%s",
+ 	            decode_boolean_bitfield(flags, 0x0001, 
+		       2*8, "Key is a Key Signing Key",
+		            "Key is a Zone Signing Key") );
 	}
 	cur_offset += 2;
 	rr_len -= 2;
 
+	if (rr_len < 1)
+	  goto bad_rr;
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 1, "Protocol: %u",
 		tvb_get_guint8(tvb, cur_offset));
 	cur_offset += 1;
 	rr_len -= 1;
 
+	if (rr_len < 1)
+	  goto bad_rr;
+	algo = tvb_get_guint8(tvb, cur_offset);
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 1, "Algorithm: %s",
-		val_to_str(tvb_get_guint8(tvb, cur_offset), algo_vals,
-	            "Unknown (0x%02X)"));
+		val_to_str(algo, algo_vals, "Unknown (0x%02X)"));
 	cur_offset += 1;
-		rr_len -= 1;
+	rr_len -= 1;
 
-	proto_tree_add_text(rr_tree, tvb, cur_offset, rr_len, "Public key");
+	key_id = compute_key_id(tvb, cur_offset-4, rr_len+4, algo);
+	proto_tree_add_text(rr_tree, tvb, 0, 0, "Key id: %u", key_id);
+
+	if (rr_len != 0)
+	  proto_tree_add_text(rr_tree, tvb, cur_offset, rr_len, "Public key");
       }
     }
     break;
@@ -1336,7 +1447,7 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
       int pname_len;
       int a6_offset;
       int suf_offset;
-      guint8 suffix[16];
+      struct e_in6_addr suffix;
 
       a6_offset = cur_offset;
       pre_len = tvb_get_guint8(tvb, cur_offset);
@@ -1345,10 +1456,10 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
       suf_octet_count = suf_len ? (suf_len - 1) / 8 + 1 : 0;
       /* Pad prefix */
       for (suf_offset = 0; suf_offset < 16 - suf_octet_count; suf_offset++) {
-        suffix[suf_offset] = 0;
+        suffix.s6_addr8[suf_offset] = 0;
       }
       for (; suf_offset < 16; suf_offset++) {
-        suffix[suf_offset] = tvb_get_guint8(tvb, cur_offset);
+        suffix.s6_addr8[suf_offset] = tvb_get_guint8(tvb, cur_offset);
         cur_offset++;
       }
 
@@ -1363,7 +1474,7 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
       if (cinfo != NULL) {
         col_append_fstr(cinfo, COL_INFO, " %d %s %s",
                         pre_len,
-                        ip6_to_str((struct e_in6_addr *)&suffix),
+                        ip6_to_str(&suffix),
                         pname);
       }
       if (dns_tree != NULL) {
@@ -1373,7 +1484,7 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
         if (suf_len) {
           proto_tree_add_text(rr_tree, tvb, a6_offset, suf_octet_count,
                               "Address suffix: %s",
-                              ip6_to_str((struct e_in6_addr *)&suffix));
+                              ip6_to_str(&suffix));
           a6_offset += suf_octet_count;
         }
         if (pre_len > 0) {
@@ -1382,7 +1493,7 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
         }
         proto_item_append_text(trr, ", addr %d %s %s",
                             pre_len,
-                            ip6_to_str((struct e_in6_addr *)&suffix),
+                            ip6_to_str(&suffix),
                             pname);
       }
     }
@@ -1441,7 +1552,55 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	} else
 	  proto_tree_add_text(rr_tree, tvb, cur_offset, data_len, "Data");
       }
-      break;
+    }
+    break;
+
+  case T_NSEC:
+    {
+      int rr_len = data_len;
+      char next_domain_name[MAXDNAME];
+      int next_domain_name_len;
+      int rr_type;
+      guint8 bits;
+      int mask, blockbase, blocksize;
+      int i;
+
+      next_domain_name_len = get_dns_name(tvb, cur_offset, dns_data_offset,
+			next_domain_name, sizeof(next_domain_name));
+      if (cinfo != NULL)
+	col_append_fstr(cinfo, COL_INFO, " %s", next_domain_name);
+      if (dns_tree != NULL) {
+	proto_item_append_text(trr, ", next domain name %s",
+		     next_domain_name);
+	proto_tree_add_text(rr_tree, tvb, cur_offset, next_domain_name_len,
+			"Next domain name: %s", next_domain_name);
+	cur_offset += next_domain_name_len;
+	rr_len -= next_domain_name_len;
+	rr_type = 0;
+	while (rr_len != 0) {
+	  blockbase = tvb_get_guint8(tvb, cur_offset);
+	  blocksize = tvb_get_guint8(tvb, cur_offset + 1);
+	  cur_offset += 2;
+	  rr_len -= 2;
+	  rr_type = blockbase * 256;
+	  for( ; blocksize; blocksize-- ) {	    
+  	       bits = tvb_get_guint8(tvb, cur_offset);
+	       mask = 1<<7;
+	       for (i = 0; i < 8; i++) {
+		 if (bits & mask) {
+		   proto_tree_add_text(rr_tree, tvb, cur_offset, 1,
+			"RR type in bit map: %s (%s)",
+			dns_type_name(rr_type),
+			dns_long_type_name(rr_type));
+		 }
+		 mask >>= 1;
+		 rr_type++;
+	       }
+	       cur_offset += 1;
+	       rr_len -= 1;
+	  }
+        }
+      }
     }
     break;
 
@@ -1513,28 +1672,36 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
       guint8 cert_keyalg;
       int rr_len = data_len;
 
-      cert_type = tvb_get_ntohs(tvb, cur_offset);
-      cur_offset += 2;
-      rr_len -= 2;
-      cert_keytag = tvb_get_ntohs(tvb, cur_offset);
-      cur_offset += 2;
-      rr_len -= 2;
-      cert_keyalg = tvb_get_guint8(tvb, cur_offset);
-      cur_offset += 1;
-      rr_len -= 1;
-
       if (dns_tree != NULL) {
-	proto_tree_add_text(rr_tree, tvb, cur_offset, 1, "Type: %s",
-		val_to_str(cert_keyalg, cert_vals,
+	if (rr_len < 2)
+	  goto bad_rr;
+	cert_type = tvb_get_ntohs(tvb, cur_offset);
+	proto_tree_add_text(rr_tree, tvb, cur_offset, 2, "Type: %s",
+		val_to_str(cert_type, cert_vals,
 	            "Unknown (0x%02X)"));
+	cur_offset += 2;
+	rr_len -= 2;
+
+	if (rr_len < 2)
+	  goto bad_rr;
+	cert_keytag = tvb_get_ntohs(tvb, cur_offset);
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 2, "Key footprint: 0x%04x",
 		cert_keytag);
+	cur_offset += 2;
+	rr_len -= 2;
+
+	if (rr_len < 1)
+	  goto bad_rr;
+	cert_keyalg = tvb_get_guint8(tvb, cur_offset);
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 1, "Algorithm: %s",
 		val_to_str(cert_keyalg, algo_vals,
 	            "Unknown (0x%02X)"));
-	proto_tree_add_text(rr_tree, tvb, cur_offset, rr_len, "Public key");
-      }
+	cur_offset += 1;
+	rr_len -= 1;
 
+	if (rr_len != 0)
+	  proto_tree_add_text(rr_tree, tvb, cur_offset, rr_len, "Public key");
+      }
     }
     break;
 
@@ -1543,6 +1710,51 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
       proto_tree_add_text(rr_tree, tvb, cur_offset, data_len, "Data");
     break;
 
+  case T_DS:
+    {
+      guint16 keytag, digest_data_size = -1;
+      guint8  ds_algorithm, ds_digest;
+      int rr_len = data_len;
+
+      static const value_string tds_digests[] = {
+	{ TDSDIGEST_RESERVED, "Reserved digest" },
+	{ TDSDIGEST_SHA1,     "SHA-1" },
+	{ 0, NULL }
+      };
+
+      if (dns_tree != NULL) {
+	if (rr_len < 2)
+	  goto bad_rr;
+	keytag = tvb_get_ntohs(tvb, cur_offset);
+	proto_tree_add_text(rr_tree, tvb, cur_offset, 2, "Key id: %04u", keytag);
+	cur_offset += 2;
+	rr_len -= 2;
+
+	if (rr_len < 1)
+	  goto bad_rr;
+	ds_algorithm = tvb_get_guint8(tvb, cur_offset);
+	proto_tree_add_text(rr_tree, tvb, cur_offset, 1, "Algorithm: %s", val_to_str(ds_algorithm, algo_vals,"Unknown (0x%02X)") );
+	cur_offset += 1;
+	rr_len -= 1;
+
+	if (rr_len < 1)
+	  goto bad_rr;
+	ds_digest = tvb_get_guint8(tvb, cur_offset);
+	proto_tree_add_text(rr_tree, tvb, cur_offset, 1, "Digest type: %s", val_to_str(ds_digest, tds_digests, "Unknown (0x%02X)"));
+	cur_offset += 1;
+	rr_len -= 1;
+
+	if (ds_digest == TDSDIGEST_SHA1)
+	  digest_data_size = 20; /* SHA1 key is always 20 bytes long */
+	if (digest_data_size > 0) {
+	  if (rr_len < digest_data_size)
+	    goto bad_rr;
+	  proto_tree_add_text(rr_tree, tvb, cur_offset, digest_data_size, "Public key"); 
+	}
+      }
+    }
+    break;
+
   case T_TKEY:
     {
       char tkey_algname[MAXDNAME];
@@ -1568,6 +1780,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += tkey_algname_len;
 	rr_len -= tkey_algname_len;
 
+	if (rr_len < 4)
+	  goto bad_rr;
 	nstime.secs = tvb_get_ntohl(tvb, cur_offset);
 	nstime.nsecs = 0;
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "Signature inception: %s",
@@ -1575,6 +1789,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += 4;
 	rr_len -= 4;
 
+	if (rr_len < 4)
+	  goto bad_rr;
 	nstime.secs = tvb_get_ntohl(tvb, cur_offset);
 	nstime.nsecs = 0;
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "Signature expiration: %s",
@@ -1582,6 +1798,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += 4;
 	rr_len -= 4;
 
+	if (rr_len < 2)
+	  goto bad_rr;
 	tkey_mode = tvb_get_ntohs(tvb, cur_offset);
 	cur_offset += 2;
 	rr_len -= 2;
@@ -1589,6 +1807,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 		val_to_str(tkey_mode, tkey_modes,
 	            "Unknown (0x%04X)"));
 
+	if (rr_len < 2)
+	  goto bad_rr;
 	tkey_error = tvb_get_ntohs(tvb, cur_offset);
 	cur_offset += 2;
 	rr_len -= 2;
@@ -1597,57 +1817,71 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 		val_to_str(tkey_error, tsigerror_vals, "Unknown error (%x)")));
 
 	tkey_keylen = tvb_get_ntohs(tvb, cur_offset);
+        proto_tree_add_text(rr_tree, tvb, cur_offset, 2, "Key Size: %u",
+            tkey_keylen);
 	cur_offset += 2;
 	rr_len -= 2;
 
-	key_item = proto_tree_add_text(
-		rr_tree, tvb, cur_offset, tkey_keylen, "Key");
+	if (tkey_keylen != 0) {
+		key_item = proto_tree_add_text(
+			rr_tree, tvb, cur_offset, tkey_keylen, "Key Data");
 
-	key_tree = proto_item_add_subtree(key_item, ett_t_key);
+		key_tree = proto_item_add_subtree(key_item, ett_t_key);
 
-	switch(tkey_mode) {
-	case TKEYMODE_GSSAPI: {
-		tvbuff_t *gssapi_tvb;
+		switch(tkey_mode) {
+		case TKEYMODE_GSSAPI: {
+			tvbuff_t *gssapi_tvb;
 
-		/*
-		 * XXX - in at least one capture, this appears to
-		 * be an NTLMSSP blob, with no ASN.1 in it, in
-		 * a query.
-		 *
-		 * See
-		 *
-		 * http://www.alternic.org/drafts/drafts-s-t/draft-skwan-gss-tsig-05.html
-		 *
-		 * which might indicate what's going on here.  (The key
-		 * is an output_token from GSS_Init_sec_context.)
-		 *
-		 * How the heck do we know what method is being used,
-		 * so we know how to decode the key?
-		 */
-		gssapi_tvb = tvb_new_subset(
-			tvb, cur_offset, tkey_keylen, tkey_keylen);
+			/*
+			 * XXX - in at least one capture, this appears to
+			 * be an NTLMSSP blob, with no ASN.1 in it, in
+			 * a query.
+			 *
+			 * See RFC 3645 which might indicate what's going
+			 * on here.  (The key is an output_token from
+			 * GSS_Init_sec_context.)
+			 *
+			 * How the heck do we know what method is being
+			 * used, so we know how to decode the key?  Do we
+			 * have to look at the algorithm name, e.g.
+			 * "gss.microsoft.com"?  The SMB dissector
+			 * checks whether the security blob begins
+			 * with "NTLMSSP" in some cases.
+			 */
+			gssapi_tvb = tvb_new_subset(
+				tvb, cur_offset, tkey_keylen, tkey_keylen);
 
-		call_dissector(gssapi_handle, gssapi_tvb, pinfo, key_tree);
+			call_dissector(gssapi_handle, gssapi_tvb, pinfo,
+				key_tree);
 
-		break;
-	}
-	default:
+			break;
+		}
+		default:
 
-		/* No dissector for this key mode */
+			/* No dissector for this key mode */
 
-		break;
+			break;
+		}
+
+		cur_offset += tkey_keylen;
+		rr_len -= tkey_keylen;
 	}
 
-	cur_offset += tkey_keylen;
-	rr_len -= tkey_keylen;
-
+	if (rr_len < 2)
+	  goto bad_rr;
 	tkey_otherlen = tvb_get_ntohs(tvb, cur_offset);
+        proto_tree_add_text(rr_tree, tvb, cur_offset, 2, "Other Size: %u",
+            tkey_otherlen);
 	cur_offset += 2;
 	rr_len -= 2;
 
-	proto_tree_add_text(rr_tree, tvb, cur_offset, tkey_otherlen, "Other");
-	cur_offset += tkey_otherlen;
-	rr_len -= tkey_otherlen;
+	if (tkey_otherlen != 0) {
+	  if (rr_len < tkey_otherlen)
+	    goto bad_rr;
+	  proto_tree_add_text(rr_tree, tvb, cur_offset, tkey_otherlen, "Other Data");
+	  cur_offset += tkey_otherlen;
+	  rr_len -= tkey_otherlen;
+	}
       }
     }
     break;
@@ -1669,6 +1903,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += tsig_algname_len;
 	rr_len -= tsig_algname_len;
 
+	if (rr_len < 6)
+	  goto bad_rr;
 	tsig_timehi = tvb_get_ntohs(tvb, cur_offset);
 	tsig_timelo = tvb_get_ntohl(tvb, cur_offset + 2);
 	nstime.secs = tsig_timelo;
@@ -1678,26 +1914,38 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += 6;
 	rr_len -= 6;
 
+	if (rr_len < 2)
+	  goto bad_rr;
 	tsig_fudge = tvb_get_ntohs(tvb, cur_offset);
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 2, "Fudge: %u",
 		tsig_fudge);
 	cur_offset += 2;
 	rr_len -= 2;
 
+	if (rr_len < 2)
+	  goto bad_rr;
 	tsig_siglen = tvb_get_ntohs(tvb, cur_offset);
 	cur_offset += 2;
 	rr_len -= 2;
 
-	proto_tree_add_text(rr_tree, tvb, cur_offset, tsig_siglen, "Signature");
-	cur_offset += tsig_siglen;
-	rr_len -= tsig_siglen;
+	if (tsig_siglen != 0) {
+	  if (rr_len < tsig_siglen)
+	    goto bad_rr;
+	  proto_tree_add_text(rr_tree, tvb, cur_offset, tsig_siglen, "Signature");
+	  cur_offset += tsig_siglen;
+	  rr_len -= tsig_siglen;
+	}
 
+	if (rr_len < 2)
+	  goto bad_rr;
 	tsig_originalid = tvb_get_ntohs(tvb, cur_offset);
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 2, "Original id: %d",
 		tsig_originalid);
 	cur_offset += 2;
 	rr_len -= 2;
 
+	if (rr_len < 2)
+	  goto bad_rr;
 	tsig_error = tvb_get_ntohs(tvb, cur_offset);
         proto_tree_add_text(rr_tree, tvb, cur_offset, 2, "Error: %s",
 		val_to_str(tsig_error, rcode_vals,
@@ -1705,13 +1953,19 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += 2;
 	rr_len -= 2;
 
+	if (rr_len < 2)
+	  goto bad_rr;
 	tsig_otherlen = tvb_get_ntohs(tvb, cur_offset);
 	cur_offset += 2;
 	rr_len -= 2;
 
-	proto_tree_add_text(rr_tree, tvb, cur_offset, tsig_otherlen, "Other");
-	cur_offset += tsig_otherlen;
-	rr_len -= tsig_otherlen;
+	if (tsig_otherlen != 0) {
+	  if (rr_len < tsig_otherlen)
+	    goto bad_rr;
+	  proto_tree_add_text(rr_tree, tvb, cur_offset, tsig_otherlen, "Other");
+	  cur_offset += tsig_otherlen;
+	  rr_len -= tsig_otherlen;
+	}
       }
     }
     break;
@@ -1725,6 +1979,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
       guint32 nservers;
 
       if (dns_tree != NULL) {
+	if (rr_len < 4)
+	  goto bad_rr;
 	local_flag = tvb_get_ntohl(tvb, cur_offset);
 	if (dns_tree != NULL) {
 	  proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "Local flag: %s",
@@ -1733,6 +1989,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += 4;
 	rr_len -= 4;
 
+	if (rr_len < 4)
+	  goto bad_rr;
 	lookup_timeout = tvb_get_ntohl(tvb, cur_offset);
 	if (dns_tree != NULL) {
 	  proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "Lookup timeout: %u seconds",
@@ -1741,6 +1999,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += 4;
 	rr_len -= 4;
 
+	if (rr_len < 4)
+	  goto bad_rr;
 	cache_timeout = tvb_get_ntohl(tvb, cur_offset);
 	if (dns_tree != NULL) {
 	  proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "Cache timeout: %u seconds",
@@ -1749,6 +2009,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	cur_offset += 4;
 	rr_len -= 4;
 
+	if (rr_len < 4)
+	  goto bad_rr;
 	nservers = tvb_get_ntohl(tvb, cur_offset);
 	if (dns_tree != NULL) {
 	  proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "Number of WINS servers: %u",
@@ -1758,6 +2020,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 	rr_len -= 4;
 
 	while (rr_len != 0 && nservers != 0) {
+	  if (rr_len < 4)
+	    goto bad_rr;
 	  if (dns_tree != NULL) {
 	    proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "WINS server address: %s",
 		     ip_to_str(tvb_get_ptr(tvb, cur_offset, 4)));
@@ -1779,6 +2043,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
       char dname[MAXDNAME];
       int dname_len;
 
+      if (rr_len < 4)
+	goto bad_rr;
       local_flag = tvb_get_ntohl(tvb, cur_offset);
       if (dns_tree != NULL) {
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "Local flag: %s",
@@ -1787,6 +2053,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
       cur_offset += 4;
       rr_len -= 4;
 
+      if (rr_len < 4)
+	goto bad_rr;
       lookup_timeout = tvb_get_ntohl(tvb, cur_offset);
       if (dns_tree != NULL) {
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "Lookup timeout: %u seconds",
@@ -1795,6 +2063,8 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
       cur_offset += 4;
       rr_len -= 4;
 
+      if (rr_len < 4)
+	goto bad_rr;
       cache_timeout = tvb_get_ntohl(tvb, cur_offset);
       if (dns_tree != NULL) {
 	proto_tree_add_text(rr_tree, tvb, cur_offset, 4, "Cache timeout: %u seconds",
@@ -1852,6 +2122,16 @@ dissect_dns_answer(tvbuff_t *tvb, int offset, int dns_data_offset,
 
   data_offset += data_len;
 
+  return data_offset - data_start;
+
+bad_rr:
+  if (dns_tree != NULL) {
+    proto_item_append_text(trr, ", bad RR length %d, too short",
+			data_len);
+  }
+
+  data_offset += data_len;
+
   return data_offset - data_start;
 }
 
@@ -1926,8 +2206,8 @@ dissect_dns_common(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree,
   /* To do: check for errs, etc. */
   id    = tvb_get_ntohs(tvb, offset + DNS_ID);
   flags = tvb_get_ntohs(tvb, offset + DNS_FLAGS);
-  opcode = (flags & F_OPCODE) >> OPCODE_SHIFT;
-  rcode = (flags & F_RCODE);
+  opcode = (guint16) ((flags & F_OPCODE) >> OPCODE_SHIFT);
+  rcode  = (guint16)  (flags & F_RCODE);
 
   if (check_col(pinfo->cinfo, COL_INFO)) {
     strcpy(buf, val_to_str(opcode, opcode_vals, "Unknown operation (%u)"));
@@ -1995,11 +2275,15 @@ dissect_dns_common(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree,
     if (flags & F_RESPONSE) {
       proto_tree_add_item(field_tree, hf_dns_flags_recavail,
 			  tvb, offset + DNS_FLAGS, 2, FALSE);
+      proto_tree_add_item(field_tree, hf_dns_flags_z,
+			 tvb, offset + DNS_FLAGS, 2, FALSE);
       proto_tree_add_item(field_tree, hf_dns_flags_authenticated,
 			  tvb, offset + DNS_FLAGS, 2, FALSE);
       proto_tree_add_item(field_tree, hf_dns_flags_rcode,
 			  tvb, offset + DNS_FLAGS, 2, FALSE);
     } else {
+      proto_tree_add_item(field_tree, hf_dns_flags_z,
+                           tvb, offset + DNS_FLAGS, 2, FALSE);
       proto_tree_add_item(field_tree, hf_dns_flags_checkdisable,
 			  tvb, offset + DNS_FLAGS, 2, FALSE);
     }
@@ -2149,6 +2433,10 @@ proto_register_dns(void)
       { "Recursion available",	"dns.flags.recavail",
 	FT_BOOLEAN, 16, TFS(&tfs_flags_recavail), F_RECAVAIL,
 	"Can the server do recursive queries?", HFILL }},
+    { &hf_dns_flags_z,
+      { "Z", "dns.flags.z",
+	FT_BOOLEAN, 16, TFS(&tfs_flags_z), F_Z,
+	"Z flag", HFILL }},
     { &hf_dns_flags_authenticated,
       { "Answer authenticated",	"dns.flags.authenticated",
 	FT_BOOLEAN, 16, TFS(&tfs_flags_authenticated), F_AUTHENTIC,
