@@ -1,7 +1,7 @@
 /* packet-ldp.c
  * Routines for LDP (RFC 3036) packet disassembly
  *
- * $Id: packet-ldp.c,v 1.43 2002/08/28 21:00:20 jmayer Exp $
+ * $Id: packet-ldp.c,v 1.50 2004/01/17 12:51:00 ulfl Exp $
  *
  * Copyright (c) November 2000 by Richard Sharpe <rsharpe@ns.aus.com>
  *
@@ -43,6 +43,7 @@
 #include "afn.h"
 
 #include "packet-frame.h"
+#include "packet-diffserv-mpls-common.h"
 
 #define TCP_PORT_LDP 646
 #define UDP_PORT_LDP 646
@@ -136,6 +137,7 @@ static int hf_ldp_tlv_fec_vc_groupid = -1;
 static int hf_ldp_tlv_fec_vc_vcid = -1;
 static int hf_ldp_tlv_fec_vc_intparam_length = -1;
 static int hf_ldp_tlv_fec_vc_intparam_mtu = -1;
+static int hf_ldp_tlv_fec_vc_intparam_tdmbps = -1;
 static int hf_ldp_tlv_fec_vc_intparam_id = -1;
 static int hf_ldp_tlv_fec_vc_intparam_maxcatmcells = -1;
 static int hf_ldp_tlv_fec_vc_intparam_desc = -1;
@@ -170,6 +172,16 @@ static int hf_ldp_tlv_set_prio = -1;
 static int hf_ldp_tlv_hold_prio = -1;
 static int hf_ldp_tlv_route_pinning = -1;
 static int hf_ldp_tlv_resource_class = -1;
+static int hf_ldp_tlv_diffserv = -1;
+static int hf_ldp_tlv_diffserv_type = -1;
+static int hf_ldp_tlv_diffserv_mapnb = -1;
+static int hf_ldp_tlv_diffserv_map = -1;
+static int hf_ldp_tlv_diffserv_map_exp = -1;
+static int hf_ldp_tlv_diffserv_phbid = -1;
+static int hf_ldp_tlv_diffserv_phbid_dscp = -1;
+static int hf_ldp_tlv_diffserv_phbid_code = -1;
+static int hf_ldp_tlv_diffserv_phbid_bit14 = -1;
+static int hf_ldp_tlv_diffserv_phbid_bit15 = -1;
 static int ett_ldp = -1;
 static int ett_ldp_header = -1;
 static int ett_ldp_ldpid = -1;
@@ -178,6 +190,8 @@ static int ett_ldp_tlv = -1;
 static int ett_ldp_tlv_val = -1;
 static int ett_ldp_fec = -1;
 static int ett_ldp_fec_vc_interfaceparam = -1;
+static int ett_ldp_diffserv_map = -1;
+static int ett_ldp_diffserv_map_phbid = -1;
 
 static int tcp_port = 0;
 static int udp_port = 0;
@@ -187,8 +201,8 @@ static gboolean ldp_desegment = FALSE;
 
 /* Add your functions here */
 
-static int global_ldp_tcp_port = TCP_PORT_LDP;
-static int global_ldp_udp_port = UDP_PORT_LDP;
+static guint32 global_ldp_tcp_port = TCP_PORT_LDP;
+static guint32 global_ldp_udp_port = UDP_PORT_LDP;
 
 /*
  * The following define all the TLV types I know about
@@ -224,7 +238,7 @@ static int global_ldp_udp_port = UDP_PORT_LDP;
 #define TLV_LSPID                  0x0821
 #define TLV_RESOURCE_CLASS         0x0822
 #define TLV_ROUTE_PINNING          0x0823
-
+#define TLV_DIFFSERV               0x0901
 #define TLV_VENDOR_PRIVATE_START   0x3E00
 #define TLV_VENDOR_PRIVATE_END     0x3EFF
 #define TLV_EXPERIMENTAL_START     0x3F00
@@ -261,6 +275,7 @@ static const value_string tlv_type_names[] = {
   { TLV_ER_HOP_LSPID,              "ER hop LSPID prefix TLV"},
   { TLV_RESOURCE_CLASS,            "Resource Class (Color) TLV"},
   { TLV_ROUTE_PINNING,             "Route Pinning TLV"},
+  { TLV_DIFFSERV,                  "Diff-Serv TLV"},
   { TLV_VENDOR_PRIVATE_START,	"Vendor Private TLV"},
   { TLV_EXPERIMENTAL_START,	"Experimental TLV"},
   { 0, NULL}
@@ -345,10 +360,15 @@ static const value_string fec_vc_types_vals[] = {
   {0x0005, "Ethernet"},
   {0x0006, "HDLC"},
   {0x0007, "PPP"},
-  {0x0009, "ATM VCC cell transport"},
-  {0x8008, "CEM"},
-  {0x000A, "ATM VPC cell transport"},
-  {0x000B, "Ethernet VPLS"},
+  {0x0008, "SONET/SDH Circuit Emulation Service"},
+  {0x0009, "ATM n-to-one VCC cell transport"},
+  {0x000A, "ATM n-to one VPC cell transport"},
+  {0x000B, "IP layer2 transport"},
+  {0x000C, "ATM one-to-one VCC Cell Mode"},
+  {0x000D, "ATM one-to-one VPC Cell Mode"},
+  {0x000E, "ATM AAL5 PDU VCC transport"},
+  {0x000F, "Frame-Relay Port mode"},
+  {0x0010, "SONET/SDH Circuit Emulation over Packet"},
   {0, NULL}
 };
 
@@ -359,6 +379,9 @@ static const value_string fec_vc_types_vals[] = {
 #define FEC_VC_INTERFACEPARAM_CEMBYTES     0x04
 #define FEC_VC_INTERFACEPARAM_CEMOPTIONS   0x05
 #define FEC_VC_INTERFACEPARAM_VPNID        0x06
+#define FEC_VC_INTERFACEPARAM_TDMBPS       0x07
+#define FEC_VC_INTERFACEPARAM_FRDLCILEN    0x08
+
 
 
 static const value_string fec_vc_interfaceparm[] = {
@@ -367,7 +390,10 @@ static const value_string fec_vc_interfaceparm[] = {
   {FEC_VC_INTERFACEPARAM_DESCRIPTION, "Interface Description"},
   {FEC_VC_INTERFACEPARAM_CEMBYTES, "CEM Payload Bytes"},
   {FEC_VC_INTERFACEPARAM_CEMOPTIONS, "CEM options"},
-  {FEC_VC_INTERFACEPARAM_VPNID, "VPN Id"}
+  {FEC_VC_INTERFACEPARAM_VPNID, "VPN Id"},
+  {FEC_VC_INTERFACEPARAM_TDMBPS, "CEP/TDM bit-rate"},
+  {FEC_VC_INTERFACEPARAM_FRDLCILEN, "Frame-Relay DLCI Length"},
+  {0, NULL},
 };
 
 static const true_false_string fec_vc_cbit = {
@@ -421,6 +447,12 @@ static const value_string route_pinning_vals[] = {
   {0, NULL}
 };
 
+static const value_string diffserv_type_vals[] = {
+  {0, "E-LSP"},
+  {1, "L-LSP"},
+  {0, NULL}
+};
+
 static const value_string ldp_loose_vals[] = {
   {0, "strict hop"},
   {1, "loose hop"},
@@ -514,6 +546,18 @@ static const value_string tlv_status_data[] = {
   {0, NULL}
 };
 
+/* Define storage class for a string handler function
+ * with a const guint8 * argument, and returning a gchar *
+ */
+typedef gchar *(string_handler_func)(const guint8 *);
+
+/* Default handler for address to string conversion */
+static gchar *
+default_str_handler(const guint8 * bytes _U_)
+{
+	return "<Support for this Address Family not implemented>";
+}
+	
 /* Dissect FEC TLV */
 
 static void
@@ -523,7 +567,7 @@ dissect_tlv_fec(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem)
 	guint16	family, ix=1, ax;
 	guint8	addr_size=0, *addr, implemented, prefix_len_octets, prefix_len, host_len, vc_len;
 	guint8  intparam_len;
-	void *str_handler=NULL;
+	string_handler_func *str_handler = default_str_handler;
 	char *str;
 
 	if (tree) {
@@ -560,7 +604,7 @@ dissect_tlv_fec(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem)
 						break;
 					case AFNUM_INET6: /*IPv6*/
 						addr_size=16;
-						str_handler=ip6_to_str;
+						str_handler = (string_handler_func *) ip6_to_str;
 						break;
 					default:
 						implemented=0;
@@ -614,7 +658,7 @@ dissect_tlv_fec(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem)
 				if( prefix_len % 8 )
 					addr[ax-1] = addr[ax-1]&(0xFF<<(8-prefix_len%8));
 
-				str = (* (char* (*)(guint8 *))str_handler)(addr);
+				str = str_handler((const guint8 *)addr);
 				proto_tree_add_string_format(fec_tree, hf_ldp_tlv_fec_pfval, tvb, offset, prefix_len_octets, str, "Prefix: %s", str);
 
 				offset += prefix_len_octets;
@@ -638,7 +682,7 @@ dissect_tlv_fec(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem)
 						break;
 					case AFNUM_INET6: /*IPv6*/
 						addr_size=16;
-						str_handler=ip6_to_str;
+						str_handler = (string_handler_func *) ip6_to_str;
 						break;
 					default:
 						implemented=0;
@@ -690,7 +734,7 @@ dissect_tlv_fec(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem)
 				for(ax=0; ax+1 <= host_len; ax++)
 					addr[ax]=tvb_get_guint8(tvb, offset+ax);
 
-				str = (* (char* (*)(guint8 *))str_handler)(addr);
+				str = str_handler((const guint8 *)addr);
 				proto_tree_add_string_format(fec_tree, hf_ldp_tlv_fec_hoval, tvb, offset, host_len, str, "Address: %s", str);
 
 				offset += host_len;
@@ -745,6 +789,10 @@ dissect_tlv_fec(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem)
 			      proto_item_append_text(ti,": MTU %u", tvb_get_ntohs(tvb,offset+2));
 			      proto_tree_add_item(vcintparam_tree,hf_ldp_tlv_fec_vc_intparam_mtu,tvb, offset+2, 2, FALSE);
 			      break;
+			    case FEC_VC_INTERFACEPARAM_TDMBPS:
+			      proto_item_append_text(ti,": BPS %u", tvb_get_ntohs(tvb,offset+2));
+			      proto_tree_add_item(vcintparam_tree,hf_ldp_tlv_fec_vc_intparam_tdmbps,tvb, offset+2, 2, FALSE);
+			      break;
 			    case FEC_VC_INTERFACEPARAM_MAXCATMCELLS:
 			      proto_item_append_text(ti,": Max ATM Concat Cells %u", tvb_get_ntohs(tvb,offset+2));
 			      proto_tree_add_item(vcintparam_tree,hf_ldp_tlv_fec_vc_intparam_maxcatmcells,tvb, offset+2, 2, FALSE);
@@ -800,7 +848,7 @@ dissect_tlv_address_list(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem)
 	proto_tree *ti = NULL, *val_tree = NULL;
 	guint16	family, ix;
 	guint8	addr_size, *addr;
-	void *str_handler;
+	string_handler_func *str_handler = default_str_handler;
 	char *str;
 
 	if (tree) {
@@ -821,7 +869,7 @@ dissect_tlv_address_list(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem)
 				break;
 			case AFNUM_INET6: /*IPv6*/
 				addr_size=16;
-				str_handler=ip6_to_str;
+				str_handler = (string_handler_func *) ip6_to_str;
 				break;
 			default:
 				proto_tree_add_text(tree, tvb, offset+2, rem-2,
@@ -846,7 +894,7 @@ dissect_tlv_address_list(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem)
 							 == NULL)
 				break;
 
-			str = (* (char* (*)(guint8 *))str_handler)(addr);
+			str = str_handler((const guint8 *)addr);
 			proto_tree_add_string_format(val_tree,
 			    hf_ldp_tlv_addrl_addr, tvb, offset, addr_size, str,
 			    "Address %u: %s", ix, str);
@@ -1590,6 +1638,51 @@ dissect_tlv_preemption(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem)
 	}
 }
 
+
+static void
+dissect_tlv_diffserv(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem)
+{
+    int type, mapnb, count;
+    int *hfindexes[] = {
+	&hf_ldp_tlv_diffserv_map,
+	&hf_ldp_tlv_diffserv_map_exp,
+	&hf_ldp_tlv_diffserv_phbid,
+	&hf_ldp_tlv_diffserv_phbid_dscp,
+	&hf_ldp_tlv_diffserv_phbid_code,
+	&hf_ldp_tlv_diffserv_phbid_bit14,
+	&hf_ldp_tlv_diffserv_phbid_bit15
+    };
+    gint *etts[] = {
+	&ett_ldp_diffserv_map,
+	&ett_ldp_diffserv_map_phbid
+    };
+
+    if (rem < 4) {
+	proto_tree_add_text(tree, tvb, offset, rem,
+	    "Error processing Diff-Serv TLV: length is %d, should be >= 4", rem);
+	return;
+    }
+    proto_tree_add_uint(tree, hf_ldp_tlv_diffserv_type, tvb, offset, 1,
+			type = tvb_get_guint8(tvb, offset));
+    type = (type >> 7) + 1;
+    if (type == 1) {
+        /* E-LSP */
+	offset += 3;
+	proto_tree_add_uint(tree, hf_ldp_tlv_diffserv_mapnb, tvb, offset,
+			    1, mapnb = tvb_get_guint8(tvb, offset) & 15);
+	offset += 1;
+	for (count = 0; count < mapnb; count++) {
+	    dissect_diffserv_mpls_common(tvb, tree, type, offset, hfindexes, etts);
+	    offset += 4;
+	}
+    }
+    else if (type == 2) {
+        /* L-LSP */
+	dissect_diffserv_mpls_common(tvb, tree, type, offset + 2, hfindexes, etts);
+    }
+}
+
+
 static int
 dissect_tlv(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem);
 
@@ -1832,6 +1925,10 @@ dissect_tlv(tvbuff_t *tvb, guint offset, proto_tree *tree, int rem)
 			dissect_tlv_route_pinning(tvb, offset +4, tlv_tree, length);
 			break;
 
+		case TLV_DIFFSERV:
+			dissect_tlv_diffserv(tvb, offset +4, tlv_tree, length);
+			break;
+
 		case TLV_VENDOR_PRIVATE_START:
 			if( length < 4 ) /*error, at least Vendor ID*/
 				proto_tree_add_text(tlv_tree, tvb, offset + 4, length,
@@ -2088,7 +2185,7 @@ dissect_ldp_tcp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 				 */
 				pinfo->desegment_offset = offset;
 				pinfo->desegment_len = 4 - length_remaining;
-				return -pinfo->desegment_len;
+				return -((gint32) pinfo->desegment_len);
 			}
 		}
 
@@ -2116,7 +2213,7 @@ dissect_ldp_tcp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 				pinfo->desegment_offset = offset;
 				pinfo->desegment_len =
 				    (plen + 4) - length_remaining;
-				return -pinfo->desegment_len;
+				return -((gint32) pinfo->desegment_len);
 			}
 		}
 
@@ -2235,7 +2332,7 @@ proto_register_ldp(void)
       { "Targeted Hello", "ldp.msg.tlv.hello.targeted", FT_BOOLEAN, 16, TFS(&hello_targeted_vals), 0x8000, "Hello Common Parameters Targeted Bit", HFILL }},
 
     { &hf_ldp_tlv_val_request,
-      { "Hello Requested", "ldp,msg.tlv.hello.requested", FT_BOOLEAN, 16, TFS(&hello_requested_vals), 0x4000, "Hello Common Parameters Hello Requested Bit", HFILL }},
+      { "Hello Requested", "ldp.msg.tlv.hello.requested", FT_BOOLEAN, 16, TFS(&hello_requested_vals), 0x4000, "Hello Common Parameters Hello Requested Bit", HFILL }},
 
     { &hf_ldp_tlv_val_res,
       { "Reserved", "ldp.msg.tlv.hello.res", FT_UINT16, BASE_HEX, NULL, 0x3FFF, "Hello Common Parameters Reserved Field", HFILL }},
@@ -2399,7 +2496,7 @@ proto_register_ldp(void)
       { "Returned PDU Label Space ID", "ldp.msg.tlv.returned.ldpid.lsid", FT_UINT16, BASE_HEX, NULL, 0x0, "LDP Label Space ID", HFILL }},
 
     { &hf_ldp_tlv_returned_msg_ubit,
-      { "Returned Message Unknown bit", "ldp.msg.tlv.returned.msg.ubit", FT_UINT8, BASE_HEX, TFS(&ldp_message_ubit), 0x80, "Message Unknown bit", HFILL }},
+      { "Returned Message Unknown bit", "ldp.msg.tlv.returned.msg.ubit", FT_BOOLEAN, BASE_HEX, TFS(&ldp_message_ubit), 0x80, "Message Unknown bit", HFILL }},
 
     { &hf_ldp_tlv_returned_msg_type,
       { "Returned Message Type", "ldp.msg.tlv.returned.msg.type", FT_UINT16, BASE_HEX, VALS(ldp_message_types), 0x7FFF, "LDP message type", HFILL }},
@@ -2434,6 +2531,9 @@ proto_register_ldp(void)
     {&hf_ldp_tlv_fec_vc_intparam_mtu,
      {"MTU", "ldp.msg.tlv.fec.vc.intparam.mtu", FT_UINT16, BASE_DEC, NULL, 0x0, "VC FEC Interface Paramater MTU", HFILL }},
 
+    {&hf_ldp_tlv_fec_vc_intparam_tdmbps,
+     {"BPS", "ldp.msg.tlv.fec.vc.intparam.tdmbps", FT_UINT16, BASE_DEC, NULL, 0x0, "VC FEC Interface Parameter CEP/TDM bit-rate", HFILL }},
+
     {&hf_ldp_tlv_fec_vc_intparam_id,
      {"ID", "ldp.msg.tlv.fec.vc.intparam.id", FT_UINT8, BASE_HEX, VALS(fec_vc_interfaceparm), 0x0, "VC FEC Interface Paramater ID", HFILL }},
 
@@ -2535,6 +2635,46 @@ proto_register_ldp(void)
     { &hf_ldp_tlv_resource_class,
       { "Resource Class", "ldp.msg.tlv.resource_class", FT_UINT32, BASE_HEX, NULL, 0, "Resource Class (Color)", HFILL}},
 
+    { &hf_ldp_tlv_diffserv,
+      { "Diff-Serv TLV", "ldp.msg.tlv.diffserv", FT_NONE, BASE_NONE, NULL,
+	0, "Diffserv TLV", HFILL}},
+
+    { &hf_ldp_tlv_diffserv_type,
+      { "LSP Type", "ldp.msg.tlv.diffserv.type", FT_UINT8, BASE_DEC,
+	VALS(diffserv_type_vals), 0x80, "LSP Type", HFILL}},
+
+    { &hf_ldp_tlv_diffserv_mapnb,
+      { "MAPnb", "ldp.msg.tlv.diffserv.mapnb", FT_UINT8, BASE_DEC, NULL,
+	0, MAPNB_DESCRIPTION, HFILL}},
+
+    { &hf_ldp_tlv_diffserv_map,
+      { "MAP", "ldp.msg.tlv.diffserv.map", FT_NONE, BASE_NONE, NULL,
+	0, MAP_DESCRIPTION, HFILL}},
+
+    { &hf_ldp_tlv_diffserv_map_exp,
+      { "EXP", "ldp.msg.tlv.diffserv.map.exp", FT_UINT8, BASE_DEC, NULL,
+	0, EXP_DESCRIPTION, HFILL}},
+
+    { &hf_ldp_tlv_diffserv_phbid,
+      { "PHBID", "ldp.msg.tlv.diffserv.phbid", FT_NONE, BASE_NONE, NULL,
+	0, PHBID_DESCRIPTION, HFILL}},
+
+    { &hf_ldp_tlv_diffserv_phbid_dscp,
+      { "DSCP", "ldp.msg.tlv.diffserv.phbid.dscp", FT_UINT16, BASE_DEC,
+	NULL, PHBID_DSCP_MASK, PHBID_DSCP_DESCRIPTION, HFILL}},
+
+    { &hf_ldp_tlv_diffserv_phbid_code,
+      { "PHB id code", "ldp.msg.tlv.diffserv.phbid.code", FT_UINT16, BASE_DEC,
+	NULL, PHBID_CODE_MASK, PHBID_CODE_DESCRIPTION, HFILL}},
+
+    { &hf_ldp_tlv_diffserv_phbid_bit14,
+      { "Bit 14", "ldp.msg.tlv.diffserv.phbid.bit14", FT_UINT16, BASE_DEC,
+	VALS(phbid_bit14_vals), PHBID_BIT14_MASK, PHBID_BIT14_DESCRIPTION, HFILL}},
+
+    { &hf_ldp_tlv_diffserv_phbid_bit15,
+      { "Bit 15", "ldp.msg.tlv.diffserv.phbid.bit15", FT_UINT16, BASE_DEC,
+	VALS(phbid_bit15_vals), PHBID_BIT15_MASK, PHBID_BIT15_DESCRIPTION, HFILL}}
+
   };
 
   static gint *ett[] = {
@@ -2545,7 +2685,9 @@ proto_register_ldp(void)
     &ett_ldp_tlv,
     &ett_ldp_tlv_val,
     &ett_ldp_fec,
-    &ett_ldp_fec_vc_interfaceparam
+    &ett_ldp_fec_vc_interfaceparam,
+    &ett_ldp_diffserv_map,
+    &ett_ldp_diffserv_map_phbid
   };
   module_t *ldp_module;
 
@@ -2570,7 +2712,7 @@ proto_register_ldp(void)
 				 10, &global_ldp_udp_port);
 
   prefs_register_bool_preference(ldp_module, "desegment_ldp_messages",
-    "Desegment all LDP messages spanning multiple TCP segments",
+    "Desegment all LDP messages\nspanning multiple TCP segments",
     "Whether the LDP dissector should desegment all messages spanning multiple TCP segments",
     &ldp_desegment);
 }
