@@ -1,9 +1,15 @@
 /* packet-mtp3.c
  * Routines for Message Transfer Part Level 3 dissection
- * Copyright 2001, Michael Tuexen <Michael.Tuexen@icn.siemens.de>
- * Updated for ANSI support by Jeff Morriss <jeff.morriss[AT]ulticom.com>
  *
- * $Id: packet-mtp3.c,v 1.14 2003/01/28 23:56:39 guy Exp $
+ * It is (hopefully) compliant to:
+ *   ANSI T1.111.4-1996
+ *   ITU-T Q.704 7/1996
+ *   GF 001-9001 (Chinese ITU variant)
+ *
+ * Copyright 2001, Michael Tuexen <tuexen [AT] fh-muenster.de>
+ * Updated for ANSI and Chinese ITU support by Jeff Morriss <jeff.morriss[AT]ulticom.com>
+ *
+ * $Id: packet-mtp3.c,v 1.26 2004/02/29 08:30:10 guy Exp $
  *
  * Ethereal - Network traffic analyzer
  * By Gerald Combs <gerald@ethereal.com>
@@ -48,18 +54,23 @@ static int hf_mtp3_network_indicator = -1;
 static int hf_mtp3_itu_spare = -1;
 static int hf_mtp3_ansi_priority = -1;
 static int hf_mtp3_itu_opc = -1;
+static int hf_mtp3_24bit_opc = -1;
 static int hf_mtp3_ansi_opc = -1;
+static int hf_mtp3_chinese_opc = -1;
 static int hf_mtp3_opc_network = -1;
 static int hf_mtp3_opc_cluster = -1;
 static int hf_mtp3_opc_member = -1;
 static int hf_mtp3_itu_dpc = -1;
+static int hf_mtp3_24bit_dpc = -1;
 static int hf_mtp3_ansi_dpc = -1;
+static int hf_mtp3_chinese_dpc = -1;
 static int hf_mtp3_dpc_network = -1;
 static int hf_mtp3_dpc_cluster = -1;
 static int hf_mtp3_dpc_member = -1;
 static int hf_mtp3_itu_sls = -1;
 static int hf_mtp3_ansi_5_bit_sls = -1;
 static int hf_mtp3_ansi_8_bit_sls = -1;
+static int hf_mtp3_chinese_itu_sls = -1;
 
 /* Initialize the subtree pointers */
 static gint ett_mtp3 = -1;
@@ -73,7 +84,9 @@ static dissector_table_t mtp3_sio_dissector_table;
 #include <packet-mtp3.h>
 Standard_Type mtp3_standard = ITU_STANDARD;
 
-gboolean mtp3_use_ansi_5_bit_sls = FALSE;
+static gboolean mtp3_use_ansi_5_bit_sls = FALSE;
+static mtp3_net_addr_fmt_e mtp3_net_addr_fmt = MTP3_NET_ADDR_FMT_DASHED;
+static mtp3_addr_pc_t mtp3_addr_dpc, mtp3_addr_opc;
 
 #define SIO_LENGTH                1
 
@@ -109,6 +122,7 @@ gboolean mtp3_use_ansi_5_bit_sls = FALSE;
 #define ANSI_MEMBER_MASK               0xFF0000
 #define ANSI_5BIT_SLS_MASK             0x1F
 #define ANSI_8BIT_SLS_MASK             0xFF
+#define CHINESE_ITU_SLS_MASK           0xF
 
 static const value_string service_indicator_code_vals[] = {
 	{ 0x0,	"Signalling Network Management Message (SNM)" },
@@ -140,6 +154,95 @@ static const value_string network_indicator_vals[] = {
 
 static dissector_handle_t data_handle;
 
+
+/*
+ * helper routine to format address to string
+ */
+void
+mtp3_addr_to_str_buf(
+  const guint8		*data,
+  gchar			*buf)
+{
+  mtp3_addr_pc_t	*addr_pc_p = (mtp3_addr_pc_t *) data;
+
+  switch (mtp3_net_addr_fmt)
+  {
+  case MTP3_NET_ADDR_FMT_DEC:
+    switch (addr_pc_p->type)
+    {
+    case ITU_STANDARD:
+      sprintf(buf, "%u", addr_pc_p->pc & ITU_PC_MASK);
+      break;
+    default:
+      /* assuming 24-bit */
+      sprintf(buf, "%u", addr_pc_p->pc & ANSI_PC_MASK);
+      break;
+    }
+    break;
+
+  case MTP3_NET_ADDR_FMT_HEX:
+    switch (addr_pc_p->type)
+    {
+    case ITU_STANDARD:
+      sprintf(buf, "%x", addr_pc_p->pc & ITU_PC_MASK);
+      break;
+    default:
+      /* assuming 24-bit */
+      sprintf(buf, "%x", addr_pc_p->pc & ANSI_PC_MASK);
+      break;
+    }
+    break;
+
+  case MTP3_NET_ADDR_FMT_NI_DEC:
+    switch (addr_pc_p->type)
+    {
+    case ITU_STANDARD:
+      sprintf(buf, "%u:%u", addr_pc_p->ni, addr_pc_p->pc & ITU_PC_MASK);
+      break;
+    default:
+      /* assuming 24-bit */
+      sprintf(buf, "%u:%u", addr_pc_p->ni, addr_pc_p->pc & ANSI_PC_MASK);
+      break;
+    }
+    break;
+
+  case MTP3_NET_ADDR_FMT_NI_HEX:
+    switch (addr_pc_p->type)
+    {
+    case ITU_STANDARD:
+      sprintf(buf, "%u:%x", addr_pc_p->ni, addr_pc_p->pc & ITU_PC_MASK);
+      break;
+    default:
+      /* assuming 24-bit */
+      sprintf(buf, "%u:%x", addr_pc_p->ni, addr_pc_p->pc & ANSI_PC_MASK);
+      break;
+    }
+    break;
+
+  default:
+    /* FALLTHRU */
+
+  case MTP3_NET_ADDR_FMT_DASHED:
+    switch (addr_pc_p->type)
+    {
+    case ITU_STANDARD:
+      sprintf(buf, "%u-%u-%u",
+	(addr_pc_p->pc & 0x3800) >> 11,
+	((addr_pc_p->pc & 0x7f8) >> 3),
+	(addr_pc_p->pc & 0x07));
+      break;
+    default:
+      /* assuming 24-bit */
+      sprintf(buf, "%u-%u-%u",
+	(addr_pc_p->pc & ANSI_NETWORK_MASK),
+	((addr_pc_p->pc & ANSI_CLUSTER_MASK) >> 8),
+	((addr_pc_p->pc & ANSI_MEMBER_MASK) >> 16));
+      break;
+    }
+    break;
+  }
+}
+
 static void
 dissect_mtp3_sio(tvbuff_t *tvb, packet_info *pinfo, proto_tree *mtp3_tree)
 {
@@ -155,12 +258,16 @@ dissect_mtp3_sio(tvbuff_t *tvb, packet_info *pinfo, proto_tree *mtp3_tree)
   proto_tree_add_uint(sio_tree, hf_mtp3_network_indicator, tvb, SIO_OFFSET,
 		      SIO_LENGTH, sio);
 
+  mtp3_addr_opc.ni = (sio & NETWORK_INDICATOR_MASK) >> 6;
+  mtp3_addr_dpc.ni = (sio & NETWORK_INDICATOR_MASK) >> 6;
+
   switch(mtp3_standard){
   case ANSI_STANDARD:
     proto_tree_add_uint(sio_tree, hf_mtp3_ansi_priority, tvb, SIO_OFFSET,
 			SIO_LENGTH, sio);
     break;
   case ITU_STANDARD:
+  case CHINESE_ITU_STANDARD:
     proto_tree_add_uint(sio_tree, hf_mtp3_itu_spare, tvb, SIO_OFFSET,
 			SIO_LENGTH, sio);
     break;
@@ -174,13 +281,16 @@ dissect_mtp3_sio(tvbuff_t *tvb, packet_info *pinfo, proto_tree *mtp3_tree)
 }
 
 static void
-dissect_mtp3_routing_label(tvbuff_t *tvb, proto_tree *mtp3_tree)
+dissect_mtp3_routing_label(tvbuff_t *tvb, packet_info *pinfo, proto_tree *mtp3_tree)
 {
-  guint32 label, dpc, opc;
+  guint32 label, dpc = 0, opc = 0;
   guint8 sls;
   proto_item *label_item, *label_dpc_item, *label_opc_item;
   proto_tree *label_tree, *label_dpc_tree, *label_opc_tree;
   char pc[ANSI_PC_STRING_LENGTH];
+  int *hf_dpc_string;
+  int *hf_opc_string;
+
 
   switch (mtp3_standard) {
   case ITU_STANDARD:
@@ -189,43 +299,60 @@ dissect_mtp3_routing_label(tvbuff_t *tvb, proto_tree *mtp3_tree)
     label_tree = proto_item_add_subtree(label_item, ett_mtp3_label);
 
     label = tvb_get_letohl(tvb, ITU_ROUTING_LABEL_OFFSET);
-    sls   = tvb_get_guint8(tvb, ITU_ROUTING_LABEL_OFFSET);
+    sls   = tvb_get_guint8(tvb, ITU_SLS_OFFSET);
+
+    opc = (label & ITU_OPC_MASK) >> 14;
+    dpc = label & ITU_DPC_MASK;
+
+    label_dpc_item =
+	proto_tree_add_uint(label_tree, hf_mtp3_itu_dpc, tvb,
+			    ITU_ROUTING_LABEL_OFFSET, ITU_ROUTING_LABEL_LENGTH,
+			    label);
+    proto_item_append_text(label_dpc_item, " (%u-%u-%u)",
+			   (dpc & 0x3800) >> 11,
+			   ((dpc & 0x7f8) >> 3),
+			   (dpc & 0x07));
+
+    label_opc_item =
+	proto_tree_add_uint(label_tree, hf_mtp3_itu_opc, tvb,
+			    ITU_ROUTING_LABEL_OFFSET, ITU_ROUTING_LABEL_LENGTH,
+			    label);
+    proto_item_append_text(label_opc_item, " (%u-%u-%u)",
+			   (opc & 0x3800) >> 11,
+			   ((opc & 0x7f8) >> 3),
+			   (opc & 0x07));
 
-    proto_tree_add_uint(label_tree, hf_mtp3_itu_dpc, tvb,
-			ITU_ROUTING_LABEL_OFFSET, ITU_ROUTING_LABEL_LENGTH,
-			label);
-    proto_tree_add_uint(label_tree, hf_mtp3_itu_opc, tvb,
-			ITU_ROUTING_LABEL_OFFSET, ITU_ROUTING_LABEL_LENGTH,
-			label);
     proto_tree_add_uint(label_tree, hf_mtp3_itu_sls, tvb, ITU_SLS_OFFSET,
 			ITU_SLS_LENGTH, sls);
     break;
 
   case ANSI_STANDARD:
+  case CHINESE_ITU_STANDARD:
+    if (mtp3_standard == ANSI_STANDARD)
+    {
+      hf_dpc_string = &hf_mtp3_ansi_dpc;
+      hf_opc_string = &hf_mtp3_ansi_opc;
+    } else /* CHINESE_ITU_STANDARD */ {
+      hf_dpc_string = &hf_mtp3_chinese_dpc;
+      hf_opc_string = &hf_mtp3_chinese_opc;
+    }
+
     /* Create the Routing Label Tree */
     label_item = proto_tree_add_text(mtp3_tree, tvb, ANSI_ROUTING_LABEL_OFFSET,
 				     ANSI_ROUTING_LABEL_LENGTH,
 				     "Routing label");
     label_tree = proto_item_add_subtree(label_item, ett_mtp3_label);
 
-    /* SLS */
-    if (mtp3_use_ansi_5_bit_sls)
-      proto_tree_add_item(label_tree, hf_mtp3_ansi_5_bit_sls, tvb,
-			  ANSI_SLS_OFFSET, ANSI_SLS_LENGTH, TRUE);
-    else
-      proto_tree_add_item(label_tree, hf_mtp3_ansi_8_bit_sls, tvb,
-			  ANSI_SLS_OFFSET, ANSI_SLS_LENGTH, TRUE);
-
     /* create the DPC tree */
     dpc = tvb_get_ntoh24(tvb, ANSI_DPC_OFFSET);
     snprintf(pc, sizeof(pc), "%d-%d-%d", (dpc & ANSI_NETWORK_MASK),
 					 ((dpc & ANSI_CLUSTER_MASK) >> 8),
 					 ((dpc & ANSI_MEMBER_MASK) >> 16));
 
-    label_dpc_item = proto_tree_add_string_format(label_tree, hf_mtp3_ansi_dpc,
+    label_dpc_item = proto_tree_add_string_format(label_tree, *hf_dpc_string,
 						  tvb, ANSI_DPC_OFFSET,
 						  ANSI_PC_LENGTH, pc,
-						  "DPC (%s)", pc);
+						  "DPC (%s) (%u)", pc, dpc);
 
     label_dpc_tree = proto_item_add_subtree(label_dpc_item, ett_mtp3_label_dpc);
 
@@ -239,15 +366,20 @@ dissect_mtp3_routing_label(tvbuff_t *tvb, proto_tree *mtp3_tree)
 			ANSI_DPC_OFFSET + ANSI_NETWORK_OFFSET, ANSI_NCM_LENGTH,
 			dpc);
 
+
+    /* add full integer values of DPC as hidden for filtering purposes */
+    proto_tree_add_uint_hidden(label_dpc_tree, hf_mtp3_24bit_dpc, tvb, ANSI_DPC_OFFSET,
+			       ANSI_PC_LENGTH, dpc);
+
     /* create the OPC tree */
     opc = tvb_get_ntoh24(tvb, ANSI_OPC_OFFSET);
     snprintf(pc, sizeof(pc), "%d-%d-%d", (opc & ANSI_NETWORK_MASK),
 					 ((opc & ANSI_CLUSTER_MASK) >> 8),
 					 ((opc & ANSI_MEMBER_MASK) >> 16));
-    label_opc_item = proto_tree_add_string_format(label_tree, hf_mtp3_ansi_opc,
+    label_opc_item = proto_tree_add_string_format(label_tree, *hf_opc_string,
 						  tvb, ANSI_OPC_OFFSET,
 						  ANSI_PC_LENGTH, pc,
-						  "OPC (%s)", pc);
+						  "OPC (%s) (%u)", pc, opc);
 
     label_opc_tree = proto_item_add_subtree(label_opc_item, ett_mtp3_label_opc);
 
@@ -260,8 +392,35 @@ dissect_mtp3_routing_label(tvbuff_t *tvb, proto_tree *mtp3_tree)
     proto_tree_add_uint(label_opc_tree, hf_mtp3_opc_network,tvb,
 			ANSI_OPC_OFFSET + ANSI_NETWORK_OFFSET, ANSI_NCM_LENGTH,
 			opc);
+
+    /* add full integer values of OPC as hidden for filtering purposes */
+    proto_tree_add_uint_hidden(label_opc_tree, hf_mtp3_24bit_opc, tvb, ANSI_OPC_OFFSET,
+			       ANSI_PC_LENGTH, opc);
+
+    /* SLS */
+    if (mtp3_standard == ANSI_STANDARD)
+    {
+	if (mtp3_use_ansi_5_bit_sls)
+	    proto_tree_add_item(label_tree, hf_mtp3_ansi_5_bit_sls, tvb,
+				ANSI_SLS_OFFSET, ANSI_SLS_LENGTH, TRUE);
+	else
+	    proto_tree_add_item(label_tree, hf_mtp3_ansi_8_bit_sls, tvb,
+				ANSI_SLS_OFFSET, ANSI_SLS_LENGTH, TRUE);
+    } else /* CHINESE_ITU_STANDARD */ {
+	proto_tree_add_item(label_tree, hf_mtp3_chinese_itu_sls, tvb,
+			    ANSI_SLS_OFFSET, ITU_SLS_LENGTH, FALSE);
+    }
+
     break;
   }
+
+  mtp3_addr_opc.type = mtp3_standard;
+  mtp3_addr_opc.pc = opc;
+  SET_ADDRESS(&pinfo->net_src, AT_SS7PC, sizeof(mtp3_addr_opc), (guint8 *) &mtp3_addr_opc);
+
+  mtp3_addr_dpc.type = mtp3_standard;
+  mtp3_addr_dpc.pc = dpc;
+  SET_ADDRESS(&pinfo->net_dst, AT_SS7PC, sizeof(mtp3_addr_dpc), (guint8 *) &mtp3_addr_dpc);
 }
 
 static void
@@ -279,16 +438,17 @@ dissect_mtp3_payload(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
     payload_tvb = tvb_new_subset(tvb, ITU_MTP_PAYLOAD_OFFSET, -1, -1);
     break;
   case ANSI_STANDARD:
+  case CHINESE_ITU_STANDARD:
     payload_tvb = tvb_new_subset(tvb, ANSI_MTP_PAYLOAD_OFFSET, -1, -1);
     break;
   }
 
+  if (check_col(pinfo->cinfo, COL_INFO))
+      col_set_str(pinfo->cinfo, COL_INFO, "DATA ");
+
   if (!dissector_try_port(mtp3_sio_dissector_table, service_indicator,
-			  payload_tvb, pinfo, tree)) {
+			  payload_tvb, pinfo, tree))
     call_dissector(data_handle, payload_tvb, pinfo, tree);
-    if (check_col(pinfo->cinfo, COL_INFO))
-      col_append_str(pinfo->cinfo, COL_INFO, "DATA ");
-  }
 }
 
 /* Code to actually dissect the packets */
@@ -300,9 +460,19 @@ dissect_mtp3(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
   proto_item *mtp3_item = NULL;
   proto_tree *mtp3_tree = NULL;
 
-  /* Make entries in Protocol column and Info column on summary display */
+  /* Make entries in Protocol column on summary display */
   if (check_col(pinfo->cinfo, COL_PROTOCOL))
-    col_set_str(pinfo->cinfo, COL_PROTOCOL, "MTP3");
+    switch(mtp3_standard) {
+      case ITU_STANDARD:
+        col_set_str(pinfo->cinfo, COL_PROTOCOL, "MTP3 (Int. ITU)");
+        break;
+      case ANSI_STANDARD:
+        col_set_str(pinfo->cinfo, COL_PROTOCOL, "MTP3 (ANSI)");
+        break;
+      case CHINESE_ITU_STANDARD:
+        col_set_str(pinfo->cinfo, COL_PROTOCOL, "MTP3 (Chin. ITU)");
+        break;
+    };      
 
   if (tree) {
     /* create display subtree for the protocol */
@@ -312,6 +482,7 @@ dissect_mtp3(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 				      ITU_HEADER_LENGTH, TRUE);
       break;
     case ANSI_STANDARD:
+    case CHINESE_ITU_STANDARD:
       mtp3_item = proto_tree_add_item(tree, proto_mtp3, tvb, 0,
 				      ANSI_HEADER_LENGTH, TRUE);
       break;
@@ -320,10 +491,10 @@ dissect_mtp3(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 
   }
 
-  /* Dissect the packet (even if !tree so can call sub-dissectors) */
+  /* Dissect the packet (even if !tree so can call sub-dissectors and update
+   * the source and destination address columns) */
   dissect_mtp3_sio(tvb, pinfo, mtp3_tree);
-  if (tree)
-    dissect_mtp3_routing_label(tvb, mtp3_tree);
+  dissect_mtp3_routing_label(tvb, pinfo, mtp3_tree);
   dissect_mtp3_payload(tvb, pinfo, tree);
 }
 
@@ -353,10 +524,18 @@ proto_register_mtp3(void)
       { "OPC", "mtp3.opc",
 	      FT_UINT32, BASE_DEC, NULL, ITU_OPC_MASK,
 	      "", HFILL }},
+    { &hf_mtp3_24bit_opc,
+      { "OPC", "mtp3.opc",
+	      FT_UINT32, BASE_DEC, NULL, ANSI_PC_MASK,
+	      "", HFILL }},
     { &hf_mtp3_ansi_opc,
       { "DPC", "mtp3.ansi_opc",
 	      FT_STRING, BASE_NONE, NULL, 0x0,
 	      "", HFILL }},
+    { &hf_mtp3_chinese_opc,
+      { "DPC", "mtp3.chinese_opc",
+	      FT_STRING, BASE_NONE, NULL, 0x0,
+	      "", HFILL }},
     { &hf_mtp3_opc_network,
      { "OPC Network", "mtp3.opc.network",
 	     FT_UINT24, BASE_DEC, NULL, ANSI_NETWORK_MASK,
@@ -373,10 +552,18 @@ proto_register_mtp3(void)
       { "DPC", "mtp3.dpc",
 	      FT_UINT32, BASE_DEC, NULL, ITU_DPC_MASK,
 	      "", HFILL }},
+    { &hf_mtp3_24bit_dpc,
+      { "DPC", "mtp3.dpc",
+	      FT_UINT32, BASE_DEC, NULL, ANSI_PC_MASK,
+	      "", HFILL }},
     { &hf_mtp3_ansi_dpc,
       { "DPC", "mtp3.ansi_dpc",
 	      FT_STRING, BASE_NONE, NULL, 0x0,
 	      "", HFILL }},
+    { &hf_mtp3_chinese_dpc,
+      { "DPC", "mtp3.chinese_dpc",
+	      FT_STRING, BASE_NONE, NULL, 0x0,
+	      "", HFILL }},
     { &hf_mtp3_dpc_network,
       { "DPC Network", "mtp3.dpc.network",
 	      FT_UINT24, BASE_DEC, NULL, ANSI_NETWORK_MASK,
@@ -400,6 +587,10 @@ proto_register_mtp3(void)
     { &hf_mtp3_ansi_8_bit_sls,
       { "Signalling Link Selector", "mtp3.sls",
 	      FT_UINT8, BASE_DEC, NULL, ANSI_8BIT_SLS_MASK,
+	      "", HFILL }},
+    { &hf_mtp3_chinese_itu_sls,
+      { "Signalling Link Selector", "mtp3.sls",
+	      FT_UINT8, BASE_DEC, NULL, CHINESE_ITU_SLS_MASK,
 	      "", HFILL }}
   };
 
@@ -413,11 +604,21 @@ proto_register_mtp3(void)
   };
 
   static enum_val_t mtp3_options[] = {
-    { "ITU",  ITU_STANDARD },
-    { "ANSI", ANSI_STANDARD },
+    { "ITU",         ITU_STANDARD },
+    { "ANSI",        ANSI_STANDARD },
+    { "Chinese ITU", CHINESE_ITU_STANDARD },
     { NULL, 0 }
   };
 
+  static enum_val_t mtp3_net_addr_fmt_str_e[] = {
+    { "Decimal",		MTP3_NET_ADDR_FMT_DEC },
+    { "Hexadecimal",		MTP3_NET_ADDR_FMT_HEX },
+    { "NI-Decimal",		MTP3_NET_ADDR_FMT_NI_DEC },
+    { "NI-Hexadecimal",		MTP3_NET_ADDR_FMT_NI_HEX },
+    { "Dashed",			MTP3_NET_ADDR_FMT_DASHED },
+    { NULL,			0 }
+  };
+
   /* Register the protocol name and description */
   proto_mtp3 = proto_register_protocol("Message Transfer Part Level 3",
 				       "MTP3", "mtp3");
@@ -441,10 +642,20 @@ proto_register_mtp3(void)
 				 "Use 5-bit SLS (ANSI only)",
 				 "Use 5-bit (instead of 8-bit) SLS in ANSI MTP3 packets",
 				 &mtp3_use_ansi_5_bit_sls);
+
+  prefs_register_enum_preference(mtp3_module, "net_addr_format", "Network Address Format",
+				 "Format for point code in the network address columns",
+				 (gint *)&mtp3_net_addr_fmt, mtp3_net_addr_fmt_str_e, FALSE);
 }
 
 void
 proto_reg_handoff_mtp3(void)
 {
+  dissector_handle_t mtp3_handle; 	 
+     	 
+  mtp3_handle = create_dissector_handle(dissect_mtp3, proto_mtp3); 	 
+        	 
+  dissector_add("wtap_encap", WTAP_ENCAP_MTP3, mtp3_handle); 	 
+	   	 
   data_handle = find_dissector("data");
 }
