@@ -1,8 +1,9 @@
 /* packet-ntlmssp.c
  * Routines for NTLM Secure Service Provider
  * Devin Heitmueller <dheitmueller@netilla.com>
+ * Copyright 2003, Tim Potter <tpot@samba.org>
  *
- * $Id: packet-ntlmssp.c,v 1.35 2003/03/04 20:52:33 guy Exp $
+ * $Id: packet-ntlmssp.c,v 1.47 2004/02/25 09:31:06 guy Exp $
  *
  * Ethereal - Network traffic analyzer
  * By Gerald Combs <gerald@ethereal.com>
@@ -38,7 +39,7 @@
 #include "crypt-rc4.h"
 #include "crypt-md4.h"
 #include "crypt-des.h"
-#include "packet-ntlmssp.h"
+#include "packet-dcerpc.h"
 
 /* Message types */
 
@@ -92,7 +93,6 @@ static const value_string ntlmssp_message_types[] = {
 #define NTLMSSP_NEGOTIATE_KEY_EXCH         0x40000000
 #define NTLMSSP_NEGOTIATE_80000000         0x80000000
 
-
 static int proto_ntlmssp = -1;
 static int hf_ntlmssp = -1;
 static int hf_ntlmssp_auth = -1;
@@ -161,6 +161,10 @@ static int hf_ntlmssp_address_list_server_nb = -1;
 static int hf_ntlmssp_address_list_domain_nb = -1;
 static int hf_ntlmssp_address_list_server_dns = -1;
 static int hf_ntlmssp_address_list_domain_dns = -1;
+static int hf_ntlmssp_address_list_terminator = -1;
+static int hf_ntlmssp_address_list_item_type = -1;
+static int hf_ntlmssp_address_list_item_len = -1;
+static int hf_ntlmssp_address_list_item_content = -1;
 static int hf_ntlmssp_verf = -1;
 static int hf_ntlmssp_verf_vers = -1;
 static int hf_ntlmssp_verf_body = -1;
@@ -174,7 +178,7 @@ static gint ett_ntlmssp_negotiate_flags = -1;
 static gint ett_ntlmssp_string = -1;
 static gint ett_ntlmssp_blob = -1;
 static gint ett_ntlmssp_address_list = -1;
-static gint ett_ntlmssp_decrypted_tree = -1;
+static gint ett_ntlmssp_address_list_item = -1;
 
 /* Configuration variables */
 static char *nt_password = NULL;
@@ -271,7 +275,7 @@ create_ntlmssp_v1_key(const char *nt_password, const guint8 *challenge,
   memset(lm_password_upper, 0, sizeof(lm_password_upper));
 
   /* Create a Lan Manager hash of the input password */
-  if (nt_password) {
+  if (nt_password[0] != '\0') {
     password_len = strlen(nt_password);
     /* Truncate password if too long */
     if (password_len > 16)
@@ -295,7 +299,7 @@ create_ntlmssp_v1_key(const char *nt_password, const guint8 *challenge,
   memset(pw21, 0xBD, sizeof(pw21));
   memcpy(pw21, lm_password_hash, sizeof(lm_password_hash));
 
-  /* Only the first eight bytes of challenge_reponse is used */
+  /* Only the first eight bytes of challenge_response is used */
   crypt_des_ecb(rc4key, lm_challenge_response, pw21, 1);
   crypt_des_ecb(rc4key + 8, lm_challenge_response, pw21 + 7, 1);
   crypt_des_ecb(rc4key + 16, lm_challenge_response, pw21 + 14, 1);
@@ -422,6 +426,11 @@ dissect_ntlmssp_blob (tvbuff_t *tvb, int offset,
       tvb_memcpy(tvb, result->contents, blob_offset, blob_length);
   }
 
+  /* If we are dissecting the NTLM response and it is a NTLMv2
+     response call the appropriate dissector. */
+
+  if (blob_hf == hf_ntlmssp_auth_ntresponse && blob_length > 24)
+	  dissect_ntlmv2_response(tvb, tree, blob_offset, blob_length);
 
   return offset;
 }
@@ -577,9 +586,11 @@ dissect_ntlmssp_address_list (tvbuff_t *tvb, int offset,
   guint16 list_maxlen = tvb_get_letohs(tvb, offset+2);
   guint32 list_offset = tvb_get_letohl(tvb, offset+4);
   guint16 item_type, item_length;
-  int item_offset;
+  guint32 item_offset;
   proto_item *tf = NULL;
   proto_tree *tree = NULL;
+  proto_item *addr_tf = NULL;
+  proto_tree *addr_tree = NULL;
 
   /* the address list is just a blob */
   if (0 == list_length) {
@@ -605,49 +616,79 @@ dissect_ntlmssp_address_list (tvbuff_t *tvb, int offset,
 		      tvb, offset, 4, list_offset);
   offset += 4;
 
+  /* Now enumerate through the individual items in the list */
   item_offset = list_offset;
-  item_type = tvb_get_letohs(tvb, item_offset);
-  item_offset += 2;
-  item_length = tvb_get_letohs(tvb, item_offset);
-  item_offset += 2;
-  while (item_type) {
-    guint16 bc;
-    int result_length;
-    const char *text;
-    bc = item_length;
+
+  while (item_offset < (list_offset + list_length)) {
+    const char *text=NULL;
+    guint32 content_offset;
+    guint16 content_length;
+    guint32 type_offset;
+    guint32 len_offset;
+
+    /* Content type */
+    type_offset = item_offset;
+    item_type = tvb_get_letohs(tvb, type_offset);
+
+    /* Content length */
+    len_offset = type_offset + 2;
+    content_length = tvb_get_letohs(tvb, len_offset);
+
+    /* Content value */
+    content_offset = len_offset + 2;
+    item_length = content_length + 4;
 
     /* Strings are always in unicode regardless of the negotiated
        string type. */
+    if (content_length > 0) {
+      guint16 bc;
+      int result_length;
+      int item_offset_int;
 
-    text = get_unicode_or_ascii_string(tvb, &item_offset,
-				       TRUE, &result_length,
-				       FALSE, FALSE, &bc);
+      item_offset_int = content_offset;
+      bc = content_length;
+      text = get_unicode_or_ascii_string(tvb, &item_offset_int,
+					 TRUE, &result_length,
+					 FALSE, FALSE, &bc);
+    }
 
     if (!text) text = ""; /* Make sure we don't blow up below */
 
     switch(item_type) {
-    case 1:
-      proto_tree_add_string(tree, hf_ntlmssp_address_list_server_nb,
-			    tvb, item_offset, item_length, text);
+    case NTLM_NAME_NB_HOST:
+      addr_tf = proto_tree_add_string(tree, hf_ntlmssp_address_list_server_nb,
+				      tvb, item_offset, item_length, text);
       break;
-    case 2:
-      proto_tree_add_string(tree, hf_ntlmssp_address_list_domain_nb,
-			    tvb, item_offset, item_length, text);
+    case NTLM_NAME_NB_DOMAIN:
+      addr_tf = proto_tree_add_string(tree, hf_ntlmssp_address_list_domain_nb,
+				      tvb, item_offset, item_length, text);
       break;
-    case 3:
-      proto_tree_add_string(tree, hf_ntlmssp_address_list_server_dns,
-			    tvb, item_offset, item_length, text);
+    case NTLM_NAME_DNS_HOST:
+      addr_tf = proto_tree_add_string(tree, hf_ntlmssp_address_list_server_dns,
+				      tvb, item_offset, item_length, text);
       break;
-    case 4:
-      proto_tree_add_string(tree, hf_ntlmssp_address_list_domain_dns,
-			    tvb, item_offset, item_length, text);
+    case NTLM_NAME_DNS_DOMAIN:
+      addr_tf = proto_tree_add_string(tree, hf_ntlmssp_address_list_domain_dns,
+				      tvb, item_offset, item_length, text);
+      break;
+    case NTLM_NAME_END:
+      addr_tf = proto_tree_add_item(tree, hf_ntlmssp_address_list_terminator,
+				    tvb, item_offset, item_length, TRUE);
+    }
+
+    /* Now show the actual bytes that made up the summary line */
+    addr_tree = proto_item_add_subtree (addr_tf, 
+					ett_ntlmssp_address_list_item);
+    proto_tree_add_item (addr_tree, hf_ntlmssp_address_list_item_type,
+			 tvb, type_offset, 2, TRUE);
+    proto_tree_add_item (addr_tree, hf_ntlmssp_address_list_item_len,
+			 tvb, len_offset, 2, TRUE);
+    if (content_length > 0) {
+      proto_tree_add_string(addr_tree, hf_ntlmssp_address_list_item_content,
+			    tvb, content_offset, content_length, text);
     }
 
     item_offset += item_length;
-    item_type = tvb_get_letohs(tvb, item_offset);
-    item_offset += 2;
-    item_length = tvb_get_letohs(tvb, item_offset);
-    item_offset += 2;
   }
 
   *end = list_offset + list_length;
@@ -825,7 +866,7 @@ dissect_ntlmssp_auth (tvbuff_t *tvb, packet_info *pinfo, int offset,
 				hf_ntlmssp_auth_ntresponse,
 				&item_end,
 				conv_ntlmssp_info == NULL ? NULL :
-				    &conv_ntlmssp_info->ntlm_response);
+				&conv_ntlmssp_info->ntlm_response);
   data_start = MIN(data_start, item_start);
   data_end = MAX(data_end, item_end);
 
@@ -1167,12 +1208,11 @@ dissect_ntlmssp_verf(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
   return offset;
 }
 
-
-static int
-dissect_ntlmssp_encrypted_payload(tvbuff_t *tvb, 
-				  packet_info *pinfo, proto_tree *tree)
+static tvbuff_t *
+dissect_ntlmssp_encrypted_payload(tvbuff_t *tvb, int offset,
+				  packet_info *pinfo, 
+				  dcerpc_auth_info *auth_info _U_)
 {
-  volatile int offset = 0;
   tvbuff_t *decr_tvb; /* Used to display decrypted buffer */
   guint8 *peer_block;
   conversation_t *conversation;
@@ -1181,8 +1221,6 @@ dissect_ntlmssp_encrypted_payload(tvbuff_t *tvb,
   rc4_state_struct *rc4_state_peer;
   ntlmssp_info *conv_ntlmssp_info = NULL;
   ntlmssp_packet_info *packet_ntlmssp_info = NULL;
-  proto_item *it;
-  static ntlmssp_decrypted_info_t ndi;
 
   encrypted_block_length = tvb_length_remaining (tvb, offset);
 
@@ -1202,14 +1240,14 @@ dissect_ntlmssp_encrypted_payload(tvbuff_t *tvb,
 				     pinfo->destport, 0);
     if (conversation == NULL) {
       /* There is no conversation, thus no encryption state */
-      return offset + encrypted_block_length;
+      return NULL;
     }
     
     conv_ntlmssp_info = conversation_get_proto_data(conversation,
 						    proto_ntlmssp);
     if (conv_ntlmssp_info == NULL) {
       /* There is no NTLMSSP state tied to the conversation */
-      return offset + encrypted_block_length;
+	    return NULL;
     }
     
     /* Get the pair of RC4 state structures.  One is used for to decrypt the
@@ -1225,16 +1263,15 @@ dissect_ntlmssp_encrypted_payload(tvbuff_t *tvb,
     
     if (rc4_state == NULL || rc4_state_peer == NULL) {
       /* There is no encryption state, so we cannot decrypt */
-      return offset + encrypted_block_length;
+      return NULL;
     }
 
     /* Store the decrypted contents in the packet state struct
        (of course at this point, they aren't decrypted yet) */
-    packet_ntlmssp_info->decrypted_payload = g_malloc (encrypted_block_length);
+    packet_ntlmssp_info->decrypted_payload = tvb_memdup(tvb, offset,
+                                                        encrypted_block_length);
     decrypted_payloads = g_slist_prepend(decrypted_payloads,
                                          packet_ntlmssp_info->decrypted_payload);
-    tvb_memcpy(tvb, packet_ntlmssp_info->decrypted_payload, 
-	       offset, encrypted_block_length);
     
     /* Do the decryption of the payload */
     crypt_rc4(rc4_state, packet_ntlmssp_info->decrypted_payload, 
@@ -1255,21 +1292,12 @@ dissect_ntlmssp_encrypted_payload(tvbuff_t *tvb,
   decr_tvb = tvb_new_real_data(packet_ntlmssp_info->decrypted_payload,
 			       encrypted_block_length,
 			       encrypted_block_length);
-  tvb_set_child_real_data_tvbuff(tvb, decr_tvb);
-  add_new_data_source(pinfo, decr_tvb,
-		      "Decrypted NTLMSSP block");
-  
-  /* Show the decrypted payload in the tree */
-  it=proto_tree_add_text(tree, decr_tvb, 0, -1,
-			"Decrypted stub data (%d byte%s)",
-			encrypted_block_length, 
-			plurality(encrypted_block_length, "", "s"));
-  ndi.decr_tree=proto_item_add_subtree(it, ett_ntlmssp_decrypted_tree);
-  ndi.decr_tvb=decr_tvb;    
-  pinfo->decrypted_data=&ndi;
 
+  tvb_set_child_real_data_tvbuff(tvb, decr_tvb);
+  
   offset += encrypted_block_length;
-  return offset;
+
+  return decr_tvb;
 }
 
 static void
@@ -1441,6 +1469,12 @@ proto_register_ntlmssp(void)
       { "Maxlen", "ntlmssp.challenge.addresslist.maxlen", FT_UINT16, BASE_DEC, NULL, 0x0, "", HFILL}},
     { &hf_ntlmssp_address_list_offset,
       { "Offset", "ntlmssp.challenge.addresslist.offset", FT_UINT32, BASE_DEC, NULL, 0x0, "", HFILL}},
+    { &hf_ntlmssp_address_list_item_type,
+      { "Target item type", "ntlmssp.targetitemtype", FT_UINT16, BASE_HEX, VALS(ntlm_name_types), 0x0, "", HFILL }},
+    { &hf_ntlmssp_address_list_item_len,
+      { "Target item Length", "ntlmssp.challenge.addresslist.item.length", FT_UINT16, BASE_DEC, NULL, 0x0, "", HFILL}},
+    { &hf_ntlmssp_address_list_item_content,
+      { "Target item Content", "ntlmssp.challenge.addresslist.item.content", FT_STRING, BASE_NONE, NULL, 0x0, "", HFILL}},
     { &hf_ntlmssp_address_list_server_nb,
       { "Server NetBIOS Name", "ntlmssp.challenge.addresslist.servernb", FT_STRING, BASE_NONE, NULL, 0x0, "", HFILL }},
     { &hf_ntlmssp_address_list_domain_nb,
@@ -1449,7 +1483,8 @@ proto_register_ntlmssp(void)
       { "Server DNS Name", "ntlmssp.challenge.addresslist.serverdns", FT_STRING, BASE_NONE, NULL, 0x0, "", HFILL }},
     { &hf_ntlmssp_address_list_domain_dns,
       { "Domain DNS Name", "ntlmssp.challenge.addresslist.domaindns", FT_STRING, BASE_NONE, NULL, 0x0, "", HFILL }},
-
+    { &hf_ntlmssp_address_list_terminator,
+      { "List Terminator", "ntlmssp.challenge.addresslist.terminator", FT_NONE, BASE_NONE, NULL, 0x0, "", HFILL }},
     { &hf_ntlmssp_verf,
       { "NTLMSSP Verifier", "ntlmssp.verf", FT_NONE, BASE_NONE, NULL, 0x0, "NTLMSSP Verifier", HFILL }},
     { &hf_ntlmssp_verf_vers,
@@ -1464,7 +1499,6 @@ proto_register_ntlmssp(void)
       { "Verifier CRC32", "ntlmssp.verf.crc32", FT_UINT32, BASE_HEX, NULL, 0x0, "", HFILL }},
     { &hf_ntlmssp_verf_sequence,
       { "Verifier Sequence Number", "ntlmssp.verf.sequence", FT_UINT32, BASE_DEC, NULL, 0x0, "", HFILL }}
-
   };
 
 
@@ -1474,7 +1508,7 @@ proto_register_ntlmssp(void)
     &ett_ntlmssp_string,
     &ett_ntlmssp_blob,
     &ett_ntlmssp_address_list,
-    &ett_ntlmssp_decrypted_tree
+    &ett_ntlmssp_address_list_item
   };
   module_t *ntlmssp_module;
   
@@ -1496,10 +1530,54 @@ proto_register_ntlmssp(void)
 
   register_dissector("ntlmssp", dissect_ntlmssp, proto_ntlmssp);
   new_register_dissector("ntlmssp_verf", dissect_ntlmssp_verf, proto_ntlmssp);
-  new_register_dissector("ntlmssp_encrypted_payload", 
-			 dissect_ntlmssp_encrypted_payload, proto_ntlmssp);
 }
 
+static int wrap_dissect_ntlmssp(tvbuff_t *tvb, int offset, packet_info *pinfo, 
+				proto_tree *tree, guint8 *drep _U_)
+{
+	tvbuff_t *auth_tvb;
+
+	auth_tvb = tvb_new_subset(
+		tvb, offset, tvb_length_remaining(tvb, offset),
+		tvb_length_remaining(tvb, offset));
+	
+	dissect_ntlmssp(auth_tvb, pinfo, tree);
+
+	return tvb_length_remaining(tvb, offset);
+}
+
+static int wrap_dissect_ntlmssp_verf(tvbuff_t *tvb, int offset, packet_info *pinfo, 
+				     proto_tree *tree, guint8 *drep _U_)
+{
+	tvbuff_t *auth_tvb;
+
+	auth_tvb = tvb_new_subset(
+		tvb, offset, tvb_length_remaining(tvb, offset),
+		tvb_length_remaining(tvb, offset));
+	
+	return dissect_ntlmssp_verf(auth_tvb, pinfo, tree);
+}
+
+static dcerpc_auth_subdissector_fns ntlmssp_sign_fns = {
+	wrap_dissect_ntlmssp, 			/* Bind */
+	wrap_dissect_ntlmssp,			/* Bind ACK */
+	wrap_dissect_ntlmssp,			/* AUTH3 */
+	wrap_dissect_ntlmssp_verf,		/* Request verifier */
+	wrap_dissect_ntlmssp_verf,		/* Response verifier */
+	NULL,			                /* Request data */
+	NULL			                /* Response data */
+};
+
+static dcerpc_auth_subdissector_fns ntlmssp_seal_fns = {
+	wrap_dissect_ntlmssp, 			/* Bind */
+	wrap_dissect_ntlmssp,			/* Bind ACK */
+	wrap_dissect_ntlmssp,			/* AUTH3 */
+	wrap_dissect_ntlmssp_verf, 		/* Request verifier */
+	wrap_dissect_ntlmssp_verf,		/* Response verifier */
+	dissect_ntlmssp_encrypted_payload,	/* Request data */
+	dissect_ntlmssp_encrypted_payload	/* Response data */
+};
+
 void
 proto_reg_handoff_ntlmssp(void)
 {     
@@ -1512,4 +1590,26 @@ proto_reg_handoff_ntlmssp(void)
   gssapi_init_oid("1.3.6.1.4.1.311.2.2.10", proto_ntlmssp, ett_ntlmssp, 
 		  ntlmssp_handle, ntlmssp_wrap_handle,
 		  "NTLMSSP - Microsoft NTLM Security Support Provider");
+
+  /* Register authenticated pipe dissector */
+
+  /*
+   * XXX - the verifiers here seem to have a version of 1 and a body of all
+   * zeroes.
+   *
+   * XXX - DCE_C_AUTHN_LEVEL_CONNECT is, according to the DCE RPC 1.1
+   * spec, upgraded to DCE_C_AUTHN_LEVEL_PKT.  Should we register
+   * any other levels here?
+   */
+  register_dcerpc_auth_subdissector(DCE_C_AUTHN_LEVEL_CONNECT,
+				    DCE_C_RPC_AUTHN_PROTOCOL_NTLMSSP,
+				    &ntlmssp_sign_fns);
+
+  register_dcerpc_auth_subdissector(DCE_C_AUTHN_LEVEL_PKT_INTEGRITY,
+				    DCE_C_RPC_AUTHN_PROTOCOL_NTLMSSP,
+				    &ntlmssp_sign_fns);
+
+  register_dcerpc_auth_subdissector(DCE_C_AUTHN_LEVEL_PKT_PRIVACY,
+				    DCE_C_RPC_AUTHN_PROTOCOL_NTLMSSP,
+				    &ntlmssp_seal_fns);
 }
