@@ -1,12 +1,12 @@
 /* packet-rtcp.c
  *
- * $Id: packet-rtcp.c,v 1.35 2002/08/28 21:00:29 jmayer Exp $
+ * $Id: packet-rtcp.c,v 1.40 2004/02/25 09:31:06 guy Exp $
  *
  * Routines for RTCP dissection
  * RTCP = Real-time Transport Control Protocol
  *
  * Copyright 2000, Philips Electronics N.V.
- * Written by Andreas Sikkema <andreas.sikkema@philips.com>
+ * Written by Andreas Sikkema <h323@ramdyne.nl>
  *
  * Ethereal - Network traffic analyzer
  * By Gerald Combs <gerald@ethereal.com>
@@ -230,7 +230,7 @@ static void rtcp_init( void )
 	fake_addr.type = AT_IPv4;
 	fake_addr.len = 4;
 
-	tmp_data = malloc( fake_addr.len );
+	tmp_data = g_malloc( fake_addr.len );
 	for ( i = 0; i < fake_addr.len; i++) {
 		tmp_data[i] = 0;
 	}
@@ -282,7 +282,7 @@ static int
 dissect_rtcp_nack( tvbuff_t *tvb, int offset, proto_tree *tree )
 {
 	/* Packet type = FIR (H261) */
-	proto_tree_add_uint( tree, hf_rtcp_rc, tvb, offset, 1, tvb_get_guint8( tvb, offset ) & 31 );
+	proto_tree_add_uint( tree, hf_rtcp_rc, tvb, offset, 1, tvb_get_guint8( tvb, offset ) );
 	offset++;
 	/* Packet type, 8 bits  = APP */
 	proto_tree_add_item( tree, hf_rtcp_pt, tvb, offset, 1, FALSE );
@@ -311,7 +311,7 @@ static int
 dissect_rtcp_fir( tvbuff_t *tvb, int offset, proto_tree *tree )
 {
 	/* Packet type = FIR (H261) */
-	proto_tree_add_uint( tree, hf_rtcp_rc, tvb, offset, 1, tvb_get_guint8( tvb, offset ) & 31 );
+	proto_tree_add_uint( tree, hf_rtcp_rc, tvb, offset, 1, tvb_get_guint8( tvb, offset ) );
 	offset++;
 	/* Packet type, 8 bits  = APP */
 	proto_tree_add_item( tree, hf_rtcp_pt, tvb, offset, 1, FALSE );
@@ -369,7 +369,6 @@ dissect_rtcp_bye( tvbuff_t *tvb, int offset, proto_tree *tree,
 {
 	unsigned int chunk          = 1;
 	unsigned int reason_length  = 0;
-	unsigned int counter = 0;
 	char* reason_text = NULL;
 
 	while ( chunk <= count ) {
@@ -385,12 +384,9 @@ dissect_rtcp_bye( tvbuff_t *tvb, int offset, proto_tree *tree,
 		proto_tree_add_item( tree, hf_rtcp_ssrc_length, tvb, offset, 1, FALSE );
 		offset++;
 
-		reason_text = ( char* ) malloc( reason_length + 1 );
-		for ( counter = 0; counter < reason_length; counter++ ) reason_text[ counter ] = tvb_get_guint8( tvb, offset + counter );
-		/* strncpy( reason_text, pd + offset, reason_length ); */
-		reason_text[ reason_length ] = '\0';
+		reason_text = tvb_get_string(tvb, offset, reason_length);
 		proto_tree_add_string( tree, hf_rtcp_ssrc_text, tvb, offset, reason_length, reason_text );
-		free( reason_text );
+		g_free( reason_text );
 		offset += reason_length;
 	}
 
@@ -462,24 +458,24 @@ dissect_rtcp_sdes( tvbuff_t *tvb, int offset, proto_tree *tree,
 				proto_tree_add_item( sdes_item_tree, hf_rtcp_ssrc_prefix_len, tvb, offset, 1, FALSE );
 				offset++;
 
-				prefix_string = ( char * ) malloc( prefix_len + 1 );
+				prefix_string = g_malloc( prefix_len + 1 );
 				for ( counter = 0; counter < prefix_len; counter++ )
 					prefix_string[ counter ] =
 					    tvb_get_guint8( tvb, offset + counter );
 				/* strncpy( prefix_string, pd + offset, prefix_len ); */
 				prefix_string[ prefix_len ] = '\0';
 				proto_tree_add_string( sdes_item_tree, hf_rtcp_ssrc_prefix_string, tvb, offset, prefix_len, prefix_string );
-				free( prefix_string );
+				g_free( prefix_string );
 				offset += prefix_len;
 			}
-			prefix_string = ( char * ) malloc( item_len + 1 );
+			prefix_string = g_malloc( item_len + 1 );
 			for ( counter = 0; counter < item_len; counter++ )
 			    prefix_string[ counter ] =
 			        tvb_get_guint8( tvb, offset + counter );
 			/* strncpy( prefix_string, pd + offset, item_len ); */
 			prefix_string[ item_len] = 0;
 			proto_tree_add_string( sdes_item_tree, hf_rtcp_ssrc_text, tvb, offset, item_len, prefix_string );
-			free( prefix_string );
+			g_free( prefix_string );
 			offset += item_len;
 		}
 
@@ -693,17 +689,17 @@ dissect_rtcp( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree )
 			temp_byte = tvb_get_guint8( tvb, offset );
 
 			proto_tree_add_uint( rtcp_tree, hf_rtcp_version, tvb,
-			    offset, 1, RTCP_VERSION( temp_byte ) );
+			    offset, 1, temp_byte);
 			padding_set = RTCP_PADDING( temp_byte );
 			proto_tree_add_boolean( rtcp_tree, hf_rtcp_padding, tvb,
-			    offset, 1, padding_set );
+			    offset, 1, temp_byte );
 			elem_count = RTCP_COUNT( temp_byte );
 
 			switch ( packet_type ) {
 				case RTCP_SR:
 				case RTCP_RR:
 					/* Receiver report count, 5 bits */
-					proto_tree_add_uint( rtcp_tree, hf_rtcp_rc, tvb, offset, 1, elem_count );
+					proto_tree_add_uint( rtcp_tree, hf_rtcp_rc, tvb, offset, 1, temp_byte );
 					offset++;
 					/* Packet type, 8 bits */
 					proto_tree_add_item( rtcp_tree, hf_rtcp_pt, tvb, offset, 1, FALSE );
@@ -720,7 +716,7 @@ dissect_rtcp( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree )
 					break;
 				case RTCP_SDES:
 					/* Source count, 5 bits */
-					proto_tree_add_uint( rtcp_tree, hf_rtcp_sc, tvb, offset, 1, elem_count );
+					proto_tree_add_uint( rtcp_tree, hf_rtcp_sc, tvb, offset, 1, temp_byte );
 					offset++;
 					/* Packet type, 8 bits */
 					proto_tree_add_item( rtcp_tree, hf_rtcp_pt, tvb, offset, 1, FALSE );
@@ -733,7 +729,7 @@ dissect_rtcp( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree )
 					break;
 				case RTCP_BYE:
 					/* Source count, 5 bits */
-					proto_tree_add_uint( rtcp_tree, hf_rtcp_sc, tvb, offset, 1, elem_count );
+					proto_tree_add_uint( rtcp_tree, hf_rtcp_sc, tvb, offset, 1, temp_byte );
 					offset++;
 					/* Packet type, 8 bits */
 					proto_tree_add_item( rtcp_tree, hf_rtcp_pt, tvb, offset, 1, FALSE );
@@ -800,7 +796,7 @@ proto_register_rtcp(void)
 				FT_UINT8,
 				BASE_DEC,
 				VALS(rtcp_version_vals),
-				0x0,
+				0xC0,
 				"", HFILL
 			}
 		},
@@ -810,9 +806,9 @@ proto_register_rtcp(void)
 				"Padding",
 				"rtcp.padding",
 				FT_BOOLEAN,
-				BASE_NONE,
+				8,
 				NULL,
-				0x0,
+				0x20,
 				"", HFILL
 			}
 		},
@@ -824,7 +820,7 @@ proto_register_rtcp(void)
 				FT_UINT8,
 				BASE_DEC,
 				NULL,
-				0x0,
+				0x1F,
 				"", HFILL
 			}
 		},
@@ -836,7 +832,7 @@ proto_register_rtcp(void)
 				FT_UINT8,
 				BASE_DEC,
 				NULL,
-				0x0,
+				0x1F,
 				"", HFILL
 			}
 		},
