@@ -3,8 +3,9 @@
  * as described in RFC 2478.
  * Copyright 2002, Tim Potter <tpot@samba.org>
  * Copyright 2002, Richard Sharpe <rsharpe@ns.aus.com>
+ * Copyright 2003, Richard Sharpe <rsharpe@richardsharpe.com>
  *
- * $Id: packet-spnego.c,v 1.42 2003/02/17 17:32:59 sharpe Exp $
+ * $Id: packet-spnego.c,v 1.51 2003/07/17 22:17:01 sharpe Exp $
  *
  * Ethereal - Network traffic analyzer
  * By Gerald Combs <gerald@ethereal.com>
@@ -75,6 +76,13 @@ static int hf_spnego_krb5_seal_alg = -1;
 static int hf_spnego_krb5_snd_seq = -1;
 static int hf_spnego_krb5_sgn_cksum = -1;
 static int hf_spnego_krb5_confounder = -1;
+static int hf_gssapi_reqflags_deleg = -1;
+static int hf_gssapi_reqflags_mutual = -1;
+static int hf_gssapi_reqflags_replay = -1;
+static int hf_gssapi_reqflags_sequence = -1;
+static int hf_gssapi_reqflags_anon = -1;
+static int hf_gssapi_reqflags_conf = -1;
+static int hf_gssapi_reqflags_integ = -1;
 
 static gint ett_spnego = -1;
 static gint ett_spnego_negtokeninit = -1;
@@ -85,6 +93,7 @@ static gint ett_spnego_mechlistmic = -1;
 static gint ett_spnego_responsetoken = -1;
 static gint ett_spnego_wraptoken = -1;
 static gint ett_spnego_krb5 = -1;
+static gint ett_spnego_reqflags = -1;
 
 static const value_string spnego_negResult_vals[] = {
   { SPNEGO_negResult_accept_completed,   "Accept Completed" },
@@ -93,6 +102,45 @@ static const value_string spnego_negResult_vals[] = {
   { 0, NULL}
 };
 
+/*
+ * These should be in the GSSAPI dissector ... XXX
+ */
+
+static const true_false_string tfs_reqflags_deleg = {
+  "Delegation Requested",
+  "Delegation NOT Requested"
+};
+
+static const true_false_string tfs_reqflags_mutual = {
+  "Mutual Authentication Requested",
+  "Mutual Authentication NOT Requested"
+};
+
+static const true_false_string tfs_reqflags_replay = {
+  "Replay Detection Requested",
+  "Replay Detection NOT Requested"
+};
+
+static const true_false_string tfs_reqflags_sequence = {
+  "Out-of-sequence Detection Requested",
+  "Out-of-sequence Detection NOT Requested"
+};
+
+static const true_false_string tfs_reqflags_anon = {
+  "Anonymous Authentication Requested",
+  "Anonymous Authentication NOT Requested"
+};
+
+static const true_false_string tfs_reqflags_conf = {
+  "Per-message Confidentiality Requested",
+  "Per-message Confidentiality NOT Requested"
+};
+
+static const true_false_string tfs_reqflags_integ = {
+  "Per-message Integrity Requested",
+  "Per-message Integrity NOT Requested"
+};
+
 /* Display an ASN1 parse error.  Taken from packet-snmp.c */
 
 static dissector_handle_t data_handle;
@@ -167,6 +215,11 @@ static const value_string spnego_krb5_seal_alg_vals[] = {
  * than designating SPNEGO as the mechanism, offering Kerberos V5, and
  * getting it accepted.
  */
+static int
+dissect_spnego_krb5_getmic_base(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *tree);
+static int
+dissect_spnego_krb5_wrap_base(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *tree);
+
 static void
 dissect_spnego_krb5(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 {
@@ -176,6 +229,7 @@ dissect_spnego_krb5(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 	ASN1_SCK hnd;
 	gboolean def;
 	guint len1, cls, con, tag, oid_len, nbytes;
+	guint16 token_id;
 	subid_t *oid;
 	gchar *oid_string;
 	gssapi_oid_value *value;
@@ -269,8 +323,9 @@ dissect_spnego_krb5(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 
 		/* Next, the token ID ... */
 
-		proto_tree_add_item(subtree, hf_spnego_krb5_tok_id, tvb, offset, 2,
-				    TRUE);
+		token_id = tvb_get_letohs(tvb, offset);
+		proto_tree_add_uint(subtree, hf_spnego_krb5_tok_id, tvb, offset, 2,
+				    token_id);
 
 		hnd.offset += 2;
 
@@ -280,7 +335,13 @@ dissect_spnego_krb5(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 
 	    case 14:	/* [APPLICATION 14] */
 	    case 15:	/* [APPLICATION 15] */
-		break;
+		/*
+		 * No token ID - just dissect as a Kerberos message and
+		 * return.
+		 */
+		krb5_tvb = tvb_new_subset(tvb, offset, -1, -1); 
+		offset = dissect_kerberos_main(krb5_tvb, pinfo, subtree, FALSE);
+		return;
 
 	    default:
 		proto_tree_add_text(subtree, tvb, offset, 0,
@@ -291,22 +352,177 @@ dissect_spnego_krb5(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
 	} else {
 	    /* Next, the token ID ... */
 
-	    proto_tree_add_item(subtree, hf_spnego_krb5_tok_id, tvb, offset, 2,
-				TRUE);
+	    token_id = tvb_get_letohs(tvb, offset);
+	    proto_tree_add_uint(subtree, hf_spnego_krb5_tok_id, tvb, offset, 2,
+				token_id);
 
 	    hnd.offset += 2;
 
 	    offset += 2;
 	}
 
-	krb5_tvb = tvb_new_subset(tvb, offset, -1, -1); 
+	switch (token_id) {
 
-	offset = dissect_kerberos_main(krb5_tvb, pinfo, subtree, FALSE);
+	case KRB_TOKEN_AP_REQ:
+	case KRB_TOKEN_AP_REP:
+	case KRB_TOKEN_AP_ERR:
+	  krb5_tvb = tvb_new_subset(tvb, offset, -1, -1); 
+	  offset = dissect_kerberos_main(krb5_tvb, pinfo, subtree, FALSE);
+	  break;
+
+	case KRB_TOKEN_GETMIC:
+	  offset = dissect_spnego_krb5_getmic_base(tvb, offset, pinfo, subtree); 
+	  break;
+
+	case KRB_TOKEN_WRAP:
+          offset = dissect_spnego_krb5_wrap_base(tvb, offset, pinfo, subtree);
+	  break;
+
+	case KRB_TOKEN_DELETE_SEC_CONTEXT:
+
+	  break;
+
+	default:
+
+	  break;
+	}
 
  done:
 	return;
 }
 
+/*
+ * XXX - This is for GSSAPI Wrap tokens ...
+ */
+static int
+dissect_spnego_krb5_wrap_base(tvbuff_t *tvb, int offset, packet_info *pinfo _U_, proto_tree *tree)
+{
+	guint16 sgn_alg;
+
+	/*
+	 * The KRB5 blob conforms to RFC1964:
+	 *   USHORT (0x0102 == GSS_Wrap)
+	 *   and so on } 
+	 */
+
+	/* Now, the sign and seal algorithms ... */
+
+	sgn_alg = tvb_get_letohs(tvb, offset);
+	proto_tree_add_uint(tree, hf_spnego_krb5_sgn_alg, tvb, offset, 2,
+			    sgn_alg);
+
+	offset += 2;
+
+	proto_tree_add_item(tree, hf_spnego_krb5_seal_alg, tvb, offset, 2,
+			    TRUE);
+
+	offset += 2;
+
+	/* Skip the filler */
+
+	offset += 2;
+
+	/* Encrypted sequence number */
+
+	proto_tree_add_item(tree, hf_spnego_krb5_snd_seq, tvb, offset, 8,
+			    TRUE);
+
+	offset += 8;
+
+	/* Checksum of plaintext padded data */
+
+	proto_tree_add_item(tree, hf_spnego_krb5_sgn_cksum, tvb, offset, 8,
+			    TRUE);
+
+	offset += 8;
+
+	/*
+	 * At least according to draft-brezak-win2k-krb-rc4-hmac-04,
+	 * if the signing algorithm is KRB_SGN_ALG_HMAC, there's an
+	 * extra 8 bytes of "Random confounder" after the checksum.
+	 * It certainly confounds code expecting all Kerberos 5
+	 * GSS_Wrap() tokens to look the same....
+	 */
+	if (sgn_alg == KRB_SGN_ALG_HMAC) {
+	  proto_tree_add_item(tree, hf_spnego_krb5_confounder, tvb, offset, 8,
+			      TRUE);
+
+	  offset += 8;
+	}
+
+	/*
+	 * Return the offset past the checksum, so that we know where
+	 * the data we're wrapped around starts.  Also, set the length
+	 * of our top-level item to that offset, so it doesn't cover
+	 * the data we're wrapped around.
+	 */
+	return offset;
+}
+
+/*
+ * XXX - This is for GSSAPI GetMIC tokens ...
+ */
+static int
+dissect_spnego_krb5_getmic_base(tvbuff_t *tvb, int offset, packet_info *pinfo _U_, proto_tree *tree)
+{
+	guint16 sgn_alg;
+
+	/*
+	 * The KRB5 blob conforms to RFC1964:
+	 *   USHORT (0x0101 == GSS_GetMIC)
+	 *   and so on } 
+	 */
+
+	/* Now, the sign algorithm ... */
+
+	sgn_alg = tvb_get_letohs(tvb, offset);
+	proto_tree_add_uint(tree, hf_spnego_krb5_sgn_alg, tvb, offset, 2,
+			    sgn_alg);
+
+	offset += 2;
+
+	/* Skip the filler */
+
+	offset += 4;
+
+	/* Encrypted sequence number */
+
+	proto_tree_add_item(tree, hf_spnego_krb5_snd_seq, tvb, offset, 8,
+			    TRUE);
+
+	offset += 8;
+
+	/* Checksum of plaintext padded data */
+
+	proto_tree_add_item(tree, hf_spnego_krb5_sgn_cksum, tvb, offset, 8,
+			    TRUE);
+
+	offset += 8;
+
+	/*
+	 * At least according to draft-brezak-win2k-krb-rc4-hmac-04,
+	 * if the signing algorithm is KRB_SGN_ALG_HMAC, there's an
+	 * extra 8 bytes of "Random confounder" after the checksum.
+	 * It certainly confounds code expecting all Kerberos 5
+	 * GSS_Wrap() tokens to look the same....
+	 */
+	if (sgn_alg == KRB_SGN_ALG_HMAC) {
+	  proto_tree_add_item(tree, hf_spnego_krb5_confounder, tvb, offset, 8,
+			      TRUE);
+
+	  offset += 8;
+	}
+
+	/*
+	 * Return the offset past the checksum, so that we know where
+	 * the data we're wrapped around starts.  Also, set the length
+	 * of our top-level item to that offset, so it doesn't cover
+	 * the data we're wrapped around.
+	 */
+
+	return offset;
+}
+
 /*
  * XXX - is this for SPNEGO or just GSS-API?
  * RFC 1964 is "The Kerberos Version 5 GSS-API Mechanism"; presumably one
@@ -320,7 +536,6 @@ dissect_spnego_krb5_wrap(tvbuff_t *tvb, packet_info *pinfo _U_, proto_tree *tree
 	proto_item *item;
 	proto_tree *subtree;
 	int offset = 0;
-	guint16 sgn_alg;
 
 	item = proto_tree_add_item(tree, hf_spnego_krb5, tvb, 0, -1, FALSE);
 
@@ -339,52 +554,9 @@ dissect_spnego_krb5_wrap(tvbuff_t *tvb, packet_info *pinfo _U_, proto_tree *tree
 
 	offset += 2;
 
-	/* Now, the sign and seal algorithms ... */
+        offset = dissect_spnego_krb5_wrap_base(tvb, offset, pinfo, subtree);
 
-	sgn_alg = tvb_get_letohs(tvb, offset);
-	proto_tree_add_uint(subtree, hf_spnego_krb5_sgn_alg, tvb, offset, 2,
-			    sgn_alg);
-
-	offset += 2;
-
-	proto_tree_add_item(subtree, hf_spnego_krb5_seal_alg, tvb, offset, 2,
-			    TRUE);
-
-	offset += 2;
-
-	/* Skip the filler */
-
-	offset += 2;
-
-	/* Encrypted sequence number */
-
-	proto_tree_add_item(subtree, hf_spnego_krb5_snd_seq, tvb, offset, 8,
-			    TRUE);
-
-	offset += 8;
-
-	/* Checksum of plaintext padded data */
-
-	proto_tree_add_item(subtree, hf_spnego_krb5_sgn_cksum, tvb, offset, 8,
-			    TRUE);
-
-	offset += 8;
-
-	/*
-	 * At least according to draft-brezak-win2k-krb-rc4-hmac-04,
-	 * if the signing algorithm is KRB_SGN_ALG_HMAC, there's an
-	 * extra 8 bytes of "Random confounder" after the checksum.
-	 * It certainly confounds code expecting all Kerberos 5
-	 * GSS_Wrap() tokens to look the same....
-	 */
-	if (sgn_alg == KRB_SGN_ALG_HMAC) {
-	  proto_tree_add_item(subtree, hf_spnego_krb5_confounder, tvb, offset, 8,
-			      TRUE);
-
-	  offset += 8;
-	}
-
-	/*
+        /*
 	 * Return the offset past the checksum, so that we know where
 	 * the data we're wrapped around starts.  Also, set the length
 	 * of our top-level item to that offset, so it doesn't cover
@@ -495,8 +667,10 @@ dissect_spnego_reqFlags(tvbuff_t *tvb, int offset, packet_info *pinfo _U_,
 			proto_tree *tree, ASN1_SCK *hnd)
 {
 	gboolean def;
-	guint len1, cls, con, tag;
+	guint len1, cls, con, tag, flags;
 	int ret;
+        proto_item *item;
+        proto_tree *subtree;
 
  	ret = asn1_header_decode(hnd, &cls, &con, &tag, &def, &len1);
 
@@ -518,8 +692,24 @@ dissect_spnego_reqFlags(tvbuff_t *tvb, int offset, packet_info *pinfo _U_,
 
 	offset = hnd->offset;
 
-	proto_tree_add_item(tree, hf_spnego_reqflags, tvb, offset, len1,
-			    FALSE);
+        flags = tvb_get_guint8(tvb, offset);
+
+	item = proto_tree_add_item(tree, hf_spnego_reqflags, tvb, offset, len1,
+            FALSE);
+
+        subtree = proto_item_add_subtree(item, ett_spnego_reqflags);
+
+        /*
+         * Now, the bits. XXX: Assume 8 bits. FIXME.
+         */
+
+        proto_tree_add_boolean(subtree, hf_gssapi_reqflags_deleg, tvb, offset, len1, flags);
+        proto_tree_add_boolean(subtree, hf_gssapi_reqflags_mutual, tvb, offset, len1, flags);
+        proto_tree_add_boolean(subtree, hf_gssapi_reqflags_replay, tvb, offset, len1, flags);
+        proto_tree_add_boolean(subtree, hf_gssapi_reqflags_sequence, tvb, offset, len1, flags);
+        proto_tree_add_boolean(subtree, hf_gssapi_reqflags_anon, tvb, offset, len1, flags);
+        proto_tree_add_boolean(subtree, hf_gssapi_reqflags_conf, tvb, offset, len1, flags);
+        proto_tree_add_boolean(subtree, hf_gssapi_reqflags_integ, tvb, offset, len1, flags);
 
 	hnd->offset += len1;
 
@@ -916,19 +1106,26 @@ dissect_spnego_responseToken(tvbuff_t *tvb, int offset, packet_info *pinfo _U_,
 
 	offset = hnd->offset;
 
-	item = proto_tree_add_item(tree, hf_spnego_responsetoken, tvb, offset, 
-				   nbytes, FALSE); 
+	item = proto_tree_add_item(tree, hf_spnego_responsetoken, tvb, offset -2 , 
+				   nbytes + 2, FALSE); 
 
 	subtree = proto_item_add_subtree(item, ett_spnego_responsetoken);
 
+
 	/*
 	 * Now, we should be able to dispatch after creating a new TVB.
+	 * However, we should make sure that there is something in the 
+	 * response token ...
 	 */
 
-	token_tvb = tvb_new_subset(tvb, offset, nbytes, -1);
-	if (next_level_dissector)
-	  call_dissector(next_level_dissector, token_tvb, pinfo, subtree);
-
+	if (nbytes) {
+	  token_tvb = tvb_new_subset(tvb, offset, nbytes, -1);
+	  if (next_level_dissector)
+	    call_dissector(next_level_dissector, token_tvb, pinfo, subtree);
+	}
+	else {
+	  proto_tree_add_text(subtree, tvb, offset-2, 2, "<Empty String>");
+	}
 	hnd->offset += nbytes; /* Update this ... */
 
  done:
@@ -1362,6 +1559,31 @@ proto_register_spnego(void)
 		{ &hf_spnego_reqflags, 
 		  { "reqFlags", "spnego.negtokeninit.reqflags", FT_BYTES,
 		    BASE_HEX, NULL, 0, "reqFlags", HFILL }},
+                { &hf_gssapi_reqflags_deleg,
+                  { "Delegation", "gssapi.reqflags.deleg", FT_BOOLEAN, 8,
+                    TFS(&tfs_reqflags_deleg), 0x01, "Delegation", HFILL }},
+                { &hf_gssapi_reqflags_mutual,
+                  { "Mutual Authentication", "gssapi.reqflags.mutual", FT_BOOLEAN,
+                    8, TFS(&tfs_reqflags_mutual), 0x02, "Mutual Authentication", HFILL}},
+                { &hf_gssapi_reqflags_replay,
+                  { "Replay Detection", "gssapi.reqflags.replay", FT_BOOLEAN,
+                    8, TFS(&tfs_reqflags_replay), 0x04, "Replay Detection", HFILL}},
+                { &hf_gssapi_reqflags_sequence,
+                  { "Out-of-sequence Detection", "gssapi.reqflags.sequence",
+                    FT_BOOLEAN, 8, TFS(&tfs_reqflags_sequence), 0x08, 
+                    "Out-of-sequence Detection", HFILL}},
+                { &hf_gssapi_reqflags_anon,
+                  { "Anonymous Authentication", "gssapi.reqflags.anon", 
+                    FT_BOOLEAN, 8, TFS(&tfs_reqflags_anon), 0x10,
+                    "Anonymous Authentication", HFILL}},
+                { &hf_gssapi_reqflags_conf,
+                  { "Per-message Confidentiality", "gssapi.reqflags.conf",
+                    FT_BOOLEAN, 8, TFS(&tfs_reqflags_conf), 0x20, 
+                    "Per-message Confidentiality", HFILL}},
+                { &hf_gssapi_reqflags_integ,
+                  { "Per-message Integrity", "gssapi.reqflags.integ", 
+                    FT_BOOLEAN, 8, TFS(&tfs_reqflags_integ), 0x40,
+                    "Per-message Integrity", HFILL}},
 		{ &hf_spnego_wraptoken,
 		  { "wrapToken", "spnego.wraptoken",
 		    FT_NONE, BASE_NONE, NULL, 0x0, "SPNEGO wrapToken",
@@ -1445,4 +1667,12 @@ proto_reg_handoff_spnego(void)
 	gssapi_init_oid("1.2.840.113554.1.2.2", proto_spnego_krb5, ett_spnego_krb5,
 			spnego_krb5_handle, spnego_krb5_wrap_handle,
 			"KRB5 - Kerberos 5");
+	gssapi_init_oid("1.2.840.113554.1.2.2.3", proto_spnego_krb5, ett_spnego_krb5,
+			spnego_krb5_handle, spnego_krb5_wrap_handle,
+			"KRB5 - Kerberos 5 - User to User");
+
+	/*
+	 * Find the data handle for some calls
+	 */
+	data_handle = find_dissector("data");
 }
