@@ -2,7 +2,7 @@
  * Routines for mgcp packet disassembly
  * RFC 2705
  *
- * $Id: packet-mgcp.c,v 1.37 2003/03/06 19:15:09 guy Exp $
+ * $Id: packet-mgcp.c,v 1.43 2003/12/05 09:25:41 guy Exp $
  *
  * Copyright (c) 2000 by Ed Warnicke <hagbard@physics.rutgers.edu>
  *
@@ -48,7 +48,7 @@
 
 #include "plugins/plugin_api_defs.h"
 
-#ifndef __ETHEREAL_STATIC__
+#ifndef ENABLE_STATIC
 G_MODULE_EXPORT const gchar version[] = VERSION;
 #endif
 
@@ -189,7 +189,7 @@ static dissector_handle_t sdp_handle;
 /*
  * Init Hash table stuff
  */
- 
+
 typedef struct _mgcp_call_info_key {
 	guint32	transid;
 	conversation_t *conversation;
@@ -259,42 +259,39 @@ dissect_mgcp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
    * but its cheap and its better than nothing.
    */
   if(is_mgcp_verb(tvb,0,tvb_len) || is_mgcp_rspcode(tvb,0,tvb_len)){
-
-    /* Build the info tree if we've been given a root */
-    if (tree || global_mgcp_message_count == TRUE) {
-      /*
-       * Loop through however many mgcp messages may be stuck in
-       * this packet using piggybacking
-       */
-      do{
-	num_messages++;
-	if(tree){
-	  /* Create out mgcp subtree */
-	  ti = proto_tree_add_item(tree,proto_mgcp,tvb,0,0, FALSE);
-	  mgcp_tree = proto_item_add_subtree(ti, ett_mgcp);
-	}
-
-	sectionlen = tvb_find_dot_line(tvb, tvb_sectionbegin, -1,
-				       &tvb_sectionend);
-	if( sectionlen != -1){
-	  dissect_mgcp_message(tvb_new_subset(tvb, tvb_sectionbegin,
-					      sectionlen, -1),
-			       pinfo, tree, mgcp_tree,ti);
-	  tvb_sectionbegin = tvb_sectionend;
-	}
-	else {
-	  break;
-	}
-      } while(tvb_sectionend < tvb_len );
-      if(mgcp_tree){
-	proto_tree_add_uint_hidden(mgcp_tree, hf_mgcp_messagecount, tvb,
-				   0 ,0 , num_messages);
+    /*
+     * Loop through however many mgcp messages may be stuck in
+     * this packet using piggybacking
+     */
+    do{
+      num_messages++;
+      if(tree){
+	/* Create out mgcp subtree */
+	ti = proto_tree_add_item(tree,proto_mgcp,tvb,0,0, FALSE);
+	mgcp_tree = proto_item_add_subtree(ti, ett_mgcp);
       }
+
+      sectionlen = tvb_find_dot_line(tvb, tvb_sectionbegin, -1,
+				     &tvb_sectionend);
+      if( sectionlen != -1){
+	dissect_mgcp_message(tvb_new_subset(tvb, tvb_sectionbegin,
+					    sectionlen, -1),
+			     pinfo, tree, mgcp_tree,ti);
+	tvb_sectionbegin = tvb_sectionend;
+      }
+      else {
+	break;
+      }
+    } while(tvb_sectionend < tvb_len );
+    if(mgcp_tree){
+      proto_tree_add_uint_hidden(mgcp_tree, hf_mgcp_messagecount, tvb,
+				 0 ,0 , num_messages);
     }
 
     /*
      * Add our column information we do this after dissecting SDP
      * in order to prevent the column info changing to reflect the SDP.
+     * XXX - can we do this with a fence?
      */
     tvb_sectionbegin = 0;
     if (check_col(pinfo->cinfo, COL_PROTOCOL)){
@@ -329,7 +326,7 @@ dissect_mgcp_message(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree,
   gint tvb_sectionend,tvb_sectionbegin, tvb_len, tvb_current_len;
   tvbuff_t *next_tvb;
   static mgcp_info_t mi;
-    
+
   /* Initialize variables */
   tvb_sectionend = 0;
   tvb_sectionbegin = tvb_sectionend;
@@ -343,51 +340,46 @@ dissect_mgcp_message(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree,
    * but its cheap and its better than nothing.
    */
   if(is_mgcp_verb(tvb,0,tvb_len) || is_mgcp_rspcode(tvb,0,tvb_len)){
+    /* dissect first line */
+    tvb_sectionbegin = 0;
+    tvb_current_len = tvb_len;
+    tvb_sectionend = tvb_sectionbegin;
+    sectionlen = tvb_find_line_end(tvb,0,-1,&tvb_sectionend,FALSE);
+    if( sectionlen > 0){
+      dissect_mgcp_firstline(tvb_new_subset(tvb, tvb_sectionbegin,
+					    sectionlen,-1), pinfo,
+			     mgcp_tree, &mi);
+    }
+    tvb_sectionbegin = tvb_sectionend;
 
-    /* Build the info tree if we've been given a root */
-    if (tree && mgcp_tree) {
-
-      /* dissect first line */
-      tvb_sectionbegin = 0;
-      tvb_current_len = tvb_len;
-      tvb_sectionend = tvb_sectionbegin;
-      sectionlen = tvb_find_line_end(tvb,0,-1,&tvb_sectionend,FALSE);
-      if( sectionlen > 0){
-	dissect_mgcp_firstline(tvb_new_subset(tvb, tvb_sectionbegin,
-					      sectionlen,-1), pinfo,
-			       mgcp_tree, &mi);
-      }
-      tvb_sectionbegin = tvb_sectionend;
-
-      /* dissect params */
-      if(tvb_sectionbegin < tvb_len){
-	sectionlen = tvb_find_null_line(tvb, tvb_sectionbegin, -1,
-					&tvb_sectionend);
-	dissect_mgcp_params(tvb_new_subset(tvb, tvb_sectionbegin,
-					   sectionlen, -1),
-			    mgcp_tree);
-	tvb_sectionbegin = tvb_sectionend;
-      }
-
-      /* set the mgcp payload length correctly so we don't include the
-       * encapsulated SDP
-       */
-      sectionlen = tvb_sectionend;
-      proto_item_set_len(ti,sectionlen);
-
-      /* Display the raw text of the mgcp message if desired */
-
-      /* Do we want to display the raw text of our MGCP packet? */
-      if(global_mgcp_raw_text){
-	mgcp_raw_text_add(tvb_new_subset(tvb,0,tvb_len,-1),
+    /* dissect params */
+    if(tvb_sectionbegin < tvb_len){
+      sectionlen = tvb_find_null_line(tvb, tvb_sectionbegin, -1,
+				      &tvb_sectionend);
+      dissect_mgcp_params(tvb_new_subset(tvb, tvb_sectionbegin,
+					 sectionlen, -1),
 			  mgcp_tree);
-      }
+      tvb_sectionbegin = tvb_sectionend;
+    }
 
-      /* dissect sdp payload */
-      if( tvb_sectionend < tvb_len && global_mgcp_dissect_tree == TRUE){
-	next_tvb = tvb_new_subset(tvb, tvb_sectionend, -1, -1);
-	call_dissector(sdp_handle, next_tvb, pinfo, tree);
-      }
+    /* set the mgcp payload length correctly so we don't include the
+     * encapsulated SDP
+     */
+    sectionlen = tvb_sectionend;
+    proto_item_set_len(ti,sectionlen);
+
+    /* Display the raw text of the mgcp message if desired */
+
+    /* Do we want to display the raw text of our MGCP packet? */
+    if(global_mgcp_raw_text) {
+      if (tree)
+	mgcp_raw_text_add(tvb, mgcp_tree);
+    }
+
+    /* dissect sdp payload */
+    if( tvb_sectionend < tvb_len){
+      next_tvb = tvb_new_subset(tvb, tvb_sectionend, -1, -1);
+      call_dissector(sdp_handle, next_tvb, pinfo, tree);
     }
   }
 }
@@ -441,7 +433,7 @@ mgcp_init_protocol(void)
 	mgcp_call_info_value_chunk = g_mem_chunk_new("call_info_value_chunk",
 	    sizeof(mgcp_call_t),
 	    200 * sizeof(mgcp_call_t),
-	    G_ALLOC_ONLY);	
+	    G_ALLOC_ONLY);
 }
 
 /* Register all the bits needed with the filtering engine */
@@ -456,10 +448,10 @@ proto_register_mgcp(void)
     { &hf_mgcp_rsp,
       { "Response", "mgcp.rsp", FT_BOOLEAN, BASE_NONE, NULL, 0x0,
 	"TRUE if MGCP response", HFILL }},
-    { &hf_mgcp_req_frame, 
+    { &hf_mgcp_req_frame,
       {	"Request Frame", "mgcp.reqframe", FT_FRAMENUM, BASE_NONE, NULL, 0,
         "Request Frame", HFILL }},
-    { &hf_mgcp_rsp_frame, 
+    { &hf_mgcp_rsp_frame,
       {	"Response Frame", "mgcp.rspframe", FT_FRAMENUM, BASE_NONE, NULL, 0,
         "Response Frame", HFILL }},
     { &hf_mgcp_time,
@@ -638,7 +630,7 @@ proto_register_mgcp(void)
                                  "Display the number of MGCP messages "
                                  "found in a packet in the protocol column.",
                                  &global_mgcp_message_count);
-  
+
   mgcp_tap = register_tap("mgcp");
 }
 
@@ -697,7 +689,7 @@ static gboolean is_mgcp_verb(tvbuff_t *tvb, gint offset, gint maxlength){
   int returnvalue = FALSE;
   guint8 word[5];
 
-  if(( maxlength >= 4) && tvb_get_nstringz0(tvb,offset,4,word)){
+  if(( maxlength >= 4) && tvb_get_nstringz0(tvb,offset,sizeof(word),word)){
     if (strncasecmp(word, "EPCF", 4) == 0 ||
 	strncasecmp(word, "CRCX", 4) == 0 ||
 	strncasecmp(word, "MDCX", 4) == 0 ||
@@ -739,7 +731,7 @@ static gboolean is_mgcp_rspcode(tvbuff_t *tvb, gint offset, gint maxlength){
   int returnvalue = FALSE;
   guint8 word[4];
   if(maxlength >= 3){
-    tvb_get_nstringz0(tvb,offset,3,word);
+    tvb_get_nstringz0(tvb,offset,sizeof(word),word);
     if( isdigit(word[0]) &&
 	isdigit(word[1]) &&
 	isdigit(word[2])){
@@ -968,7 +960,7 @@ static void dissect_mgcp_firstline(tvbuff_t *tvb, packet_info *pinfo,
   mgcp_call_info_key *new_mgcp_call_key = NULL;
   mgcp_call_t *mgcp_call = NULL;
   nstime_t delta;
-  
+
   static address null_address = { AT_NONE, 0, NULL };
   proto_item* (*my_proto_tree_add_string)(proto_tree*, int, tvbuff_t*, gint,
 					  gint, const char*);
@@ -1001,7 +993,6 @@ static void dissect_mgcp_firstline(tvbuff_t *tvb, packet_info *pinfo,
 	tokenlen = tvb_current_offset - tvb_previous_offset;
       }
       if(tokennum == 0){
-        code = g_malloc(tokenlen);
         code = tvb_format_text(tvb,tvb_previous_offset,tokenlen);
         strncpy(mi->code,code,4);
         mi->code[4] = '\0';
@@ -1022,7 +1013,6 @@ static void dissect_mgcp_firstline(tvbuff_t *tvb, packet_info *pinfo,
 	}
       }
       if(tokennum == 1){
-      	transid = g_malloc(tokenlen);
         transid = tvb_format_text(tvb,tvb_previous_offset,tokenlen);
         /* XXX - what if this isn't a valid text string? */
         mi->transid = atol(transid);
@@ -1073,6 +1063,7 @@ static void dissect_mgcp_firstline(tvbuff_t *tvb, packet_info *pinfo,
       tokennum++;
     } while( tvb_current_offset < tvb_len && tvb_previous_offset < tvb_len
 	     && tokennum <= 3);
+
     switch (mgcp_type){
     case MGCP_RESPONSE:
 	proto_tree_add_boolean_hidden(tree, hf_mgcp_rsp, tvb, 0, 0, TRUE);
@@ -1119,6 +1110,7 @@ static void dissect_mgcp_firstline(tvbuff_t *tvb, packet_info *pinfo,
 			if(mgcp_call->req_num){
 				mi->request_available = TRUE;
 				mgcp_call->responded = TRUE;
+				strcpy(mi->code,mgcp_call->code);
 				proto_tree_add_uint_format(tree, hf_mgcp_req_frame,
 				    tvb, 0, 0, mgcp_call->req_num,
 				    "This is a response to a request in frame %u",
@@ -1144,7 +1136,7 @@ static void dissect_mgcp_firstline(tvbuff_t *tvb, packet_info *pinfo,
 				if (mgcp_call->rsp_num != pinfo->fd->num) {
 					/* No, so it's a duplicate response.
 					   Mark it as such. */
-					mi->is_duplicate = TRUE;   
+					mi->is_duplicate = TRUE;
 					if (check_col(pinfo->cinfo, COL_INFO)) {
 						col_append_fstr(pinfo->cinfo, COL_INFO,
 							", Duplicate Response %u",mi->transid);
@@ -1206,11 +1198,11 @@ static void dissect_mgcp_firstline(tvbuff_t *tvb, packet_info *pinfo,
 			    pinfo->destport, 0);
 		}
 	}
-	
+
 	/* prepare the key data */
 	mgcp_call_key.transid = mi->transid;
 	mgcp_call_key.conversation = conversation;
-	
+
 	/* look up the request */
 	mgcp_call = g_hash_table_lookup(mgcp_calls, &mgcp_call_key);
 	if (mgcp_call != NULL) {
@@ -1223,11 +1215,11 @@ static void dissect_mgcp_firstline(tvbuff_t *tvb, packet_info *pinfo,
 			mi->is_duplicate = TRUE;
 			if (check_col(pinfo->cinfo, COL_INFO)) {
 				col_append_fstr(pinfo->cinfo, COL_INFO,
-					", Duplicate Request %u",mi->transid);	
+					", Duplicate Request %u",mi->transid);
 				if (tree) {
-					proto_tree_add_uint_hidden(tree, 
+					proto_tree_add_uint_hidden(tree,
 						hf_mgcp_dup, tvb, 0,0, mi->transid);
-					proto_tree_add_uint_hidden(tree, 
+					proto_tree_add_uint_hidden(tree,
 						hf_mgcp_req_dup, tvb, 0,0, mi->transid);
 				}
 			}
@@ -1249,6 +1241,7 @@ static void dissect_mgcp_firstline(tvbuff_t *tvb, packet_info *pinfo,
 		mgcp_call->req_time.secs=pinfo->fd->abs_secs;
 		mgcp_call->req_time.nsecs=pinfo->fd->abs_usecs*1000;
 		strcpy(mgcp_call->code,mi->code);
+
 		/* store it */
 		g_hash_table_insert(mgcp_calls, new_mgcp_call_key, mgcp_call);
 	}
@@ -1532,7 +1525,7 @@ static gint tvb_find_dot_line(tvbuff_t* tvb, gint offset,
 
 /* Start the functions we need for the plugin stuff */
 
-#ifndef __ETHEREAL_STATIC__
+#ifndef ENABLE_STATIC
 
 G_MODULE_EXPORT void
 plugin_reg_handoff(void){
